//  Copyright Michael S. Briggs, University of Alabama in Huntsville.
//  For NASA's Fermi Gamma-ray Burst Monitor (GBM).
//  2016 Jan 26.  Always attempt to read the trigger time.
//  2008 October 26 -- Nov 11 &  2009 January 30--31.

//  Revised 2015 May 20.  Check that the number of energy channels has the
//  expected value of 128.
//  Revised 2015 May 5.  split the function definitions that should be
//  externally visible out to a separate file, Read_Lvl1_TTE.h.  That way
//  that file can be included by a C file that uses those routines, Read_Lvl1_TTE
//  and Cleanup_TTE_FitsFile.

//  Revised 2014 February 11.    The previous approach was that routine Read_Lvl1_TTE
//  was called to allocate memory, open FITS file, read FITS file and place info into
//  arguments.   Later Cleanup_TTE_FitsFile is called to deallocate FITS file and close
//  the FITS file.  Now the FITS file is automatically closed in Read_Lvl1_TTE.  Routine
//  Cleanup_TTE_FitsFile should still be called to deallocate memory.  This change was
//  made because we are finished with the FITS file after it is processed by Read_Lvl1_TTE
//  and so there is no reason to keep it open.   It is better to close the FITS file
//  ASAP and free memory.  The changes are minimal, in particular, the file handle is still
//  a global variable, which is not necessary in the new design.   The previous code is
//  removed by commenting out with "//>> 2014Feb11".  New code is similarly marked.
//  Implementing these changes in this way allows easy reversion, should we change our
//  mind.   MSB.

//  Revised 2011 March 5: add variables FileStartTime & FileStopTime: time range of the file --
//  the FITS keywords TSTART & TSTOP.
//  Revised 2010 Nov 6: add output of the FileName to some error messages.

//  Also add debug subroutine ReportSharedVar.

//  Revised 2010 May 29 to use stdbool.h
//  Revised 2009 December 24: reduce the normal amount of output.  Verbose and Debug level
//  output may be selected with compile time options VERBOSE_READ_TTE and DEBUG_READ_TTE,
//  e.g., -DVERBOSE_READ_TTE or -DDEBUG_READ_TTE.
//  Revised 2009 October 25 to translate cfitsio status returns into strings, etc.
//  Revised 2009 March 1 to also read the energy edges of the channels.



//  Subroutine that uses cfitsio to read GBM Level 1 TTE FITS files.

//  Callable from Fortran 95/2003 via an interface that I have written,
//  also using the ISO C Binding.


// Headers

#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>

#include "fitsio.h"


//  define functions below that should be externally accessible:

#include "Read_Lvl1_TTE.h"


//  typedef

typedef struct Shared_Between_Routines_type {

   _Bool  InUse;
   fitsfile * fits_file_ptr;
   short int **  tte_chan_ptr_to_ptr;
   double    **  tte_time_ptr_to_ptr;

} Shared_Between_Routines_type;



//  local function prototypes:


void  StatusValue_2_String (

   const int  status,
   char InputFileName [FILE_NAME_MAX_LEN]

);


void  ReportSharedVar (

   long int  DetNum

);




//  file-scope globals shared between the two functions:
//  (As file-scope, values are preserved across calls.)


//  Routine Read_Lvl1_TTE allocates files and arrays, while
//  routine Cleanup_TTE_FitsFile deallocates / frees the
//  files and arrays.   To keep it simple for the calling routine,
//  this "module" remembers the pointers to the files & arrays.
//  This is "information hiding" where this "module" hides most
//  of the information from the calling routine and instead
//  provides a much simpler method for the calling routine to
//  help the two routines of this file communicate with each other.
//  The pointers are indexed by detector number -- we know that
//  GBM has exactly 14 detectors.  The calling routine indicates
//  which file & arrays it no longer needs by detector number and
//  Cleanup_TTE_FitsFile looks up the pointers in these arrays
//  (structure members) by using the detector number as index.

//  "static" so that the variable is local to this file.

static  Shared_Between_Routines_type  Shared_Between_Routines [DET_NUMBER];


// ****************************************************************************************


//  Re the output argument TriggerTime:
//  TriggerTime is only valid for trigger events, in which case the value returned will
//  be larger than 237427201.0 (2008 July 11, the day of GBM trigger enable).
//  If the data file is for continuous (i.e., non-trigger) data, then -99.99 is returned.

void  Read_Lvl1_TTE (

   // input arguments:
   char InputFileName [FILE_NAME_MAX_LEN],
   long int  DetNum,

   //  output arguments:
   double  * FileStartTime_ptr,         // time range of file.  FITS keywords TSTART & TSTOP
   double  * FileStopTime_ptr,
   double  * TriggerTime_ptr,           // pointer to scaler
   long int * Num_TTE_ptr,              // pointer to scaler
   char * DataSum_ptr,                  // pointer to char array of length 11
   short int **  tte_chan_ptr_to_ptr,   // pointer to pointer to variable length array
   double    **  tte_time_ptr_to_ptr,
   float  * EMin_of_chan_ptr,           // pointer to array of length NUM_SPEC_CHAN
   float  * EMax_of_chan_ptr

) {

   unsigned int  idet;

   fitsfile * fits_file_ptr;
   int  status = 0;

   float  fnulval;
   double  dnulval;
   short int snulval;
   int anynul;

   int  NumHeaders;
   int  Check_NumChan;
   char Comment [FLEN_COMMENT];


   #ifdef DEBUG_READ_TTE
      long int ichan;
   #endif


   // static to preserve value across calls:
   static  _Bool  FirstCall = true;


   // * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   //  First call initializations:

   if ( FirstCall ) {   // FirstCall ?

      FirstCall = false;
      for ( idet=0;  idet<DET_NUMBER;  idet++ )  Shared_Between_Routines [idet] .InUse = false;

      #ifdef DEBUG_READ_TTE
         printf ( "\nFIRST CALL\n" );
      #endif

   } else {   // FirstCall ?
      #ifdef DEBUG_READ_TTE
         printf ( "\nNOT FIRST CALL\n" );
      #endif
   }   // FirstCall ?


   if ( DetNum < 0  ||  DetNum >= DET_NUMBER ) {
      printf ( "\nBad input for Detector Number: %ld\n", DetNum );
      exit (21);
   }


   if ( Shared_Between_Routines [DetNum] .InUse )
      printf ( "\n\nProgram ERROR: Memory Leak: Reallocating files/arrays beforing freeing.  Detector %ld !!\n", DetNum );


   // Record that file and arrays very soon will be allocated for detector DetNum:
   Shared_Between_Routines [DetNum] .InUse = true;


   //  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^


   //  Open the FITS file

   fits_open_file ( &fits_file_ptr, InputFileName, READONLY, &status );

   if ( status != 0 ) {
      StatusValue_2_String  ( status, InputFileName );
      printf ( "==> Failed to open file '%s'\n", InputFileName );
      exit (22);
   }


   //  Store pointer for use by Cleanup_TTE_FitsFile:
   Shared_Between_Routines [DetNum] .fits_file_ptr = fits_file_ptr;


   // optional, extra test -- does the file have the expected number of headers?

   fits_get_num_hdus ( fits_file_ptr, &NumHeaders, &status );
   if ( status != 0 ) {
      StatusValue_2_String  ( status, InputFileName );
      printf ( "==> Failed to obtain number of Headers.\n" );
      exit (23);
   }
   if ( NumHeaders != 4 ) {
      printf ( "\nNumber of headers has unexpected value !!: %d\n", NumHeaders );
      printf ( "This error occurred while processing file '%s'\n", InputFileName );
   }


   //  Read keywords from the Primary Header:

   //  Read TSTART and TSTOP

   fits_read_key ( fits_file_ptr, TDOUBLE, "TSTART", FileStartTime_ptr, Comment, &status );

   if ( status != 0 ) {
      StatusValue_2_String  ( status, InputFileName );
      printf ( "==> Read of the TSTART KeyWord of Primary Header FAILED!\n" );
      exit (61);
   }

   fits_read_key ( fits_file_ptr, TDOUBLE, "TSTOP", FileStopTime_ptr, Comment, &status );

   if ( status != 0 ) {
      StatusValue_2_String  ( status, InputFileName );
      printf ( "==> Read of the TSTOP KeyWord of Primary Header FAILED!\n" );
      exit (62);
   }


   //  Read the TRIGTIME Keyword.
   //  This item is only present in the file for triggers so not finding it is OK !
   //  If not found, indicate with a negative value.

   *TriggerTime_ptr = 0.0;

   fits_read_key ( fits_file_ptr, TDOUBLE, "TRIGTIME", TriggerTime_ptr, Comment, &status );

   if ( status != 0 ) {
       status = 0;    // clear the error
      *TriggerTime_ptr = -99.99;
   }


   #if defined (VERBOSE_READ_TTE) || defined (DEBUG_READ_TTE)
      printf  ( "\nTrigger Time is %17.6lf\n", *TriggerTime_ptr );
   #endif


   //  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^


   //  Move to EBOUNDS Header.

   fits_movnam_hdu ( fits_file_ptr, BINARY_TBL, "EBOUNDS", 0, &status );
   if ( status != 0 ) {
      StatusValue_2_String  ( status, InputFileName );
      printf ( "==> Failed to locate EBOUNDS header\n" );
      exit (25);
   }


   //  Verify 128 energy (output) channels

   fits_read_key ( fits_file_ptr, TINT, "NAXIS2", &Check_NumChan, Comment, &status );

   if ( status != 0 ) {
      StatusValue_2_String  ( status, InputFileName );
      printf ( "==> Read of the NAXIS2 KeyWord of EBOUNDS Header FAILED!\n" );
      exit (75);
   }
   if ( Check_NumChan != 128 ) {
      printf ( "\nUnexpected number of energy channels -- only support 128 !!: %d\n", Check_NumChan );
      printf ( "This error occurred while processing file '%s'\n", InputFileName );
      exit (77);
   }


   //  Read column 2 = E_MIN,
   //  Read column 3 = E_MAX

   fnulval = -7.77F;
   fits_read_col ( fits_file_ptr, TFLOAT, 2, 1LL, 1LL, NUM_SPEC_CHAN, &fnulval, EMin_of_chan_ptr, &anynul, &status );
   if ( status != 0 ) {
      StatusValue_2_String  ( status, InputFileName );
      printf ( "==> Failed to read Column 2 with E_MIN values.\n" );
      exit (26);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for Emin!\n" );

   fits_read_col ( fits_file_ptr, TFLOAT, 3, 1LL, 1LL, NUM_SPEC_CHAN, &fnulval, EMax_of_chan_ptr, &anynul, &status );
   if ( status != 0 ) {
      StatusValue_2_String  ( status, InputFileName );
      printf ( "==> Failed to read Column 3 with E_MAX values.\n" );
      exit (27);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for Emax!\n" );


   #ifdef DEBUG_READ_TTE
     for (ichan=0; ichan<NUM_SPEC_CHAN; ichan++ )
         printf ( "%ld  %15.5e   %15.5e\n", ichan, EMin_of_chan_ptr[ichan], EMax_of_chan_ptr[ichan] );
   #endif



   //  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^


   //  Move to EVENTS Header.

   fits_movnam_hdu ( fits_file_ptr, BINARY_TBL, "EVENTS", 0, &status );
   if ( status != 0 ) {
      StatusValue_2_String  ( status, InputFileName );
      printf ( "==> Failed to locate EVENTS header\n" );
      exit (28);
   }


   //   How many TTE events are in the file?
   //   Equivalent to how many rows are in the EVENTS Table:

   fits_get_num_rows ( fits_file_ptr, Num_TTE_ptr, &status );
   if ( status != 0 ) {
      StatusValue_2_String  ( status, InputFileName );
      printf ( "==> Failed to read number of TTE events.\n" );
      exit (29);
   }


   #if defined (VERBOSE_READ_TTE) || defined (DEBUG_READ_TTE)
      printf ( "\nThere are %ld TTE Events in file '%s'.\n", *Num_TTE_ptr, InputFileName );
   #endif


   //   Read the DATASUM keyword from the EVENTS header:

   fits_read_key ( fits_file_ptr, TSTRING, "DATASUM", DataSum_ptr, Comment, &status );

   if ( status != 0 ) {
      StatusValue_2_String  ( status, InputFileName );
      printf ( "==> Failed to read number DATASUM keyword.\n" );
      DataSum_ptr = "FailedRead";
      status = 0;    // clear the error
   }


   //   Allocate memory for the TTE data in this file:

   *tte_chan_ptr_to_ptr = malloc ( (size_t) *Num_TTE_ptr * sizeof (short int) );
   *tte_time_ptr_to_ptr = malloc ( (size_t) *Num_TTE_ptr * sizeof (double) );


   #ifdef DEBUG_READ_TTE
      printf ( "\nAllocated sizes: %ld  %ld\n", *Num_TTE_ptr * sizeof (short int), *Num_TTE_ptr * sizeof (double) );
      printf ( "The two addresses created by malloc: %p  %p\n", (void *) *tte_chan_ptr_to_ptr, (void *) *tte_time_ptr_to_ptr );
   #endif


   //  Store pointers for use by Cleanup_TTE_FitsFile:
   Shared_Between_Routines [DetNum] .tte_chan_ptr_to_ptr = tte_chan_ptr_to_ptr;
   Shared_Between_Routines [DetNum] .tte_time_ptr_to_ptr = tte_time_ptr_to_ptr;

   if ( *tte_chan_ptr_to_ptr == NULL  ||  *tte_time_ptr_to_ptr == NULL ) {
      printf ( "\n\nFailed to allocate memory for the TTE data !\n" );
      exit (30);
   }


   //  Read column 1 = TIME,
   //  Read column 2 = PHA

   dnulval = -7.77;
   fits_read_col ( fits_file_ptr, TDOUBLE, 1, 1LL, 1LL, (long long) *Num_TTE_ptr, &dnulval, *tte_time_ptr_to_ptr, &anynul, &status );
   if ( status != 0 ) {
      StatusValue_2_String  ( status, InputFileName );
      printf ( "==> Failed to read Column 1 with TTE Times.\n" );
      exit (31);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for TTE Time!\n" );

   snulval = -7;
   fits_read_col ( fits_file_ptr, TSHORT, 2, 1LL, 1LL, (long long) *Num_TTE_ptr, &snulval, *tte_chan_ptr_to_ptr, &anynul, &status );
   if ( status != 0 ) {
      StatusValue_2_String  ( status, InputFileName );
      printf ( "==> Failed to read Column 2 with TTE Channels.\n" );
      exit (32);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for TTE Channel!\n" );


   #ifdef DEBUG_READ_TTE
      printf ( "1st TTE event: time, chan: %lf  %d\n", (*tte_time_ptr_to_ptr) [0] - *TriggerTime_ptr, (*tte_chan_ptr_to_ptr) [0] );
      printf ( "2nd TTE event: time, chan: %lf  %d\n", (*tte_time_ptr_to_ptr) [1] - *TriggerTime_ptr, (*tte_chan_ptr_to_ptr) [1] );
      printf ( "Last TTE event: chan, time: %lf  %d\n",
            (*tte_time_ptr_to_ptr) [*Num_TTE_ptr-1] - *TriggerTime_ptr, (*tte_chan_ptr_to_ptr) [*Num_TTE_ptr-1] );
   #endif



   //>> 2014Feb11: Revised so that this routine closes the file after it has
   // read the information that it places into its arguments, rather than
   // having the calling program call Cleanup_TTE_FitsFile to have that done.

   fits_close_file ( Shared_Between_Routines [DetNum] .fits_file_ptr, &status );  //>> 2014Feb11



   return;

}  // Read_Lvl1_TTE ()



// ****************************************************************************************


//>> 2014Feb11:  Revised 2014 February 11: This routine no longer closes the FITS file.
//  The new approach is that Read_Lvl1_TTE does that automatically.  This supersedes
//  the comments below.  The old approach is commented out with "//>> 2014Feb11" should
//  we wish to revert.

//   DetNum is an input argument -- it indicates which file and memory to deallocate.
//   The file and memory have been previously allocated by Read_Lvl1_TTE and the
//   information stored in file-scope shared variables that are indexed by DetNum.

//  Cleanup_TTE_FitsFile can (should) be called after calling Read_Lvl1_TTE -- it
//  deallocates one FITS file and two malloc'd arrays.
//  It can be called after additional calls to Read_Lvl1_TTE have been made, but must
//  be called before a second call to Read_Lvl1_TTE is made for the same detector --
//  the file-scope shared variables can only remember one set of "file and two arrays
//  per detector".   If Read_Lvl1_TTE is called twice for one particular detector
//  with no intervening call to Cleanup_TTE_FitsFile, there will be lost, unfreed memory.


void  Cleanup_TTE_FitsFile (

   long int  DetNum

) {

   //>> 2014Feb11:  int  status = 0;


   // * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   if ( DetNum < 0  ||  DetNum >= DET_NUMBER ) {
      printf ( "\nBad input for Detector Number: %ld\n", DetNum );
      exit (41);
   }

   #ifdef DEBUG_READ_TTE
      printf ( "Cleanup_TTE_FitsFile: about to deallocate open FITS file and malloc'd arrays for detector %ld\n", DetNum );
      printf ( "\nThe two addresses about to be freed: %p  %p\n",
           (void *) * (Shared_Between_Routines [DetNum] .tte_chan_ptr_to_ptr),
           (void *) * (Shared_Between_Routines [DetNum] .tte_time_ptr_to_ptr) );
   #endif

   if ( Shared_Between_Routines [DetNum] .InUse ) {

      //>> 2014Feb11: fits_close_file ( Shared_Between_Routines [DetNum] .fits_file_ptr, &status );

      free ( * (Shared_Between_Routines [DetNum] .tte_chan_ptr_to_ptr) );
      free ( * (Shared_Between_Routines [DetNum] .tte_time_ptr_to_ptr) );

      Shared_Between_Routines [DetNum] .InUse = false;

   } else {

      printf ( "Error calling Cleanup_TTE_FitsFile: Detector %ld is not in use!\n", DetNum );

   }


   return;

}  // Cleanup_TTE_FitsFile ()



// ****************************************************************************************

//  Call this if an error has occurred on a call to a cfitsio routine -- this routine
//  will convert the status numeric value to a string and output the string.

//  2010 Nov 6: add output of the name of the file -- while reading this file, the
//  cfitsio error "status" occurred.


void  StatusValue_2_String (

   const int  status,
   char InputFileName [FILE_NAME_MAX_LEN]

) {

    char ErrorText [30];


   // * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   fits_get_errstatus  ( status, ErrorText );

   printf ( "\nA cfitsio routine returned an error.  The numeric value is %d\n", status );
   printf ( "This translates to: '%s'\n", ErrorText );
   printf ( "This error occurred while processing file '%s'\n", InputFileName );

   return;

}  // StatusValue_2_String ()



// ****************************************************************************************

//  Debug routine to check on the contents of "Shared_Between_Routines".
//  There is no need to call this routine for normal usage of this code.

void  ReportSharedVar (

   long int  DetNum

) {


   // * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   if ( DetNum < 0  ||  DetNum >= DET_NUMBER ) {
      printf ( "\nBad input for Detector Number: %ld\n", DetNum );
      exit (53);
   }


   if ( Shared_Between_Routines [DetNum] .InUse ) {
      printf ( "Good call to ReportSharedVar: Detector %ld IS in use!\n", DetNum );
   } else {
      printf ( "Error calling ReportSharedVar: Detector %ld is not in use!\n", DetNum );
   }

   printf ( "ReportSharedVar: reporting addresses for detector %ld\n", DetNum );
   printf ( "The two addresses are: %p  %p\n",
        (void *) * (Shared_Between_Routines [DetNum] .tte_chan_ptr_to_ptr),
        (void *) * (Shared_Between_Routines [DetNum] .tte_time_ptr_to_ptr) );

   return;

}  // ReportSharedVar ()


// ****************************************************************************************

