!  Copyright Michael S. Briggs, University of Alabama in Huntsville.
!  For NASA's Fermi Gamma-ray Burst Monitor (GBM).
!  copyright 2009 February 9 -- 2016.

!  First release: version 0.9
!  Version 0.9.1, 2009 Feb 12: add PNG output option.
!  Version 0.9.5, 2009 March 1: upgraded subroutine Read_Lvl1_TTE to read energy edges.
!  Version 0.9.6, 2009 March 7: moved all user input to main program, moved from
!     subroutine MakeMultiHistoPlot.
!  Version 0.99 by MSB, 2009 March 22: call widget routine to select energy channels
!  Version 1.1 by MSB, 2009 May 29: processing of TTE data factored out into subroutine Read_and_Bin_TTE.
!  Version 1.2 by MSB, 2009 June 21: automatic defaults for some options.
!  Version 1.3 by MSB, 2009 August 16:
!    a) provide over-ridable default for x tick mark spacing (xdiv),
!    b) EPS output option vis Get_GraphFormat and changes in subroutine MakeMultiHistoPlot.

!  Program TTE_Histo_MultiGraphs: plots histograms of GBM TTE data.  Black curve is observed (raw)
!  counts, optional red histogram is deadtime corrected.
!  One graph is made for each of the 14 GBM detectors.

!  Usage: TTE_Histo_MultiGraphs.exe  FileName.


! #################################################################################################


program TTE_Histo_MultiGraphs

   use MSB_Constants

   use, intrinsic :: iso_c_binding, only: C_CHAR

   use TTE_subroutines, only:  TTE_GetCmdLineInput_AllFileNames,  MakeUpperCase,  MakeLowerCase,      &
               Read_and_Bin_TTE, MakeMultiHistoPlot,  Select_EChannels,                               &
               auto_ymax_value,  auto_ydiv_value,  Get_GraphFormat,  auto_xdiv_value, MakeUpperCase

   use TTE_constant_defs
   use TTE_Histogram_Arrays

   implicit none

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   character (len=FILE_NAME_MAX_LEN) :: NamePortion
   character (kind=C_CHAR, len=FILE_NAME_MAX_LEN), dimension(0: MAX_GBM_DET-1 ) :: InputFileNames_for_C


   real (real64), dimension (2) :: GraphTimeRange_s
   real (real64) :: BinWidth_s
   integer (int32), dimension (2) :: ChannelRange


   real (real32), dimension ( 0: NUM_SPEC_CHAN-1, 0: MAX_GBM_DET-1 ) :: EMin_of_chan
   real (real32), dimension ( 0: NUM_SPEC_CHAN-1, 0: MAX_GBM_DET-1 ) :: EMax_of_chan

   integer :: i
   logical, dimension ( 0: MAX_GBM_DET-1 ), parameter :: RequestDetectors = (/  ( .TRUE., i=1, MAX_GBM_DET )  /)
   integer (int32) :: idet
   integer :: TimeMode

   character (len=2) :: TimeLabel, BinLabel
   integer (int32), dimension (2) :: MaxBinValueRaw
   real (real32), dimension (2) :: MaxBinValueDeadTime

   real (real32) :: xdiv
   real (real32) :: ydiv
   real (real32) :: ymax
   real (real32) :: ymax_data
   character (len=1) :: PlotDeadtimeCorr

   character (len=19) :: ChannelRangeLabel
   logical :: Single_DetType_Selected

   character (len=1) :: PlotLayout
   character (len=3) :: GraphFormat

   logical, parameter :: Determine_Channel_Range = .FALSE.
   character (len=40) :: string

   real (real32) :: Xaxis_range

   character (len=9) :: TimeRefType

   !  This is the TriggerTime, read from the FITS file, for trigger data.
   !  For continuous / non-trigger data, it is a user input value.
   real (real128) :: ReferenceTime


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   write ( *, '( / "TTE_Histo_MultiGraphs by MSB, Version 1.5, 2016 Jan 26." / )' )


   !  read command line input, obtain name of datafiles to read:

   call TTE_GetCmdLineInput_AllFileNames ( NamePortion, InputFileNames_for_C )



   !  more input from user:

   1010 continue
   write ( *, '( / "Select Plot Layout:")' )
   write ( *, '( """L"" = Large plots spread across several pages," )' )
   write ( *, '( """S"" = Small plots on one page." )' )
   write ( *, '( "Input ""L"" or ""S"" or ENTER for S: " )', advance='no' )
   read ( *, '(A)', err=1010 )  PlotLayout
   if ( len_trim ( PlotLayout ) == 0 )  PlotLayout = "S"
   call MakeUpperCase ( PlotLayout )
   if ( PlotLayout /= "L"  .and.  PlotLayout /= "S" ) goto 1010


   1020 continue
   write ( *, '( / "There are three modes for the units for inputting time selection and labeling the plot:" )' )
   write ( *, ' ( "     x-axis   bin widths" / "1:     s        ms" / "2:    ms        ms" / "3:    ms       microsec" )' )
   write ( *, '( "Select input/label mode: " )', advance='no' )
   read ( *, *, err=1020 )  TimeMode
   if ( TimeMode /= 1  .and.  TimeMode /= 2  .and.  TimeMode /= 3 ) goto 1020

   if ( TimeMode == 1 ) then
      TimeLabel = "s "
   else
      TimeLabel = "ms"
   end if
   if ( TimeMode /= 3 ) then
      BinLabel = "ms"
   else
      BinLabel = "us"
   end if


   1030 continue
   write ( *, '( / "Input Time Range to plot for file ", A, " in ", A, ": " )', advance='no' )  trim(NamePortion), TimeLabel
   read ( *, *, err=1030 )  GraphTimeRange_s (1), GraphTimeRange_s (2)

   !  If user input time range in ms, convert to seconds, the units of GraphTimeRange_s:
   if ( TimeLabel == "ms" )  GraphTimeRange_s = GraphTimeRange_s / 1000.0_real64


   write ( *, '( "The time range of the graph will be", F8.3, " to", F8.3, " s." )' )  GraphTimeRange_s (1), GraphTimeRange_s (2)
   write ( *, '( "or", F8.1, " to", F8.1, " ms." )' )    &
          GraphTimeRange_s (1) * 1000.0_real64, GraphTimeRange_s (2) * 1000.0_real64

   1040 continue
   write ( *, '( / "Input Histogram Bin Width in ", A, ": " )', advance='no' )  BinLabel
   read ( *, *, err=1040 )  BinWidth_s
   if ( BinLabel == "ms" )  then
      BinWidth_s = BinWidth_s / 1000.0_real64
   else
      BinWidth_s = BinWidth_s / 1.0E+6_real64
   end if

   write ( 6, '( / "BinWidth in s:", 1PG13.5 )' ) BinWidth_s


   !  Because this program always plots both detector types, i.e., both NaI and BGO data,
   !  so that Single_DetType_Selected = .false, the arrays EMin_of_chan and EMax_of_chan
   !  are not used by subroutine Select_Echannels_Widget.  Therefore we can load those
   !  arrays with fake values rather than being forced to wait to call Select_EChannels_Widget
   !  after the data of the first detector has been read:

   Single_DetType_Selected = .false.
   EMin_of_chan = 0.0
   EMax_of_chan = 0.0

   call Select_EChannels  ( Single_DetType_Selected, EMin_of_chan, EMax_of_chan, "Select Energy Channels:", ChannelRange )

   write ( ChannelRangeLabel, '( "Channels ", I3, " to ", I3 )' )  ChannelRange



   !  Read the TTE data from the Level 1 Fits file, then bin it into the shared arrays of
   !  the module TTE_Histogram_Arrays.

   call Read_and_Bin_TTE  ( InputFileNames_for_C, GraphTimeRange_s, BinWidth_s, RequestDetectors,      &
            Single_DetType_Selected, Determine_Channel_Range, 0.0_real32, .FALSE.,                     &
            ChannelRange, EMin_of_chan, EMax_of_chan, ReferenceTime, TimeRefType )



   !  Questions for the plots -- we only want to ask once, so ask here:

   Xaxis_range = real (  GraphTimeRange_s (2) - GraphTimeRange_s (1),  real32 )
   if ( TimeLabel == "ms" )  Xaxis_range = Xaxis_range * 1000.0_real32

   xdiv = auto_xdiv_value ( Xaxis_range )
   1060 continue
   write ( *, '( / "Input X-axis (time) major tick spacing in ", A, " or ENTER to keep default of ", 1P,G10.4, ": " )', &
         advance='no' )  TimeLabel, xdiv
   read ( *, '(A)', err=1060 )  string
   if ( len_trim (string) /= 0 )  read ( string, *, err=1060 )  xdiv


   1070 continue
   write ( *, '( // "Do you want to overplot with a deadtime corrected histogram?" / "Y, N or ENTER for N: " )', advance='no' )
   read ( *, '(A)', err=1070 )  PlotDeadtimeCorr
   if ( len_trim (PlotDeadtimeCorr) == 0 )  PlotDeadtimeCorr = 'N'
   call MakeUpperCase ( PlotDeadtimeCorr )


   ! max bin value of raw data, separately for NaI and BGO:
   MaxBinValueRaw (1) = maxval ( GBL_Data_Histogram_ByDet ( :, 0:11) )
   MaxBinValueRaw (2) = maxval ( GBL_Data_Histogram_ByDet ( :, 12:13) )

   !  Deadtime corrected max may not be in the same bin as the raw max
   !  because events in the overflow channel have a larger deadtime.
   if ( PlotDeadtimeCorr == 'Y' ) then
      MaxBinValueDeadTime (1) = maxval ( GBL_Deadtime_Histogram_ByDet ( :, 0:11) )
      MaxBinValueDeadTime (2) = maxval ( GBL_Deadtime_Histogram_ByDet ( :, 12:13) )
   end if


   write ( *, '( / "Maximum values observed in histogram bins are:" / "     Observed      Deadtime" )' )
   write ( *, '( "NaI: ", I6 )', advance='no' )  MaxBinValueRaw (1)
   ymax_data = MaxBinValueRaw (1)
   if ( PlotDeadtimeCorr == 'Y' ) then
      write ( *, '( 6X, F9.2 )' )  MaxBinValueDeadTime (1)
      ymax_data = MaxBinValueDeadTime (1)
   else
      write ( *, * )
   end if
   write ( *, '( "BGO: ", I6 )', advance='no' )  MaxBinValueRaw (2)
   ymax_data = max ( ymax_data, real (MaxBinValueRaw (2), real32) )
   if ( PlotDeadtimeCorr == 'Y' ) then
      write ( *, '( 6X, F9.2 )' )  MaxBinValueDeadTime (2)
      ymax_data = max ( ymax_data, MaxBinValueDeadTime (2) )
   else
      write ( *, * )
   end if

   1080 continue
   ymax = auto_ymax_value  ( ymax_data )
   write ( *, '( "Input the value to use for the top of the Y-axis or ENTER to keep", 1P, G13.5, ": " )', advance='no' )  ymax
   read ( *, '(A)' )  string
   if ( len_trim (string) /= 0 )  read ( string, *, err=1080 )  ymax

   ydiv = auto_ydiv_value  ( ymax )
   write ( *, '( "Input the Y-axis major (numbered) tick spacing or ENTER to keep", 1P, G13.5, ": " )', advance='no' )  ydiv
   read ( *, '(A)' )  string
   if ( len_trim (string) /= 0 )  read ( string, *, err=1080 )  ydiv


   call  Get_GraphFormat  ( GraphFormat )



   !  Plot a Histograms of the TTE events, one for each detector !!!
   !  Two arrays obtained from module TTE_Histogram_Arrays

   do idet=0, MAX_GBM_DET - 1

      call MakeMultiHistoPlot  ( idet, GBL_Data_Histogram_ByDet (:,idet), GBL_Deadtime_Histogram_ByDet(:,idet),         &
            GraphTimeRange_s, BinWidth_s, .TRUE., NamePortion, ChannelRangeLabel, TimeLabel, BinLabel,                  &
            xdiv, ydiv, ymax, PlotLayout, GraphFormat, PlotDeadtimeCorr )

   end do


   !  Cleanup
   !  Deallocate global arrays allocated in Fortran, the arrays of module TTE_Histogram_Arrays.
   !  (Not really necessary since this program is about to exit, but might be useful in
   !   other programs.)

   deallocate (GBL_Data_Histogram_SummedDets)
   deallocate (GBL_Deadtime_Histogram_SummedDets)
   deallocate (GBL_Data_Histogram_ByDet)
   deallocate (GBL_Deadtime_Histogram_ByDet)


   stop

end program TTE_Histo_MultiGraphs


! #################################################################################################


