!  Copyright Michael S. Briggs, University of Alabama in Huntsville.
!  For NASA's Fermi Gamma-ray Burst Monitor (GBM).
!  2010 -- 2016.


! #################################################################################################


program TTE_Output_Merged

   use MSB_Constants

   use, intrinsic :: iso_c_binding, only: C_CHAR, C_DOUBLE

   use TTE_subroutines, only : TTE_GetCmdLineInput_AllFileNames, MakeOutputFileName, Return_all_TTE,     &
          Select_Detectors, Output_CSV_TitleLines, MET_Corr_for_ClockDrift, GapTimeQuestion,             &
          Determine_ReferenceTime, Output_E_edges

   use TTE_constant_defs

   use Type_for_all_TTE_Events

   use Time_and_Coord_Routines, only: MET_to_UTC


   implicit none

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   character (len=FILE_NAME_MAX_LEN) :: NamePortion
   character (kind=C_CHAR, len=FILE_NAME_MAX_LEN), dimension(0: MAX_GBM_DET-1 ) :: InputFileNames_for_C

   real (real64) :: OutputTimeStart_s, OutputTimeEnd_s

   integer (int32) :: NumRequestDetectors
   integer (int32) :: SingleRequestDetector    ! only applicable if user requests a single detector
   logical :: Single_DetType_Selected
   character (len=DETECTOR_SELECTION_LABEL_LEN) :: DetectorSelectionLabel

   real (real64), parameter :: OutputTimeStart_s_DEFAULT = -0.1_real64
   real (real64), parameter :: OutputTimeEnd_s_DEFAULT = 0.05_real64


   type (All_TTE_Events_type), dimension (0: MAX_GBM_DET-1) :: TTE_Data

   integer :: out_LUN
   integer :: ldet
   integer :: output_counter = 0

   character (len=FILE_NAME_MAX_LEN) :: OutputFileName

   logical, dimension ( 0: MAX_GBM_DET-1 ) :: RequestDetectors

   integer :: time_idx_by_det ( 0: MAX_GBM_DET - 1 )

   integer :: EarliestDet
   real (C_DOUBLE) :: EarliestTime

   real (kind=real32) :: GapTime
   logical :: ClockDrift_Flag

   character (len=1) :: Reply

   real (kind=real128) :: Raw_MET, Corr_MET

   real (kind=real128) :: JulianDay

   integer (kind=int32) :: year
   integer (kind=int32) :: month
   integer (kind=int32) :: day
   integer (kind=int32) :: hours
   integer (kind=int32) :: minutes
   real (kind=real64) :: seconds

   real (C_DOUBLE) :: TriggerTime
   real (kind=real128) :: ReferenceTime
   real (kind=real128) :: CorrReferenceTime
   real (kind=real128) :: RelativeTime

   character (len=9) :: TimeRefType

   character (len=100), parameter :: ProgVersion = "TTE_Output_Merged by MSB, Version 1.4, 2016 Jan 27"


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   write ( *, '( / A / )' )  trim (ProgVersion)



   !  read command line input, obtain name of datafiles to read:

   call TTE_GetCmdLineInput_AllFileNames ( NamePortion, InputFileNames_for_C )


   !  Determine which detectors should have their data read and merged into the output file:

   call Select_Detectors    &
      ( RequestDetectors, NumRequestDetectors, SingleRequestDetector, Single_DetType_Selected, DetectorSelectionLabel )



   !  Call a subroutine to read and return the TTE for all of the requested detectors.
   !  An alternative design, which was originally used, is to read the data directly in
   !  this routine -- this allows reading and plotting the data within the same ldet loop --
   !  in this implementation one need only store the data for one detector at a time, rather
   !  than having to store the data for all of the detectors, as is done in the current
   !  implementation.  The advantage of the current implementation is that the TTE-reading
   !  code has been factored out into a subroutine.  We are not concerned about the increased
   !  memory consumption.

   call Return_all_TTE  ( InputFileNames_for_C, RequestDetectors, TTE_Data, TriggerTime, TimeRefType )



   call Determine_ReferenceTime  ( RequestDetectors, TriggerTime, TTE_Data, ReferenceTime )


   do ldet=0, MAX_GBM_DET - 1
      if ( RequestDetectors (ldet) ) then
         !  output the energy edges of the channels, each requested detector, separate files:
         call Output_E_edges ( ldet, TTE_Data )
      end if
   end do



   OutputTimeStart_s = OutputTimeStart_s_DEFAULT
   OutputTimeEnd_s = OutputTimeEnd_s_DEFAULT

   write (*, '( / " This program will output data from", F7.2, " to", F7.2, " s relative to the ", A, " Time of", F17.6 )' )  &
           OutputTimeStart_s, OutputTimeEnd_s, trim (TimeRefType), ReferenceTime

   write (*, '( "Change this range?  Y/N: " )', advance="no" )
   read (*, '(A)' )  Reply
   if ( Reply == "Y"  .OR.  Reply == "y" ) then
      write (*, '( "Input the time range to output: " )', advance="no" )
      read (*, *)  OutputTimeStart_s, OutputTimeEnd_s
   end if


   call GapTimeQuestion  ( ClockDrift_Flag, GapTime )


  !  Create output file name and open the output file:

   OutputFileName = trim (NamePortion) // "_Merged.csv"
   open ( file=OutputFileName, newunit=out_LUN, status="replace", form="formatted", access="sequential", action="write" )


   !  Advance indices, by detector, into the TTE data, until they are within the requested time range:

   time_idx_by_det = 1

   do ldet=0, MAX_GBM_DET - 1

      if ( RequestDetectors (ldet) ) then

         TooEarly: do

            if ( TTE_Data (ldet) % tte_times ( time_idx_by_det (ldet) )   >=  ReferenceTime + OutputTimeStart_s )  exit TooEarly

            time_idx_by_det (ldet) = time_idx_by_det (ldet) + 1

         end do TooEarly

      end if

   end do    ! ldet


   WhileEventsAvail: do

      !  Find which detector has the earliest event currently pointed by the
      !  index array time_idx_by_det:

      EarliestDet = 0
      EarliestTime = 9.9E+20

      do ldet=0, MAX_GBM_DET - 1

         if ( RequestDetectors (ldet) ) then

            if ( time_idx_by_det (ldet) > 0 ) then     ! data avail?

               if ( TTE_Data (ldet) % tte_times ( time_idx_by_det (ldet) )  <  EarliestTime ) then

                  EarliestDet = ldet
                  EarliestTime = TTE_Data (ldet) % tte_times ( time_idx_by_det (ldet) )

               end if

            end if     ! data avail?

         end if

      end do    ! ldet


      !  We have found the earliest event still available from any detector --
      !  is that event past the requested time interval?   If so, we are done
      !  outputting and should exit the loop.

      if ( EarliestTime  >  ReferenceTime + OutputTimeEnd_s )  exit WhileEventsAvail


      !  OK, 1) we have the earliest currently available TTE event, and
      !  2) it is within the requested time window, so:
      !  I) output it, and
      !  II) advance the time bin index for that detector since we have used the event:

      output_counter = output_counter + 1

      Raw_MET = EarliestTime


      if ( ClockDrift_Flag)  then
         Corr_MET = MET_Corr_for_ClockDrift  ( Raw_MET, GapTime )
      else
         Corr_MET = Raw_MET
      end if


      call MET_to_UTC ( Corr_MET, JulianDay, year, month, day, hours, minutes, seconds )


      if ( output_counter == 1 ) then

         if ( ClockDrift_Flag) then
            CorrReferenceTime = MET_Corr_for_ClockDrift  ( ReferenceTime, GapTime )
         else
            CorrReferenceTime = ReferenceTime
         end if

         call Output_CSV_TitleLines ( out_LUN, ProgVersion, TimeRefType, ReferenceTime, CorrReferenceTime,    &
                 ClockDrift_Flag, GapTime, NumRequestDetectors, DetectorSelectionLabel, .TRUE. )

      end if


      RelativeTime =  Corr_MET - CorrReferenceTime


      write ( out_LUN, '( F17.6, ", ", F17.6, ", ", F10.6, ", ", F10.6, ", ", I3, ", ", I4 )', round="nearest" )     &
              Raw_MET, Corr_MET, seconds, RelativeTime,                 &
                      EarliestDet, TTE_Data (EarliestDet) % tte_channels ( time_idx_by_det (EarliestDet) )


      !  advance time bin index for this detector only if more data is available:

      if ( time_idx_by_det (EarliestDet) < TTE_Data(EarliestDet) % Num_TTE ) then
         time_idx_by_det (EarliestDet) = time_idx_by_det (EarliestDet) + 1
      else
         time_idx_by_det (EarliestDet) = -1   ! mark that no more data is available
      end if


   end do WhileEventsAvail



   write ( *, '( I8, " TTE events output to file ''", A, "'' ." )' )  output_counter,    &
                     OutputFileName ( 1: len_trim (OutputFileName) )


   !  Cleanup -- deallocate the Fortran TTE arrays:

   do ldet=0, MAX_GBM_DET - 1

      if ( RequestDetectors (ldet) ) then

         deallocate  ( TTE_Data (ldet) % tte_channels )
         deallocate  ( TTE_Data (ldet) % tte_times )

      end if

   end do    ! ldet


   stop


end program TTE_Output_Merged
