!  Copyright Michael S. Briggs, University of Alabama in Huntsville.
!  For NASA's Fermi Gamma-ray Burst Monitor (GBM).
!  2010 -- 2016.

! #################################################################################################


program TTE_Output_CSV

   use MSB_Constants

   use, intrinsic :: iso_c_binding, only: C_CHAR, C_DOUBLE

   use TTE_subroutines, only : TTE_GetCmdLineInput_AllFileNames, MakeOutputFileName, Return_all_TTE,    &
              Output_CSV_TitleLines, MET_Corr_for_ClockDrift, GapTimeQuestion, Determine_ReferenceTime, &
              Output_E_edges

   use TTE_constant_defs

   use Type_for_all_TTE_Events

   use Time_and_Coord_Routines, only: MET_to_UTC


   implicit none

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   character (len=FILE_NAME_MAX_LEN) :: NamePortion
   character (kind=C_CHAR, len=FILE_NAME_MAX_LEN), dimension(0: MAX_GBM_DET-1 ) :: InputFileNames_for_C

   real (real64) :: OutputTimeStart_s, OutputTimeEnd_s

   real (real64), parameter :: OutputTimeStart_s_DEFAULT = -0.1_real64
   real (real64), parameter :: OutputTimeEnd_s_DEFAULT = 0.05_real64


   type (All_TTE_Events_type), dimension (0: MAX_GBM_DET-1) :: TTE_Data

   integer :: out_LUN
   integer :: ldet
   integer :: output_counter

   character (len=FILE_NAME_MAX_LEN) :: OutputFileName

   character (len=DETECTOR_SELECTION_LABEL_LEN) :: DetectorSelectionLabel

   integer (int64) :: j_tte
   integer :: i
   logical, dimension ( 0: MAX_GBM_DET-1 ), parameter :: RequestDetectors =  [  ( .TRUE., i=1, MAX_GBM_DET )  ]


   real (real32) :: GapTime
   logical :: ClockDrift_Flag

   character (len=1) :: Reply

   real (real128) :: Raw_MET, Corr_MET

   real (real128) :: JulianDay

   integer (int32) :: year
   integer (int32) :: month
   integer (int32) :: day
   integer (int32) :: hours
   integer (int32) :: minutes
   real (real64) :: seconds

   real (C_DOUBLE) :: TriggerTime
   real (real128) :: ReferenceTime
   real (real128) :: CorrReferenceTime
   real (real128) :: RelativeTime

   integer (int32), parameter :: NumRequestDetectors = 1
   character (len=9) :: TimeRefType

   character (len=100), parameter :: ProgVersion = "TTE_Output_CSV by MSB, Version 1.4, 2016 Jan 27"


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   write ( *, '( / A / )' )  trim (ProgVersion)



   !  read command line input, obtain name of datafiles to read:

   call TTE_GetCmdLineInput_AllFileNames ( NamePortion, InputFileNames_for_C )



   !  Call a subroutine to read and return the TTE for all of the requested detectors.
   !  An alternative design, which was originally used, is to read the data directly in
   !  this routine -- this allows reading and plotting the data within the same ldet loop --
   !  in this implementation one need only store the data for one detector at a time, rather
   !  than having to store the data for all of the detectors, as is done in the current
   !  implementation.  The advantage of the current implementation is that the TTE-reading
   !  code has been factored out into a subroutine.  We are not concerned about the increased
   !  memory consumption.

   call Return_all_TTE  ( InputFileNames_for_C, RequestDetectors, TTE_Data, TriggerTime, TimeRefType )



   call Determine_ReferenceTime  ( RequestDetectors, TriggerTime, TTE_Data, ReferenceTime )



   OutputTimeStart_s = OutputTimeStart_s_DEFAULT
   OutputTimeEnd_s = OutputTimeEnd_s_DEFAULT


   write (*, '( / " This program will output data from", F7.2, " to", F7.2, " s relative to the ", A, " Time of", F17.6 )' )  &
           OutputTimeStart_s, OutputTimeEnd_s, trim (TimeRefType), ReferenceTime

   write (*, '( "Change this range?  Y/N: " )', advance="no" )
   read (*, '(A)' )  Reply
   if ( Reply == "Y"  .OR.  Reply == "y" ) then
      write (*, '( "Input the time range to output: " )', advance="no" )
      read (*, *)  OutputTimeStart_s, OutputTimeEnd_s
   end if


   call GapTimeQuestion  ( ClockDrift_Flag, GapTime )



   do ldet=0, MAX_GBM_DET - 1

      !  output the energy edges of the channels, each detector, separate files:
      call Output_E_edges ( ldet, TTE_Data )

      !  Create (primary) output file name and open the (primary) output file:

      call MakeOutputFileName ( InputFileNames_for_C (ldet), "csv", OutputFileName )
      open ( file=OutputFileName, newunit=out_LUN, status="replace", form="formatted", access="sequential", action="write" )


      !  Write TTE events from the requested time range to the (primary) output file:

      output_counter = 0

      do j_tte=1, TTE_Data (ldet) % Num_TTE

         if ( TTE_Data (ldet) % tte_times (j_tte)  >  ReferenceTime + OutputTimeStart_s  .AND.   &
              TTE_Data (ldet) % tte_times (j_tte)  <  ReferenceTime + OutputTimeEnd_s ) then

            output_counter = output_counter + 1

            Raw_MET = TTE_Data (ldet) % tte_times (j_tte)


            if ( ClockDrift_Flag)  then
               Corr_MET = MET_Corr_for_ClockDrift  ( Raw_MET, GapTime )
            else
               Corr_MET = Raw_MET
            end if


            call MET_to_UTC ( Corr_MET, JulianDay, year, month, day, hours, minutes, seconds )


            if ( output_counter == 1 ) then

               if ( ClockDrift_Flag) then
                  CorrReferenceTime = MET_Corr_for_ClockDrift  ( ReferenceTime, GapTime )
               else
                  CorrReferenceTime = ReferenceTime
               end if

               if ( ldet <= 11 ) then
                  write (DetectorSelectionLabel, '( "GBM NaI ", I2 )' )  ldet
               else
                 write (DetectorSelectionLabel, '( "GBM BGO ", I1 )' )  ldet - 12
               end if
               call Output_CSV_TitleLines ( out_LUN, ProgVersion, TimeRefType, ReferenceTime, CorrReferenceTime,   &
                        ClockDrift_Flag, GapTime, NumRequestDetectors, DetectorSelectionLabel, .FALSE. )

            end if


            RelativeTime = Corr_MET - CorrReferenceTime

            write (out_LUN, '( F17.6, ", ", F17.6, ", ", F10.6, ", ", F10.6, ", ", I4 )', round="nearest" )     &
                               Raw_MET, Corr_MET, seconds, RelativeTime, TTE_Data (ldet) % tte_channels (j_tte)

         end if

      end do   ! j_tte


      write ( *, '( I8, " TTE events output to file ''", A, "'' ." )' )  output_counter,    &
                     OutputFileName ( 1: len_trim (OutputFileName) )


      !  Cleanup -- deallocate the Fortran TTE arrays:

      deallocate  ( TTE_Data (ldet) % tte_channels )
      deallocate  ( TTE_Data (ldet) % tte_times )


   end do    ! ldet



end program TTE_Output_CSV
