!  Copyright Michael S. Briggs, University of Alabama in Huntsville.
!  For NASA's Fermi Gamma-ray Burst Monitor (GBM).
!  copyright 2010  -- 2016.


!   Modified to analyze the GBM clock drift.   It plots a scatter plot against
!   absolute time, modulo one second.   "absolute" means not relative to trigger
!   time.   If the clock is slow, there will be a gap without events before zero.
!   If the clock is fast, there will be an interval past zero with an excess of
!   events.    In practice, the GBM clock is alway slow and there is a gap of
!   time without events.
!   MSB, 2010 Jan 23 & 24.

!  Michael S. Briggs, UAHuntsville & NSSTC, copyright 2009 February 21 -- 22.
!  First release: version 0.9
!  Revised: 0.9.1, 2009 February 28 -- don't ask symbol question for one detector.
!  Version 0.9.5, 2009 March 1: upgraded subroutine Read_Lvl1_TTE to read energy edges,
!     added subroutine Energy_Scale_RightAxis to plot energy scale on the right Y-axis
!     if only one detector type is being plotted.
!  Revised: 0.9.6, 2009 March 2: smaller symbols
!  Revised: 0.9.7, 2009 March 3: colors for symbols
!  Revision 0.99 by MSB, 2009 March 22: call widget routine for energy channels,
!     include energy range in channel label when possible.
!  Revision 0.99.1 by MSB, 2009 April 15-- gfortran version 4.4.0 didn't like an
!     array initialization statement with varying string lengths.

!  Revision 1.0 by MSB, 2009 May 10:
!    a) rearranged to place subroutines in a module at the top of this file -- this way
!    the routines automatically have explicit interfaces, rather than having manually
!    written interfaces.
!    b) the "bugs" that the right-axis tick-marks had to be inside the graph is now understood.
!    It was also noticed that some outside labels were missing on PNG but not on PDF graphs.
!    It turns out that the clpwin call effects these features, so they must be drawn before
!    the clpwin call -- therefore order rearranged.

!  Revision 1.2 by MSB, 2009 July 26:
!    a) add eps (encapsulated Postscript) output,
!    b) add option to suppress titles above the graph (useful for figures to be included
!       in a paper).
!  Revision 1.3 by MSB, 2009 August 4 & 15:
!    a) larger fonts and thicker lines, for better appearence of plots at publication size,
!    b) add option for user to specify name of output file.


!  Program TTE_ScatterPlot:

!  Usage: TTE_ScatterPlot.exe  bnyymmhhfff
!  OR
!  Usage: TTE_ScatterPlot.exe  bnyymmhhfff   VerNum
!  "yymmhhfff" is the trigger number, VerNum is the file version number, which may be omitted if 00.



! #################################################################################################

module Closeup_Summmed_TTE_mod

   use MSB_Constants

   implicit none

   integer (int32), parameter :: Summed_TTE_len = 40

   integer (int32), dimension ( -Summed_TTE_len: Summed_TTE_len ) :: Closeup_Summed_TTE

end module Closeup_Summmed_TTE_mod

! #################################################################################################


module TTE_ScatterPlot_Routines

    use, intrinsic :: iso_c_binding, only: C_FLOAT, C_DOUBLE, C_SHORT

    use TTE_Subroutines, only: Get_OutputFileName, Get_GraphFormat

    use MSB_Constants

    use dislin

    implicit none

contains

! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


   subroutine Setup_TTE_ScatterPlot  ( GraphTimeRange_s, ChannelRange, Single_DetType_Selected, NamePortion,     &
                 TimeLabel, DetectorSelectionLabel, ChannelRangeLabel )


      use TTE_constant_defs
      use TTE_subroutines, only : MakeUpperCase, auto_xdiv_value, Get_GraphFormat, Get_OutputFileName

      implicit none


      ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

      real (real64), dimension (2), intent (in) :: GraphTimeRange_s
      integer (int32), dimension (2), intent (in) :: ChannelRange
      logical, intent (in) :: Single_DetType_Selected
      character (len=*), intent (in) :: NamePortion
      character (len=*), intent (in) :: DetectorSelectionLabel
      character (len=*), intent (in) :: ChannelRangeLabel
      character (len=2), intent (in) :: TimeLabel


      ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

      ! internal variables:

      character (len=FILE_NAME_MAX_LEN) :: OutputFileName

      character (len=8), parameter :: Fragment = ".GapPlot"
      character (len=3) :: GraphFormat
      character (len=1) :: response

      character (len=40) :: X_AxisLabel, Y_AxisLabel

      real (real32), dimension (2) :: TimeRange_for_Call
      real (real32) :: TimeDiff


      real (real32) :: xdiv
      real (real32) :: ymin, ymax, ydiv, y_label_one

      integer :: y_int
      integer :: PltCoord_Left, PltCoord_Right, PltCoord_Bot, PltCoord_Top
      integer :: Plt_Width, Plt_Height

      character (len=40) :: input
      real (real32) :: try


      ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


      call  Get_GraphFormat  ( GraphFormat )


      !  Create output file name:
      call  Get_OutputFileName  ( NamePortion, Fragment, GraphFormat, OutputFileName )


      !  configure and initiate the plot:

      call setfil ( trim (OutputFileName) )
      call metafl ( GraphFormat )

      ! Normally PNG plots white lines and characters on black background --
      !  reverse to black on white:
      ! Also override default pixel dimensions to have same aspect ratio as logical pixel dimensions
      if ( GraphFormat == 'png' ) then
         call scrmod ( "REVERS" )
         call winsiz ( 853, 660 )  ! physical pixel dimensions, override default of 603 x 853
      end if


      call setpag ( 'USAL' )      ! US A (=Letter), Landscape, 2790 x 2160 points

      call filmod ( 'DELETE' )    ! overwrite (delete) previous file

      call disini ()

      if ( GraphFormat == 'pdf' .or. GraphFormat == 'eps' )  call psfont ( 'Helvetica' )


      !  Make all lines thicker, including symbols; default is one; doesn't apply to characters.
      call linwid ( 5 )


      !  axes and axes labels:

      if ( TimeLabel == "s " ) then
         TimeRange_for_Call = GraphTimeRange_s
      else
         !  x-axis in ms instead of s case:
         TimeRange_for_Call = GraphTimeRange_s * 1000.0_real64
      end if
      TimeDiff = TimeRange_for_Call (2) - TimeRange_for_Call (1)


      !  Determine tick-spacing of X-axis, either automatically (with user approval), or from user input:

      try = auto_xdiv_value ( TimeDiff )

      9020 continue
      write ( *, 9030, advance='no' )  TimeLabel, try, TimeLabel
      9030 format ( / "Input X (time) axis major (numbered) tick spacing in ", A, " or ENTER to use", F8.2, 1X, A, " : " )
      read ( *, '(A)', err=9020 ) input

      if ( len_trim (input) == 0 ) then
         xdiv = try
      else
         read ( input, *, err=9020 )  xdiv
      end if


      call labdig ( -2, 'X' )  ! DISLIN (>=9.4?) automatically calculates number of digits after decimal point on X-axis labels
      call labdig ( 0, 'Y' )   ! zero digits after decimal point on Y-axis labels

      Y_AxisLabel = "Energy Channel"
      write ( X_AxisLabel, '( "Time, modulo 1 second (", A, ")" )' )  trim (TimeLabel)
      call name ( X_AxisLabel,  'X' )
      call name ( Y_AxisLabel, 'Y' )


      !  increase size of axis titles (aka names) from default of 36,
      !  and from previous (long used) value of 46:
      call hname ( 55 )

      ! increase the size of characters from default of 36:
      call height ( 55 )


      !  increase length of major and minor ticks from defaults of 24 & 16:
      call ticlen ( 33, 22 )


      !  Draw graph box / axes:
      !  The x-axis is simple -- the time range.
      !  The basic y-axis is the channel range.
      !  If only one detector type (NaI or BGO) will be plotted, we also plot
      !  an energy scale on the right y-axis.


      ymin = ChannelRange (1)
      ymax = ChannelRange (2)

      !  special case: all channels plotted: expand Y-axis range to have no clipping of symbols:
      if ( ChannelRange (1) == 0  .and.  ChannelRange (2) == 127 ) then
         ymin = -2
         ymax = 129
      end if

      !   first y-axis label needs to be a multiple of 10, first value above ymin:
      y_int = 10 * (ChannelRange (1) / 10)
      if ( y_int  <  ChannelRange (1) - 0.1 )  y_int = y_int + 10
      y_label_one = y_int

      ydiv = 10.0
      if ( ChannelRange (2) - ChannelRange (1)  >  90 )  ydiv = 20


      !  If only one detector type is selected, draw just a line rather than a line with tick marks, for the
      !  right y-axis.  The tick marks are suppressed because we will draw an energy scale instead of a channel scale.
      !  So for the fourth argument, the default of "TICKS" is replaced with "LINE".

      if ( Single_DetType_Selected )  call setgrf ( "NAME", "NAME", "TICKS", "LINE" )

      call graf ( TimeRange_for_Call (1), TimeRange_for_Call (2), TimeRange_for_Call (1), xdiv, ymin, ymax, y_label_one, ydiv )




      !  special case: all channels plotted: plot dotted lines at actual limits
      if ( ChannelRange (1) == 0  .and.  ChannelRange (2) == 127 ) then
         call dot ()
         call rline ( TimeRange_for_Call (1),   0.0, TimeRange_for_Call (2),   0.0 )
         call rline ( TimeRange_for_Call (1), 127.0, TimeRange_for_Call (2), 127.0 )
         call solid ()  ! restore default
      end if


      !  add the titles:

      9040 continue
      write ( *, '( / "Output Titles above the Graph?  ""Y"" for Yes, or ENTER for Yes, or ""N"" for No: " )', advance='no' )
      read ( *, '(A)', err=9040 )  response
      if ( len_trim (response) == 0 )  response = 'Y'
      call MakeUpperCase (response)
      if ( response /= 'N' .and. response /= 'Y' ) goto 9040
      if ( response == 'Y' ) then

         call titlin ( NamePortion, 1 )
         call titlin ( DetectorSelectionLabel, 2 )
         call titlin ( ChannelRangeLabel, 3 )
         call title ()

      end if



      !  *** The various labels must be plotted before the call to 0 because the clpwin
      !  *** call applies to the labels.   (Strangely, when clpwin prevents labels outside
      !  *** of its range for PNG plots, but not for PDF.)


      !  Extend the clipping region past the top and bottom of the graph, but not past the sides.
      !  This way symbols aren't truncated when the extend past the bottom or top edges of the
      !  graph -- the truncation looks bad.

      !  First we have to figure out the plot coordinates of the graph box.
      !  To do this we use DISLIN routines to calculate plot coordinates from user coordinates,
      !  using the limits of the graph box in user coordinates.

      PltCoord_Left = NXPOSN ( TimeRange_for_Call (1) )
      PltCoord_Right = NXPOSN ( TimeRange_for_Call (2) )
      PltCoord_Bot = NYPOSN ( real (ChannelRange (1)) )
      PltCoord_Top = NYPOSN ( real (ChannelRange (2)) )

      Plt_Width = PltCoord_Right - PltCoord_Left
      Plt_Height = PltCoord_Bot - PltCoord_Top

      !  First two inputs are coordinates of upper-left of graph box -- we shift
      !  clip region upwards w.r.t. actual graph box by 1/20 of height of box.
      !  Second two inputs are width and height of clip box -- we increase height
      !  by 1/10 of the height.   This gives an extra margin of 1/20 of height at
      !  past the top and the bottom.

      call clpwin ( PltCoord_Left, PltCoord_Top - Plt_Height / 20,    &
           Plt_Width, Plt_Height + Plt_Height / 10 )
      call noclip ()   ! required to disable clipping at axis limits so that clip window controls


      return


   end subroutine Setup_TTE_ScatterPlot


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


   subroutine Plot_TTE_ScatterPlot  ( GraphTimeRange_s, ChannelRange, tte_channels, tte_times, TimeLabel )

      use Closeup_Summmed_TTE_mod

      implicit none

      ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

      real (real64), dimension (2), intent (in) :: GraphTimeRange_s
      integer (int32), dimension (2), intent (in) :: ChannelRange
      integer (C_SHORT), dimension (:), pointer, intent (in) :: tte_channels
      real (C_DOUBLE), dimension (:), pointer, intent (in) :: tte_times
      character (len=2), intent (in) :: TimeLabel


      ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

      ! internal variables:

      integer (int32) :: Num_TTE
      integer (int32) :: i_tte

      real (C_DOUBLE) :: time
      integer :: PltSymbol

      integer (int32) :: nbin


      ! local, persistent variables:

      integer, save :: CallCount = 0


      ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


      Num_TTE = size (tte_times)


      !  Decrease symbol size from default of 35 (units: plot coord.)
      call hsymbl ( 30 )


      !  Determine symbol to use
      !  Count detectors on plot by counting calls to this subroutine.
      !  Symbol also depends on whether user has selected Different symbols
      !  for different detectors, or Same symbol for all detectors.

      CallCount = CallCount + 1

      PltSymbol = 15   !  open circle -- used for "same symbol" for all dets




      !  Plot symbol for any TTE event within the requested channel ChannelRange and
      !  time range GraphTimeRange_s (modulo!):
      !  (We could use DISLIN subroutine "CURVE", but instead we select individual
      !  events that are within requested ranges .....)

      ! Note: The calculations & comparison are done in double for higher accuracy.
      ! Time is passed to rlsymb as a "single" because we are using the single-precision version of dislin.

      do i_tte=1, Num_TTE

         time = mod ( tte_times (i_tte), 1.0_C_DOUBLE )
         if ( time > 0.5_C_DOUBLE )  time = time - 1.0_C_DOUBLE
         if ( time >= GraphTimeRange_s (1)  .AND.  time <= GraphTimeRange_s (2) ) then

              if ( tte_channels (i_tte) >= ChannelRange (1)  .AND.    &
                   tte_channels (i_tte) <= ChannelRange (2) ) then

                 nbin = nint ( time / 2.0E-6_C_DOUBLE )
                 if ( abs (nbin) <= Summed_TTE_len)  Closeup_Summed_TTE (nbin) = Closeup_Summed_TTE (nbin) + 1

                 if ( TimeLabel == "ms" )  time = time * 1000.0_C_DOUBLE     ! if x-axis is in ms instead of s

                 call rlsymb ( PltSymbol, real (time, real32), real (tte_channels (i_tte), real32) )

              end if    !  channel range

         end if    !  time rang

      end do    ! loop over tte events



      return


   end subroutine Plot_TTE_ScatterPlot


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


end module TTE_ScatterPlot_Routines


! #################################################################################################


program TTE_ScatterPlot_GAP

   use MSB_Constants

   use, intrinsic :: iso_c_binding, only: C_CHAR

   use TTE_subroutines, only:  TTE_GetCmdLineInput_AllFileNames,  MakeUpperCase,  MakeLowerCase,   &
          MakeEnergyLabels, Return_all_TTE

   use TTE_constant_defs
   use Type_for_all_TTE_Events

   use TTE_ScatterPlot_Routines

   use Closeup_Summmed_TTE_mod


   implicit none

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   character (len=FILE_NAME_MAX_LEN) :: NamePortion
   character (kind=C_CHAR, len=FILE_NAME_MAX_LEN), dimension(0: MAX_GBM_DET-1 ) :: InputFileNames_for_C


   !  -0.05 to +0.03 ms in units of seconds:
   real (real64), dimension (2) :: GraphTimeRange_s = [ -0.04, +0.02 ] / 1000.0_real64

   integer (int32), dimension (2), parameter :: ChannelRange = [ 0, 127 ]


   type (All_TTE_Events_type), dimension (0: MAX_GBM_DET-1) :: TTE_Data


   integer (int32) :: idet

   character (len=2), parameter :: TimeLabel = "ms"

   integer :: i
   logical, dimension ( 0: MAX_GBM_DET-1 ), parameter :: RequestDetectors =   [  ( .TRUE., i=1, MAX_GBM_DET )  ]

   logical, parameter :: Single_DetType_Selected = .FALSE.
   character (len=DETECTOR_SELECTION_LABEL_LEN), parameter :: DetectorSelectionLabel = "All 14 GBM Detectors"

   character (len=42) :: ChannelRangeLabel


   logical :: FirstDetRead = .true.

   integer (int64) :: Tot_Num_TTE = 0_int64

   integer (int32) :: nbin
   integer (int32) :: ibin, start_bin, end_bin

   real (real64) :: MeanCounts_per_bin

   character (len=9) :: TimeRefType
   real (C_DOUBLE) :: TriggerTime

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   write ( *, '( / "TTE_ScatterPlot_GAP by MSB, 2010 Jan 23 & 24." / )' )


   Closeup_Summed_TTE = 0_int32



   !  read command line input, obtain name of datafiles to read:

   call TTE_GetCmdLineInput_AllFileNames ( NamePortion, InputFileNames_for_C )



   !  time units and time range:

   write (*, '( / "The time range of the graph will be", F8.2, " to", F8.2, " ms about the second boundaries." )' )    &
          GraphTimeRange_s (1) * 1000.0_real64, GraphTimeRange_s (2) * 1000.0_real64
   write (*, '( "This is done by plotting the data modulo one second." / )' )



   !  Call a subroutine to read and return the TTE for all of the requested detectors.
   !  An alternative design, which was originally used, is to read the data directly in
   !  this routine -- this allows reading and plotting the data within the same idet loop --
   !  in this implementation one need only store the data for one detector at a time, rather
   !  than having to store the data for all of the detectors, as is done in the current
   !  implementation.  The advantage of the current implementation is that the TTE-reading
   !  code has been factored out into a subroutine.  We are not concerned about the increased
   !  memory consumption.

   call Return_all_TTE  ( InputFileNames_for_C, RequestDetectors, TTE_Data, TriggerTime, TimeRefType )



   !  Loop through detectors, plotting the data of each detector:


   do idet=0, MAX_GBM_DET - 1


      Tot_Num_TTE = Tot_Num_TTE  +  TTE_Data (idet) % Num_TTE


      !  Some user input based on the data -- we use the data of the first requested detector:
      !  we need to know the energies of the channels, i.e., values for EMin_of_chan and EMax_of_chan,
      !  to provide that guidance to the user in making their channel selections.
      !  Only need the user input and to perform plot setup once.
      !  In this program we use the energy edges of the channels of the FIRST detector read.

      if ( FirstDetRead ) then

         FirstDetRead = .false.

         write ( ChannelRangeLabel, '( "Channels ", I3, " to ", I3 )' )  ChannelRange


         !  Setup / configure the plot:

         call Setup_TTE_ScatterPlot  ( GraphTimeRange_s, ChannelRange, Single_DetType_Selected, NamePortion,     &
              TimeLabel, DetectorSelectionLabel, ChannelRangeLabel )


      end if   ! FirstDetRead ?


      !  Plot the events of detector idet:

      call Plot_TTE_ScatterPlot  ( GraphTimeRange_s, ChannelRange,                           &
                    TTE_Data (idet) % tte_channels, TTE_Data (idet) % tte_times, TimeLabel )


      !  Cleanup -- deallocate the Fortran TTE arrays:

      deallocate  ( TTE_Data(idet)%tte_channels )
      deallocate  ( TTE_Data(idet)%tte_times )


   end do    ! detector loop


   MeanCounts_per_bin = real (Tot_Num_TTE, real64) / 500000.0_real64
   write (*, '( // "Total number of TTE events read from all detectors: ", I0 )' )  Tot_Num_TTE
   write (*, '( "Mean number per 2 microsecond bin:", F9.2 )' )  MeanCounts_per_bin


   nbin = -1
   if ( Closeup_Summed_TTE (nbin) /= 0 )  write (*, '( / "WARNING: There appears to be no gap!" )' )


   GapLenFind: do

      if (nbin < -Summed_TTE_len) then
         write (*, '( / "ERROR: The entire left side of the array lacks counts!" )' )
         nbin = nbin + 1   ! to have legal value
         exit GapLenFind
      end if

      if ( Closeup_Summed_TTE (nbin) /= 0 ) then
         nbin = nbin + 1   ! adjust to last zero value
         exit GapLenFind
      end if

      nbin = nbin - 1

   end do GapLenFind


   write (*, '( / "The last zero bin is: ", I0 )' )  nbin
   write (*, '( "The length of the gap is", I3, " microseconds." )' )  2 * (-nbin)

   if (nbin /= -Summed_TTE_len ) then
      if ( Closeup_Summed_TTE (nbin - 1)  <  MeanCounts_per_bin / 2.0_real64 ) then
         write (*, '( / "The bin immediately preceding the gap has only", I3, " counts, so")' )  Closeup_Summed_TTE (nbin - 1)
         write (*, '( "the gap should be considered to be", I3, " microseconds long.")' )  2 * (-nbin + 1)
      end if
   end if


   start_bin = max ( -Summed_TTE_len, nbin - 12 )
   end_bin = min ( 7, Summed_TTE_len )

   write (*, *)
   do ibin= start_bin, end_bin
      write (*, '( I4, 2X, I4 )' )  ibin, Closeup_Summed_TTE (ibin)
   end do


   !  Close Plot:

   call disfin ()


   !  Cleanup -- nothing more to do


   stop


end program TTE_ScatterPlot_GAP


! #################################################################################################
