!  Copyright Michael S. Briggs, University of Alabama in Huntsville.
!  For NASA's Fermi Gamma-ray Burst Monitor (GBM).
!  copyright 2009 February 21 -- 2016.

!  First release: version 0.9
!  Revised: 0.9.1, 2009 February 28 -- don't ask symbol question for one detector.
!  Version 0.9.5, 2009 March 1: upgraded subroutine Read_Lvl1_TTE to read energy edges,
!     added subroutine Energy_Scale_RightAxis to plot energy scale on the right Y-axis
!     if only one detector type is being plotted.
!  Revised: 0.9.6, 2009 March 2: smaller symbols
!  Revised: 0.9.7, 2009 March 3: colors for symbols
!  Revision 0.99 by MSB, 2009 March 22: call widget routine for energy channels,
!     include energy range in channel label when possible.
!  Revision 0.99.1 by MSB, 2009 April 15-- gfortran version 4.4.0 didn't like an
!     array initialization statement with varying string lengths.

!  Revision 1.0 by MSB, 2009 May 10:
!    a) rearranged to place subroutines in a module at the top of this file -- this way
!    the routines automatically have explicit interfaces, rather than having manually
!    written interfaces.
!    b) the "bugs" that the right-axis tick-marks had to be inside the graph is now understood.
!    It was also noticed that some outside labels were missing on PNG but not on PDF graphs.
!    It turns out that the clpwin call effects these features, so they must be drawn before
!    the clpwin call -- therefore order rearranged.

!  Revision 1.2 by MSB, 2009 July 26:
!    a) add eps (encapsulated Postscript) output,
!    b) add option to suppress titles above the graph (useful for figures to be included
!       in a paper).
!  Revision 1.3 by MSB, 2009 August 4 & 15:
!    a) larger fonts and thicker lines, for better appearence of plots at publication size,
!    b) add option for user to specify name of output file.

!   Revision 1.6 by MSB 2013 Nov 13:
!   Moved subroutines to TTE_Subroutines.f90 for use by various versions of main program, e.g.,
!   interactive (this program) and file input.


!  Program TTE_ScatterPlot:

!  Usage: TTE_ScatterPlot.exe  FileName

! #################################################################################################



program TTE_ScatterPlot

   use MSB_Constants

   use, intrinsic :: iso_c_binding, only: C_CHAR

   use dislin, only: disfin

   use TTE_subroutines, only:  TTE_GetCmdLineInput_AllFileNames,  MakeUpperCase,  MakeLowerCase,           &
          MakeEnergyLabels, Return_all_TTE, Select_Detectors,  Select_EChannels, Determine_ReferenceTime,  &
          Setup_TTE_ScatterPlot, Plot_TTE_ScatterPlot

   use TTE_constant_defs
   use Type_for_all_TTE_Events

   implicit none

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   character (len=FILE_NAME_MAX_LEN) :: NamePortion
   character (kind=C_CHAR, len=FILE_NAME_MAX_LEN), dimension(0: MAX_GBM_DET-1 ) :: InputFileNames_for_C

   real (real64), dimension (2) :: GraphTimeRange_s
   integer (int32), dimension (2) :: ChannelRange


   type (All_TTE_Events_type), dimension (0: MAX_GBM_DET-1) :: TTE_Data


   integer (int32) :: idet

   character (len=2) :: TimeLabel

   logical, dimension ( 0: MAX_GBM_DET-1 ) :: RequestDetectors
   integer (int32) :: NumRequestDetectors
   integer (int32) :: SingleRequestDetector
   logical :: Single_DetType_Selected
   character (len=DETECTOR_SELECTION_LABEL_LEN) :: DetectorSelectionLabel
   character (len=42) :: ChannelRangeLabel
   character (len=1) :: DetSymMode
   logical :: DetColorMode
   character (len=1) :: TmpString2
   character (len=ENERGY_LABEL_LEN) :: Left_Energy_Label
   character (len=ENERGY_LABEL_LEN) :: Right_Energy_Label

   logical :: FirstDetRead = .TRUE.

   character (len=9) :: TimeRefType
   real (C_DOUBLE) :: TriggerTime
   real (real128) :: ReferenceTime

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   write ( *, '( / "TTE_ScatterPlot by MSB, Version 1.7, 2016 Jan 26." / )' )


   !  read command line input, obtain name of datafiles to read:

   call TTE_GetCmdLineInput_AllFileNames ( NamePortion, InputFileNames_for_C )


   !  more input from user:

   !  Primary output, from user input: detector selection.  Secondary output: derived quantities from detector selection:

   call Select_Detectors   &
          ( RequestDetectors, NumRequestDetectors, SingleRequestDetector, Single_DetType_Selected, DetectorSelectionLabel )



  !  Ask re symbols if multiple detectors are being plotted:

   DetSymMode = "S"
   DetColorMode = .FALSE.

   if ( NumRequestDetectors > 1  .AND.  NumRequestDetectors <= 7 ) then

      1010 continue
      write ( *, '( / "Do you want to plot the ", I3, " detectors with the Same, Different symbols or Numbers?" )' )     &
         NumRequestDetectors
      write ( *, '( "Input ""S"" for Same Symbols, ""D"" for Different Symbols, or ""N"" for Numbers: " )', advance="no" )
      read ( *, '( A1 )' )  DetSymMode
      call MakeUpperCase (DetSymMode)
      if ( DetSymMode /= "S"  .AND.  DetSymMode /= "D"  .AND.  DetSymMode /= "N" )  goto 1010

      !  multiple colors are only available if different symbols were selected:
      if ( DetSymMode == "D" ) then
         1015 continue
         write ( *, '( "Also use color to distinguish the detectors?" )' )
         write ( *, '( "Input ""Y"" or ""N"": " )', advance="no" )
         read ( *, '( A1 )' )  TmpString2
         call MakeUpperCase (TmpString2)
         if ( TmpString2 /= "Y"  .AND.  TmpString2 /= "N" ) goto 1015
         if ( TmpString2 == "Y" )  DetColorMode = .TRUE.
      end if

   else

      if ( NumRequestDetectors > 7 ) then

         1017 continue
         write ( *, '( / "Do you want to plot the ", I3, " detectors with the Same symbols or Numbers?" )' )  NumRequestDetectors
         write ( *, '( "Input ""S"" for Same Symbols or ""N"" for Numbers: " )', advance="no" )
         read ( *, '( A1 )' )  DetSymMode
         call MakeUpperCase (DetSymMode)
         if ( DetSymMode /= "S"  .AND.  DetSymMode /= "N" )  goto 1017

      end if

   end if

   if ( DetSymMode == "N" )    &
     write ( *, '( // "WARNING: The positions of the numbers are not as precise as the other plotting modes !!" / )' )

   !  Ask: time units and time range:

   1020 continue
   write ( *, '( / "Do you wish to input the time range to be plotted as seconds (s) or milliseconds (ms)?" )' )
   write ( *, '( "Input either ""s"" or ""ms"": " )', advance='no' )
   read ( *, *, ERR=1020 )  TimeLabel
   call MakeLowerCase (TimeLabel)
   if ( TimeLabel /= "s "  .and.  TimeLabel /= "ms" )  goto 1020


   1030 continue
   write ( *, '( / "Input Time Range to plot for file ", A, " in ", A, ": " )', advance='no' )  trim (NamePortion), TimeLabel
   read ( *, *, err=1030 )  GraphTimeRange_s (1), GraphTimeRange_s (2)
   if ( TimeLabel == "ms" )  GraphTimeRange_s = GraphTimeRange_s / 1000.0_real64

   write ( *, '( "The time range of the graph will be", F9.4, " to", F9.4, " s." )' )  GraphTimeRange_s (1), GraphTimeRange_s (2)
   write ( *, '( "or", F8.1, " to", F8.1, " ms." )' )    &
          GraphTimeRange_s (1) * 1000.0_real64, GraphTimeRange_s (2) * 1000.0_real64



   !  Call a subroutine to read and return the TTE for all of the requested detectors.
   !  An alternative design, which was originally used, is to read the data directly in
   !  this routine -- this allows reading and plotting the data within the same idet loop --
   !  in this implementation one need only store the data for one detector at a time, rather
   !  than having to store the data for all of the detectors, as is done in the current
   !  implementation.  The advantage of the current implementation is that the TTE-reading
   !  code has been factored out into a subroutine.  We are not concerned about the increased
   !  memory consumption.

   call Return_all_TTE  ( InputFileNames_for_C, RequestDetectors, TTE_Data, TriggerTime, TimeRefType )


   call Determine_ReferenceTime  ( RequestDetectors, TriggerTime, TTE_Data, ReferenceTime )



   !  Loop through detectors, plotting the data of each detector:


   do idet=0, MAX_GBM_DET - 1

      if ( RequestDetectors (idet) ) then


         !  Some user input based on the data -- we use the data of the first requested detector:
         !  we need to know the energies of the channels, i.e., values for EMin_of_chan and EMax_of_chan,
         !  to provide that guidance to the user in making their channel selections.
         !  Only need the user input and to perform plot setup once.
         !  In this program we use the energy edges of the channels of the FIRST detector read.

         if ( FirstDetRead ) then

            call Select_EChannels  ( Single_DetType_Selected,     &
                   TTE_Data (idet) % EMin_of_chan, TTE_Data (idet) % EMax_of_chan, "Select Energy Channels:", ChannelRange )

            write ( ChannelRangeLabel, '( "Channels ", I3, " to ", I3 )' )  ChannelRange

            if ( Single_DetType_Selected ) then

               call MakeEnergyLabels ( TTE_Data (idet) % EMin_of_chan, TTE_Data (idet) % EMax_of_chan,       &
                          ChannelRange (1), ChannelRange (2), Left_Energy_Label, Right_Energy_Label )

               ChannelRangeLabel =   &
               ChannelRangeLabel (1:19) // " = " // trim(Left_Energy_Label) // " to " // trim(Right_Energy_Label)

            end if    ! Single_DetType_Selected ?


            !  Setup / configure the plot:

            call Setup_TTE_ScatterPlot  ( GraphTimeRange_s, ChannelRange, Single_DetType_Selected, NamePortion,     &
                           TimeLabel, DetectorSelectionLabel, ChannelRangeLabel,                                    &
                           TTE_Data (idet) % EMin_of_Chan, TTE_Data (idet) % EMax_of_Chan,                          &
                           TimeRefType, ReferenceTime, .TRUE. )

         end if   ! FirstDetRead ?



         !  Plot the events of detector idet:

         call Plot_TTE_ScatterPlot  ( GraphTimeRange_s, ChannelRange, TTE_Data (idet) % tte_channels,           &
                 TTE_Data (idet) % tte_times, ReferenceTime, TimeLabel, DetSymMode, DetColorMode, idet, FirstDetRead )

         FirstDetRead = .FALSE.


         !  Cleanup -- deallocate the Fortran TTE arrays:

         deallocate  ( TTE_Data (idet) % tte_channels )
         deallocate  ( TTE_Data (idet) % tte_times )


      end if     ! RequestDetector ?

   end do    ! detector loop




   !  Close Plot:

   call disfin ()


   !  Cleanup -- nothing more to do


   stop


end program TTE_ScatterPlot


! #################################################################################################
