!  Copyright Michael S. Briggs, University of Alabama in Huntsville.
!  For NASA's Fermi Gamma-ray Burst Monitor (GBM).
!  copyright 2009 February 9 -- 2016.


!  MSB, 2015 Nov 28: add auto_ymin_value

!  MSB, 2011 March 5: support continuous TTE files.

!  MSB, 2010 Sept 25.  Add subroutine CheckTriggerTime.
!  MSB,  various revisions in 2010.

!  MSB, 2009 May 30: 1) two new subroutines to read TTE data, factored out of
!  several main programs.
!  2) merged previous separate modules/tiles TTE_Subroutines_wo_DISLIN and
!  TTE_Subroutines_with_DISLIN, because they were use-ing each other, which
!  is not permitted.
!  MSB, revised March 19--22: widget to select energy channels.
!  MSB, revised March 17: Improved sharing of variables between two routines.
!  2009 March 7 -- 11.
!  Revised 2009 March 1: add additional arguments to subroutine Read_Lvl1_TTE of
!  interface-to-C Read_Lvl1_TTE_Interface.
!  Revised 2009 Feb 11: add subroutine  MakeLowerCase.


! #################################################################################################


module TTE_constant_defs

   use MSB_Constants

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   integer, parameter :: MAX_GBM_DET = 14

   integer, parameter :: MAX_NAI_DET = 12

   integer, parameter :: NUM_SPEC_CHAN = 128

   integer, parameter :: FILE_NAME_MAX_LEN = 200
   integer, parameter :: ENERGY_LABEL_LEN = 8
   integer, parameter :: DETECTOR_SELECTION_LABEL_LEN = 37

   real (real32), parameter :: DeadTime_per_Event_microsec = 2.6_real32
   real (real32), parameter :: DeadTime_per_Overflow_microsec = 10.4_real32


end module TTE_constant_defs


! #################################################################################################


module Type_for_all_TTE_Events

   use, intrinsic :: iso_c_binding
   use TTE_constant_defs

   implicit none

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   !  This is a structure for storing the "raw" event-list TTE data.

   !  If we create instance of this type that is a vector dimensioned at the number of
   !  GBM detectors -- loading the data of all GBM detectors for a trigger into the vector --
   !  then each element TriggerTime of each structure element of the vector will have
   !  the same value.

   !  The "tte_times" of this type are the raw times from the data file
   !  and are NOT corrected for the drift of the GBM DPU clock.

   type All_TTE_Events_type

      real (C_DOUBLE) :: FileStartTime, FileStopTime
      real (C_DOUBLE) :: TriggerTime
      integer (C_LONG) :: Num_TTE
      integer (C_SHORT), dimension (:), pointer :: tte_channels => null ()
      real (C_DOUBLE), dimension (:), pointer :: tte_times => null ()
      real (C_FLOAT), dimension ( 0: NUM_SPEC_CHAN-1 ) :: EMin_of_chan
      real (C_FLOAT), dimension ( 0: NUM_SPEC_CHAN-1 ) :: EMax_of_chan

   end type All_TTE_Events_type


end module Type_for_all_TTE_Events


! #################################################################################################


module TTE_Histogram_Arrays

   use MSB_Constants

   implicit none

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   !  This module declared shared storage -- arrays that hold binned TTE data.

   !  The times of the TTE data that are binned into these histograms, may, if requested
   !  by the user, by corrected by the drift of the GBM DPU clock.

   integer (int32), dimension (:), allocatable, SAVE :: GBL_Data_Histogram_SummedDets
   real (real32), dimension (:), allocatable, SAVE :: GBL_Deadtime_Histogram_SummedDets

   integer (int32), dimension (:, :), allocatable, SAVE :: GBL_Data_Histogram_ByDet
   real (real32), dimension (:, :), allocatable, SAVE :: GBL_Deadtime_Histogram_ByDet


end module TTE_Histogram_Arrays


! #################################################################################################


module TTE_subroutines


   use MSB_Constants

   use TTE_constant_defs

   use, intrinsic :: iso_c_binding, only : C_FLOAT, C_DOUBLE


   implicit none


   PRIVATE   ! change default to PRIVATE

   PUBLIC ::  Select_EChannels,  Select_Detectors, auto_xdiv_value, auto_ymax_value, auto_ydiv_value,                  &
              auto_ymin_value, MakeUpperCase,  MakeLowerCase,  TTE_GetCmdLineInput_3args,                              &
              TTE_GetCmdLineInput_AllFileNames,  TTE_GetCmdLineInput_AUTO_Histo,                                       &
              TTE_GetCmdLineInput_AUTO_DetSelection, Analyze_Detector_Selection,  MakeOutputFileName,                  &
              MakeEnergyLabels, Read_and_Bin_TTE, Return_all_TTE, Get_OutputFileName, Get_GraphFormat,                 &
              CheckTriggerTime, Output_CSV_TitleLines, MET_Corr_for_ClockDrift, GapTimeQuestion, Bin_TTE_Events,       &
              Determine_ReferenceTime, Make_X_AxisLabel, Time_to_BinNum_on_Graph,  TTE_OneHisto_2_file, Output_E_edges

#ifndef NODISLIN
   PUBLIC :: Energy_Scale_RightAxis, Get_DataFileName_Widget, Make_TTE_OneHisto,         &
             MakeMultiHistoPlot, Setup_TTE_ScatterPlot, Plot_TTE_ScatterPlot, PlotLegend

#endif



   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   !  Variables shared between routines of this module.  All are PRIVATE to this module.


   !  Variable MODSHR1_SelectDetectorWidget_Buttons_IDs (a user defined type / structure with several
   !  components) is shared between routine Select_Detectors_Widget and its callback routine
   !  Select_Detectors_Widget_CALLBACK.
   !  The argument list of subroutine Select_Detectors_Widget_CALLBACK is defined by DISLIN, so
   !  the two routines need other variables to communicate.

   !  All components of structure SelectDetectorWidget_Buttons_IDs_type are IDs of DISLIN
   !  widget buttons.

   type Detector_Widget_Buttons_IDs_type

      integer, dimension ( 0: MAX_GBM_DET-1 ) :: Particular_Detector
      integer :: All_Dets
      integer :: All_NaIs
      integer :: Both_BGOs
      integer :: PlusX_Dets
      integer :: MinusX_Dets

   end type Detector_Widget_Buttons_IDs_type

   type (Detector_Widget_Buttons_IDs_type) :: MODSHR1_Detector_Widget_Buttons_IDs



   !  MODSHR2_EChan_Widget_Buttons_IDs and MODSHR2_EChan_Widget_MiscVars (both user defined
   !  types / structures with several components) are shared between routine Select_EChannels_Widget
   !  and its callback routine Select_EChannels_Widget_CALLBACK.
   !  The argument list of subroutine Select_EChannels_Widget_CALLBACK is defined by DISLIN, so
   !  the two routines need other variables to communicate.

   !  All components of structure EChan_Widget_Buttons_IDs_type are IDs of DISLIN
   !  widget buttons.

   type EChan_Widget_Buttons_IDs_type

      integer, dimension (2) :: EChan_Label
      integer, dimension (2) :: EValue_Label
      integer, dimension (2) :: WidgetColumn
      integer, dimension (2) :: OneStepAdjustDown
      integer, dimension (2) :: TenStepAdjustDown
      integer, dimension (2) :: OneStepAdjustUp
      integer, dimension (2) :: TenStepAdjustUp

   end type EChan_Widget_Buttons_IDs_type


   !  Items shared between Select_EChannels_Widget and Select_EChannels_Widget_CALLBACK that
   !  are NOT widget button IDs:

   type EChan_Widget_MiscVars_type

      logical :: EValue_Widget_Created
      integer (int32), dimension (2) :: ChannelRange
      real (C_FLOAT), dimension (0: NUM_SPEC_CHAN-1) :: EMin_of_chan
      real (C_FLOAT), dimension (0: NUM_SPEC_CHAN-1) :: EMax_of_chan

   end type EChan_Widget_MiscVars_type


   type (EChan_Widget_Buttons_IDs_type) :: MODSHR2_EChan_Widget_Buttons_IDs

   type (EChan_Widget_MiscVars_type) :: MODSHR2_EChan_Widget_MiscVars


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


contains


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


#ifndef NODISLIN
   subroutine Setup_TTE_ScatterPlot  ( GraphTimeRange_s, ChannelRange, Single_DetType_Selected, TriggerName,     &
                 TimeLabel, DetectorSelectionLabel, ChannelRangeLabel, EMin_of_Chan, EMax_of_Chan,               &
                 TimeRefType, ReferenceTime, Interactive )


      use TTE_constant_defs

      use dislin

      implicit none


      ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

      real (real64), dimension (2), intent (in) :: GraphTimeRange_s
      integer (int32), dimension (2), intent (in) :: ChannelRange
      logical, intent (in) :: Single_DetType_Selected
      character (len=*), intent (in) :: TriggerName
      character (len=*), intent (in) :: DetectorSelectionLabel
      character (len=*), intent (in) :: ChannelRangeLabel
      character (len=2), intent (in) :: TimeLabel
      real (C_FLOAT), dimension (0: NUM_SPEC_CHAN-1), intent (in) :: EMin_of_chan
      real (C_FLOAT), dimension (0: NUM_SPEC_CHAN-1), intent (in) :: EMax_of_chan

      character (len=9), intent (in) :: TimeRefType
      real (real128), intent (in) :: ReferenceTime
      logical, intent (in) :: Interactive


      ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

      ! internal variables:

      character (len=FILE_NAME_MAX_LEN) :: OutputFileName

      character (len=12), parameter :: Fragment = ".ScatterPlot"
      character (len=3) :: GraphFormat
      character (len=1) :: response

      character (len=80) :: X_AxisLabel
      character (len=40) :: Y_AxisLabel

      real (real32), dimension (2) :: TimeRange_for_Call
      real (real32) :: TimeDiff


      real (real32) :: xdiv
      real (real32) :: ymin, ymax, ydiv, y_label_one

      integer :: y_int
      integer :: PltCoord_Left, PltCoord_Right, PltCoord_Bot, PltCoord_Top
      integer :: Plt_Width, Plt_Height

      character (len=40) :: input
      real (real32) :: try


      ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


      if ( Interactive ) then
         call  Get_GraphFormat  ( GraphFormat )
         !  Create output file name:
         call  Get_OutputFileName  ( TriggerName, Fragment, GraphFormat, OutputFileName )
      else
         GraphFormat = "pdf"
         OutputFileName = trim (TriggerName) // trim (Fragment) // ".pdf"
      end if


      !  configure and initiate the plot:

      call setfil ( trim (OutputFileName) )
      call metafl ( GraphFormat )

      ! Normally PNG plots white lines and characters on black background --
      !  reverse to black on white:
      ! Also override default pixel dimensions to have same aspect ratio as logical pixel dimensions
      if ( GraphFormat == 'png' ) then
         call scrmod ( "REVERS" )
         call winsiz ( 853, 660 )  ! physical pixel dimensions, override default of 603 x 853
      end if


      call setpag ( 'USAL' )      ! US A (=Letter), Landscape, 2790 x 2160 points

      call filmod ( 'DELETE' )    ! overwrite (delete) previous file

      call disini ()


      if ( GraphFormat == 'pdf' .or. GraphFormat == 'eps' )  call psfont ( 'Helvetica' )


      !  Make all lines thicker, including symbols; default is one; doesn't apply to characters.
      call linwid ( 5 )


      !  axes and axes labels:

      if ( TimeLabel == "s " ) then
         TimeRange_for_Call = real ( GraphTimeRange_s, real32 )
      else
         !  x-axis in ms instead of s case:
         TimeRange_for_Call = real ( GraphTimeRange_s * 1000.0_real64, real32 )
      end if
      TimeDiff = TimeRange_for_Call (2) - TimeRange_for_Call (1)


      !  Determine tick-spacing of X-axis, either automatically (with user approval), or from user input:

      try = auto_xdiv_value ( TimeDiff )
      xdiv = try

      if ( Interactive ) then
         9020 continue
         write ( *, 9030, advance='no' )  TimeLabel, try, TimeLabel
         9030 format ( / "Input X (time) axis major (numbered) tick spacing in ", A, " or ENTER to use", F8.2, 1X, A, " : " )
         read ( *, '(A)', err=9020 ) input

         if ( len_trim (input) == 0 ) then
            xdiv = try
         else
            read ( input, *, err=9020 )  xdiv
         end if
      end if


      call labdig ( -2, 'X' )  ! DISLIN (>=9.4?) automatically calculates number of digits after decimal point on X-axis labels
      call labdig ( 0, 'Y' )   ! zero digits after decimal point on Y-axis labels

      X_AxisLabel = Make_X_AxisLabel ( TimeRefType, TimeLabel, .FALSE., 0.0_real32, ReferenceTime )
      Y_AxisLabel = "Energy Channel"

      call name ( X_AxisLabel,  'X' )
      call name ( Y_AxisLabel, 'Y' )


      !  increase size of axis titles (aka names) from default of 36,
      !  and from previous (long used) value of 46.
      !  Now longer X-Axis Label, so decrease from 55.
      call hname ( 50 )

      ! increase the size of characters from default of 36:
      call height ( 50 )


      !  increase length of major and minor ticks from defaults of 24 & 16:
      call ticlen ( 33, 22 )


      !  Draw graph box / axes:
      !  The x-axis is simple -- the time range.
      !  The basic y-axis is the channel range.
      !  If only one detector type (NaI or BGO) will be plotted, we also plot
      !  an energy scale on the right y-axis.


      ymin = ChannelRange (1)
      ymax = ChannelRange (2)

      !  special case: all channels plotted: expand Y-axis range to have no clipping of symbols:
      if ( ChannelRange (1) == 0  .and.  ChannelRange (2) == 127 ) then
         ymin = -2
         ymax = 129
      end if

      !   first y-axis label needs to be a multiple of 10, first value above ymin:
      y_int = 10 * (ChannelRange (1) / 10)
      if ( y_int  <  ChannelRange (1) - 0.1 )  y_int = y_int + 10
      y_label_one = y_int

      ydiv = 10.0
      if ( ChannelRange (2) - ChannelRange (1)  >  90 )  ydiv = 20


      !  If only one detector type is selected, draw just a line rather than a line with tick marks, for the
      !  right y-axis.  The tick marks are suppressed because we will draw an energy scale instead of a channel scale.
      !  So for the fourth argument, the default of "TICKS" is replaced with "LINE".

      if ( Single_DetType_Selected )  call setgrf ( "NAME", "NAME", "TICKS", "LINE" )

      call graf ( TimeRange_for_Call (1), TimeRange_for_Call (2), TimeRange_for_Call (1), xdiv, ymin, ymax, y_label_one, ydiv )



      !  If only one detector type is being plotted, plot an energy scale on the right y-axis:

      if ( Single_DetType_Selected )  call Energy_Scale_RightAxis ( ChannelRange, EMin_of_chan, EMax_of_chan )



      !  special case: all channels plotted: plot dotted lines at actual limits
      if ( ChannelRange (1) == 0  .and.  ChannelRange (2) == 127 ) then
         call dot ()
         call rline ( TimeRange_for_Call (1),   0.0, TimeRange_for_Call (2),   0.0 )
         call rline ( TimeRange_for_Call (1), 127.0, TimeRange_for_Call (2), 127.0 )
         call solid ()  ! restore default
      end if


      !  add the titles:

      response = "Y"
      if ( Interactive ) then
         9040 continue
         write ( *, '( / "Output Titles above the Graph?  ""Y"" for Yes, or ENTER for Yes, or ""N"" for No: " )', advance='no' )
         read ( *, '(A)', err=9040 )  response
         if ( len_trim (response) == 0 )  response = 'Y'
         call MakeUpperCase (response)
         if ( response /= 'N' .and. response /= 'Y' ) goto 9040
      end if

      if ( response == 'Y' ) then

         call titlin ( TriggerName, 1 )
         call titlin ( DetectorSelectionLabel, 2 )
         call titlin ( ChannelRangeLabel, 3 )
         call title ()

      end if



      !  *** The various labels must be plotted before the call to clpwin because the clpwin
      !  *** call applies to the labels.   (Strangely, when clpwin prevents labels outside
      !  *** of its range for PNG plots, but not for PDF.)


      !  Extend the clipping region past the top and bottom of the graph, but not past the sides.
      !  This way symbols aren't truncated when the extend past the bottom or top edges of the
      !  graph -- the truncation looks bad.

      !  First we have to figure out the plot coordinates of the graph box.
      !  To do this we use DISLIN routines to calculate plot coordinates from user coordinates,
      !  using the limits of the graph box in user coordinates.

      PltCoord_Left = NXPOSN ( TimeRange_for_Call (1) )
      PltCoord_Right = NXPOSN ( TimeRange_for_Call (2) )
      PltCoord_Bot = NYPOSN ( real (ChannelRange (1)) )
      PltCoord_Top = NYPOSN ( real (ChannelRange (2)) )

      Plt_Width = PltCoord_Right - PltCoord_Left
      Plt_Height = PltCoord_Bot - PltCoord_Top


      !  Optional legend inside of the graph box:

      if ( Interactive )  call  PlotLegend  ( PltCoord_Left, PltCoord_Right, PltCoord_Bot, PltCoord_Top )



      !  First two inputs are coordinates of upper-left of graph box -- we shift
      !  clip region upwards w.r.t. actual graph box by 1/20 of height of box.
      !  Second two inputs are width and height of clip box -- we increase height
      !  by 1/10 of the height.   This gives an extra margin of 1/20 of height at
      !  past the top and the bottom.

      call clpwin ( PltCoord_Left, PltCoord_Top - Plt_Height / 20,    &
           Plt_Width, Plt_Height + Plt_Height / 10 )
      call noclip ()   ! required to disable clipping at axis limits so that clip window controls


      return


   end subroutine Setup_TTE_ScatterPlot
#endif


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


#ifndef NODISLIN
   subroutine Plot_TTE_ScatterPlot       &
         ( GraphTimeRange_s, ChannelRange, tte_channels, tte_times, ReferenceTime, TimeLabel, DetSymMode, DetColorMode,    &
           DetNum, StartPlot )

      use, intrinsic :: iso_c_binding, only: C_SHORT, C_DOUBLE

      use dislin

      implicit none

      ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

      real (real64), dimension (2), intent (in) :: GraphTimeRange_s
      integer (int32), dimension (2), intent (in) :: ChannelRange
      integer (C_SHORT), dimension (:), pointer, intent (in) :: tte_channels
      real (C_DOUBLE), dimension (:), pointer, intent (in) :: tte_times
      real (real128), intent (in) :: ReferenceTime
      character (len=2), intent (in) :: TimeLabel
      character (len=1), intent (in) :: DetSymMode
      logical, intent (in) :: DetColorMode

      integer (int32), intent (in) :: DetNum
      logical, intent (in) :: StartPlot


      ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

      ! internal variables:

      integer (int32) :: Num_TTE
      integer (int32) :: i_tte

      real (real128) :: time
      integer :: PltSymbol
      character (len=1) :: DetString


      ! local, persistent variables:

      integer, save :: CallCount = 0


      ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

      if ( StartPlot )  CallCount = 0

      Num_TTE = size (tte_times)


      !  Decrease symbol size from default of 35 (units: plot coord.)
      call hsymbl ( 30 )


      !  Determine symbol to use
      !  Count detectors on plot by counting calls to this subroutine.
      !  Symbol also depends on whether user has selected Different symbols
      !  for different detectors, or Same symbol for all detectors.

      CallCount = CallCount + 1

      PltSymbol = 15   !  open circle -- used for "same symbol" for all dets
      call color ( "FORE" )   ! default color, if "same symbol" is selected

      if ( DetSymMode == "D" ) then

         select case (CallCount)
            case (1)
               PltSymbol = 15     ! open circle
               if ( DetColorMode )  call color ( "RED" )
            case (2)
               PltSymbol = 4      ! X
               if ( DetColorMode )  call color ( "BLUE" )
            case (3)
               PltSymbol = 2      ! open triangle
               if ( DetColorMode )  call color ( "MAGENTA" )
            case (4)
               PltSymbol = 0      ! open square
               if ( DetColorMode )  call color ( "ORANGE" )
            case (5)
               PltSymbol = 3      ! +
               if ( DetColorMode )  call color ( "GREEN" )
            case (6)
               PltSymbol = 8      !  asterisk
               if ( DetColorMode )  call color ( "CYAN" )
            case (7)
               PltSymbol = 5      ! open diamond
               if ( DetColorMode )  call color ( "YELLOW" )
            case default
               PltSymbol = 21   ! filled disk -- should never occur!
               if ( DetColorMode )  call color ( "FORE" )
         end select

      end if



      !  Plot symbol for any TTE event within the requested channel ChannelRange and
      !  time range GraphTimeRange_s:
      !  (We could use DISLIN subroutine "CURVE", but instead we select individual
      !  events that are within requested ranges .....)

      ! Note: comparison is done in PastDouble for accuracy and because of the type of ReferenceTime.
      ! Time is passed to rlsymb in single because we are using the single-precision version of dislin.

      call txtjus ( "CENT" )

      call height ( 24 )

      do i_tte=1, Num_TTE

         if ( tte_times (i_tte) >= ReferenceTime + GraphTimeRange_s (1)  .AND.    &
              tte_times (i_tte) <= ReferenceTime + GraphTimeRange_s (2) ) then

              if ( tte_channels (i_tte) >= ChannelRange (1)  .AND.    &
                   tte_channels (i_tte) <= ChannelRange (2) ) then

                 time = tte_times (i_tte) - ReferenceTime
                 if ( TimeLabel == "ms" )  time = time * 1000.0_C_DOUBLE     ! if x-axis is in ms instead of s

                 if ( DetSymMode /= "N" ) then
                    call rlsymb ( PltSymbol, real (time, real32), real (tte_channels (i_tte), real32) )
                 else
                    write ( DetString, '(Z1)' )  DetNum
                    call rlmess ( DetString, real (time, real32), real (tte_channels (i_tte), real32) )
                 end if

              end if    !  channel range

         end if    !  time range

      end do    ! loop over tte events



      call color ( "FORE" )    !  restore default color

      call txtjus ( "LEFT" )   !  restore default text justification

      call height ( 36 )   ! restore default height


      return


   end subroutine Plot_TTE_ScatterPlot
#endif


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


#ifndef NODISLIN
   subroutine Energy_Scale_RightAxis ( ChannelRange, EMin_of_chan, EMax_of_chan )

      use TTE_constant_defs

      use dislin

      implicit none

      ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

      integer (int32), dimension (2), intent (in) :: ChannelRange
      real (C_FLOAT), dimension (0: NUM_SPEC_CHAN-1), intent (in) :: EMin_of_chan
      real (C_FLOAT), dimension (0: NUM_SPEC_CHAN-1), intent (in) :: EMax_of_chan


      ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

      ! internal variables:

      integer :: ichan, e_idx
      real (real32) :: mid_chan

      real (real32), dimension (1:8), parameter :: Energy_Values_List =    &
         (/     10.,       30.,     100.,    300.,      1000.,   3000.,   10000.,  30000.  /)

      character (len=7), dimension (1:8), parameter :: Energy_Labels_List =     &
         (/  "10 keV ", "30 keV ", "100 keV", "300 keV", "1 MeV  ", "3 MeV  ", "10 MeV ", "30 MeV "  /)


      ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


      call labjus ( "LEFT", "Y" )
      call ticpos ( "LABELS", "Y" )

      do ichan=ChannelRange(1), min ( ChannelRange(2), NUM_SPEC_CHAN-2 )        ! omit overflow channel

         do e_idx=1, 8

            if ( Energy_Values_List (e_idx) >= EMin_of_chan (ichan)   .AND.    &
                 Energy_Values_List (e_idx) <= EMax_of_chan (ichan) ) then

                 mid_chan = ichan + 0.5
                 call addlab ( trim (Energy_Labels_List (e_idx) ), mid_chan, 2, "YRIGHT" )

            end if

         end do
      end do


   end subroutine Energy_Scale_RightAxis
#endif


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  Divide the observed (aka raw) counts by this factor to obtain the deadtime corrected value.
!  elemetal ==> pure ==> no side effects are allowed !!

elemental  function  LivetimeFraction  ( RegularNumber, OverflowNumber, BinWidth_microsec )

   real (real32) :: LivetimeFraction     ! function return

   integer (int32), intent (in) :: RegularNumber      ! total counts in time bin over energy channels 1 to 126
   integer (int32), intent (in) :: OverflowNumber      ! total counts in time bin in energy channel 127
   real (real32), intent (in) :: BinWidth_microsec

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   ! internal variables:

   real (real32) :: DeadTime, factor

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   DeadTime = RegularNumber * DeadTime_per_Event_microsec  +  &
             OverflowNumber * DeadTime_per_Overflow_microsec

   factor = 1.0_real32 - DeadTime / BinWidth_microsec

   if ( factor <= 0.01_real32 )  factor = 0.01_real32   ! override absurd livetime fraction


   LivetimeFraction = factor


   return

end function


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  All output of subroutine Bin_TTE_Events is into the arrays of the module TTE_Histogram_Arrays.

!  Revised by M. S. Briggs, 2011 Feb 14.

!  The times received in "tte_times" are raw times as read the data file,
!  uncorrected for the drift of the GBM DPU clock.

!  The histograms output by this routine to the shared variables of
!  module TTE_Histogram_Arrays are optionally corrected by the clock drift.


subroutine  Bin_TTE_Events  ( DetNum, Num_TTE, TriggerTime, tte_channels, tte_times, GraphTimeRange_s, BinWidth_s,     &
                 ChannelRange, GapTime, ClockDrift_Flag )

   use, intrinsic :: iso_c_binding
   use TTE_Histogram_Arrays

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   integer (C_LONG), intent (in) :: DetNum
   integer (C_LONG), intent (in) :: Num_TTE
   real (real128), intent (in) :: TriggerTime
   integer (C_SHORT), dimension (:), pointer, intent (in) :: tte_channels
   real (C_DOUBLE), dimension (:), pointer, intent (in) :: tte_times
   real (real64), dimension (2), intent (in) :: GraphTimeRange_s
   real (real64), intent (in) :: BinWidth_s
   integer (int32), dimension (2), intent (in) :: ChannelRange

   real (real32), intent (in) :: GapTime
   logical, intent (in) :: ClockDrift_Flag

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   ! internal variables:

   integer (int32), dimension (:), allocatable :: Regular_Histogram_OneDet
   integer (int32), dimension (:), allocatable :: Overflow_Histogram_OneDet

   integer (C_LONG) :: i_tte
   real (real128) :: time_rel_trig    ! must do calc in double or better to avoid round off error
   integer (int32) :: BinNum
   real (real32) :: BinWidth_microsec

   real (real128) :: Corr_TriggerTime = 0.0_real128   ! initialize to suppress warning

   real (real64) :: tmp_DBL
   real (real128) :: tmp_QUAD


   !  interval variables, values preserved across calls:

   logical, SAVE :: FirstCall = .TRUE.
   integer (int32), SAVE :: Bins_in_Histogram  ! value calculated on first call is always used !!


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   tmp_DBL = BinWidth_s * 1000000.0_real64
   BinWidth_microsec = real ( tmp_DBL, real32 )

   if ( ClockDrift_Flag)  Corr_TriggerTime = MET_Corr_for_ClockDrift  ( TriggerTime, GapTime )


   if ( FirstCall ) then

      FirstCall = .FALSE.


      !  Calculate number of histogram bins, and allocate the histgram arrays.

      !  Convert to greatest integer that is equal to or greater than real argument --
      !  sometimes the real value for the number of bins will be a fraction, e.g., 31.25, if the inputs
      !  results in a fractional final bin, and this real value must be rounded up when converting to the
      !  integer number of bins (e.g., 32).  Unfortunately if the real value should be an exact integer,
      !  e.g., 31.0000, and rounding causes it to be very slightly higher, e.g., 31.0001, we will get an
      !  extra bin -- but extra is much better than too few!
      !  The variables are Doubles to reduce the probability of rounding problems -- many are caused
      !  by conversion from decimal to binary.

      Bins_in_Histogram = ceiling ( ( GraphTimeRange_s (2) - GraphTimeRange_s (1) ) / BinWidth_s )


      write ( *, '( / "Histogram will have", I6, " bins." )' )  Bins_in_Histogram
      if ( Bins_in_Histogram <= 2  .or.                         &
           GraphTimeRange_s (2) <= GraphTimeRange_s (1)  .or.   &
           BinWidth_s <= 4.0E-6 ) then

         write ( *, '( / "The time parameters of the graph are unacceptable.  Execution aborting." )' )
         write ( *, '( "Time Range of graph in seconds:", 1pG14.5, 2x, 1pG14.5 )' )  GraphTimeRange_s
         write ( *, '( "Bin width in seconds:", 1pG14.5 )' )  BinWidth_s
      end if


      !  Create and initialize arrays to hold the persistent output Histograms of the module TTE_Histogram_Arrays:
      !   GBL_Deadtime_Histogram_ByDet doesn't need to be initialized because it isn't summed into.

      allocate ( GBL_Data_Histogram_SummedDets (0: Bins_in_Histogram - 1 ) )
      allocate ( GBL_Deadtime_Histogram_SummedDets ( 0: Bins_in_Histogram - 1 ) )

      allocate ( GBL_Data_Histogram_ByDet ( 0: Bins_in_Histogram - 1, 0: MAX_GBM_DET - 1 ) )
      allocate ( GBL_Deadtime_Histogram_ByDet ( 0: Bins_in_Histogram - 1, 0: MAX_GBM_DET - 1 ) )

      GBL_Data_Histogram_SummedDets = 0
      GBL_Deadtime_Histogram_SummedDets = 0.0

      GBL_Data_Histogram_ByDet = 0

   end if


   ! Create and initialize workspace Histogram:

   allocate ( Regular_Histogram_OneDet ( 0: Bins_in_Histogram - 1 ) )
   allocate ( Overflow_Histogram_OneDet  ( 0: Bins_in_Histogram - 1 ) )

   Regular_Histogram_OneDet = 0
   Overflow_Histogram_OneDet = 0



   !  Load any TTE events in the the requested time range into the Histograms:

   do i_tte=1, Num_TTE


      !  Is the TTE event within the requested time range ?

      if ( ClockDrift_Flag) then
         time_rel_trig = MET_Corr_for_ClockDrift  ( real (tte_times (i_tte), real128), GapTime ) - Corr_TriggerTime
      else
         time_rel_trig = tte_times (i_tte) - TriggerTime
      end if


      if ( time_rel_trig  >  GraphTimeRange_s (1)  .AND.    &
           time_rel_trig  <  GraphTimeRange_s (2) ) then


         ! Calculate Bin Number of this TTE event

         tmp_QUAD = ( time_rel_trig - GraphTimeRange_s (1) ) / BinWidth_s
         BinNum = int ( tmp_QUAD, int32 )

         if ( BinNum < 0  .OR.  BinNum >= Bins_in_Histogram ) then
            write ( *, * ) "Program Failure re TTE BinNum:", i_tte, tte_times (i_tte), time_rel_trig, GraphTimeRange_s (1), BinNum
            stop
         end if


         !  Increment the bin of the Histograms for this detector:
         !  Summation has to be into the histogram for one detector, zeroed for that
         !  detector, to allow deadtime calculation -- see below.
         !  For the deadtime correction, we need the total number of events from channels 1 to 127,
         !  and the number of events in channel 127, the overflow channel.
         !  The output array, GBL_Data_Histogram_ByDet, isn't usable for the deadtime correction if
         !  the user has selected a subset of the channels.

         if ( tte_channels (i_tte) >= ChannelRange (1)  .and.  tte_channels (i_tte) <= ChannelRange (2) )  &
            GBL_Data_Histogram_ByDet (BinNum, DetNum) = GBL_Data_Histogram_ByDet (BinNum, DetNum) + 1

         if ( tte_channels (i_tte) < 127 )  Regular_Histogram_OneDet (BinNum) = Regular_Histogram_OneDet (BinNum) + 1

         if ( tte_channels (i_tte) == 127 )  Overflow_Histogram_OneDet (BinNum) = Overflow_Histogram_OneDet (BinNum) + 1

      end if


   end do    ! i_tte



   !  Calculate deadtime corrected counts per bin, using "OneDet" accumulations since only events
   !  from a particular detector contribute to the deadtime of that detector.

   GBL_Deadtime_Histogram_ByDet (:, DetNum) = GBL_Data_Histogram_ByDet (:, DetNum)  /       &
         LivetimeFraction ( Regular_Histogram_OneDet, Overflow_Histogram_OneDet, BinWidth_microsec )


   !  Create histograms summed over all detectors (for which this subroutine has been called)
   !  by adding in the histograms that we just created for detector DetNum.

   GBL_Data_Histogram_SummedDets = GBL_Data_Histogram_SummedDets + GBL_Data_Histogram_ByDet (:, DetNum)
   GBL_Deadtime_Histogram_SummedDets = GBL_Deadtime_Histogram_SummedDets + GBL_Deadtime_Histogram_ByDet (:, DetNum)


   ! cleanup

   deallocate (Regular_Histogram_OneDet)
   deallocate (Overflow_Histogram_OneDet)


   return


end subroutine Bin_TTE_Events


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  Converts any lower-case alphabet characters of string to upper-case.
!  Conversion is done in place, i.e., string is input & output argument.
!  Non-alphabet characters are ignored.
!  Algorithm is based on ASCII numeric values.

subroutine  MakeUpperCase  ( string )

   character (len=*), intent (inout) :: string

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   ! internal variables:

   integer :: i
   integer :: ascii_val

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   do i=1, len (string)

      ascii_val = iachar (string (i:i) )

      if ( ascii_val >= 97  .and.  ascii_val <= 122 ) then

         ascii_val = ascii_val - 32
         string (i:i) = achar (ascii_val)

      end if

   end do


   return


end subroutine MakeUpperCase


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  Converts any upper-case alphabet characters of string to lower-case.
!  Conversion is done in place, i.e., string is input & output argument.
!  Non-alphabet characters are ignored.
!  Algorithm is based on ASCII numeric values.

subroutine  MakeLowerCase  ( string )

   character (len=*), intent (inout) :: string

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   ! internal variables:

   integer :: i
   integer :: ascii_val

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   do i=1, len (string)

      ascii_val = iachar (string (i:i) )

      if ( ascii_val >= 65  .and.  ascii_val <= 90 ) then

         ascii_val = ascii_val + 32
         string (i:i) = achar (ascii_val)

      end if

   end do


   return


end subroutine MakeLowerCase


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


subroutine  TTE_GetCmdLineInput_3args  ( InputFileName, InputFileName_for_C, OutputTimeStart_s, OutputTimeEnd_s )

   use, intrinsic :: iso_c_binding

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   character (len=FILE_NAME_MAX_LEN), intent (out) :: InputFileName
   character (kind=C_CHAR, len=FILE_NAME_MAX_LEN), intent (out) :: InputFileName_for_C

   !  These two arguments are "inout" in order to receive the default value, which are returned
   !  back to the calling routine if overriding values are not provided on the command line:

   real (real32), intent (inout) :: OutputTimeStart_s, OutputTimeEnd_s


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   ! internal variables:

   integer :: NumberCmdArguments
   character (len=50) string


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   NumberCmdArguments = COMMAND_ARGUMENT_COUNT ()
   if ( NumberCmdArguments /= 1  .AND.  NumberCmdArguments /= 3 ) then
      write ( *, '( / "Unexpected number of arguments on command line:", I3 )' )  NumberCmdArguments
      write ( *, '( "Usage: ProgName.exe  InputFileName.fit" / "OR" )' )
      write ( *, '( "Usage: ProgName.exe  InputFileName.fit  OutputTimeStart  OutputTimeEnd" )' )
      write ( *, '( "The optional times are in seconds relative to trigger time." )' )
      write ( *, '( "Default values for the times are", 1PG12.4, " and", 1PG12.4, " s." / )' )  OutputTimeStart_s, OutputTimeEnd_s
      stop
   end if

   call GET_COMMAND_ARGUMENT ( 1, InputFileName )

   if ( NumberCmdArguments == 3 ) then

      call GET_COMMAND_ARGUMENT ( 2, string )
      read ( string, * )  OutputTimeStart_s
      call GET_COMMAND_ARGUMENT ( 3, string )
      read ( string, * )  OutputTimeEnd_s

      if ( OutputTimeStart_s >= OutputTimeEnd_s ) then
         write ( *, '( "Bad input for output time range:", 1PG13.5, 1PG13.5 )' )  OutputTimeStart_s, OutputTimeEnd_s
         stop
      end if

   end if


   !  Add terminating C Null character.

   InputFileName_for_C = trim (InputFileName) // C_NULL_CHAR


   return


end subroutine TTE_GetCmdLineInput_3args


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


subroutine  TTE_GetCmdLineInput_AllFileNames  ( NamePortion, InputFileNames_for_C,   &
                 AllowExtraArgs )

   use, intrinsic :: iso_c_binding

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   character (len=FILE_NAME_MAX_LEN), intent (out) :: NamePortion
   character (kind=C_CHAR, len=FILE_NAME_MAX_LEN), dimension(0: MAX_GBM_DET-1 ), intent (out) :: InputFileNames_for_C
   logical, intent (in), OPTIONAL :: AllowExtraArgs


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   ! internal variables:

   integer :: NumberCmdArguments
   character (len=FILE_NAME_MAX_LEN-1) :: FirstArg
   character (len=FILE_NAME_MAX_LEN-1) :: FileName
   integer :: ldet
   integer :: pos
   character (len=4) :: FileType

   logical :: ApplyArgTest

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   NumberCmdArguments = COMMAND_ARGUMENT_COUNT ()

   if ( PRESENT (AllowExtraArgs) ) then
      ApplyArgTest = AllowExtraArgs
   else
      ApplyArgTest = .TRUE.
   end if

   if ( ApplyArgTest  .AND.  NumberCmdArguments > 1 ) then
      write ( *, '( / "Unexpected number of arguments on command line:", I3 )' )  NumberCmdArguments
      write ( *, '( "Usage: ProgName.exe  FileName" )' )
      write (*, '( "The file name should be the TTE file for the b0 (BGO 0) detector." )' )
      write (*, '( "The program will automatically read the data of the other detectors.")' )
      stop
   end if


   call GET_COMMAND_ARGUMENT ( 1, FirstArg )


   !  Prepare the full set of filenames, for all detectors,
   !  and add the terminating C Null character.

   !  search from the end of the string for the detector name portion:
   pos = index ( FirstArg, "_b0_", .TRUE. )
   if ( pos == 0 )  pos = index ( FirstArg, "_B0_", .TRUE. )

   if ( pos == 0 ) then
      write (*, '( / "The program could not locate _b0_ in the filename: ", A )' )  trim (FirstArg)
      write (*, '( "You should provide the name of the BGO 0 TTE data file." )' )
      stop
   end if

   pos = pos + 1

   do ldet=0, MAX_GBM_DET - 1

      FileName = FirstArg

      write ( FileName (pos:pos+1), '( "n", z1 )' )  ldet

      !  Some (probably all) Fortran 95 compilers output hexadecimal (z) output letters in upper-case, whereas
      !  GBM TTE filenames are entirely in lower case.  (This doesn't matter for the usual filesystem of
      !  Mac OS X, but it matters for Linux filesystems.)
      call MakeLowerCase  ( FileName (pos:pos+1) )

      !  Special case code for the BGO detectors:
      if ( ldet == 12 )  FileName (pos:pos+1) = 'b0'
      if ( ldet == 13 )  FileName (pos:pos+1) = 'b1'

      !  Add terminating C Null character.
      InputFileNames_for_C (ldet) = trim (FileName) // C_NULL_CHAR

   end do  ! ldet



   !  assume a Unix OS and extract just the filename, omitting the path

   pos = index ( FirstArg, "/", .TRUE. )
   NamePortion = FirstArg (pos+1: )

   !   also remove the filetype, if we recognize it.
   !   If we don't recognize it, leave it on.  Its too risky to
   !   remove in that case.

   pos = index ( NamePortion, ".", .TRUE. )
   FileType = trim ( NamePortion (pos+1:) )
   call MakeLowerCase (FileType)

   if (  FileType == "fit" .OR. FileType == "fits"  .OR. FileType == "pha" )   &
      NamePortion = NamePortion (1:pos-1)

   return


end subroutine TTE_GetCmdLineInput_AllFileNames


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


subroutine  TTE_GetCmdLineInput_AUTO_Histo   &
     ( TriggerName, TimeMode, GraphTimeRange, BinWidth, ChannelRange, InputFileNames_for_C )

   use, intrinsic :: iso_c_binding, only: C_CHAR

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   character (len=11), intent (out) :: TriggerName
   integer, intent (out) :: TimeMode
   real (real64), dimension (2), intent (out) :: GraphTimeRange
   real (real64), intent (out) :: BinWidth
   integer (int32), dimension (2), intent (out) :: ChannelRange
   character (kind=C_CHAR, len=FILE_NAME_MAX_LEN), dimension(0: MAX_GBM_DET-1 ), intent (out) :: InputFileNames_for_C


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   ! internal variables:

   integer :: NumberCmdArguments

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   NumberCmdArguments = COMMAND_ARGUMENT_COUNT ()
   if ( NumberCmdArguments /= 8 ) then
      write ( *, '( / "Unexpected number of arguments on command line:", I3 )' )  NumberCmdArguments
      write ( *, '( "Usage: ProgName.exe  bnyymmhhfff  VerNum  TimeMode &
                  & GraphTimeStart  GraphTimeEnd  BinWidth  ChanStart  ChanEnd" )' )
      stop
   end if


   !  reads and processes command line arguments 1 to 8:

   call TTE_GetCmdLineInput_AUTO_Histo_CommonCode          &
             ( TriggerName, TimeMode, GraphTimeRange, BinWidth, ChannelRange, InputFileNames_for_C )


   return


end subroutine TTE_GetCmdLineInput_AUTO_Histo


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


subroutine  TTE_GetCmdLineInput_AUTO_DetSelection                                             &
     ( TriggerName, TimeMode, GraphTimeRange, BinWidth, ChannelRange, InputFileNames_for_C ,  &
       RequestDetectors, NumRequestDetectors, SingleRequestDetector, Single_DetType_Selected, DetectorSelectionLabel )

   use, intrinsic :: iso_c_binding, only: C_CHAR
   use TTE_constant_defs

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   character (len=11), intent (out) :: TriggerName
   integer, intent (out) :: TimeMode
   real (real64), dimension (2), intent (out) :: GraphTimeRange
   real (real64), intent (out) :: BinWidth
   integer (int32), dimension (2), intent (out) :: ChannelRange
   character (kind=C_CHAR, len=FILE_NAME_MAX_LEN), dimension(0: MAX_GBM_DET-1 ), intent (out) :: InputFileNames_for_C
   logical, dimension ( 0: MAX_GBM_DET-1 ), intent (out) :: RequestDetectors
   integer (int32), intent (out) :: NumRequestDetectors
   integer (int32), intent (out) :: SingleRequestDetector  ! only applicable if user requests a single detector
   logical, intent (out) :: Single_DetType_Selected
   character (len=DETECTOR_SELECTION_LABEL_LEN), intent (out) :: DetectorSelectionLabel

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   ! internal variables:

   integer :: NumberCmdArguments
   character (len=30) :: string
   integer :: ldet
   character (len=MAX_GBM_DET) :: DetMask

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   NumberCmdArguments = COMMAND_ARGUMENT_COUNT ()
   if ( NumberCmdArguments /= 9 ) then
      write ( *, '( / "Unexpected number of arguments on command line:", I3 )' )  NumberCmdArguments
      write ( *, '( "Usage: ProgName.exe  bnyymmhhfff  VerNum  TimeMode &
                  & GraphTimeStart  GraphTimeEnd  BinWidth  ChanStart  ChanEnd  DetMask" )' )
      stop
   end if


   !  reads and processes command line arguments 1 to 8:

   call TTE_GetCmdLineInput_AUTO_Histo_CommonCode   &
     ( TriggerName, TimeMode, GraphTimeRange, BinWidth, ChannelRange, InputFileNames_for_C )


   !  Now the work unique to this subroutine -- process the last command line argument --
   !  a mask specifying which detectors should be summed into the histogram:

   call GET_COMMAND_ARGUMENT ( 9, string )

   DetMask = trim (string)
   if ( len_trim (string) /= MAX_GBM_DET ) then
      write ( *, '( "The DetectorMask argument should have exactly 14 characters: ''", A, "''" )' )  trim (DetMask)
      stop
   end if


   !  Process the command argument into the output variable RequestDetectors
   !  Detectors are numbered from zero, and array RequestDetectors is nicely zero-indexed.
   !  But I guess character positions in a character scaler always start at 1?

   do ldet=0, MAX_GBM_DET-1
      if ( DetMask (ldet+1:ldet+1) == '0' ) then
         RequestDetectors (ldet) = .false.
      else if ( DetMask (ldet+1:ldet+1) == '1' ) then
         RequestDetectors (ldet) = .true.
      else
         write ( *, '( "Invalid character in DetectorMask -- should only contain ''0'' or ''1'' chars: ''", A, "''" )' )     &
             trim (DetMask)
         stop
      end if
   end do


   ! Calculate various derived arguments from the basic output argument RequestDetectors:

   NumRequestDetectors = count (RequestDetectors)

   if ( NumRequestDetectors == 0 ) then
      write ( *, '( // "Must select at least one detector -- DetMask can be all zeros!" / )' )
      stop
   end if


   call Analyze_Detector_Selection   &
          ( RequestDetectors, NumRequestDetectors, SingleRequestDetector, Single_DetType_Selected, DetectorSelectionLabel )


   return


end subroutine TTE_GetCmdLineInput_AUTO_DetSelection


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


subroutine  Analyze_Detector_Selection    &
    ( RequestDetectors, NumRequestDetectors, SingleRequestDetector, Single_DetType_Selected, DetectorSelectionLabel )

   use TTE_constant_defs

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   logical, dimension ( 0: MAX_GBM_DET-1 ), intent (in) :: RequestDetectors
   integer (int32), intent (out) :: NumRequestDetectors
   integer (int32), intent (out) :: SingleRequestDetector  ! only applicable if user requests a single detector
   logical, intent (out) :: Single_DetType_Selected
   character (len=DETECTOR_SELECTION_LABEL_LEN), intent (out) :: DetectorSelectionLabel

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   ! internal variables:

   integer :: ldet
   logical :: NaI_Det_Selected
   logical :: BGO_Det_Selected
   integer :: pos

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   NumRequestDetectors = count (RequestDetectors)


   NaI_Det_Selected = any ( RequestDetectors (0:11) )
   BGO_Det_Selected = RequestDetectors(12) .OR. RequestDetectors(13)
   Single_DetType_Selected = NaI_Det_Selected .NEQV. BGO_Det_Selected       ! neqv is xor


   SingleRequestDetector = -999

   if ( NumRequestDetectors == 1 ) then
      do ldet=0, MAX_GBM_DET - 1
         if ( RequestDetectors (ldet) )  SingleRequestDetector = ldet
      end do
   end if


   !  Create a label listing the detectors selected.
   !  Identify several special simple cases: All detectors, all NaI, both BGO;
   !  otherwise build up a list of the particular detectors selected.

   if ( NumRequestDetectors == 14 ) then

      !  Simple case: all detectors:
      DetectorSelectionLabel = "All 14 GBM Detectors"

   else if ( Single_DetType_Selected  .AND.  BGO_Det_Selected  .AND.  NumRequestDetectors == 2 ) then

      !  Simple Case: Both BGO
      DetectorSelectionLabel = "Both GBM BGO Detectors"

   else if ( Single_DetType_Selected  .AND. NumRequestDetectors == MAX_NAI_DET ) then

      ! Simple Case: All NaI
      DetectorSelectionlabel = "All 12 GBM NaI Detectors"

   else

      !  complex case: either: 1) not both BGO, 2) not all NaI, 3) or mixture of NaI and BGO that
      !  isn't all 14 detectors.
      !  So build up a list of the detectors that were selected:

      DetectorSelectionLabel = " "
      pos = 1

      if ( NaI_Det_Selected ) then    ! any NaIs selected ?

         DetectorSelectionLabel = "NaI"
         pos = pos + 3
         do ldet=0, 9
            if (RequestDetectors (ldet) ) then
               write ( DetectorSelectionLabel (pos : pos+1), '( I2 )' )  ldet
               pos = pos + 2
            end if
         end do
         do ldet=10, 11
            if (RequestDetectors (ldet) ) then
               write ( DetectorSelectionLabel (pos : pos+2), '( I3 )' )  ldet
               pos = pos + 3
            end if
         end do

      end if   ! any NaIs selected ?


      !  at least one NaI and at least one BGO ?
      if ( .NOT. Single_DetType_Selected ) then
         DetectorSelectionLabel (pos: pos+2) = " & "
         pos = pos + 3
      end if


      if ( BGO_Det_Selected ) then    ! any BGOs selected ?

         DetectorSelectionLabel (pos : pos+2) = "BGO"
         pos = pos + 3
         do ldet=12, 13
            if (RequestDetectors (ldet) ) then
               write ( DetectorSelectionLabel (pos : pos+1), '( I2 )' )  ldet - 12
               pos = pos + 2
            end if
         end do
      end if   ! any BGO selected ?

   end if


   return


end subroutine Analyze_Detector_Selection


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


subroutine  TTE_GetCmdLineInput_AUTO_Histo_CommonCode   &
     ( TriggerName, TimeMode, GraphTimeRange, BinWidth, ChannelRange, InputFileNames_for_C )

   use, intrinsic :: iso_c_binding, only: C_CHAR, C_NULL_CHAR

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   character (len=11), intent (out) :: TriggerName
   integer, intent (out) :: TimeMode
   real (real64), dimension (2), intent (out) :: GraphTimeRange
   real (real64), intent (out) :: BinWidth
   integer (int32), dimension (2), intent (out) :: ChannelRange
   character (kind=C_CHAR, len=FILE_NAME_MAX_LEN), dimension(0: MAX_GBM_DET-1 ), intent (out) :: InputFileNames_for_C


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   ! internal variables:

   character (len=30) FileName
   integer :: ldet
   integer :: VerNum
   character (len=30) :: string

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   call GET_COMMAND_ARGUMENT ( 1, TriggerName )

   if ( TriggerName (1:2) /= "bn" ) then
      write ( *, '( "TriggerName on command line should begin with ""bn"", received: """, A, """." )' )  TriggerName
      write ( *, '( "This program only processes TTE from Triggers!" )' )
      stop
   end if


   call GET_COMMAND_ARGUMENT ( 2, string )
   read ( string, * )  VerNum

   call GET_COMMAND_ARGUMENT ( 3, string )
   read ( string, * )  TimeMode

   call GET_COMMAND_ARGUMENT ( 4, string )
   read ( string, * )  GraphTimeRange (1)

   call GET_COMMAND_ARGUMENT ( 5, string )
   read ( string, * )  GraphTimeRange (2)

   call GET_COMMAND_ARGUMENT ( 6, string )
   read ( string, * )  BinWidth

   call GET_COMMAND_ARGUMENT ( 7, string )
   read ( string, * )  ChannelRange (1)

   call GET_COMMAND_ARGUMENT ( 8, string )
   read ( string, * )  ChannelRange (2)


   if ( GraphTimeRange (2) < GraphTimeRange (1) ) then
      write ( *, '( "Time Range of graph must increase.  Instead received:", 1pG14.5, 2x, 1pG14.5  )' )   &
           GraphTimeRange (1), GraphTimeRange (2)
      stop
   end if

   if ( BinWidth <= 0.0 ) then
      write ( *, '( "Histogram bin width must be postive.  Instead received:", 1pG14.5 )' )  BinWidth
      stop
   end if

   if ( ChannelRange (1) < 0  .or.  ChannelRange (2) >= NUM_SPEC_CHAN  .or.  &
        ChannelRange (1) > ChannelRange (2) ) then
      write ( *, '( "Invalid Channel Range:", I8, " to", I8 )' )  ChannelRange
      stop
   end if


   !  Reading command line input completed.

   !  Create the file names of the TTE fits files from the trigger name:

   do ldet=0, MAX_GBM_DET - 1

      write ( FileName, '( "glg_tte_n", z1, "_", A11, "_v", I2.2, ".fit" )' )  ldet, TriggerName, VerNum

      !  Some (probably all) Fortran 95 compilers output hexadecimal (z) output letters in upper-case, whereas
      !  GBM TTE filenames are entirely in lower case.  (This doesn't matter for the usual filesystem of
      !  Mac OS X, but it matters for Linux filesystems.)
      call MakeLowerCase  (FileName)

      !  Above code writes detector 12 as "nc", but NaI C doesn't exist, should be BGO 0 = "b0";
      !  similarly replace "nd" with "b1"
      if ( ldet == 12 )  FileName (9:10) = 'b0'
      if ( ldet == 13 )  FileName (9:10) = 'b1'

      !  Add terminating C Null character.
      InputFileNames_for_C (ldet) = trim (FileName) // C_NULL_CHAR

   end do


   return


end subroutine TTE_GetCmdLineInput_AUTO_Histo_CommonCode


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


subroutine  MakeOutputFileName ( InputFileName, Suffix, OutputFileName )

   character (len=FILE_NAME_MAX_LEN), intent (in) :: InputFileName
   character (len=*), intent (in) :: Suffix
   character (len=FILE_NAME_MAX_LEN), intent (out) :: OutputFileName

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   ! internal variables:

   integer :: pos

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   pos = index ( InputFileName, ".", back=.TRUE. )   ! find last period in InputFileName

   if ( pos == 0 )  then
      write ( *, '( "InputFileName lacks a period: ''", A, "''." )' )  InputFileName
      stop
   end if

   if ( pos + len (Suffix)  >  FILE_NAME_MAX_LEN ) then
      write ( *, '( "InputFileName is too long: ''", A, "''." )' )  InputFileName
      stop
   end if

   OutputFileName = InputFileName (1:pos) // Suffix


   return


end subroutine MakeOutputFileName


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  This subroutine creates energy channel labels in "nice" human readable form.
!  Input is the first and last energy channel of the selected range.
!  Additional input is the lower and upper energy edges of all of the energy channels.
!  This routine locates the lower energy edge of the first selected energy channel, and
!  the upper energy edge of the last selected energy channel.
!  It writes these to strings with units of either "keV" (native units of values input to
!  this routine) or "MeV" (rescaled for large values).   The output has a sensible small
!  number of digits, i.e., almost always two.    Below 10 keV, only one digit.
!  A very small number of signficant digits is appropriate because this routine is
!  commonly used to label the output of several detectors -- the detectors won't have
!  exactly the same gain and so the energy edges of the channels can't be precisely
!  defined.  The input to this routine is typically the energy edges for one of the
!  detectors.  Accuracy could be improved by averaging the edges for the various detectors,
!  but the calling code doesn't use that method.
!  Works for energies up to 99,000 keV = 99 MeV.


subroutine  MakeEnergyLabels ( EMin_of_chan, EMax_of_chan, Left_Channel, Right_Channel, Left_Energy_Label, Right_Energy_Label )

   use TTE_constant_defs
   use, intrinsic :: iso_c_binding, only: C_FLOAT

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   real (C_FLOAT), dimension (0: NUM_SPEC_CHAN-1), intent (in) :: EMin_of_chan
   real (C_FLOAT), dimension (0: NUM_SPEC_CHAN-1), intent (in) :: EMax_of_chan
   integer, intent (in) :: Left_Channel
   integer, intent (in) :: Right_Channel
   character (len=ENERGY_LABEL_LEN), intent (out) :: Left_Energy_Label
   character (len=ENERGY_LABEL_LEN), intent (out) :: Right_Energy_Label

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   ! internal variables:


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   Left_Energy_Label = EnergyLabel_Function  ( EMin_of_chan, Left_Channel )


   if ( Right_Channel == 127 ) then

      Right_Energy_Label = "infinity"

   else

      Right_Energy_Label = EnergyLabel_Function  ( EMax_of_chan, Right_Channel )

   end if


   return


end subroutine MakeEnergyLabels


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  See comments for subroutine MakeEnergyLabels.


function  EnergyLabel_Function ( Energy_of_Channel, Channel )

   use TTE_constant_defs
   use, intrinsic :: iso_c_binding, only: C_FLOAT

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   character (len=ENERGY_LABEL_LEN) :: EnergyLabel_Function      ! function return

   real (C_FLOAT), dimension (0: NUM_SPEC_CHAN-1), intent (in) :: Energy_of_Channel
   integer, intent (in) :: Channel


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   ! internal variables:

   real (C_FLOAT) :: FltTmp
   character (len=ENERGY_LABEL_LEN + 1) :: TmpString

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   if ( Energy_of_Channel (Channel) <= 0.0 ) then

      EnergyLabel_Function = "BAD"

   else

      ! nominal boundary 10,000 keV = 10 MeV, but 9,950 keV gets rounded up to 10,000 keV:
      if ( Energy_of_Channel (Channel) >= 9950.0 ) then

         write ( EnergyLabel_Function, '( F3.0, " MeV" )' )  Energy_of_Channel (Channel) / 1000.0

      ! nominal boundary 1,000 keV = 1 MeV, but 995 keV gets rounded up to 1000 keV:
      else if ( Energy_of_Channel (Channel) >= 995.0 ) then

         !  We want to use F3.1, which is actually large enough, but ifort emits an irrittating
         !  warning message -- ifort doesn't "know" that the value is positive and so needs to
         !  reserve space for a leading minus sign.   So to suppress the warning message, use F4.1
         !  into a larger string, and then copy the desired 3.1 field into the output string.

         write ( TmpString, '( F4.1, " MeV" )' )  Energy_of_Channel (Channel) / 1000.0
         EnergyLabel_Function = TmpString (2: ENERGY_LABEL_LEN + 1)

      ! nominal boundary 100 keV, but 99.5 keV gets rounded up to 100 keV:
      else if ( Energy_of_Channel (Channel) >= 99.5 ) then

         ! F4.0 format would show 3 figures, so we remove one....
         FltTmp = nint (Energy_of_Channel (Channel) / 10.0) * 10.0
         write ( EnergyLabel_Function, '( F4.0, " keV" )' )  FltTmp

      else

         write ( EnergyLabel_Function, '( F3.0, " keV" )' )  Energy_of_Channel (Channel)

      end if

   end if

   EnergyLabel_Function = adjustl (EnergyLabel_Function)


   return


 end function EnergyLabel_Function


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  Revised 2016 Jan 26.  The user need no longer inform this routine whether not
!  the input file is for trigger data. This routine determines that by attempting
!  to read the trigger time.

!  This subroutine reads Level 1 TTE by calling C-routine Read_Lvl1_TTE that calls CFITSIO.
!  It converts the C-pointers to the TTE data into pointers to Fortran arrays -- this
!  is necessary for the arrays contain the TTE channel number and time value data because
!  we don't know the number of TTE events in advance of reading the file and so these arrays
!  have to be dynamically allocated in the C-routine via malloc.   Since they can't be
!  declared in advance, they have to be handled as pointers.
!  Next, the lists of TTE events are passed to Fortran subroutine Bin_TTE_Events, which
!  bins them into histograms, including deadtime corrected histograms.
!  Finally, we deallocate all items that are no longer needed.

!  The primary output is via setting the shared arrays of the module TTE_Histogram_Arrays.
!  Also outputs energy edges of the channels of whichever detectors were read in two arguments.
!  Some callers input ChannelRange to this routine, while other callers request this routine to
!  ask for user input, so that ChannelRange is output -- thus ChannelRange is an "inout" argument
!  -- the choice is indicated by argument "Determine_Channel_Range".

!  Based on earlier code, refactored out of programs into this subroutine.
!  Michael S. Briggs, 2009 May 29.

!  Revised by MSB, 2011 March 6.   For trigger TTE, the data is binned relative to the trigger time,
!  as before.  The revision: for continuous TTE, the user is prompted for a "reference time" to
!  use as the "base time" relative to which the data will be binned.

!  Revised by M. S. Briggs, 2011 Feb 14.
!  The times read into "tte_time_Cptr" are raw times from the data file, uncorrected for the
!  drift of the GBM DPU clock.   The output histograms to the shared variables of module
!  TTE_Histogram_Arrays are optionally corrected by the clock drift -- see routine Bin_TTE_Events.


subroutine  Read_and_Bin_TTE  ( InputFileNames_for_C, GraphTimeRange_s, BinWidth_s, RequestDetectors,   &
                 Single_DetType_Selected, Determine_Channel_Range, GapTime, ClockDrift_Flag,            &
                 ChannelRange, EMin_of_chan, EMax_of_chan, ReferenceTime, TimeRefType )


   use, intrinsic :: iso_c_binding

   use Read_Lvl1_TTE_Interface

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   character (kind=C_CHAR, len=FILE_NAME_MAX_LEN), dimension(0: MAX_GBM_DET-1 ), intent (in) :: InputFileNames_for_C
   real (real64), dimension (2), intent (in) :: GraphTimeRange_s
   real (real64), intent (in) :: BinWidth_s
   logical, dimension ( 0: MAX_GBM_DET-1 ), intent (in) :: RequestDetectors
   logical, intent (in) :: Single_DetType_Selected
   logical, intent (in) :: Determine_Channel_Range

   real (real32), intent (in) :: GapTime
   logical, intent (in) :: ClockDrift_Flag

   integer (int32), dimension (2), intent (inout) :: ChannelRange

   real (real32), dimension ( 0: NUM_SPEC_CHAN-1, 0: MAX_GBM_DET-1 ), intent (out) :: EMin_of_chan
   real (real32), dimension ( 0: NUM_SPEC_CHAN-1, 0: MAX_GBM_DET-1 ), intent (out) :: EMax_of_chan

   !  This is the TriggerTime, read from the FITS file, for trigger data.
   !  For continuous / non-trigger data, it is a user input value.
   real (real128), intent (out) :: ReferenceTime

   character (len=9), intent (out) :: TimeRefType


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   real (C_DOUBLE) :: TriggerTime
   logical :: IsTrigger

   integer (C_LONG) :: ldet

   real (C_DOUBLE) :: FileStartTime, FileStopTime

   integer (C_SHORT), dimension (:), pointer :: tte_channels => null ()
   real (C_DOUBLE), dimension (:), pointer :: tte_times => null ()

   real (C_FLOAT), dimension ( 0: NUM_SPEC_CHAN-1 ) :: EMin_of_chan_1det
   real (C_FLOAT), dimension ( 0: NUM_SPEC_CHAN-1 ) :: EMax_of_chan_1det

   integer (C_LONG) :: Num_TTE
   character (kind=C_CHAR, len=11) :: FITS_DataSum
   type (C_PTR) :: tte_chan_Cptr
   type (C_PTR) :: tte_time_Cptr

   character (len=34) :: WidgetTitle

   logical :: FirstDet = .TRUE.


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   !  loop through detectors, reading TTE data files, binning into histograms,
   !  both for each detector ("ByDet") and summed over all det.  Calculate deadtime corrections.

   do ldet=0, MAX_GBM_DET - 1

      if ( RequestDetectors (ldet) ) then


         !  Read TTE via C subroutine that uses CFITSIO:

         call Read_Lvl1_TTE  ( InputFileNames_for_C  (ldet),                                &
                 ldet, FileStartTime, FileStopTime, TriggerTime, Num_TTE, FITS_DataSum,     &
                 tte_chan_Cptr, tte_time_Cptr, EMin_of_chan_1det, EMax_of_chan_1det )

         IsTrigger = .FALSE.
         TimeRefType = "Reference"
         if ( TriggerTime > 0.0_C_DOUBLE ) then
            IsTrigger = .TRUE.
            TimeRefType = "Trigger"
         end if

         EMin_of_chan ( :, ldet ) = EMin_of_chan_1det
         EMax_of_chan ( :, ldet) = EMax_of_chan_1det


         !  Tricky case -- variable length arrays that had memory allocated in the C subroutine
         !  using malloc -- from C we receive pointers to that memory -- now associate those
         !  pointers to Fortran arrays of the correct shape and size:

         call c_f_pointer ( tte_chan_Cptr, tte_channels, [ Num_TTE ]  )
         call c_f_pointer ( tte_time_Cptr, tte_times, [ Num_TTE ]  )



         !  Some callers to this routine have pre-determined ChannelRange, other callers want
         !  this routine to have the user determine ChannelRange based on the data read from the
         !  first detector.   The choice is indicated by "Determine_Channel_Range".
         !  Whether this is the first detector from which data is read is indicated by "FirstDet".
         !  In the mode that the user selects the channels, if Single_DetType_Selected is TRUE,
         !  they will be basing their choice on the energy edges of the first detector that is
         !  read.  If more than one detector is to be read, the others may have slightly different
         !  energy edge values.
         !  New feature: if this is continuous TTE, then the user needs to input a "Reference Time"
         !  to be used in lieu of the Trigger Time.

         if ( FirstDet ) then

            FirstDet = .FALSE.

            !  Input energy channel range to plot via DISLIN widget:
            write ( WidgetTitle, '( "Select Energy Channels for Det", I3, ":" )' )  ldet
            if ( Determine_Channel_Range )  call Select_EChannels  ( Single_DetType_Selected,    &
                       EMin_of_chan_1det, EMax_of_chan_1det, WidgetTitle, ChannelRange )

            ReferenceTime = TriggerTime
            if ( .NOT. IsTrigger ) then

               write (*, '( "The file for detector", I3, " contains data from", F14.3, " to", F14.3, " s." )' )    &
                   ldet, FileStartTime, FileStopTime
               write (*, '( "Input a Reference Time which will serve as a ""base"" to refer times to." )' )
               write (*, '( "For ground-search TGFs, this could be the MET of the bin of the candidate." )' )
               write (*, '( "Input the Reference Time: " )', advance="no" )
               read (*, *)  ReferenceTime

               if ( ReferenceTime < FileStartTime  .OR.  ReferenceTime > FileStopTime )    &
                  write (*, '( / "WARNING!! -- Time", F15.3, " is outside of the time range of the file !" / )' )  ReferenceTime

            else
               write (*, '( "This is a file for a triggered event. Trigger time is: ", F18.6 )' )  TriggerTime
            end if

         end if


         !  Bin the TTE events into Histograms suitable for plotting.
         !  Includes calculating the Deadtime correction.
         !  All output of subroutine Bin_TTE_Events is into the arrays of the module TTE_Histogram_Arrays.

         call Bin_TTE_Events  ( ldet, Num_TTE, ReferenceTime, tte_channels, tte_times, GraphTimeRange_s, BinWidth_s,     &
                  ChannelRange, GapTime, ClockDrift_Flag )


         !  Cleanup -- release the FITS file, free the malloc'd arrays in C to release the storage, and
         !  nullify the Fortran pointers to these arrays.
         !  It would be illegal to use "deallocate" on the Fortran arrays with the pointer attribute,
         !  tte_channels and tte_times, because they were not created with via "allocate", but
         !  were instead assocated with pre-existing C-pointers via "c_f_pointer".   Clean_TTE_FitsFile
         !  will free the corresponding memory.

         call Cleanup_TTE_FitsFile ( ldet )

         nullify (tte_channels)
         nullify (tte_times)


      end if     ! RequestDetector ?

   end do    !  ldet


   if ( IsTrigger )  call CheckTriggerTime  ( TriggerTime )


   return


end subroutine  Read_and_Bin_TTE


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


subroutine Output_E_edges ( ldet, TTE_Data )

   use Type_for_all_TTE_Events

   integer (int32), intent (in) :: ldet
   type (All_TTE_Events_type), dimension (0: MAX_GBM_DET-1), intent (in) :: TTE_Data


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   integer :: out_LUN
   character (len=200) :: FileName
   integer :: kchan

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   write (FileName, '( "ChanEdges_N", Z1, ".txt" )' ) ldet
   if (FileName (12:12) == "C" )  FileName (11:12) = "B0"
   if (FileName (12:12) == "D" )  FileName (11:12) = "B1"

   open ( newunit=out_LUN, file=FileName, status="replace", form="formatted", access="sequential", action="write" )

   write (out_LUN, '( "The energy boundaries in keV of the 128 channels of GBM detector " )', advance="no" )
   if ( ldet <= 11 ) then
      write (out_LUN, '( "NaI ", I2 )' )  ldet
   else
     write (out_LUN, '( "BGO ", I1 )' )  ldet - 12
   end if

   do kchan=0, 126

      if ( ldet <= 11 ) then
         write (out_LUN, '( I3, 2X, F6.1, " - ", F6.1 )' )    &
             kchan, TTE_Data (ldet) % EMin_of_chan (kchan), TTE_Data (ldet) % EMax_of_chan (kchan)
      else
         write (out_LUN, '( I3, 2X, F6.0, " - ", F6.0 )' )    &
             kchan, TTE_Data (ldet) % EMin_of_chan (kchan), TTE_Data (ldet) % EMax_of_chan (kchan)
      end if

   end do

   !  special case the overflow channel:
   if ( ldet <= 11 ) then
      write (out_LUN, '( I3, 2X, ">= ", F6.1  )' )    &
          127, TTE_Data (ldet) % EMin_of_chan (127)
   else
      write (out_LUN, '( I3, 2X, ">= ", F6.0 )' )    &
          127, TTE_Data (ldet) % EMin_of_chan (127)
   end if


   close (out_LUN)


end subroutine Output_E_edges


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  The subroutine reads the TTE data for the requested detectors.  Unlike Read_and_Bin_TTE,
!  it returns all of the TTE data -- the "raw" event list TTE data.   The data is returned
!  in the item TTE_Data (14) -- this is a vector, each of of which is a structure.
!  Inside the structure there are pointers to arrays listing the channels and times of individual
!  TTE events.   There are two reasons for using this data structure: 1) it is good programming
!  practice to package related data into a structure, and 2) Fortran 95 doesn't allow delaring
!  arrays of pointers, but it does allow declaring arrays of structures that contain pointers,
!  which accomplishes the same.

!  The times returned in TTE_Data are NOT corrected for the drift of the GBM DPU clock.


subroutine  Return_all_TTE  ( InputFileNames_for_C, RequestDetectors, TTE_Data, TriggerTime, TimeRefType )

   use, intrinsic :: iso_c_binding

   use Read_Lvl1_TTE_Interface
   use Type_for_all_TTE_Events

   implicit none

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   character (kind=C_CHAR, len=FILE_NAME_MAX_LEN), dimension(0: MAX_GBM_DET-1 ), intent (in) :: InputFileNames_for_C
   logical, dimension ( 0: MAX_GBM_DET-1 ), intent (in) :: RequestDetectors

   type (All_TTE_Events_type), dimension (0: MAX_GBM_DET-1), intent (out) :: TTE_Data

   integer (C_SHORT), dimension (:), pointer :: tte_channels_F95_ptr => null ()
   real (C_DOUBLE), dimension (:), pointer :: tte_times_F95_ptr => null ()

   !  positive for trigger data, negative for non-trigger (continuous) data
   real (C_DOUBLE), intent (out) :: TriggerTime

   character (len=9), intent (out) :: TimeRefType


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   integer (C_LONG) :: ldet

   character (kind=C_CHAR, len=11) :: FITS_DataSum

   type (C_PTR) :: tte_chan_Cptr
   type (C_PTR) :: tte_time_Cptr

   logical :: FirstDetRead


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   TriggerTime = -88.88

   FirstDetRead = .TRUE.

   !  Loop through detectors, the reading TTE data files....


   do ldet=0, MAX_GBM_DET - 1

      if ( RequestDetectors (ldet) ) then


         !  Read TTE via C subroutine that uses CFITSIO:

         call Read_Lvl1_TTE ( InputFileNames_for_C  (ldet), ldet,    &
                 TTE_Data (ldet) % FileStartTime, TTE_Data (ldet) % FileStopTime,                 &
                 TTE_Data (ldet) % TriggerTime, TTE_Data (ldet) % Num_TTE, FITS_DataSum,          &
                 tte_chan_Cptr, tte_time_Cptr, TTE_Data (ldet) % EMin_of_chan, TTE_Data (ldet) % EMax_of_chan )

         TriggerTime = TTE_Data (ldet) % TriggerTime   ! should be the same for all detectors of same event


         !  Tricky case -- variable length arrays that had memory allocated in the C subroutine
         !  using malloc -- from C we receive pointers to that memory -- now associate those
         !  pointers to Fortran arrays of the correct shape and size:
         !  The Fortran pointers to arrays are now stored inside a structure, because we want this
         !  subroutine to return data for (potentially) all GBM detectors, but Fortran 95 doesn't
         !  allow arrays of pointers -- the work-around is to create an arrays of structures that
         !  contain pointers.

         !  We can't associate / convert the C-address of the array created within C with a Fortran pointer/array
         !  member of a structure: 1) it doesn't work, and 2) I guess the addresses of items within
         !  structures are pre-determined and can't be set by c_f_pointer.
         !  So we have to associate the C-pointer addresses with temporary Fortran pointer/arrays,
         !  then explicity copy those arrays into the Fortran pointer/array member of the structure.
         !  This would NOT be necessary if we were not packaging the data inside of a structure / type,
         !  in order to return the data of multiple detectors.

         call c_f_pointer ( tte_chan_Cptr, tte_channels_F95_ptr,  [ TTE_Data (ldet) % Num_TTE ]  )
         call c_f_pointer ( tte_time_Cptr, tte_times_F95_ptr,  [ TTE_Data (ldet) % Num_TTE ]  )

         allocate  ( TTE_Data (ldet) % tte_channels  ( 1: TTE_Data (ldet) % Num_TTE ) )
         allocate  ( TTE_Data (ldet) % tte_times  ( 1: TTE_Data (ldet) % Num_TTE ) )

         TTE_Data (ldet) % tte_channels = tte_channels_F95_ptr
         TTE_Data (ldet) % tte_times = tte_times_F95_ptr


         !  Cleanup -- nullify the Fortran pointers which are really the C-arays that we are about to free;
         !  release the FITS file and free the malloc'd arrays in C to release the storage.
         !  It would be illegal to use "deallocate" on the Fortran arrays with the pointer attribute,
         !  tte_channels_F95_ptr and tte_times_F95_ptr, because they were not created with via "allocate", but
         !  were instead assocated with pre-existing C-pointers via "c_f_pointer".   Clean_TTE_FitsFile
         !  will free the corresponding memory.

         nullify  ( tte_channels_F95_ptr )
         nullify  ( tte_times_F95_ptr )
         call Cleanup_TTE_FitsFile ( ldet )


      end if     ! RequestDetector ?

   end do    ! detector loop


   TimeRefType = "Reference"
   if ( TriggerTime > 0.0_C_DOUBLE )  TimeRefType = "Trigger"


   return


end subroutine  Return_all_TTE


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  MSB, 2009 March 7: moved all user input to calling routine so that this subroutine
!  can be used in programs that are run automatically, without an interactive user.

! MSB, 2009 August 15: 1) eps file format option, 2) increase thickness of lines, sizes of fonts, etc.,
! for better appearence of graphs at publication size, 3) Output filename is now an input.

! MSB, 2009 Dec 2: correct bug re lower limit of right Y-axis: case of minimum other than zero wasn't
! handled.


#ifndef NODISLIN
subroutine  Make_TTE_OneHisto  ( Data_Histogram, Deadtime_Histogram, GraphTimeRange_s, BinWidth_s,                    &
             NumRequestDetectors, OutputFileName, TriggerName, TimeLabel, BinLabel, DetectorSelectionLabel,           &
             ChannelRangeLabel, xdiv, ydiv, ymin, ymax, PlotDeadtimeCorr, GraphFormat, RightAxis_is_PerDet,           &
             ClockDrift_Flag, GapTime, ReferenceTime, TimeRefType )

   use dislin


   integer (int32), dimension (0:), intent (in) :: Data_Histogram
   real (real32), dimension (0:), intent (in) :: Deadtime_Histogram
   real (real64), dimension (2), intent (in) :: GraphTimeRange_s
   real (real64), intent (in) :: BinWidth_s
   integer (int32), intent (in) :: NumRequestDetectors
   character (len=*), intent (in) :: OutputFileName
   character (len=*), intent (in) :: TriggerName
   character (len=2), intent (in) :: TimeLabel
   character (len=2), intent (in) :: BinLabel
   character (len=*), intent (in) :: DetectorSelectionLabel
   character (len=*), intent (in) :: ChannelRangeLabel
   real (real32), intent (in) :: xdiv
   real (real32), intent (in) :: ydiv
   real (real32), intent (in) :: ymin
   real (real32), intent (in) :: ymax
   character (len=1), intent (in) :: PlotDeadtimeCorr
   character (len=3), intent (inout) :: GraphFormat    ! "out" is to allow fixing case
   logical, intent (in) :: RightAxis_is_PerDet

   logical, intent (in) :: ClockDrift_Flag
   real (real32), intent (in) :: GapTime
   real (real128), intent (in) :: ReferenceTime
   character (len=9), intent (in) :: TimeRefType


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   ! internal variables:

   integer (int32) :: Bins_in_Histogram


   character (len=80) :: X_AxisLabel
   character (len=40) :: Y_AxisLabel

   integer (int32) :: BinNum

   real (real32), dimension (2) :: TimeRange_for_Call

   real (real32) :: xleft, xright

   integer (int32), parameter :: BAD_VALUE = -9999_int32
   integer (int32) :: top, top_next, top_prev

   real (real32) :: DeadtimeCorr_top, DeadtimeCorr_top_next, DeadtimeCorr_top_prev

   real (real32) :: yrate_max
   real (real32) :: ydiv_CR
   real (real32) :: factor

   integer :: NX_Left, NX_Right, NY_Bot, NY_Top, NX_Len, NY_Len

   character (len=1) :: response

   real (real32) :: Start_X_Label
   character (len=50) :: LongResponse

   real (real64) :: tmp_DBL


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   call MakeLowerCase ( GraphFormat )

   Bins_in_Histogram = size (Data_Histogram)


   !  configure and initiate the plot:

   call setfil ( trim (OutputFileName) )
   call metafl ( GraphFormat )

   ! Normally PNG plots white lines and characters on black background -- ./Mreverse to black on white:
   ! Also override default pixel dimensions to have same aspect ratio as logical pixel dimensions
   if ( GraphFormat == 'png' ) then
      call scrmod ( "REVERS" )
      call winsiz ( 853, 660 )  ! physical pixel dimensions, override default of 603 x 853
   end if


   call setpag ( 'USAL' )      ! US A (=Letter), Landscape, 2790 x 2160 points


   call filmod ( 'DELETE' )    ! overwrite (delete) previous file

   call disini ()

   if ( GraphFormat == 'pdf' .or. GraphFormat == 'eps' )  call psfont ( 'Helvetica' )


   !  Make all lines thicker, including symbols; default is one; doesn't apply to characters.
   call linwid ( 5 )


   !  axes and axes labels:

   call labdig ( -2, 'X' )  ! DISLIN (>=9.4?) automatically calculates number of digits after decimal point on X-axis labels
   call labdig ( 0, 'Y' )   ! zero digits after decimal point on Y-axis labels

   if ( BinLabel == "ms" ) then
      write ( Y_AxisLabel, '( "Counts per ", F6.1, 1X, A2 )' )  BinWidth_s * 1000.0_real64, BinLabel
   else
      write ( Y_AxisLabel, '( "Counts per ", F5.0, 1X, A2 )' )  BinWidth_s * 1.0E+6_real64, BinLabel
   end if


   X_AxisLabel = Make_X_AxisLabel ( TimeRefType, TimeLabel, ClockDrift_Flag, GapTime, ReferenceTime )


   call name ( X_AxisLabel,  'X' )
   call name ( Y_AxisLabel, 'Y' )


   !  Draw graph box / axes:

   !  Primary Y-units are counts per bin -- this is used for the LEFT axis.
   !  On the RIGHT axis alternative units are plotted: count rate.


   !  increase size of axis titles (aka names) from default of 36,
   !  and from previous (long used) value of 46:
   !  2011 March 6 ... the X-axis label is now much longer, so decrease from 55.
   call hname ( 50 )

   ! increase the size of characters from default of 36:
   call height ( 50 )


   !  increase length of major and minor ticks from defaults of 24 & 16:
   call ticlen ( 33, 22 )


  if ( TimeLabel == "s " ) then
     TimeRange_for_Call = real ( GraphTimeRange_s, real32 )
     factor = 1.0_real32
  else
     !  X-axis will be in ms -- all input and calculations are in seconds -- the next two
     !  variables take care of converions to make the plot output correctly:
     TimeRange_for_Call = real ( GraphTimeRange_s * 1000.0_real64, real32 )
     factor = 1000.0_real32
   end if

   !  Suppress drawing the RIGHT Y-axis so that we can replace it with a different axis, with count-rate units.
   call setgrf ( "NAME", "NAME", "TICKS", "NONE" )

   !  Draw the graph box: both X-axes, and the LEFT Y-axis with counts per bin units.  These Y-units are
   !  the actual units of the plot, used to draw data.

   9030 continue
   Start_X_Label = TimeRange_for_Call (1)
   write (*, '( / "Input value for first tick label on time axis, or ENTER to use", 1PG14.4, ": " )', advance='no' ) Start_X_Label
   read (*, '(A)', err=9030 )  LongResponse
   if ( len_trim (LongResponse) /= 0 ) then
      read (LongResponse, *, err=9030)  Start_X_Label
      if ( Start_X_Label < TimeRange_for_Call (1)  .OR.  Start_X_Label > TimeRange_for_Call (2) ) goto 9030
   end if

   call graf ( TimeRange_for_Call (1), TimeRange_for_Call (2), Start_X_Label, xdiv, ymin, ymax, ymin, ydiv )


   !  plot secondary Y-axis on the LEFT with units of count rate.  These doesn't change the units that
   !  are used to draw inside the graph box -- those are counts per bin, as established by the values used
   !  in the call to graf.

   !  Determine positions / lengths in plot coordinates by converting from user units:

   NX_Left =  NXPOSN ( TimeRange_for_Call (1) )
   NX_Right = NXPOSN ( TimeRange_for_Call (2) )
   NY_Bot = NYPOSN (ymin)
   NY_Top = NYPOSN (ymax)

   NX_Len = NX_Right - NX_Left
   NY_Len = NY_Bot - NY_Top


   !  Determine parameters of the RIGHT y-axis scale.
   !  The maximum value, at the top of the scale, will be whatever value in count rate units
   !  that corresponds to the the top of the left y-axis scale.   The maximum value
   !  at the top of the axis might be an irregular value that won't get a tick mark.
   !  If "RightAxis_is_PerDet" is TRUE, the units are per detector, otherwise the
   !  units are count rate summed over the detectors.
   !  Of course, the distinction doesn't exist is there is only one detector.

   yrate_max = real ( ymax / BinWidth_s, real32 )
   if ( RightAxis_is_PerDet)  yrate_max = yrate_max / NumRequestDetectors

   !  Logic to:
   !  Rescale the count rate units between Hz (the starting point), kHz and MHz,
   !  and to select a "nice" value for the label spacing on the RIGHT y-axis.


   if ( yrate_max > 2000000.0 ) then

      yrate_max = yrate_max / 1.0E+6
      Y_AxisLabel = "Count Rate (MHz)"
      if ( NumRequestDetectors > 1  .AND.  RightAxis_is_PerDet )  Y_AxisLabel = "Count Rate per Detector (MHz)"

      !  For the right y-axis, when using MHz units, switch from zero digits after the decimal point
      !  on Y-axis labels to having dislin make the decision.   The maximum value for a "MHz" scale might
      !  be 1.5 or 2.0, and tick marks with non-integer values are extremely likely.
      call labdig ( -2, 'Y' )

   else if ( yrate_max > 12000.0 ) then

      yrate_max = yrate_max / 1000.0
      Y_AxisLabel = "Count Rate (kHz)"
      if ( NumRequestDetectors > 1  .AND.  RightAxis_is_PerDet )  Y_AxisLabel = "Count Rate per Detector (kHz)"

   else

      Y_AxisLabel = "Count Rate (Hz)"
      if ( NumRequestDetectors > 1  .AND.  RightAxis_is_PerDet )  Y_AxisLabel = "Count Rate per Detector (Hz)"

   end if


   if ( yrate_max > 8000.0 ) then
      ydiv_CR = 2000.0
   else if ( yrate_max > 4000.0 ) then
      ydiv_CR = 1000.0
   else if ( yrate_max > 2000.0 ) then
      ydiv_CR = 500.0
   else if ( yrate_max > 800.0 ) then
      ydiv_CR = 200.0
   else if ( yrate_max > 400.0 ) then
      ydiv_CR = 100.0
   else if ( yrate_max > 200.0 ) then
      ydiv_CR = 50.0
   else if ( yrate_max > 80.0 ) then
      ydiv_CR = 20.0
   else if ( yrate_max > 40.0 ) then
      ydiv_CR = 10.0
   else if ( yrate_max > 20.0 ) then
      ydiv_CR = 5.0
   else if ( yrate_max > 8.0 ) then
      ydiv_CR = 2.0
   else if ( yrate_max > 4.0 ) then
      ydiv_CR = 1.0
   else if ( yrate_max > 2.0 ) then
      ydiv_CR = 0.5
   else
      ydiv_CR = 0.2
   end if


   !  Finally draw the RIGHT Y-axis, with count-rate units.
   !  The units of this axis have nothing to do with how things are plotted in the graph
   !  within this subroutine.

   call yaxis ( 0.0, yrate_max, 0.0, ydiv_CR, NY_Len, Y_AxisLabel, 0, NX_Right, NY_Bot )


   !  Optionally plot a legend inside of the graph box:

   call PlotLegend  ( NX_Left, NX_Right, NY_Bot, NY_Top )



   !  Draw the histogram :

   do BinNum=0, Bins_in_Histogram - 1

      if ( Data_Histogram (BinNum) > 0 ) then   ! non-zero bin ?

         tmp_DBL = GraphTimeRange_s (1) + BinNum * BinWidth_s
         xleft = real ( tmp_DBL, real32 )

         tmp_DBL = xleft + BinWidth_s
         xright = real (tmp_DBL, real32 )


         !  last bin may be partial width:
         xright = min ( xright, real (GraphTimeRange_s (2), real32) )


         ! unit conversion if time axis is ms:
         xleft = xleft * factor
         xright = xright * factor

         !  First calculate the quantities for measured (non-deadtime corrected) histogram --
         !  but plotting is delayed until after the deadtime corrected histogram (RED)
         !  so that the measured (BLACK) histogram will be on top.
         !  These quantities are calculated in before they are plotted because they
         !  are plotted (BLACK) because they are also used to calculate the deadtime-
         !  corrected (RED) histogram.

         ! horizontal bar:
         top = Data_Histogram (BinNum)
         DeadtimeCorr_top = Deadtime_Histogram (BinNum)

         !  right side of bar:
         top_next = BAD_VALUE
         if ( BinNum /= Bins_in_Histogram - 1 ) then
            top_next = Data_Histogram (BinNum+1)
            DeadtimeCorr_top_next = Deadtime_Histogram (BinNum+1)
         end if

         ! The left side of the bar for this bin will already have been drawn as the
         ! right side of the previous bar, unless that bar wasn't drawn because
         ! that bin had zero counts:

         top_prev = BAD_VALUE
         if ( BinNum > 0 ) then
            if ( Data_Histogram (BinNum-1) == 0 ) then
               top_prev = Data_Histogram (BinNum-1)
               DeadtimeCorr_top_prev = Deadtime_Histogram (BinNum-1)
            end if
         end if



         if ( PlotDeadtimeCorr == 'Y' ) then   ! plot deadtime-corrected histogram ?

            !  Now plot the deadtime corrected histogram in RED:

            call color ( "RED" )

            ! horizontal bar:
            call rline ( xleft, DeadtimeCorr_top, xright, DeadtimeCorr_top )

            !  right side of bar:
            if ( top_next /= BAD_VALUE )  call rline ( xright, DeadtimeCorr_top, xright, DeadtimeCorr_top_next )

            ! left side of the bar:
            if ( top_prev /= BAD_VALUE )  call rline ( xleft, DeadtimeCorr_top_prev, xleft, DeadtimeCorr_top )

         end if    ! plot deadtime-corrected histogram ?


         call color ( "FORE" )   ! restore default color -- BLACK


         !  Finally plot the measure / non-deadtime corrected (BLACK) histogram.
         !  It is delayed until the end so that it will be on top of
         !  the deadtime-corrected (RED) histogram.

         ! horizontal bar:
         call rline ( xleft, real(top), xright, real(top) )

         !  right side of bar:
         if ( top_next /= BAD_VALUE )  call rline ( xright, real(top), xright, real(top_next) )


         !  left side of the bar:
         if ( top_prev /= BAD_VALUE )  call rline ( xleft, real(top_prev), xleft, real(top) )


      end if   ! non-zero bin ?

   end do    ! loop over Histogram bins



   !  add the titles:

   9010 continue
   write ( *, '( / "Output Titles above the Graph?  ""Y"" for Yes, or ENTER for Yes, or ""N"" for No: " )', advance='no' )
   read ( *, '(A)', err=9010 )  response
   if ( len_trim (response) == 0 )  response = 'Y'
   call MakeUpperCase (response)
   if ( response /= 'N' .and. response /= 'Y' ) goto 9010
   if ( response == 'Y' ) then

      call titlin ( TriggerName, 1 )
      call titlin ( DetectorSelectionLabel, 2 )
      call titlin ( ChannelRangeLabel, 3 )
      call title ()

   end if


   call disfin ()    ! finish plot


   return


end subroutine Make_TTE_OneHisto
#endif


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


subroutine TTE_OneHisto_2_file ( Data_Histogram, GraphTimeRange_s, BinWidth_s, NamePortion )

   integer (int32), dimension (0:), intent (in) :: Data_Histogram
   real (real64), dimension (2), intent (in) :: GraphTimeRange_s
   real (real64), intent (in) :: BinWidth_s
   character (len=11) :: NamePortion

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   character (len=21) :: FileName
   integer :: histo_LUN
   integer :: BinNum

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   FileName = trim (NamePortion) // ".Histo.txt"

   open ( file=FileName, newunit=histo_LUN, status="replace", action="write", access="sequential", form="formatted" )


   do BinNum=0, size (Data_Histogram) - 1

      write ( histo_LUN, '( I6, 4X, F18.6, 2X, F18.6, 4X, I10 )' )  BinNum, GraphTimeRange_s (1) + BinNum * BinWidth_s,    &
          GraphTimeRange_s (1) + (BinNum+1) * BinWidth_s, Data_Histogram (BinNum)

   end do


   close ( unit=histo_LUN )


end subroutine TTE_OneHisto_2_file


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  This subroutine is a callback routine used by Select_EChannels_Widget -- when
!  DISLIN push button widgets created by subroutine Select_EChannels_Widget are clicked
!  by the user, this subroutine is called.   Its single argument is the ID of the
!  widget that was clicked.

!  Subroutine Select_EChannels_Widget communicates "on the side" with this subroutine
!  via "module-scope" variables MODSHR2_EChan_Widget_Buttons_IDs and
!  MODSHR2_EChan_Widget_MiscVars.  Both variables are user defined type / structure with
!  several components -- see the top of this module for the type definitions.   This method
!  is necessary because DISLIN defines the argument list of this routine, so additional
!  information must be exchanged outside of the argument list.  The MODSHR2_EChan_Widget_Buttons_IDs
!  variable is used to inform this routine of the widget IDs of buttons that were created by
!  Select_EChannels_Widget.   Then when this routine receives a widget ID as an argument,
!  indicating that the push button with that widget ID was clicked, it can compare the value of
!  the argument to the component values in MODSHR2_EChan_Widget_Buttons_IDs to identify which push
!  button was clicked.
!  Having identified which push button was clicked, this routine can perform the appropriate actions.
!  Depending on the push button that was clicked, it alters the current channel range selection --
!  either the value of the first or last channel, and either by 1 or 10 channels.
!  (If the change results in an illegal channel range, the change is modified to be legal.)
!  This routine also updates the labels on the widget to show the revised channel range.


!  input:
!  argument ID_of_Activated_Widget,
!  shared variable MODSHR2_EChan_Widget_Buttons_IDs and MODSHR2_EChan_Widget_MiscVars.

!  output:
!  changing the values of MODSHR2_EChan_Widget_MiscVars%ChannelRange.
!  additional action:
!  rewriting the labels on the widget to reflect the revised channel range.
!  Also rewrites the energy range, if that label was created.


#ifndef NODISLIN
subroutine  Select_EChannels_Widget_CALLBACK ( ID_of_Activated_Widget )

   use dislin

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   integer, intent (in) :: ID_of_Activated_Widget

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   ! internal variables:

   integer :: jcolumn
   integer :: ReqChanStep = 999   ! to suppress warning msg
   integer :: ReqColumn = 999     ! to suppress warning msg
   logical :: RecognizedFlag

   character (len=16) :: Left_Channel_Label, Right_Channel_Label
   character (len=ENERGY_LABEL_LEN) :: Left_Energy_Label, Right_Energy_Label


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   RecognizedFlag = .false.


   !  Match the ID of push button widget that was clicked, ID_of_Activated_Widget (=input
   !  argument received by this subroutine), to IDs of buttons by name,
   !  MODSHR2_EChan_Widget_Buttons_IDs%xxxxx.   Then having recognized which
   !  push button was clicked, setup the changes to make according to that choice.

   do jcolumn=1, 2

      if ( ID_of_Activated_Widget == MODSHR2_EChan_Widget_Buttons_IDs%OneStepAdjustDown (jcolumn) ) then
         ReqChanStep = -1
         ReqColumn = jcolumn
         RecognizedFlag = .true.
      end if

      if ( ID_of_Activated_Widget == MODSHR2_EChan_Widget_Buttons_IDs%TenStepAdjustDown (jcolumn) ) then
         ReqChanStep = -10
         ReqColumn = jcolumn
         RecognizedFlag = .true.
      end if

      if ( ID_of_Activated_Widget == MODSHR2_EChan_Widget_Buttons_IDs%OneStepAdjustUp (jcolumn) ) then
         ReqChanStep = +1
         ReqColumn = jcolumn
         RecognizedFlag = .true.
      end if

      if ( ID_of_Activated_Widget == MODSHR2_EChan_Widget_Buttons_IDs%TenStepAdjustUp (jcolumn) ) then
         ReqChanStep = +10
         ReqColumn = jcolumn
         RecognizedFlag = .true.
      end if

   end do   ! jcolumn


   if ( .not. RecognizedFlag ) then
      write ( *, * )  "Subroutine Select_EChannels_Widget_CALLBACK: widget ID", ID_of_Activated_Widget, " not recognized!"
      stop
   end if


   if ( ReqColumn == 1 ) then

      ! implement user requested change, first channel case:
      MODSHR2_EChan_Widget_MiscVars%ChannelRange (1) = MODSHR2_EChan_Widget_MiscVars%ChannelRange (1) + ReqChanStep

      ! if user requested change resulted in illegal channel value, fix the problem:

      if ( MODSHR2_EChan_Widget_MiscVars%ChannelRange (1)  <  0 )  MODSHR2_EChan_Widget_MiscVars%ChannelRange (1) = 0

      if ( MODSHR2_EChan_Widget_MiscVars%ChannelRange (1)  >  MODSHR2_EChan_Widget_MiscVars%ChannelRange (2) )    &
           MODSHR2_EChan_Widget_MiscVars%ChannelRange (1)  =  MODSHR2_EChan_Widget_MiscVars%ChannelRange (2)

   else

      ! implement user requested change, last channel case:
      MODSHR2_EChan_Widget_MiscVars%ChannelRange (2) = MODSHR2_EChan_Widget_MiscVars%ChannelRange (2) + ReqChanStep

      ! if user requested change resulted in illegal channel value, fix the problem:

      if ( MODSHR2_EChan_Widget_MiscVars%ChannelRange (2)  >=  NUM_SPEC_CHAN )   &
           MODSHR2_EChan_Widget_MiscVars%ChannelRange (2) = NUM_SPEC_CHAN - 1

      if ( MODSHR2_EChan_Widget_MiscVars%ChannelRange (2)  <  MODSHR2_EChan_Widget_MiscVars%ChannelRange (1) )    &
           MODSHR2_EChan_Widget_MiscVars%ChannelRange (2)  =  MODSHR2_EChan_Widget_MiscVars%ChannelRange (1)

   end if


   !  Modify the labels on the widget to reflect the choice just made:

   write ( Left_Channel_Label,  '( "First Chan =", I4 )' )  MODSHR2_EChan_Widget_MiscVars%ChannelRange (1)
   write ( Right_Channel_Label, '( "Last Chan =", I4 )' )   MODSHR2_EChan_Widget_MiscVars%ChannelRange (2)

   call swgtxt ( MODSHR2_EChan_Widget_Buttons_IDs%EChan_Label (1), Left_Channel_Label )
   call swgtxt ( MODSHR2_EChan_Widget_Buttons_IDs%EChan_Label (2), Right_Channel_Label )


   !  Only re-write the energy value labels if that label was created:

   if ( MODSHR2_EChan_Widget_MiscVars%EValue_Widget_Created ) then

      call MakeEnergyLabels ( MODSHR2_EChan_Widget_MiscVars%EMin_of_chan, MODSHR2_EChan_Widget_MiscVars%EMax_of_chan,   &
              MODSHR2_EChan_Widget_MiscVars%ChannelRange (1), MODSHR2_EChan_Widget_MiscVars%ChannelRange (2),    &
                      Left_Energy_Label, Right_Energy_Label )

      call swgtxt ( MODSHR2_EChan_Widget_Buttons_IDs%EValue_Label (1), Left_Energy_Label )
      call swgtxt ( MODSHR2_EChan_Widget_Buttons_IDs%EValue_Label (2), Right_Energy_Label )

   end if


   return


end subroutine Select_EChannels_Widget_CALLBACK
#endif


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  This subroutine uses DISLIN to offer the user a widget to select the Energy Channel
!  range for the plot (or perhaps from some other use).
!  The initial range is 0 to NUM_SPEC_CHAN - 1 = 127.

!  The widget has two columns, which are very similar.
!  The left column is for the starting channel of the selection, the right column for the
!  ending channel of the selection.
!  From top to bottom, each column has:
!  Label with current value of channel selection,
!  if it makes sense, label with energy value of channel selection.
!     Energy value labels are only provided if the detector selection consists of detectors
!     of a single type -- if the detector selection has both NaI and BGO detectors,
!     there is no consistent meaning of energy versus channel number.
!     For the starting channel, the energy is the left edge, for the ending channel,
!     the energy value is the right edge.
!  Push buttons attached to callback routine Select_EChannels_Widget_CALLBACK as follows:
!  "<", "<<", ">" and ">>".
!  When a push button is clicked by the user, the callback routine is called,
!  and it modifies the current channel selection.
!  "<" causes the selection to decrease by 1, while "<<" causes a decrease of 10, etc.
!  When the "OK" button is clicked, execution advances in this routine past the call to wgfin.
!  This routine copies the current channel selection from the shared variable
!  MODSHR2_EChan_Widget_MiscVars%ChannelRange to its output argument ChannelRange.


!  Input arguments: Single_DetType_Selected -- whether a single detector type was selected.
!      EMin_of_chan and EMax_of_chan -- energy edges of the channels.
!  Output argument: ChannelRange -- the Channel range selected by the user.

!  Additional input: shared variable MODSHR2_EChan_Widget_MiscVars%ChannelRange, which is both
!  input and output.  The callback routine Select_EChannels_Widget_CALLBACK may modify the value.
!  Additional output: the shared variables MODSHR2_EChan_Widget_Buttons_IDs and
!  MODSHR2_EChan_Widget_Buttons_IDs which are used to communicate information to the
!  callback routine Select_EChannels_Widget_CALLBACK.


#ifndef NODISLIN
subroutine  Select_EChannels_Widget  ( Single_DetType_Selected, EMin_of_chan, EMax_of_chan, WidgetTitle, ChannelRange )


   use dislin

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   !  Input arguments:

   logical, intent (in) :: Single_DetType_Selected
   real (C_FLOAT), dimension (0: NUM_SPEC_CHAN-1), intent (in) :: EMin_of_chan
   real (C_FLOAT), dimension (0: NUM_SPEC_CHAN-1), intent (in) :: EMax_of_chan
   character (len=*), intent (in) :: WidgetTitle

   ! Output argument, an array indicating the Energy Channel Range selected by the user:

   integer (int32), dimension (2), intent (out) :: ChannelRange


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   ! internal variables:

   integer :: widget_parent_id, ok_widget_id

   integer :: jcolumn

   character (len=*), parameter :: HelpString =   &
      'The current (SPEC) Energy Channel Selection, and the corresponding Energy Range,|&
      &are shown at the top.   The energy values are for the first selected.|&
      &Energy values are omitted if both NaI and BGO detectors were selected.|&
      &Alter a Channel Selection by clicking on buttons below the Channel:|&
      &"<" or "<<" to decrease the value, ">" or ">>" to increase the value.|&
      &The buttons "<" and ">" change the channel by one; "<<" and ">>" by ten.|&
      &When done click "OK".'

   character (len=ENERGY_LABEL_LEN) :: Left_Energy_Label, Right_Energy_Label


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   !  Transfer information to the callback routine for this widget routine,
   !  Select_EChannels_Widget_CALLBACK.

   !  Initial values for the channel selection -- entire range.
   !  This subroutine will not further modify this variable.  The user will click on
   !  push buttons, which will cause the callback routine Select_EChannels_Widget_CALLBACK
   !  to change the values.  At the end of this routine, these variables are read by this
   !  routine and the values copied into the output arguments ChannelRange.

   MODSHR2_EChan_Widget_MiscVars%ChannelRange (1) = 0
   MODSHR2_EChan_Widget_MiscVars%ChannelRange (2) = NUM_SPEC_CHAN - 1

   !  Inform the callback routine of the energy boundaries of the channels:

   MODSHR2_EChan_Widget_MiscVars%EMin_of_chan = EMin_of_chan
   MODSHR2_EChan_Widget_MiscVars%EMax_of_chan = EMax_of_chan


   !  Setup the overall title, background & foreground color, a help string to display
   !  in the help box if Help is clicked, configure the parent widget:

   call swgtit ( WidgetTitle )
   call swgclr ( 0.55, 0.55, 0.85, 'BACK' )
   call swgclr ( 0.00, 0.00, 0.00, 'FORE' )
   call swghlp ( HelpString )

   call wgini ( "HORI", widget_parent_id )


   !  Configure two container widgets to contain other widgets in columns:

   call wgbas ( widget_parent_id, "VERT", MODSHR2_EChan_Widget_Buttons_IDs%WidgetColumn (1) )
   call wgbas ( widget_parent_id, "VERT", MODSHR2_EChan_Widget_Buttons_IDs%WidgetColumn (2) )


   !  Create label widgets with the (current) channel selections;

   call wglab ( MODSHR2_EChan_Widget_Buttons_IDs%WidgetColumn (1), "First Chan = 0",  &
                MODSHR2_EChan_Widget_Buttons_IDs%EChan_Label (1) )

   call wglab ( MODSHR2_EChan_Widget_Buttons_IDs%WidgetColumn (2), "Last Chan = 127",  &
                MODSHR2_EChan_Widget_Buttons_IDs%EChan_Label (2) )


   !  Additionally make label widgets with the energy ranges of the channel selections,
   !  if doing so makes sense.
   !  It only makes sense to display energy values if the detectors of a single detector type
   !  are being plotted.   If both NaI and BGO detectors are being plotted, then channels don't
   !  have the same energy values for the two detector types and we can't plot a value that
   !  has meaning for both detector types.

   if ( Single_DetType_Selected ) then

      MODSHR2_EChan_Widget_MiscVars%EValue_Widget_Created = .true.

      call MakeEnergyLabels ( EMin_of_chan, EMax_of_chan, 0, 127, Left_Energy_Label, Right_Energy_Label )

      call wglab ( MODSHR2_EChan_Widget_Buttons_IDs%WidgetColumn (1), Left_Energy_Label,  &
                   MODSHR2_EChan_Widget_Buttons_IDs%EValue_Label (1) )

      call wglab ( MODSHR2_EChan_Widget_Buttons_IDs%WidgetColumn (2), Right_Energy_Label,  &
                   MODSHR2_EChan_Widget_Buttons_IDs%EValue_Label (2) )

   else

      MODSHR2_EChan_Widget_MiscVars%EValue_Widget_Created = .false.

      !MODSHR2_EChan_Widget_Buttons_IDs%EValue_Label (1) = -999
      !MODSHR2_EChan_Widget_Buttons_IDs%EValue_Label (2) = -999

   end if


   !  Create push button widgets that offer user choices in altering the channel selections:
   !  These widgets won't be read below, but cause actions when the buttons are clicked on,
   !  as they are about to be connected to a callback routine:

   do jcolumn=1, 2
      call wgpbut ( MODSHR2_EChan_Widget_Buttons_IDs%WidgetColumn (jcolumn), "<",   &
                    MODSHR2_EChan_Widget_Buttons_IDs%OneStepAdjustDown (jcolumn) )

      call wgpbut ( MODSHR2_EChan_Widget_Buttons_IDs%WidgetColumn (jcolumn), "<<",  &
                    MODSHR2_EChan_Widget_Buttons_IDs%TenStepAdjustDown (jcolumn) )

      call wgpbut ( MODSHR2_EChan_Widget_Buttons_IDs%WidgetColumn (jcolumn), ">",   &
                    MODSHR2_EChan_Widget_Buttons_IDs%OneStepAdjustUp (jcolumn) )

      call wgpbut ( MODSHR2_EChan_Widget_Buttons_IDs%WidgetColumn (jcolumn), ">>",  &
                    MODSHR2_EChan_Widget_Buttons_IDs%TenStepAdjustUp (jcolumn) )
   end do

   !  Connect the EChannel adjustment push button widgets to a call back routine.
   !  If a push button is pushed, the call back routine
   !  is called -- the callback routine will change the EChannel values.

   do jcolumn=1, 2
      call swgcbk ( MODSHR2_EChan_Widget_Buttons_IDs%OneStepAdjustDown (jcolumn), Select_EChannels_Widget_CALLBACK )
      call swgcbk ( MODSHR2_EChan_Widget_Buttons_IDs%TenStepAdjustDown (jcolumn), Select_EChannels_Widget_CALLBACK )
      call swgcbk ( MODSHR2_EChan_Widget_Buttons_IDs%OneStepAdjustUp   (jcolumn), Select_EChannels_Widget_CALLBACK )
      call swgcbk ( MODSHR2_EChan_Widget_Buttons_IDs%TenStepAdjustUp   (jcolumn), Select_EChannels_Widget_CALLBACK )
   end do


   !  finalize and output the widget with wgfin:

   call wgok ( MODSHR2_EChan_Widget_Buttons_IDs%WidgetColumn (1), ok_widget_id )
   call wgfin


   !  When the user clicks OK, execution advances past wgfin --
   !  at which point we copy the Channel Range Selection, MODSHR2_EChan_Widget_MiscVars%ChannelRange,
   !  shared between this routine and its callback routine, Select_EChannels_Widget_CALLBACK,
   !  into the output argument of this routine.   The callback routine has modified
   !  MODSHR2_EChan_Widget_MiscVars%ChannelRange according to the push buttons that were clicked
   !  by the user.

   ChannelRange = MODSHR2_EChan_Widget_MiscVars%ChannelRange


   return


end subroutine Select_EChannels_Widget
#endif


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


subroutine  Select_EChannels  ( Single_DetType_Selected, EMin_of_chan, EMax_of_chan, WidgetTitle, ChannelRange )


   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   !  Input arguments:

   logical, intent (in) :: Single_DetType_Selected
   real (C_FLOAT), dimension (0: NUM_SPEC_CHAN-1), intent (in) :: EMin_of_chan
   real (C_FLOAT), dimension (0: NUM_SPEC_CHAN-1), intent (in) :: EMax_of_chan
   character (len=*), intent (in) :: WidgetTitle

   ! Output argument, an array indicating the Energy Channel Range selected by the user:

   integer (int32), dimension (2), intent (out) :: ChannelRange


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

#if  defined (NODISLIN)  ||  (defined NOGUI)
   character (len=50) :: string
#endif

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


#if  defined (NODISLIN)  ||  (defined NOGUI)

   !  primitive method if there are problems with dislin and the screen, i.e., problems
   !  with dislin & OpenMotif

   SelectEchan: do

      write (*, '( / "Select energy channel range within 0 to 127, or <CR> for 0 to 127: " )', advance='no' )

      read (*, '(A)' )  string
      if ( len_trim (string) == 0 ) then
         ChannelRange = [ 0, 127 ]
      else
         read (string, *)  ChannelRange
      end if

      if ( ChannelRange (1) >= 0  .AND.  ChannelRange (2) < NUM_SPEC_CHAN  .AND.   &
           ChannelRange (1) <= ChannelRange (2) ) then

         exit SelectEchan   ! good input

      else
         write (*, '( I0, ": ", I0, "are illegal values, please try again!" / )' )  ChannelRange
      end if

    end do SelectEchan

#else

   !  The preferred method of user input -- GUI menu -- this has a major advantage --
   !  if one detector type is selected, the user sees both channels and the energy values
   !  of those channels.

   call Select_EChannels_Widget  ( Single_DetType_Selected, EMin_of_chan, EMax_of_chan, WidgetTitle, ChannelRange )

#endif

   return

end subroutine Select_EChannels


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  MSB, rev. 2010 May 21 to add the Plus-X and Minus-X choices.

!  This subroutine is a callback routine used by Select_Detectors_Widget -- when
!  DISLIN push button widgets created by subroutine Select_Detectors_Widget are clicked
!  by the user, this subroutine is called.   Its single argument is the ID of the
!  widget that was clicked.

!  Subroutine Select_Detectors_Widget communicates "on the side" with this subroutine
!  via "module-scope" variable MODSHR1_Detector_Widget_Buttons_IDs.   This
!  variable is a user defined type / structure with several components -- see the
!  top of this module for the type definition.   This method is necessary because DISLIN
!  defines the argument list of this routine, so additional information must be
!  exchanged outside of the argument list.  The MODSHR1_Detector_Widget_Buttons_IDs variable
!  is used to inform this routine of the widget IDs of buttons that were created by Select_Detectors_Widget.
!  Then when this routine receives a widget ID as an argument, indicating that the
!  push button with that widget ID was clicked, it can compare the value of the argument
!  to the component values in MODSHR1_Detector_Widget_Buttons_IDs to identify which push
!  button was clicked.  Having identified which push button was clicked, this routine can perform
!  the appropriate action -- depending on the push button that was clicked, it sets and unsets
!  the values of the detector on/off buttons.
!  This routine also learns of the IDs of the detector on/off buttons via a component of
!  MODSHR1_Detector_Widget_Buttons_IDs.

!  input:
!  argument ID_of_Activated_Widget,
!  shared variable MODSHR1_Detector_Widget_Buttons_IDs

!  action:
!  changing the settings of the detector on/off buttons on the Detector Selection
!  Widget.    No other output.


#ifndef NODISLIN
subroutine  Select_Detectors_Widget_CALLBACK ( ID_of_Activated_Widget )


   use dislin

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   integer, intent (in) :: ID_of_Activated_Widget

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   ! internal variables:

   integer :: ldet

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   !  Match the ID of push button widget that was clicked, ID_of_Activated_Widget (=input
   !  argument received by this subroutine), to IDs of buttons by name,
   !  MODSHR1_Detector_Widget_Buttons_IDs%xxxxx.   Having recognized a
   !  widget button by its ID, perform the appropriate action -- set/unset the
   !  detector on/off buttons according to that name.

   if ( ID_of_Activated_Widget == MODSHR1_Detector_Widget_Buttons_IDs%All_Dets ) then

      do ldet=0, MAX_GBM_DET - 1
         call swgbut ( MODSHR1_Detector_Widget_Buttons_IDs%Particular_Detector (ldet), 1 )
      end do

   else if ( ID_of_Activated_Widget ==  MODSHR1_Detector_Widget_Buttons_IDs%All_NaIs ) then

      do ldet=0, 11
         call swgbut ( MODSHR1_Detector_Widget_Buttons_IDs%Particular_Detector (ldet), 1 )
      end do
      call swgbut ( MODSHR1_Detector_Widget_Buttons_IDs%Particular_Detector (12), 0 )
      call swgbut ( MODSHR1_Detector_Widget_Buttons_IDs%Particular_Detector (13), 0 )

   else if ( ID_of_Activated_Widget == MODSHR1_Detector_Widget_Buttons_IDs%Both_BGOs ) then

      do ldet=0, 11
         call swgbut ( MODSHR1_Detector_Widget_Buttons_IDs%Particular_Detector (ldet), 0 )
      end do
      call swgbut ( MODSHR1_Detector_Widget_Buttons_IDs%Particular_Detector (12), 1 )
      call swgbut ( MODSHR1_Detector_Widget_Buttons_IDs%Particular_Detector (13), 1 )

   else if ( ID_of_Activated_Widget == MODSHR1_Detector_Widget_Buttons_IDs%PlusX_Dets ) then

      ! Plus X side of the Spacecraft is NaI detectors 0 to 5 and BGO 0, which is detector 12
      do ldet=0, 5
         call swgbut ( MODSHR1_Detector_Widget_Buttons_IDs%Particular_Detector (ldet), 1 )
      end do
      do ldet=6, 11
         call swgbut ( MODSHR1_Detector_Widget_Buttons_IDs%Particular_Detector (ldet), 0 )
      end do
      call swgbut ( MODSHR1_Detector_Widget_Buttons_IDs%Particular_Detector (12), 1 )
      call swgbut ( MODSHR1_Detector_Widget_Buttons_IDs%Particular_Detector (13), 0 )

   else if ( ID_of_Activated_Widget == MODSHR1_Detector_Widget_Buttons_IDs%MinusX_Dets ) then

      ! Minus X side of the Spacecraft is NaI detectors 6 to 11 and BGO 1, which is detector 13
      do ldet=0, 5
         call swgbut ( MODSHR1_Detector_Widget_Buttons_IDs%Particular_Detector (ldet), 0 )
      end do
      do ldet=6, 11
         call swgbut ( MODSHR1_Detector_Widget_Buttons_IDs%Particular_Detector (ldet), 1 )
      end do
      call swgbut ( MODSHR1_Detector_Widget_Buttons_IDs%Particular_Detector (12), 0 )
      call swgbut ( MODSHR1_Detector_Widget_Buttons_IDs%Particular_Detector (13), 1 )

   else

      write ( *, '( "Select_Detectors_Widget_CALLBACK: Widget ID not recognized!" )' )
      write ( *, * )  ID_of_Activated_Widget,  MODSHR1_Detector_Widget_Buttons_IDs%All_Dets,   &
             MODSHR1_Detector_Widget_Buttons_IDs%All_NaIs, MODSHR1_Detector_Widget_Buttons_IDs%Both_BGOs
      stop

   end if


   return


end subroutine Select_Detectors_Widget_CALLBACK
#endif


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  MSB, rev. 2010 May 21 to add the Plus-X and Minus-X choices.

!  This subroutine uses DISLIN to offer the user a widget to select GBM Detectors.
!  1) There are on/off status buttons for the 14 GBM detectors, which are labeled
!  NaI 0 to 11 and "12: BGO 0" and "13: BGO 1".   These on/off status of these buttons
!  indicate which detectors will be plotted.
!  (These detector on/off buttons are capable in DISLIN of being connected to callback
!   routines but this has not been done in this design.)
!  2) There are also five "common choice" push buttons which offer the user common detector
!  selections: all detectors, all NaI detectors, both BGO detectors, the detectors on
!  the plus X-side of the Spacecraft, and the detectors on the minus X-side.  Clicking a
!  "common choice" push button invokes a callback routine, Select_Detectors_Widget_CALLBACK,
!  which resets the detector buttons to the values of corresponding to the
!  selected "common choice" push button.
!  3) When the user is done, they click the "OK" button,
!  which causes this routine to read the values of the detector on/off status buttons and
!  to store the results in output argument RequestDetectors.
!  4) This subroutine also performs some simple analyses of the selection to provide
!  additional outputs.

!  This routine communicates the widget IDs of the on/off buttons and the push buttons with
!  its callback routine Select_Detectors_Widget_CALLBACK using the "module-scope" variable
!  MODSHR1_Detector_Widget_Buttons_IDs -- this variable is a user defined type / structure with
!  components -- see the top of this module for the definition.
!  When a push button is pressed, the callback routine Select_Detectors_Widget_CALLBACK is
!  called, receiving a single argument, the ID of the button that was pressed.  It uses the
!  variables in MODSHR1_Detector_Widget_Buttons_IDs to identify which of the three push buttons
!  was pressed.  It then sets/unsets the detector on/off status buttons, as appropriate
!  for that push button.

!  When the user finally clicks "OK", the code of this routine advances past "wgfin" and
!  this routine reads the statuses of the on/off detector buttons.  Their final values
!  become the values of the output argument RequestDetectors.   These values may have been
!  set by the user clicking on:
!      detector on/off status buttons, or common choice push buttons.
!  This includes modifying a common choice by clicking on detector buttons after a common
!  choice push button.   All that matters is the configuration of the on/off detector
!  buttons at the time the "OK" button is pressed.

!  All arguments are output, created from interactive user input.
!  Additional output: the shared variable MODSHR1_Detector_Widget_Buttons_IDs.


#ifndef NODISLIN
subroutine  Select_Detectors_Widget  ( RequestDetectors )


   use dislin

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


   ! The primary output, an array indicating which detectors were selected by the user:

   logical, dimension ( 0: MAX_GBM_DET-1 ), intent (out) :: RequestDetectors

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   ! internal variables:

   integer (int32) :: NumRequestDetectors

   integer :: id_label_widget, widget_parent_id, ok_widget_id

   integer, dimension ( 0: MAX_GBM_DET-1 ) :: response
   integer :: initial_value = 0
   character (len=9) :: ButtonLabel
   character (len=60) :: WidgetTitle, WidgetLabel
   integer :: ldet

   character (len=*), parameter :: HelpString =   &
      'Select at least one detector -- data from the selected detectors will be included in the plot.|&
      &Select particular detectors by clicking the box before the detector label, or|&
      &Select a common group of detectors by clicking the named buttons at the bottom.|&
      &After selecting a common group, the selection may be modified by clicking on particular detectors.|&
      &When finished, click "OK".'


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   WidgetTitle = "GBM Detectors"
   WidgetLabel = "Select Detectors to plot:"


   !  Execute this do loop, creating and obtaining user input from a widget on each iteration,
   !  until the user has selected at least one detector -- the IF statement below will cause
   !  the EXIT from the loop.   With a sensible user, only one iteration will occur.

   GetDetectorSelection: do

      !  Overall title, background & foreground color, help string to display in help box if Help is clicked,
      !  configure the parent widget:

      call swgtit ( WidgetTitle )
      call swgclr ( 0.55, 0.55, 0.85, 'BACK' )
      call swgclr ( 0.00, 0.00, 0.00, 'FORE' )
      call swghlp ( HelpString )

      call wgini ( "VERT", widget_parent_id )
      call wglab ( widget_parent_id, WidgetLabel, id_label_widget )


      !  Configure the on/off status detector buttons:
      !  We don't connect these to a callback routine.
      !  They will be read at the end.

      do ldet=0, MAX_GBM_DET - 1

         write ( ButtonLabel, '( "NaI ", I2 )' )  ldet
         if ( ldet == 12 )  ButtonLabel = "12: BGO 0"
         if ( ldet == 13 )  ButtonLabel = "13: BGO 1"

         call wgbut ( widget_parent_id, ButtonLabel, initial_value,    &
                        MODSHR1_Detector_Widget_Buttons_IDs%Particular_Detector (ldet) )

      end do


      !  Create push button widgets that offer common detector selection choices:
      !  These widgets won't be read below, but cause actions when the buttons are clicked on,
      !  via the callback routine:

      call wgpbut ( widget_parent_id, "All detectors", MODSHR1_Detector_Widget_Buttons_IDs%All_Dets )
      call wgpbut ( widget_parent_id, "All NaI detectors", MODSHR1_Detector_Widget_Buttons_IDs%All_NaIs )
      call wgpbut ( widget_parent_id, "Both BGO detectors", MODSHR1_Detector_Widget_Buttons_IDs%Both_BGOs )
      call wgpbut ( widget_parent_id, "Plus-X detectors", MODSHR1_Detector_Widget_Buttons_IDs%PlusX_Dets )
      call wgpbut ( widget_parent_id, "Minus-X detectors", MODSHR1_Detector_Widget_Buttons_IDs%MinusX_Dets )


      !  Connect the "common choice" push button widgets to a call back routine.
      !  If a push button is pushed, the call back routine
      !  is called -- the callback routine will change the values of the detector buttons.

      call swgcbk ( MODSHR1_Detector_Widget_Buttons_IDs%All_Dets, Select_Detectors_Widget_CALLBACK )
      call swgcbk ( MODSHR1_Detector_Widget_Buttons_IDs%All_NaIs, Select_Detectors_Widget_CALLBACK )
      call swgcbk ( MODSHR1_Detector_Widget_Buttons_IDs%Both_BGOs, Select_Detectors_Widget_CALLBACK )
      call swgcbk ( MODSHR1_Detector_Widget_Buttons_IDs%PlusX_Dets, Select_Detectors_Widget_CALLBACK )
      call swgcbk ( MODSHR1_Detector_Widget_Buttons_IDs%MinusX_Dets, Select_Detectors_Widget_CALLBACK )


      !  configure an "OK" button, finalize and output the widget with wgfin:

      call wgok ( widget_parent_id, ok_widget_id )
      call wgfin


      !  When the user clicks OK, execution advances past wgfin --
      !  at which point we read the final detector button settings using gwgbut
      !  and store the user choices in the primary output argument RequestDetectors.

      do ldet=0, MAX_GBM_DET - 1

         call gwgbut ( MODSHR1_Detector_Widget_Buttons_IDs%Particular_Detector (ldet), response (ldet) )
         if ( response (ldet) == 0 ) then
            RequestDetectors (ldet) = .false.
         else
            RequestDetectors (ldet) = .true.
         end if

      end do


      !  Determine whether the user has selected at least one detector -- if so
      !  exit the loop.
      !  This also produces one of the secondary output arguments, NumRequestDetectors.

      NumRequestDetectors = count (RequestDetectors)

      if ( NumRequestDetectors == 0 ) then
         write ( *, '( // "Please SELECT AT LEAST ONE DETECTOR !" / )' )
         WidgetLabel = "SELECT >= 1 DETECTOR:"      ! change widget label because of invalid user input
      else

         exit GetDetectorSelection        !  *****  User has selected >= 1 detectors -- exit loop !  *****

      end if

   end do GetDetectorSelection


   return


end subroutine Select_Detectors_Widget
#endif


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^



subroutine  Select_Detectors    &
      ( RequestDetectors, NumRequestDetectors, SingleRequestDetector, Single_DetType_Selected, DetectorSelectionLabel )


   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


   ! The primary output, an array indicating which detectors were selected by the user:

   logical, dimension ( 0: MAX_GBM_DET-1 ), intent (out) :: RequestDetectors

   ! Secondary outputs created from RequestDetectors:

   integer (int32), intent (out) :: NumRequestDetectors
   integer (int32), intent (out) :: SingleRequestDetector    ! only applicable if user requests a single detector
   logical, intent (out) :: Single_DetType_Selected
   character (len=DETECTOR_SELECTION_LABEL_LEN), intent (out) :: DetectorSelectionLabel


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

#if  defined (NODISLIN)  ||  (defined NOGUI)
   integer (int32) :: ldet
   character (len=1) :: reply
#endif

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


#if  defined (NODISLIN)  ||  (defined NOGUI)

    !  primitive method if there are problems with dislin and the screen, i.e., problems
    !  with dislin & OpenMotif

   SelectDet: do


      write (*, '( // "Select detectors for inclusion in the plot.  Reply ""Y"" (=<CR>) or ""N"":" / )' )

      do ldet=0, MAX_GBM_DET - 1

         1010 continue
         if ( ldet <= 11 ) then
            write (*, '( "Include NaI ", I2, "? " )', advance='no' )  ldet
         else
            write (*, '( "Include BGO ", I2, "? " )', advance='no' )  ldet - 12
         end if

         read (*, '(A)' )  reply

         if ( reply == 'Y'  .OR.  reply == 'y'  .OR.  len_trim (reply) == 0 ) then
            RequestDetectors (ldet) = .TRUE.
         else
            if ( reply == 'N'  .OR.  reply == 'n' ) then
                RequestDetectors (ldet) = .FALSE.
            else
               write (*, '( ">> bad input ... try again.")' )
               goto 1010
            end if
         end if

      end do   ! ldet

      NumRequestDetectors = count (RequestDetectors)

      if ( NumRequestDetectors == 0 )  then
         write ( *, '( // "Please select at least one detector!" / )' )
      else

         exit SelectDet    ! > 0 detectors selected ... done

      end if


    end do SelectDet

#else

    !  preferred method ... GUI menu:

    call Select_Detectors_Widget  ( RequestDetectors )

#endif


   !  user input completed, process primary output argument RequestDetectors to produce
   !  additional (secondary) output arguments:

   call Analyze_Detector_Selection    &
           ( RequestDetectors, NumRequestDetectors, SingleRequestDetector, Single_DetType_Selected, DetectorSelectionLabel )


   write (*, '( / "Detectors selected: ", A )' )  trim (DetectorSelectionLabel)


   return

   end subroutine  Select_Detectors



! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


#ifndef NODISLIN
subroutine  MakeMultiHistoPlot  ( DetNum, Data_Histogram, Deadtime_Histogram, GraphTimeRange_s, BinWidth_s,    &
       InteractiveCall, TriggerName, ChannelRangeLabel, TimeLabel, BinLabel, xdiv, ydiv, ymax,                 &
       PlotLayout_Input, GraphFormat_Input, PlotDeadtimeCorr )

   use dislin


   integer (int32), intent (in) :: DetNum
   integer (int32), dimension (0:), intent (in) :: Data_Histogram
   real (real32), dimension (0:), intent (in) :: Deadtime_Histogram
   real (real64), dimension (2), intent (in) :: GraphTimeRange_s
   real (real64), intent (in) :: BinWidth_s
   logical, intent (in) :: InteractiveCall
   character (len=*), intent (in) :: TriggerName
   character (len=*), intent (in) :: ChannelRangeLabel
   character (len=2), intent (in) :: TimeLabel
   character (len=2), intent (in) :: BinLabel
   real (real32), intent (in) :: xdiv
   real (real32), intent (in) :: ydiv
   real (real32), intent (in) :: ymax
   character (len=1), intent (in) :: PlotLayout_Input    ! shouldn't change between calls of a run
   character (len=3), intent (in) :: GraphFormat_Input     ! shouldn't change between calls of a run
   character (len=1), intent (in) :: PlotDeadtimeCorr


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   ! internal variables:

   integer (int32) :: Bins_in_Histogram

   real (real64) :: OutputWidth

   character (len=FILE_NAME_MAX_LEN) :: OutputFileName
   character (len=FILE_NAME_MAX_LEN), dimension (4) :: OutputFileNames

   character (len=40) :: X_AxisLabel, Y_AxisLabel, X_AxisLabelMaster, Y_AxisLabelMaster
   character (len=5) :: XAxisNumberingStyle, YAxisNumberingStyle

   integer (int32) :: BinNum

   real (real32), dimension (2) :: TimeRange_for_Call

   real (real32) :: xleft, xright

   integer (int32), parameter :: BAD_VALUE = -9999_int32
   integer (int32) :: top, top_next, top_prev

   real (real32) :: DeadtimeCorr_top, DeadtimeCorr_top_next, DeadtimeCorr_top_prev

   real (real32) :: factor

   character (len=2) DetName
   real (real32) :: x_pos, y_pos

   integer :: x_loc, y_loc

   character (len=12), parameter :: Fragment = ".TTE_EachDet"

   real (real64) :: tmp_DBL


   ! local, persistent variables:

   character (len=1), SAVE :: PlotLayout
   character (len=3), SAVE :: GraphFormat

   integer, SAVE :: PageNum = 0
   integer, SAVE :: graph_in_page = 0

   logical, SAVE :: FirstCall = .TRUE.
   logical, SAVE :: AllPlots_in_OneFile = .TRUE.


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   if ( DetNum < 0  .or.  DetNum >= MAX_GBM_DET ) then
      write ( *, '( "Bad Input for DetNum:", I6 )' )  DetNum
      stop
   end if


   Bins_in_Histogram = size (Data_Histogram)


   if ( BinLabel == "ms" ) then
      OutputWidth = BinWidth_s * 1000.0_real64
   else
      OutputWidth = BinWidth_s * 1.0E+6_real64
   end if



   if ( OutputWidth > 999. ) then
      write ( Y_AxisLabelMaster, '( "Counts per ", 1P, G8.3, 0P, 1X, A2 )' )  OutputWidth, BinLabel
   else if ( OutputWidth > 99.9 ) then
      write ( Y_AxisLabelMaster, '( "Counts per", F5.0, 1X, A2 )' )  OutputWidth, BinLabel
   else if ( OutputWidth > 3. ) then
      write ( Y_AxisLabelMaster, '( "Counts per", F5.1, 1X, A2 )' )  OutputWidth, BinLabel
   else if ( OutputWidth > 0.3 ) then
      write ( Y_AxisLabelMaster, '( "Counts per", F5.2, 1X, A2 )' )  OutputWidth, BinLabel
   else if ( OutputWidth > 0.025 ) then
     write ( Y_AxisLabelMaster, '( "Counts per", F6.3, 1X, A2 )' )  OutputWidth, BinLabel
   else
      write ( Y_AxisLabelMaster, '( "Counts per", F7.4, 1X, A2 )' )  OutputWidth, BinLabel
   end if


   write ( X_AxisLabelMaster, '( "Time (", A, ")" )' )  trim (TimeLabel)



   !  On first call, initialize the overall plot / DISLIN:

   if ( FirstCall ) then


      FirstCall = .FALSE.


      GraphFormat = GraphFormat_Input     !  Choices are made on 1st call; cannot change
      call MakeLowerCase (GraphFormat)

      PlotLayout = PlotLayout_Input

      !  PlotLayout "S" has only one page so all graph formats need only one file.
      !  Of the graph formats PDF, EPS, and PNG, only PDF can support multiple pages
      !  within one file -- therefore if multiple pages are requested for a non-PDF format,
      !  we must use multiple files.
      if ( PlotLayout == "L"  .and.  GraphFormat /= "pdf" )  AllPlots_in_OneFile = .FALSE.


      if ( AllPlots_in_OneFile ) then

         !  all plots on one page -- so simple case of only one output file name --
         !  if this is an interactive call, give the user the option of selecting the output file name:
         if ( InteractiveCall ) then
            call  Get_OutputFileName  ( TriggerName, Fragment, GraphFormat, OutputFileName )
         else
            OutputFileName = trim (TriggerName) // ".TTE_EachDet." // GraphFormat
         end if

         call setfil ( trim (OutputFileName) )

      else   ! AllPlots_in_OneFile ?

         !  The output file names for multiple pages/files of output are always automatically generated:
         OutputFileNames (1) = trim (TriggerName) // ".TTE_Dets.A." // GraphFormat
         OutputFileNames (2) = trim (TriggerName) // ".TTE_Dets.B." // GraphFormat
         OutputFileNames (3) = trim (TriggerName) // ".TTE_Dets.C." // GraphFormat
         OutputFileNames (4) = trim (TriggerName) // ".TTE_Dets.D." // GraphFormat

         call setfil ( trim ( OutputFileNames (1) ) )    ! for PNG, other pages below

      end if   ! AllPlots_in_OneFile ?


      call metafl ( GraphFormat )


      ! Normally PNG plots white lines and characters on black background --
      !  reverse to black on white.
      !  For PNG, override default pixel dimensions to have same aspect ratio as logical pixel dimensions.
      if ( GraphFormat == 'png' )  call scrmod ( "REVERS" )


      if ( PlotLayout == "L" ) then
         call setpag ( 'USAP' )      ! US A (=Letter), Portrait,  2160 x 2790 points
         if ( GraphFormat == 'png' )  call winsiz ( 660, 853 )  ! physical pixel dimensions, override default of 603 x 853
      else
         call setpag ( 'USAL' )      ! US A (=Letter), Landscape,  2790 x 2160 points
         if ( GraphFormat == 'png' )  call winsiz ( 853, 660 )  ! physical pixel dimensions, override default of 603 x 853
      end if

      call filmod ( 'DELETE' )    ! overwrite (delete) previous file

      call disini ()

      if ( GraphFormat == 'pdf' .or. GraphFormat == 'eps' )  call psfont ( 'Helvetica' )

   end if   ! FirstCall ?




   !  configure and initiate the each graph:

   !   PlotLayout "L" is Large Plots on multiple pages.
   !   For EPS & PNG plots, multiple pages are equivalent to multiple files;
   !   for a PDF plot, multiple pages are done by page breaks within a single file.
   !   This code section 1) this code takes care of page breaks
   !   (different actions for PNG and PDF), 2) sizes and
   !   positions the graph frames, and 3) decides which graph frames get
   !   axis labels.  For "L", all graph frames have both axes numbered.

   !  For mode "L", the dislin close of the previous graph was already done at the bottom of this routine.

   if ( PlotLayout == "L" ) then

      if ( mod (DetNum, 4 ) == 0 ) then

         PageNum = PageNum + 1
         graph_in_page = 0

         !  New page after the first:
         !  1) for PDF output, simply a newpage within one file,
         !  2) for EPS & PNG output, the previous graph was finalized, so must reinitilize --
         !     this somewhat duplicates code in the FirstCall section -- could refactor....

         if ( PageNum /= 1 ) then

            if ( AllPlots_in_OneFile ) then
               call newpag    ! advances PDF page within one file
            else

               !  reach this branch for PNG and EPS
               call setfil ( trim ( OutputFileNames (PageNum) ) )
               call metafl ( GraphFormat )
               if ( GraphFormat == 'png' )  call scrmod ( "REVERS" )
               call setpag ( 'USAP' )      ! US A (=Letter), Portrait,  2160 x 2790 points -- logical size
               if ( GraphFormat == 'png' )  call winsiz ( 660, 853 )  ! physical pixel dimensions, override default of 603 x 853
               call filmod ( 'DELETE' )    ! overwrite (delete) previous file
               call disini ()
               if ( GraphFormat == 'eps' )  call psfont ( 'Helvetica' )

            end if

         end if

      end if


      graph_in_page = graph_in_page + 1

      call axslen ( 1650, 420 )
      call axspos ( 350, 700 + (graph_in_page-1) * 590 )

      X_AxisLabel = ''
      if ( graph_in_page == 4  .or.  DetNum == MAX_GBM_DET - 1 )  X_AxisLabel = X_AxisLabelMaster
      Y_AxisLabel = Y_AxisLabelMaster

      XAxisNumberingStyle = "float"
      YAxisNumberingStyle = "float"

   end if


   !  PlotLayout "S" is Small Plots, all on a single page.
   !   This code section 1) sizes and positions the graph frames,
   !   2) decides which graph frames get axis labels, and
   !   3) decides which graph frames have numbered axes.
   !   4) Also decrease tick lengths in plot mode S -- with the smaller graph
   !      box size of this mode, the default tick sizes are too large.

   if ( PlotLayout == "S" ) then

      ! decrease lengths of major and minor ticks from defaults of 24 & 16:
      call ticlen ( 16, 10 )

      if ( DetNum <= 11 ) then
         if ( mod ( DetNum, 4) == 0 )  x_loc = 300    ! NaI column 1
         if ( mod ( DetNum, 4) == 1 )  x_loc = 910    ! NaI column 2
         if ( mod ( DetNum, 4) == 2 )  x_loc = 1520   ! NaI column 3
         if ( mod ( DetNum, 4) == 3 )  x_loc = 2130   ! NaI column 4
      end if

      if ( DetNum >= 0  .and.  DetNum <=  3 )  y_loc = 1120    ! NaI row 1
      if ( DetNum >= 4  .and.  DetNum <=  7 )  y_loc = 1500    ! NaI row 2
      if ( DetNum >= 8  .and.  DetNum <= 11 )  y_loc = 1880    ! NaI row 3

      if ( DetNum == 12 ) then
         x_loc = 555
         y_loc = 620
      end if
      if ( DetNum == 13 ) then
         x_loc = 1770
         y_loc = 620
      end if

      if ( DetNum <= 11 ) then
         call axslen ( 530, 300 )
      else
         call axslen ( 636, 360 )
      end if

      call axspos ( x_loc, y_loc )

      X_AxisLabel = ''
      Y_AxisLabel = ''

      if ( DetNum >= 8  .and.  DetNum <= 13 )  X_AxisLabel = X_AxisLabelMaster    ! NaI row 3 plus BGOs
      if ( DetNum == 4  .or.  DetNum == 12  .or.  DetNum == 13 )  Y_AxisLabel = Y_AxisLabelMaster

      XAxisNumberingStyle = "none"
      YAxisNumberingStyle = "none"

      if ( DetNum >= 8  .and.  DetNum <= 13 )  XAxisNumberingStyle = "float"    ! NaI row 3 plus BGOs
      if ( mod ( DetNum, 4) == 0  .or.  DetNum == 12  .or.  DetNum == 13 )  YAxisNumberingStyle = "float"   ! NaI col 1 & BGOs

   end if


   !  axes and axes labels:

   call labdig ( -2, 'X' )  ! DISLIN (>=9.4?) automatically calculates number of digits after decimal point on X-axis labels
   call labdig ( 0, 'Y' )   ! zero digits after decimal point on Y-axis labels

   if ( GraphFormat /= 'png' ) then
      call hname ( 30 )   ! height of axis labels
   else
      call hname ( 24 )   ! height of axis labels
   end if

   !  labels may have content, or may be empty, as configured above:
   call name ( X_AxisLabel,  'X' )
   call name ( Y_AxisLabel, 'Y' )



   !  Draw graph box / axes:


  if ( TimeLabel == "s " ) then
     TimeRange_for_Call = real ( GraphTimeRange_s, real32 )
     factor = 1.0_real32
  else
     !  X-axis will be in ms -- all input and calculations are in seconds -- the next two
     !  variables take care of converions to make the plot output correctly:
     TimeRange_for_Call = real ( GraphTimeRange_s * 1000.0_real64, real32 )
     factor = 1000.0_real32
   end if

   ! instruct whether to number axes:
   call labels ( XAxisNumberingStyle, 'X' )
   call labels ( YAxisNumberingStyle, 'Y' )

   call height ( 30 )
   call graf ( TimeRange_for_Call (1), TimeRange_for_Call (2), TimeRange_for_Call (1), xdiv, 0.0, ymax, 0.0, ydiv )
   call height ( 36 )    ! restore default


   !  label in upper-right corner of box with detector number:

   if ( DetNum <= 11 ) then
      write ( DetName, '( "N", Z1 )' )  DetNum
   else
      if ( DetNum == 12 ) then
         DetName = "B0"
      else
         DetName = "B1"
      end if
   end if

   x_pos = TimeRange_for_Call (1) + 0.03 * ( TimeRange_for_Call (2) - TimeRange_for_Call (1) )
   y_pos = 0.9 * ymax

   call rlmess ( DetName, x_pos, y_pos )


   !  Draw the histogram :

   do BinNum=0, Bins_in_Histogram - 1

      if ( Data_Histogram (BinNum) > 0 ) then   ! non-zero bin ?

         tmp_DBL = GraphTimeRange_s (1) + BinNum * BinWidth_s
         xleft = real ( tmp_DBL, real32 )

         tmp_DBL = xleft + BinWidth_s
         xright = real ( tmp_DBL, real32 )

         ! unit conversion if time axis is ms:
         xleft = xleft * factor
         xright = xright * factor

         !  First calculate the quantities for measured (non-deadtime corrected) histogram --
         !  but plotting is delayed until after the deadtime corrected histogram (RED)
         !  so that the measured (BLACK) histogram will be on top.
         !  These quantities are calculated in before they are plotted because they
         !  are plotted (BLACK) because they are also used to calculate the deadtime-
         !  corrected (RED) histogram.

         ! horizontal bar:
         top = Data_Histogram (BinNum)
         DeadtimeCorr_top = Deadtime_Histogram (BinNum)

         !  right side of bar:
         top_next = BAD_VALUE
         if ( BinNum /= Bins_in_Histogram - 1 ) then
            top_next = Data_Histogram (BinNum+1)
            DeadtimeCorr_top_next = Deadtime_Histogram (BinNum+1)
         end if

         ! The left side of the bar for this bin will already have been drawn as the
         ! right side of the previous bar, unless that bar wasn't drawn because
         ! that bin had zero counts:

         top_prev = BAD_VALUE
         if ( BinNum > 0 ) then
            if ( Data_Histogram (BinNum-1) == 0 ) then
               top_prev = Data_Histogram (BinNum-1)
               DeadtimeCorr_top_prev = Deadtime_Histogram (BinNum-1)
            end if
         end if



         if ( PlotDeadtimeCorr == 'Y' ) then   ! plot deadtime-corrected histogram ?

            !  Now plot the deadtime corrected histogram in RED:

            call color ( "RED" )

            ! horizontal bar:
            call rline ( xleft, DeadtimeCorr_top, xright, DeadtimeCorr_top )

            !  right side of bar:
            if ( top_next /= BAD_VALUE )  call rline ( xright, DeadtimeCorr_top, xright, DeadtimeCorr_top_next )

            ! left side of the bar:
            if ( top_prev /= BAD_VALUE )  call rline ( xleft, DeadtimeCorr_top_prev, xleft, DeadtimeCorr_top )

         end if    ! plot deadtime-corrected histogram ?


         call color ( "FORE" )   ! restore default color -- BLACK


         !  Finally plot the measure / non-deadtime corrected (BLACK) histogram.
         !  It is delayed until the end so that it will be on top of
         !  the deadtime-corrected (RED) histogram.

         ! horizontal bar:
         call rline ( xleft, real(top), xright, real(top) )

         !  right side of bar:
         if ( top_next /= BAD_VALUE )  call rline ( xright, real(top), xright, real(top_next) )


         !  left side of the bar:
         if ( top_prev /= BAD_VALUE )  call rline ( xleft, real(top_prev), xleft, real(top) )


      end if   ! non-zero bin ?

   end do    ! loop over Histogram bins



   !  add one title per page:

   if ( ( PlotLayout == "L"  .and.  graph_in_page == 1 ) ) then
      call titlin ( ChannelRangeLabel, 3 )
      call titlin ( TriggerName, 2 )
      call title ()
   end if

   if ( PlotLayout == "S"  .and.  DetNum == 12 ) then
      call titlin ( TriggerName, 3 )
      call title ()
   end if
   if ( PlotLayout == "S"  .and.  DetNum == 13 ) then
      call titlin ( ChannelRangeLabel, 3 )
      call title ()
   end if

   call endgrf      ! finish this graph


   !  Finish group of plots when have drawn the last graph in the file:
   !  If we are making multiple files (Large Plot / PNG case),
   !  then at the end of each page we close of the plot.
   !  Graphs in page are numbered 1, 2, 3, and 4.
   !  Easy case: when have drawn last graph, i.e., detector
   !  MAX_GBM_DET - 1 = 13, then finish plots.

   if ( ( .not. AllPlots_in_OneFile  .and.  graph_in_page == 4 )  .or.  DetNum == MAX_GBM_DET - 1 ) then

      call disfin ()    ! finish plot

   end if


   return


end subroutine MakeMultiHistoPlot
#endif


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  Input "Xaxis_Range" is the range of an X-axis, Xmax - Xmin.
!  Supported range of Xaxis_Range is about .04 to 20,000.
!  This function will return a "nice" value to use as a tick mark
!  spacing for plotting a X-axis with this range.


function  auto_xdiv_value  ( Xaxis_range )

   real (real32) :: auto_xdiv_value     ! function return

   real (real32), intent (in) :: Xaxis_range

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   real (real32) :: xdiv

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   if ( Xaxis_range > 8000.0 ) then
      xdiv = 2000.0
   else if ( Xaxis_range >= 4000.0 ) then
      xdiv = 1000.0
   else if ( Xaxis_range >= 2000.0 ) then
      xdiv = 500.0
   else if ( Xaxis_range >= 800.0 ) then
      xdiv = 200.0
   else if ( Xaxis_range >= 400.0 ) then
      xdiv = 100.0
   else if ( Xaxis_range >= 200.0 ) then
      xdiv = 50.0
   else if ( Xaxis_range >= 80.0 ) then
      xdiv = 20.0
   else if ( Xaxis_range >= 40.0 ) then
      xdiv = 10.0
   else if ( Xaxis_range >= 20.0 ) then
      xdiv = 5.0
   else if ( Xaxis_range >= 8.0 ) then
      xdiv = 2.0
   else if ( Xaxis_range >= 4.0 ) then
      xdiv = 1.0
   else if ( Xaxis_range >= 2.0 ) then
      xdiv = 0.5
   else if ( Xaxis_range >= 1.0 ) then
      xdiv = 0.25
   else if ( Xaxis_range >= 0.5 ) then
      xdiv = 0.1
   else if ( Xaxis_range >= 0.2 ) then
      xdiv = 0.05
   else if ( Xaxis_range >= 0.1 ) then
      xdiv = 0.025
   else
      xdiv = 0.01
   end if

   !  fail safe, but may obtain peculiar value:
   if ( xdiv > Xaxis_range / 3. )  xdiv = Xaxis_range / 10.


   auto_xdiv_value = xdiv


   return


end function auto_xdiv_value


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  Input is maximum y-data value.   Since y-values are counts, the maximum value should
!  always be >= 1.    Since the function processes the value by scaling, then unscaling,
!  it can handle any value larger than 1.
!  It returns, as the function return, a "nice" "round" value to use as the
!  maximum value of the y-axis on a plot.


function auto_ymax_value  ( MaxDataValue )

   real (real32) :: auto_ymax_value     ! function return

   real (real32), intent (in) :: MaxDataValue


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   real (real32) :: ymax_scaled
   real (real32) :: scale_factor
   real (real32) :: ymax

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   ymax_scaled = MaxDataValue
   scale_factor = 1.0_real32

   find_scale: do
      if ( ymax_scaled > 9.9_real32 ) then
         ymax_scaled = ymax_scaled / 10.0_real32
         scale_factor = scale_factor * 10.0_real32
      else
         exit find_scale
      end if
   end do find_scale

   if ( ymax_scaled < 1.2 ) then
      ymax = 1.2
   else if ( ymax_scaled < 1.4 ) then
      ymax = 1.4
   else if ( ymax_scaled < 1.6) then
      ymax = 1.6
   else if ( ymax_scaled < 2.0 ) then
      ymax = 2.0
   else if ( ymax_scaled < 2.4 ) then
      ymax = 2.4
   else if ( ymax_scaled < 3.0 ) then
      ymax = 3.0
   else if ( ymax_scaled < 4.0 ) then
      ymax = 4.0
   else if ( ymax_scaled < 5.0 ) then
      ymax = 5.0
   else if ( ymax_scaled < 6.0 ) then
      ymax = 6.0
   else if ( ymax_scaled < 8.0 ) then
      ymax = 8.0
   else
      ymax = 10.0
   end if

   ymax = ymax * scale_factor


   auto_ymax_value = ymax

   return


end function auto_ymax_value


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  Input is minimum y-data value.   The values are assumed to be counts
!  and are thus non-negative.  If the input is zero, the obvious answer
!  is zero.
!  Otherwise we handle by scaling and rouding to a "nice" "round" value.


function auto_ymin_value  ( MinDataValue )

   real (real32) :: auto_ymin_value     ! function return

   real (real32), intent (in) :: MinDataValue


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   real (real32) :: ymin_scaled
   real (real32) :: scale_factor
   real (real32) :: ymin

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   if ( MinDataValue <= 1.9 ) then

      auto_ymin_value = 0.0
      return   ! EARLY RETURN

   end if

   ymin_scaled = MinDataValue
   scale_factor = 1.0_real32

   find_scale: do
      if ( ymin_scaled > 10.0_real32 ) then
         ymin_scaled = ymin_scaled / 10.0_real32
         scale_factor = scale_factor * 10.0_real32
      else
         exit find_scale
      end if
   end do find_scale


   if ( ymin_scaled > 8.2 ) then
      ymin = 8.0
   else if ( ymin_scaled > 6.2 ) then
      ymin = 6.0
   else if ( ymin_scaled > 4.2 ) then
      ymin = 4.0
   else if ( ymin_scaled > 2.7 ) then
      ymin = 2.5
   else if ( ymin_scaled > 2.1 ) then
      ymin = 2.0
   else if ( ymin_scaled > 1.5 ) then
      ymin = 1.4
   else
      ymin = 1.0
   end if


   ymin = ymin * scale_factor


   auto_ymin_value = ymin

   return


end function auto_ymin_value


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  The input is the maximum value of the y-axis of a plot.
!  The function return is a "nice" value to use as a tick spacing.
!  Since the y-axis is for counts, ymax must be >= 1.
!  Since the function scales, then unscales, the maximum value, it can handle
!  any ymax value larger than 1.


function auto_ydiv_value  ( ymax )

   real (real32) :: auto_ydiv_value    ! function return

   real (real32), intent (in) :: ymax

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   real (real32) :: ymax_scaled
   real (real32) :: scale_factor
   real (real32) :: ydiv

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   ymax_scaled = ymax
   scale_factor = 1.0_real32

   find_scale: do
      if ( ymax_scaled > 1000.0_real32 ) then
         ymax_scaled = ymax_scaled / 1000.0_real32
         scale_factor = scale_factor * 1000.0_real32
      else
         exit find_scale
      end if
   end do find_scale


   if ( ymax_scaled > 800.0 ) then
      ydiv = 200.0
   else if ( ymax_scaled >= 400.0 ) then
      ydiv = 100.0
   else if ( ymax_scaled >= 200.0 ) then
      ydiv = 50.0
   else if ( ymax_scaled >= 80.0 ) then
      ydiv = 20.0
   else if ( ymax_scaled >= 40.0 ) then
      ydiv = 10.0
   else if ( ymax_scaled >= 20.0 ) then
      ydiv = 5.0
   else if ( ymax_scaled >= 13.0 ) then
      ydiv = 4.0
   else if  ( ymax_scaled >= 7.0 ) then
      ydiv = 2.0
   else
      ydiv = 1.0
   end if

   ydiv = ydiv * scale_factor


   auto_ydiv_value = ydiv

   return


end function auto_ydiv_value


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


subroutine  Get_OutputFileName  ( NamePortion, Fragment, GraphFormat, OutputFileName )

   character (len=*), intent (in) :: NamePortion
   character (len=*), intent (in) :: Fragment
   character (len=3), intent (in) :: GraphFormat
   character (len=FILE_NAME_MAX_LEN), intent (out) :: OutputFileName

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   character (len=FILE_NAME_MAX_LEN) :: TempName

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   1010 continue
   TempName = trim (NamePortion) // trim (Fragment)
   write ( *, '( / "Input name to use for the output file, or ENTER to use the default of """, A, """." )' )  trim (TempName)
   write ( *, '("The filetype of "".", A, """ will be automatically appended.   Input name: ")', advance='no' )  GraphFormat
   read ( *, '(A)', err=1010 )  OutputFileName
   if ( len_trim (OutputFileName) == 0 )  OutputFileName = TempName
   OutputFileName = trim (OutputFileName) // "." // GraphFormat
   write ( *, '( "The output file will be """, A, """." / )' )  trim (OutputFileName)

   return

end subroutine  Get_OutputFileName


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


subroutine  Get_GraphFormat  ( GraphFormat )

   character (len=3), intent (out) :: GraphFormat

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   1010 continue
   write ( *, '( / "Input graph type, ""pdf"", ""eps"" or ""png"", or ENTER for default of ""pdf"": " )', advance='no' )
   read ( *, '(A)', err=1010 )  GraphFormat
   if ( len_trim (GraphFormat) == 0 )  GraphFormat = 'pdf'
   call MakeLowerCase (GraphFormat)
   if ( GraphFormat /= 'pdf' .and. GraphFormat /= 'eps' .and. GraphFormat /= 'png' ) then
      write ( *, '("Invalid Graph Type: ", A )' )  GraphFormat
      goto 1010
   end if

   return

end subroutine  Get_GraphFormat


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  Optionally plot a legend inside of the graph box:

#ifndef NODISLIN
subroutine  PlotLegend  ( NX_Left, NX_Right, NY_Bot, NY_Top )

   use dislin

   !  coordinates of graph box (axes) in plot coordinates, rather than user units:

   integer, intent (in) :: NX_Left, NX_Right, NY_Bot, NY_Top


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   ! internal variables:


   character (len=15) :: LegendString
   character (len=1) :: LegendPos

   integer :: NX_Len, NY_Len
   integer :: NX_SFL, NY_SFL

   real :: offset


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   write (*, '( / "Input an optional legend, such as GRB/TGF name, up to 15 characters, or ENTER to omit: " )', advance="no" )
   read (*, '(A15)' )  LegendString

   if ( len_trim (LegendString)  /=  0 ) then

      NX_Len = NX_Right - NX_Left
      NY_Len = NY_Bot - NY_Top


      999 continue
      write (*, '( / "Input a position for the legend: R for right, L for left, or C for Custom: " )', advance="no" )
      read (*, '(A1)', err=999 )  LegendPos
      call MakeUpperCase  ( LegendPos )
      if ( LegendPos /= "L"  .AND.  LegendPos /= "R"  .AND.  LegendPos /= "C" ) goto 999

      !  all cases, slightly down from the top:
      NY_SFL = int ( NY_Top + 0.07 * NY_Len, int32 )


      select case ( LegendPos )

         case ( "L" )    ! left
            NX_SFL = int ( NX_Left + 0.06 * NX_Len, int32 )
            LegendString = adjustl (LegendString)

         case ( "R" )    ! right
            call txtjus ( 'right' )  ! right justification: NX_SFL will designate the end of string, not the start
            NX_SFL = int ( NX_Left + 0.94 * NX_Len, int32 )
            LegendString = adjustr (LegendString)    ! adjust non-blank chars to the end (right) of the string

         case ( "C" )
            LegendString = adjustl (LegendString)
            888 continue
            write (*, '( / "Input the starting position of the legend, from the left, between 0.0 and 1.0: " )', advance='no' )
            read (*, *, err=888) offset
            if ( offset < 0.0  .OR.  offset > 1.0 ) goto 888
            NX_SFL = int ( NX_Left + offset * NX_Len, int32 )

      end select


      call messag  ( LegendString, NX_SFL, NY_SFL )
      call txtjus ( 'left' )   ! restore default justification

   end if


   return


end subroutine  PlotLegend
#endif


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  Use dislin for a GUI widget to select a GBM TTE data file.
!  Based on the input detector number, a file name filter is
!  created that will offer the user a list of TTE data files for
!  just that detector.   Or if the detector value is -1, all TTE
!  files will be listed.

#ifndef NODISLIN
subroutine  Get_DataFileName_Widget  ( Which_Det, DataFileName )


   use dislin

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   integer, intent (in) :: Which_Det    ! 0 to 13.

   character (len=FILE_NAME_MAX_LEN), intent (out) :: DataFileName


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   ! internal variables:

   character (len=12) :: FileNameFilter

   integer :: widget_parent_id, ok_widget_id, file_widget_id, some_id

   character (len=*), parameter :: WidgetTitle = "Select Data File:"

   character (len=150) :: HelpString =   &
      'Select the TTE data file for DET ?? for to be searched.  Click on File, Use the GUI menu, then when done click "OK".'


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   if ( Which_Det < -1  .OR.  Which_Det > 13 ) then
      write (*, '( I0, " is an illegal GBM detector number." )' )  Which_Det
      stop
   end if


   !  Setup the HelpString and FileNameFilter based upon the input detector number:

   if ( Which_Det == -1 ) then

      ! all detectors case
      HelpString (30:35) = "AnyDet"
      FileNameFilter = "glg_tte_*"

   else

      if ( Which_Det <= 11 ) then

         ! NaI case
         write (HelpString (30:35), '( "NaI ", I2)' )  Which_Det

         if ( Which_Det <= 11 ) then
            write (FileNameFilter, '( "glg_tte_n", I1, "_*" )' )  Which_Det
         else if ( Which_Det == 10 ) then
            FileNameFilter = "glg_tte_na_*"
         else
            FileNameFilter = "glg_tte_nb_*"
         end if

      else

         !  BGO case
         write (HelpString (30:35), '( "BGO ", I2)' )  Which_Det - 12
         write (FileNameFilter, '( "glg_tte_b", I1, "_*" )' )  Which_Det - 12

      end if

   end if


   !  Setup the overall title, background & foreground color, a help string to display
   !  in the help box if Help is clicked, configure the parent widget, create a
   !  label widget

   call swgtit ( WidgetTitle )
   call swgclr ( 0.55, 0.55, 0.85, 'BACK' )
   call swgclr ( 0.00, 0.00, 0.00, 'FORE' )
   call swghlp ( HelpString )

   call wgini ( "VERT", widget_parent_id )

   call wglab ( widget_parent_id, 'Click on "File", then "OK":', some_id )


   !  The file widget:

   call wgfil  ( widget_parent_id, 'Click this to get a GUI menu!', ' ', FileNameFilter, file_widget_id )


   !  finalize and output the widget with wgfin:

   call wgok ( widget_parent_id, ok_widget_id )
   call wgfin


   !  get the result

   call gwgfil  ( file_widget_id, DataFileName)


   return


end subroutine Get_DataFileName_Widget
#endif


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  When the pipeline software cannot determine the trigger time it "guesses".
!  This is a bad design -- it should abort.
!  The "guessed" values tend to be multiples of 1 or 0.1 seconds.
!  A Trigger Time (measured to 2 microseconds) that is an exact multiple of 0.1
!  seconds is quite improbable, so if such a value is encountered, is almost
!  certainly means that the Trigger Time is erroneous, resulting from the
!  pipelines software bug.   This routine checks for that pattern, and if
!  is found, warns of it.


subroutine  CheckTriggerTime  ( TriggerTime )

   use, intrinsic :: iso_c_binding, only: C_DOUBLE

   real (C_DOUBLE), intent (in) :: TriggerTime

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   real (C_DOUBLE) :: TriggerTime_x10


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   !  Multiply by x10 so that the pattern becomes whether or not the value is an integer.
   !  Then remove integer part and test whether the remaining fractional value is zero or not.
   !  Since TriggerTime is provided to 2 microseconds, TriggerTime_x10 should have digits
   !  to 2E-5.

   TriggerTime_x10 = 10.0_C_DOUBLE * TriggerTime

   TriggerTime_x10 = TriggerTime_x10  -  floor (TriggerTime_x10, int64)

   if ( abs (TriggerTime_x10) < 3.0E-6_C_DOUBLE ) then

      write (*, '( // "******************************************************************************************" )' )
      write (*, '( "**** Possible ERROR in the DATA: Trigger Time", F18.6, " is a multiple of 0.1 s." )' )  TriggerTime
      write (*, '( "**** This value is suspicious -- sometimes produced by a bug in the pipeline software." )' )
      write (*, '( "******************************************************************************************" // )' )

   end if


   return

end subroutine  CheckTriggerTime


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


subroutine  Output_CSV_TitleLines ( out_LUN, ProgVersion, TimeRefType, RawReferenceTime, CorrReferenceTime,   &
      ClockDrift_Flag, GapTime, NumRequestDetectors, DetectorSelectionLabel, merged )

   use Time_and_Coord_Routines, only: MET_to_UTC

   integer, intent (in) :: out_LUN
   character (len=*), intent (in) :: ProgVersion

   character (len=9), intent (in):: TimeRefType

   real (real128), intent (in) :: RawReferenceTime
   real (real128), intent (in) :: CorrReferenceTime
   logical, intent (in) :: ClockDrift_Flag
   real (real32), intent (in) :: GapTime
   integer (int32), intent (in) :: NumRequestDetectors
   character (len=DETECTOR_SELECTION_LABEL_LEN), intent (in) :: DetectorSelectionLabel
   logical, intent (in) :: merged

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   character (len=8) :: CurDate
   character (len=10) :: CurTime

   real (real128) :: JulianDay

   integer (int32) :: year
   integer (int32) :: month
   integer (int32) :: day
   integer (int32) :: hours
   integer (int32) :: minutes
   real (real64) :: seconds

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   call date_and_time ( CurDate, CurTime )

   write (out_LUN, '( "Data from program: ", A )' )  trim (ProgVersion)
   write (out_LUN, '( "Produced on ", A4, "-", A2, "-", A2, 1X, A2, ":", A2, ":", A6 )' )     &
       CurDate (1:4), CurDate (5:6), CurDate (7:8), CurTime (1:2), CurTime (3:4), CurTime (5:10)


   call MET_to_UTC ( CorrReferenceTime, JulianDay, year, month, day, hours, minutes, seconds )


   write (out_LUN, '( "The first and second column columns are GBM MET" )' )
   write (out_LUN,       &
'( "The third column is seconds w.r.t. year=", I5, " month=", I3, " day =", I3, " hours=", I3, " and minutes=", I3, " UT"  )' ) &
             year, month, day, hours, minutes

   write (out_LUN, '( "The uncorrected ", A, " Time is", F17.6 )' )  trim (TimeRefType), RawReferenceTime

   if ( ClockDrift_Flag ) then
      write (out_LUN, '( "The fourth colum is w.r.t. to the Corrected ", A, " Time of ", F17.6, "," )' )   &
       trim (TimeRefType), CorrReferenceTime
   else
      !  The next output line was calculated from "CorrReferenceTime", but the output will imply that they are are
      !  for "RawReferenceTime".   These two variables should be equal for this case ... validate!
      if ( abs (CorrReferenceTime - RawReferenceTime)  >  2.0E-6 ) then
         write (*, '( "Program Logic failure.  Corrected and Raw Reference Times are not equal!" )' )
         write (*, *)   RawReferenceTime, CorrReferenceTime
         stop
      end if
   end if

   write (out_LUN,    &
     '( "which is year=", I5, " month=", I3, " day =", I3, " hours=", I3, " minutes=", I3, " and seconds=", F10.6, " UTC" )' )   &
             year, month, day, hours, minutes, seconds

   if ( ClockDrift_Flag ) then
      write (out_LUN, '( "All times in the output columns EXCEPT the first column are corrected for the GBM DPU clock drift." )' )
      write (out_LUN,   &
        '( "The ''gap'' of slowness per second of the GBM DPU clock was input as", F6.2, " micoseconds per s." )' )  GapTime
   else
      write (out_LUN, '( "No correction was applied for the drift of the GBM DPU clock !" )' )
   end if


   if ( TimeRefType == "Trigger" ) then
      if ( ClockDrift_Flag ) then
         write (out_LUN, '( "The fourth column are times relative to the Trigger Time, corrected for GBM clock drift." )' )
      else
         write (out_LUN, '( "The fourth column are times relative to the Trigger Time, UNcorrected for GBM clock drift." )' )
      end if
   end if
   if ( TimeRefType == "Reference" ) then
      if ( ClockDrift_Flag ) then
         write (out_LUN, '( "The fourth column are times relative to the Reference Time, corrected for GBM clock drift." )' )
      else
         write (out_LUN, '( "The fourth column are times relative to the Reference Time, UNcorrected for GBM clock drift." )' )
      end if
   end if

   if ( merged ) then
      write (out_LUN, '( "The fifth column is the number of the detector that detected the event, and" )' )
      write (out_LUN, '( "the sixth column is the energy channel number of the event." )' )
   else
      write (out_LUN, '( "The fifth column is the energy channel number of the event." )' )
   end if

   if ( NumRequestDetectors == 1 ) then
      write (out_LUN, '( "Data for detector: ", A)' )  DetectorSelectionLabel
   else
      write (out_LUN, '( "Detectors merged into this file: ", A)' )  DetectorSelectionLabel
   end if


   return

end subroutine  Output_CSV_TitleLines


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  The GBM DPU clock drifts from x.0 seconds to x.999998 seconds, then re-sychronizes at the second
!  boundary.   The clock is slow, never reaching x.999998, therefore the correction value should
!  be added.   Therefore the correction is ADDED in proportion to how far the clock has advanced from
!  x.0 to x.999998 s.


function  MET_Corr_for_ClockDrift  ( MET, GapTime )

   real (real128) :: MET_Corr_for_ClockDrift

   real (real128), intent (in) :: MET     !  GBM MET, in seconds
   real (real32), intent (in) :: GapTime    ! in microseconds


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   MET_Corr_for_ClockDrift  =  MET  +  ( MET - FLOOR (MET, int64) ) * GapTime * 1.0E-6_real128


   return

end  function  MET_Corr_for_ClockDrift


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


subroutine  GapTimeQuestion  ( ClockDrift_Flag, GapTime )

   real (real32), intent (out) :: GapTime
   logical, intent (out) :: ClockDrift_Flag


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   character (len=1) :: Reply

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   ClockDrift_Flag = .FALSE.
   GapTime = 0.0


   write (*, '( / "Do you wish to correct the output TTE times for the drift of the GBM DPU clock?" )' )
   write (*, '( "Input Y/N, or ENTER for N: " )', advance="no" )
   read (*, '(A)' )  Reply
   if ( Reply == "Y"  .OR.  Reply == "y" ) then
      ClockDrift_Flag = .TRUE.
      write (*, '( "Input the time, in microseconds, by which the DPU clock falls short each second." )' )
      write (*, '( "Also known as the ""gap"" time.  Input positive value: " )' )
      read (*, *)  GapTime
      if ( GapTime < 0.0  .OR.  GapTime > 50.0 ) then
         write ( *, '( 1G12.4, " is an unreasonable value." )' )  GapTime
         stop
      end if
   end if


   return

end subroutine  GapTimeQuestion


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


subroutine  Determine_ReferenceTime  ( RequestDetectors, TriggerTime, TTE_Data, ReferenceTime )

   use, intrinsic :: iso_c_binding, only: C_DOUBLE

   use Type_for_all_TTE_Events


   logical, dimension ( 0: MAX_GBM_DET-1 ), intent (in) :: RequestDetectors

   !  positive for trigger data, negative for non-trigger (continuous) data
   real (C_DOUBLE), intent (in) :: TriggerTime

   type (All_TTE_Events_type), dimension (0: MAX_GBM_DET-1), intent (in) :: TTE_Data

   real (real128), intent (out) :: ReferenceTime


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   integer :: ldet
   real (C_DOUBLE) :: StartVal, StopVal

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   StartVal = -9.9E+30
   StopVal = +9.9E+30

   do ldet=0, MAX_GBM_DET-1
      if (RequestDetectors (ldet) ) then
         StartVal = max (StartVal, TTE_Data (ldet) % FileStartTime)
         StopVal  = min (StopVal,  TTE_Data (ldet) % FileStopTime)
      end if
   end do

   write (*, '( "The files contain data from", F14.3, " to", F14.3, " s." )' )  StartVal, StopVal

   if ( TriggerTime > 0.0_C_DOUBLE ) then

      FindUsedDet: do ldet=0, MAX_GBM_DET-1
         if ( RequestDetectors (ldet) ) then
            ReferenceTime = TTE_Data (ldet) % TriggerTime
            exit FindUsedDet
         end if
      end do FindUsedDet

   else

      write (*, '( "Input a Reference Time which will serve as a ""base"" to refer times to." )' )
      write (*, '( "For ground-search TGFs, this could be the MET of the bin of the candidate." )' )
      write (*, '( "Input the Reference Time: " )', advance="no" )
      read (*, *)  ReferenceTime

      if ( ReferenceTime < StartVal  .OR.  ReferenceTime > StopVal )    &
         write (*, '( / "WARNING!! -- Time", F15.3, " appears to be outside of the time range of the file !" / )' )  ReferenceTime

   end if


   return

end subroutine  Determine_ReferenceTime


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


function Make_X_AxisLabel ( TimeRefType, TimeLabel, ClockDrift_Flag, GapTime, ReferenceTime )

   character (len=80) :: Make_X_AxisLabel   ! function return

   character (len=9), intent (in) :: TimeRefType
   character (len=2), intent (in) :: TimeLabel
   logical, intent (in) :: ClockDrift_Flag
   real (real32), intent (in) :: GapTime
   real (real128), intent (in) :: ReferenceTime


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   character (len=80) :: X_AxisLabel

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   if ( ClockDrift_Flag ) then
      write ( X_AxisLabel, '( "Time (", A, ") relative to corrected ", A, " Time", F18.6 )' )     &
            trim (TimeLabel), trim (TimeRefType), MET_Corr_for_ClockDrift ( ReferenceTime, GapTime )
   else
      write ( X_AxisLabel, '( "Time (", A, ") relative to raw ", A, " Time", F18.6 )' )     &
           trim (TimeLabel), trim (TimeRefType), ReferenceTime
   end if


   Make_X_AxisLabel = X_AxisLabel

   return

end function Make_X_AxisLabel


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  Determines bin number that corresponds to time in seconds.

function  Time_to_BinNum_on_Graph ( time, BinWidth_s, GraphTimeRange_s )


   integer (int32) :: Time_to_BinNum_on_Graph

   real (real64), intent (in) :: time
   real (real64), intent (in) :: BinWidth_s
   real (real64), dimension (2) :: GraphTimeRange_s



   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   Time_to_BinNum_on_Graph = floor ( ( time - GraphTimeRange_s (1) ) / BinWidth_s )

   return

end function  Time_to_BinNum_on_Graph



! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^



end module TTE_subroutines

