//  Michael S. Briggs, 2010 Feb 15.    UA Huntsville.   Copyright 2010.
//  Revised 2012 Nov 25: close the file when done !!

//  Revised 2011 February 9: add output of Quaternion and Angular Velocity.

//  Adapted from 2008 October 26 -- Nov 11 to  2009 December 24.

//  Subroutine Read_Lvl1_SC_PosHist_SV uses cfitsio to read
//  GBM Level 1 Spacecraft Position History files.   This version
//  returns selected information -- the values from the file before &
//  after an input time.
//  "Read_Lvl1_SC_PosHist_SV": "SV" = "Selected Values".

//  2011 May 15: add a special case ... if the requested time is slightly
//  past the last time in the file, up to 3 seconds, then this routine
//  will return the last two records.   Of course, they no longer
//  bracket the requested time.   This is done because SC Position History
//  files are known to end before the end of the time, such as at
//  23h 59m 57.x s.


//  Subroutine Read_Lvl1_SC_PosHist uses cfitsio to read
//  GBM Level 1 Spacecraft Position History files.   This version returns,
//  for certain items such as Spacecraft position X, Y, & Z, vectors with
//  all values contained in the file.

//  For subroutine Read_Lvl1_SC_PosHist (not Read_Lvl1_SC_PosHist_SV),
//  when you are done with the vectors, call Cleanup_Read_Lvl1_SC_PosHist
//  to free the memory.

//  Callable from Fortran 95/2003 via an interface that I have written,
//  using the ISO C Binding.

//  There are three macro symbols that may be specified on the compile command
//  with '-D' to include additional code / features:
//  VERBOSE_READ_SCPOS, DEBUG_READ_SCPOS, and FIND_ALT_RANGE.
//  FIND_ALT_RANGE is only available in Read_Lvl1_SC_PosHist_SV.


// Headers

#include <stdio.h>
#include <math.h>

#include "fitsio.h"



//  #define symbols:

#define  FILE_NAME_MAX_LEN  200




//  local function prototypes:


void  Output_StatusValue_as_Msg (

   const int  status

);


// ****************************************************************************************


void  Read_Lvl1_SC_PosHist_SV (

   // input arguments:
   char InputFileName [FILE_NAME_MAX_LEN],
   double Request_MET,

   //  output arguments:

   //  scalers:
   double * Before_MET_ptr,   // seconds
   double * After_MET_ptr,

   // 4-vectors:
   double * Before_QSJ_ptr,
   double * After_QSJ_ptr,

   // 3-vectors:
   double * Before_WSJ_ptr,   // rad / s
   double * After_WSJ_ptr,

   float * Before_Position_ptr,   // m
   float * After_Position_ptr,

   float * Before_Velocity_ptr,   //  m / s
   float * After_Velocity_ptr

) {


   fitsfile * fits_file_ptr;
   int  status = 0;

   long int  Num_History_Recs;

   // pointers to pointers to vectors of length Num_History_Recs

   double * MET_ptr;

   double * QSJ_1_ptr;
   double * QSJ_2_ptr;
   double * QSJ_3_ptr;
   double * QSJ_4_ptr;

   double * WSJ_1_ptr;    // rad / s
   double * WSJ_2_ptr;
   double * WSJ_3_ptr;

   float * Position_X_ptr;    //  m
   float * Position_Y_ptr;
   float * Position_Z_ptr;

   float * Velocity_X_ptr;    //  m / s
   float * Velocity_Y_ptr;
   float * Velocity_Z_ptr;


   float  fnulval;
   double  dnulval;
   int anynul;

   int  NumHeaders;
   long int  i;

   #ifdef FIND_ALT_RANGE
      float  altitude;
      float  alt_min = +9.9E+10;
      float  alt_max = -9.9E+10;
   #endif


   // * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   //  Open the FITS file

   fits_open_file ( &fits_file_ptr, InputFileName, READONLY, &status );

   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to open file '%s'\n", InputFileName );
      exit (22);
   }



   // optional, extra test -- does the file have the expected number of headers?

   fits_get_num_hdus ( fits_file_ptr, &NumHeaders, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to obtain number of Headers.\n" );
      exit (23);
   }
   if ( NumHeaders != 2 ) {
      printf ( "\nNumber of headers has unexpected value !!: %d\n", NumHeaders );
   }


   //  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^


   //  Move to "GLAST POS HIST" Header.

   fits_movnam_hdu ( fits_file_ptr, BINARY_TBL, "GLAST POS HIST", 0, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to locate GLAST POS HIST header\n" );
      exit (25);
   }



   //   How many history records are in the file?
   //   Equivalent to how many rows are in this Table:

   fits_get_num_rows ( fits_file_ptr, &Num_History_Recs, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read number of Position History Records events.\n" );
      exit (29);
   }


   #if defined (VERBOSE_READ_SCPOS) || defined (DEBUG_READ_SCPOS)
      printf ( "\nThere are %ld Position History Records in file '%s'.\n", Num_History_Recs, InputFileName );
   #endif



   //   Allocate memory for the Position data:

   MET_ptr = malloc ( Num_History_Recs * sizeof (double) );

   QSJ_1_ptr = malloc ( Num_History_Recs * sizeof (double) );
   QSJ_2_ptr = malloc ( Num_History_Recs * sizeof (double) );
   QSJ_3_ptr = malloc ( Num_History_Recs * sizeof (double) );
   QSJ_4_ptr = malloc ( Num_History_Recs * sizeof (double) );

   WSJ_1_ptr = malloc ( Num_History_Recs * sizeof (double) );
   WSJ_2_ptr = malloc ( Num_History_Recs * sizeof (double) );
   WSJ_3_ptr = malloc ( Num_History_Recs * sizeof (double) );

   Position_X_ptr = malloc ( Num_History_Recs * sizeof (float) );
   Position_Y_ptr = malloc ( Num_History_Recs * sizeof (float) );
   Position_Z_ptr = malloc ( Num_History_Recs * sizeof (float) );

   Velocity_X_ptr = malloc ( Num_History_Recs * sizeof (float) );
   Velocity_Y_ptr = malloc ( Num_History_Recs * sizeof (float) );
   Velocity_Z_ptr = malloc ( Num_History_Recs * sizeof (float) );


   #ifdef DEBUG_READ_SCPOS
      printf ( "\nAllocated sizes: %ld %ld \n", Num_History_Recs * sizeof (double), Num_History_Recs * sizeof (float) );
      printf ( "\nOne T address created by malloc: %p\n", MET_ptr );
      printf ( "\nFour Q addresses created by malloc: %p  %p  %p  %p\n", QSJ_1_ptr, QSJ_2_ptr, QSJ_3_ptr, QSJ_4_ptr );
      printf ( "\nThree W addresses created by malloc: %p  %p  %p\n", WSJ_1_ptr, WSJ_2_ptr, WSJ_3_ptr );
      printf ( "\nThree P addresses created by malloc: %p  %p  %p\n", Position_X_ptr, Position_Y_ptr, Position_Z_ptr );
      printf ( "\nThree V addresses created by malloc: %p  %p  %p\n", Velocity_X_ptr, Velocity_Y_ptr, Velocity_Z_ptr );
   #endif


   if ( MET_ptr == NULL  ||
        QSJ_1_ptr == NULL  ||  QSJ_2_ptr == NULL  ||  QSJ_3_ptr == NULL  ||  QSJ_4_ptr == NULL  ||
        WSJ_1_ptr == NULL  ||  WSJ_2_ptr == NULL  ||  WSJ_3_ptr == NULL  ||
        Position_X_ptr == NULL  ||  Position_Y_ptr == NULL  ||  Position_Z_ptr == NULL  ||
        Velocity_X_ptr == NULL  ||  Velocity_Y_ptr == NULL  ||  Velocity_Z_ptr == NULL ) {
      printf ( "\n\nFailed to allocate memory for the Position History data !\n" );
      exit (30);
   }


   //  Read column 1 = TIME

   dnulval = -7.77E+300;
   fits_read_col ( fits_file_ptr, TDOUBLE, 1, 1, 1, Num_History_Recs, &dnulval, MET_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 1 with the Times.\n" );
      exit (31);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for Time!\n" );


   //  Read column 2 = Quaternion 1

   dnulval = -7.77E+300;
   fits_read_col ( fits_file_ptr, TDOUBLE, 2, 1, 1, Num_History_Recs, &dnulval, QSJ_1_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 2 with the QSJ 1.\n" );
      exit (41);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for QSJ 1!\n" );


   //  Read column 3 = Quaternion 2

   dnulval = -7.77E+300;
   fits_read_col ( fits_file_ptr, TDOUBLE, 3, 1, 1, Num_History_Recs, &dnulval, QSJ_2_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 3 with the QSJ 2.\n" );
      exit (42);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for QSJ 2!\n" );


   //  Read column 4 = Quaternion 3

   dnulval = -7.77E+300;
   fits_read_col ( fits_file_ptr, TDOUBLE, 4, 1, 1, Num_History_Recs, &dnulval, QSJ_3_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 4 with the QSJ 3.\n" );
      exit (43);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for QSJ 3!\n" );


   //  Read column 5 = Quaternion 4

   dnulval = -7.77E+300;
   fits_read_col ( fits_file_ptr, TDOUBLE, 5, 1, 1, Num_History_Recs, &dnulval, QSJ_4_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 5 with the QSJ 4.\n" );
      exit (44);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for QSJ 4!\n" );


   //  Read column 6 = Angular Velocity 1

   dnulval = -7.77E+300;
   fits_read_col ( fits_file_ptr, TDOUBLE, 6, 1, 1, Num_History_Recs, &dnulval, WSJ_1_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 6 with the WSJ 1.\n" );
      exit (45);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for WSJ 1!\n" );


   //  Read column 7 = Angular Velocity 2

   dnulval = -7.77E+300;
   fits_read_col ( fits_file_ptr, TDOUBLE, 7, 1, 1, Num_History_Recs, &dnulval, WSJ_2_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 7 with the WSJ 2.\n" );
      exit (46);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for WSJ 2!\n" );


   //  Read column 8 = Angular Velocity 3

   dnulval = -7.77E+300;
   fits_read_col ( fits_file_ptr, TDOUBLE, 8, 1, 1, Num_History_Recs, &dnulval, WSJ_3_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 8 with the WSJ 3.\n" );
      exit (47);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for WSJ 3!\n" );


   //  Read column 9 = Position X

   fnulval = -7.77E+37;
   fits_read_col ( fits_file_ptr, TFLOAT, 9, 1, 1, Num_History_Recs, &fnulval, Position_X_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 9 with Position X.\n" );
      exit (32);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for Position X!\n" );


   //  Read column 10 = Position Y

   fnulval = -7.77E+37;
   fits_read_col ( fits_file_ptr, TFLOAT, 10, 1, 1, Num_History_Recs, &fnulval, Position_Y_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 10 with Position Y.\n" );
      exit (33);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for Position Y!\n" );


   //  Read column 11 = Position Z

   fnulval = -7.77E+37;
   fits_read_col ( fits_file_ptr, TFLOAT, 11, 1, 1, Num_History_Recs, &fnulval, Position_Z_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 11 with Position Z.\n" );
      exit (34);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for Position Z!\n" );


   //  Read column 12 = Velocity X

   fnulval = -7.77E+37;
   fits_read_col ( fits_file_ptr, TFLOAT, 12, 1, 1, Num_History_Recs, &fnulval, Velocity_X_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 12 with Velocity X.\n" );
      exit (35);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for Velocity X!\n" );


   //  Read column 13 = Velocity Y

   fnulval = -7.77E+37;
   fits_read_col ( fits_file_ptr, TFLOAT, 13, 1, 1, Num_History_Recs, &fnulval, Velocity_Y_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 13 with Velocity Y.\n" );
      exit (36);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for Velocity Y!\n" );


   //  Read column 14 = Velocity Z

   fnulval = -7.77E+37;
   fits_read_col ( fits_file_ptr, TFLOAT, 14, 1, 1, Num_History_Recs, &fnulval, Velocity_Z_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 14 with Velocity Z.\n" );
      exit (37);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for Velocity Z!\n" );



   #ifdef DEBUG_READ_SCPOS
      printf ( "\n1st Record: time: %lf\n", MET_ptr [0] );

      printf ( "Quaternion: %lf  %lf  %lf  %lf\n", QSJ_1_ptr [0], QSJ_2_ptr [0], QSJ_3_ptr [0], QSJ_4_ptr [0] );
      printf ( "Angular Velocity: %lf  %lf  %lf\n", WSJ_1_ptr [0], WSJ_2_ptr [0], WSJ_3_ptr [0] );
      printf ( "Position: %f  %f  %f\n", Position_X_ptr [0], Position_Y_ptr [0], Position_Z_ptr [0] );
      printf ( "Velocity: %f  %f  %f\n", Velocity_X_ptr [0], Velocity_Y_ptr [0], Velocity_Z_ptr [0] );

      printf ( "\n2nd Record: time: %lf\n", MET_ptr [1]  );
      printf ( "Quaternion: %lf  %lf  %lf  %lf\n", QSJ_1_ptr [1], QSJ_2_ptr [1], QSJ_3_ptr [1], QSJ_4_ptr [1] );
      printf ( "Angular Velocity: %lf  %lf  %lf\n", WSJ_1_ptr [1], WSJ_2_ptr [1], WSJ_3_ptr [1] );
      printf ( "Position: %f  %f  %f\n", Position_X_ptr [1], Position_Y_ptr [1], Position_Z_ptr [1] );
      printf ( "Velocity: %f  %f  %f\n", Velocity_X_ptr [1], Velocity_Y_ptr [1], Velocity_Z_ptr [1] );

      printf ( "\nLast record: time: %lf\n", MET_ptr [Num_History_Recs-1] );
      printf ( "Quaternion: %lf  %lf  %lf  %lf\n", QSJ_1_ptr [Num_History_Recs - 1], QSJ_2_ptr [Num_History_Recs - 1], QSJ_3_ptr [Num_History_Recs - 1], QSJ_4_ptr [Num_History_Recs - 1] );
      printf ( "Angular Velocity: %lf  %lf  %lf\n", WSJ_1_ptr [Num_History_Recs - 1], WSJ_2_ptr [Num_History_Recs - 1], WSJ_3_ptr [Num_History_Recs - 1] );
      printf ( "Position: %f  %f  %f\n",
          Position_X_ptr [Num_History_Recs-1], Position_Y_ptr [Num_History_Recs-1], Position_Z_ptr [Num_History_Recs-1] );
      printf ( "Velocity: %f  %f  %f\n",
          Velocity_X_ptr [Num_History_Recs-1], Velocity_Y_ptr [Num_History_Recs-1], Velocity_Z_ptr [Num_History_Recs-1] );
   #endif



   //  Extract the data of most interest -- locate the two records that
   //  bracket the requested time.  We allow times slightly before or
   //  after the start or end of the file.


   if ( Request_MET <= MET_ptr [0] - 2.5  ||  Request_MET > MET_ptr [Num_History_Recs-1] + 3.0 ) {
      printf ( "\nERROR: Requested time is not within the input Position History File!\n" );
      printf ( "Requested time: %lf\n", Request_MET );
      printf ( "Time range of the file: %lf to %lf\n", MET_ptr [0], MET_ptr [Num_History_Recs-1] );
      exit (39);
   }


   if ( Request_MET > MET_ptr [Num_History_Recs-1] ) {   // just past last file time ?

      // Special case ... the requested time is just past the last time in the file.
      // We know that this occurs because SC Position History files are observed
      // to end at times such as 23h 59m 57.x s.  In this case return the
      // last two records in the file.  However, these records won't bracket
      // the requested time.

      i = Num_History_Recs-1;

   } else if ( Request_MET <= MET_ptr [0] ) {

      // Special case ... the requested time is just before the first time in the file.
      // This might occur because of a data gap in the file.  In this case return the
      // first two records in the file.  However, these records won't bracket
      // the requested time.

      i = 1;

   } else {

      //  The requested time is within the file.  Normal case.
      //  Tests above guarantee that the while loop will terminate.
      //  Test "Request_MET <= MET_ptr [0]" guarantees that i will
      //  advance to at least 1, which is needed by the code below.

      i = 0;
      while ( Request_MET > MET_ptr [i] )  i++;

   }  // just past last file time ?


   *Before_MET_ptr = MET_ptr [i-1];

   Before_QSJ_ptr [0] = QSJ_1_ptr [i-1];
   Before_QSJ_ptr [1] = QSJ_2_ptr [i-1];
   Before_QSJ_ptr [2] = QSJ_3_ptr [i-1];
   Before_QSJ_ptr [3] = QSJ_4_ptr [i-1];

   Before_WSJ_ptr [0] = WSJ_1_ptr [i-1];
   Before_WSJ_ptr [1] = WSJ_2_ptr [i-1];
   Before_WSJ_ptr [2] = WSJ_3_ptr [i-1];

   Before_Position_ptr [0] = Position_X_ptr [i-1];
   Before_Position_ptr [1] = Position_Y_ptr [i-1];
   Before_Position_ptr [2] = Position_Z_ptr [i-1];

   Before_Velocity_ptr [0] = Velocity_X_ptr [i-1];
   Before_Velocity_ptr [1] = Velocity_Y_ptr [i-1];
   Before_Velocity_ptr [2] = Velocity_Z_ptr [i-1];


   *After_MET_ptr = MET_ptr [i];

   After_QSJ_ptr [0] = QSJ_1_ptr [i];
   After_QSJ_ptr [1] = QSJ_2_ptr [i];
   After_QSJ_ptr [2] = QSJ_3_ptr [i];
   After_QSJ_ptr [3] = QSJ_4_ptr [i];

   After_WSJ_ptr [0] = WSJ_1_ptr [i];
   After_WSJ_ptr [1] = WSJ_2_ptr [i];
   After_WSJ_ptr [2] = WSJ_3_ptr [i];

   After_Position_ptr [0] = Position_X_ptr [i];
   After_Position_ptr [1] = Position_Y_ptr [i];
   After_Position_ptr [2] = Position_Z_ptr [i];

   After_Velocity_ptr [0] = Velocity_X_ptr [i];
   After_Velocity_ptr [1] = Velocity_Y_ptr [i];
   After_Velocity_ptr [2] = Velocity_Z_ptr [i];



   #ifdef FIND_ALT_RANGE
      for (i=0;  i < Num_History_Recs;  i++ ) {
         altitude = sqrt ( Position_X_ptr [i] * Position_X_ptr [i]  +
                           Position_Y_ptr [i] * Position_Y_ptr [i]  +
                           Position_Z_ptr [i] * Position_Z_ptr [i] );
         if ( altitude > alt_max )  alt_max = altitude;
         if ( altitude < alt_min )  alt_min = altitude;
      }
      //  calculation: convert from m to km, subtract mean radius of the Earth.
      printf ( "\nThe range of altitudes in the file is %7.2f to %7.2f km\n",
        alt_min / 1000.0 - 6371.0, alt_max / 1000.0 - 6371.0 );
   #endif


   //  We've extracted the data that we want, so free the big arrays:

   free ( MET_ptr );

   free (QSJ_1_ptr);
   free (QSJ_2_ptr);
   free (QSJ_3_ptr);
   free (QSJ_4_ptr);

   free (WSJ_1_ptr);
   free (WSJ_2_ptr);
   free (WSJ_3_ptr);

   free ( Position_X_ptr );
   free ( Position_Y_ptr );
   free ( Position_Z_ptr );

   free ( Velocity_X_ptr );
   free ( Velocity_Y_ptr );
   free ( Velocity_Z_ptr );


   //  Close the FITS file

   fits_close_file ( fits_file_ptr, &status );

   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to close file '%s'\n", InputFileName );
      exit (41);
   }


   return;

}  // Read_Lvl1_SC_PosHist_SV ()



// ****************************************************************************************

void  Read_Lvl1_SC_PosHist (

   // input arguments:
   char InputFileName [FILE_NAME_MAX_LEN],

   //  output arguments:

   //  scalers:
   long int * Num_History_Recs_ptr,

   // pointers to vectors of length *Num_History_Recs_ptr
   double ** MET_ptr_to_ptr,

   double ** QSJ_1_ptr_to_ptr,
   double ** QSJ_2_ptr_to_ptr,
   double ** QSJ_3_ptr_to_ptr,
   double ** QSJ_4_ptr_to_ptr,

   double ** WSJ_1_ptr_to_ptr,    // rad / s
   double ** WSJ_2_ptr_to_ptr,
   double ** WSJ_3_ptr_to_ptr,

   float ** Position_X_ptr_to_ptr,    //  m
   float ** Position_Y_ptr_to_ptr,
   float ** Position_Z_ptr_to_ptr,

   float ** Velocity_X_ptr_to_ptr,    //  m / s
   float ** Velocity_Y_ptr_to_ptr,
   float ** Velocity_Z_ptr_to_ptr

) {


   fitsfile * fits_file_ptr;
   int  status = 0;


   float  fnulval;
   double  dnulval;
   int anynul;

   int  NumHeaders;

   #ifdef DEBUG_READ_SCPOS
		double * MET_ptr;
      double * QSJ_1_ptr;
      double * QSJ_2_ptr;
      double * QSJ_3_ptr;
      double * QSJ_4_ptr;
      double * WSJ_1_ptr;
      double * WSJ_2_ptr;
      double * WSJ_3_ptr;
		float * Position_X_ptr;
		float * Position_Y_ptr;
		float * Position_Z_ptr;
		float * Velocity_X_ptr;
		float * Velocity_Y_ptr;
		float * Velocity_Z_ptr;
   #endif


   // * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   //  Open the FITS file

   fits_open_file ( &fits_file_ptr, InputFileName, READONLY, &status );

   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to open file '%s'\n", InputFileName );
      exit (22);
   }



   // optional, extra test -- does the file have the expected number of headers?

   fits_get_num_hdus ( fits_file_ptr, &NumHeaders, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to obtain number of Headers.\n" );
      exit (23);
   }
   if ( NumHeaders != 2 ) {
      printf ( "\nNumber of headers has unexpected value !!: %d\n", NumHeaders );
   }


   //  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^  ^


   //  Move to "GLAST POS HIST" Header.

   fits_movnam_hdu ( fits_file_ptr, BINARY_TBL, "GLAST POS HIST", 0, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to locate GLAST POS HIST header\n" );
      exit (25);
   }



   //   How many history records are in the file?
   //   Equivalent to how many rows are in this Table:

   fits_get_num_rows ( fits_file_ptr, Num_History_Recs_ptr, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read number of Position History Records events.\n" );
      exit (29);
   }


   #if defined (VERBOSE_READ_SCPOS) || defined (DEBUG_READ_SCPOS)
      printf ( "\nThere are %ld Position History Records in file '%s'.\n", *Num_History_Recs_ptr, InputFileName );
   #endif



   //   Allocate memory for the Position data:

   *MET_ptr_to_ptr = malloc ( *Num_History_Recs_ptr * sizeof (double) );

   *QSJ_1_ptr_to_ptr = malloc ( *Num_History_Recs_ptr * sizeof (double) );
   *QSJ_2_ptr_to_ptr = malloc ( *Num_History_Recs_ptr * sizeof (double) );
   *QSJ_3_ptr_to_ptr = malloc ( *Num_History_Recs_ptr * sizeof (double) );
   *QSJ_4_ptr_to_ptr = malloc ( *Num_History_Recs_ptr * sizeof (double) );

   *WSJ_1_ptr_to_ptr = malloc ( *Num_History_Recs_ptr * sizeof (double) );
   *WSJ_2_ptr_to_ptr = malloc ( *Num_History_Recs_ptr * sizeof (double) );
   *WSJ_3_ptr_to_ptr = malloc ( *Num_History_Recs_ptr * sizeof (double) );

   *Position_X_ptr_to_ptr = malloc ( *Num_History_Recs_ptr * sizeof (float) );
   *Position_Y_ptr_to_ptr = malloc ( *Num_History_Recs_ptr * sizeof (float) );
   *Position_Z_ptr_to_ptr = malloc ( *Num_History_Recs_ptr * sizeof (float) );

   *Velocity_X_ptr_to_ptr = malloc ( *Num_History_Recs_ptr * sizeof (float) );
   *Velocity_Y_ptr_to_ptr = malloc ( *Num_History_Recs_ptr * sizeof (float) );
   *Velocity_Z_ptr_to_ptr = malloc ( *Num_History_Recs_ptr * sizeof (float) );


   #ifdef DEBUG_READ_SCPOS
      printf ( "\nAllocated sizes: %ld %ld \n", *Num_History_Recs_ptr * sizeof (double), *Num_History_Recs_ptr * sizeof (float) );
      printf ( "\nOne T address created by malloc: %p\n", *MET_ptr_to_ptr );
      printf ( "\nFour Q addresses created by malloc: %p  %p  %p  %p\n", *QSJ_1_ptr_to_ptr, *QSJ_2_ptr_to_ptr, *QSJ_3_ptr_to_ptr, *QSJ_4_ptr_to_ptr );
      printf ( "\nThree W addresses created by malloc: %p  %p  %p\n", *WSJ_1_ptr_to_ptr, *WSJ_2_ptr_to_ptr, *WSJ_3_ptr_to_ptr );
      printf ( "\nThree P addresses created by malloc: %p  %p  %p\n", *Position_X_ptr_to_ptr, *Position_Y_ptr_to_ptr, *Position_Z_ptr_to_ptr );
      printf ( "\nThree V addresses created by malloc: %p  %p  %p\n", *Velocity_X_ptr_to_ptr, *Velocity_Y_ptr_to_ptr, *Velocity_Z_ptr_to_ptr );
   #endif


   if ( *MET_ptr_to_ptr == NULL  ||
        *QSJ_1_ptr_to_ptr == NULL  ||  *QSJ_2_ptr_to_ptr == NULL  ||  *QSJ_3_ptr_to_ptr == NULL  ||  *QSJ_4_ptr_to_ptr == NULL  ||
        *WSJ_1_ptr_to_ptr == NULL  ||  *WSJ_2_ptr_to_ptr == NULL  ||  *WSJ_3_ptr_to_ptr == NULL  ||
        *Position_X_ptr_to_ptr == NULL  ||  *Position_Y_ptr_to_ptr == NULL  ||  *Position_Z_ptr_to_ptr == NULL  ||
        *Velocity_X_ptr_to_ptr == NULL  ||  *Velocity_Y_ptr_to_ptr == NULL  ||  *Velocity_Z_ptr_to_ptr == NULL ) {
      printf ( "\n\nFailed to allocate memory for the Position History data !\n" );
      exit (30);
   }


   //  Read column 1 = TIME

   dnulval = -7.77E+300;
   fits_read_col ( fits_file_ptr, TDOUBLE, 1, 1, 1, *Num_History_Recs_ptr, &dnulval, *MET_ptr_to_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 1 with the Times.\n" );
      exit (31);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for Time!\n" );


   //  Read column 2 = Quaternion 1

   dnulval = -7.77E+300;
   fits_read_col ( fits_file_ptr, TDOUBLE, 2, 1, 1, *Num_History_Recs_ptr, &dnulval, *QSJ_1_ptr_to_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 2 with the QSJ 1.\n" );
      exit (41);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for QSJ 1!\n" );


   //  Read column 3 = Quaternion 2

   dnulval = -7.77E+300;
   fits_read_col ( fits_file_ptr, TDOUBLE, 3, 1, 1, *Num_History_Recs_ptr, &dnulval, *QSJ_2_ptr_to_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 3 with the QSJ 2.\n" );
      exit (42);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for QSJ 2!\n" );


   //  Read column 4 = Quaternion 3

   dnulval = -7.77E+300;
   fits_read_col ( fits_file_ptr, TDOUBLE, 4, 1, 1, *Num_History_Recs_ptr, &dnulval, *QSJ_3_ptr_to_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 4 with the QSJ 3.\n" );
      exit (43);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for QSJ 3!\n" );


   //  Read column 5 = Quaternion 4

   dnulval = -7.77E+300;
   fits_read_col ( fits_file_ptr, TDOUBLE, 5, 1, 1, *Num_History_Recs_ptr, &dnulval, *QSJ_4_ptr_to_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 5 with the QSJ 4.\n" );
      exit (44);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for QSJ 4!\n" );


   //  Read column 6 = Angular Velocity 1

   dnulval = -7.77E+300;
   fits_read_col ( fits_file_ptr, TDOUBLE, 6, 1, 1, *Num_History_Recs_ptr, &dnulval, *WSJ_1_ptr_to_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 6 with the WSJ 1.\n" );
      exit (45);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for WSJ 1!\n" );


   //  Read column 7 = Angular Velocity 2

   dnulval = -7.77E+300;
   fits_read_col ( fits_file_ptr, TDOUBLE, 7, 1, 1, *Num_History_Recs_ptr, &dnulval, *WSJ_2_ptr_to_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 7 with the WSJ 2.\n" );
      exit (46);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for WSJ 2!\n" );


   //  Read column 8 = Angular Velocity 3

   dnulval = -7.77E+300;
   fits_read_col ( fits_file_ptr, TDOUBLE, 8, 1, 1, *Num_History_Recs_ptr, &dnulval, *WSJ_3_ptr_to_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 8 with the WSJ 3.\n" );
      exit (47);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for WSJ 3!\n" );


   //  Read column 9 = Position X

   fnulval = -7.77E+37;
   fits_read_col ( fits_file_ptr, TFLOAT, 9, 1, 1, *Num_History_Recs_ptr, &fnulval, *Position_X_ptr_to_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 9 with Position X.\n" );
      exit (32);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for Position X!\n" );


   //  Read column 10 = Position Y

   fnulval = -7.77E+37;
   fits_read_col ( fits_file_ptr, TFLOAT, 10, 1, 1, *Num_History_Recs_ptr, &fnulval, *Position_Y_ptr_to_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 10 with Position Y.\n" );
      exit (33);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for Position Y!\n" );


   //  Read column 11 = Position Z

   fnulval = -7.77E+37;
   fits_read_col ( fits_file_ptr, TFLOAT, 11, 1, 1, *Num_History_Recs_ptr, &fnulval, *Position_Z_ptr_to_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 11 with Position Z.\n" );
      exit (34);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for Position Z!\n" );


   //  Read column 12 = Velocity X

   fnulval = -7.77E+37;
   fits_read_col ( fits_file_ptr, TFLOAT, 12, 1, 1, *Num_History_Recs_ptr, &fnulval, *Velocity_X_ptr_to_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 12 with Velocity X.\n" );
      exit (35);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for Velocity X!\n" );


   //  Read column 13 = Velocity Y

   fnulval = -7.77E+37;
   fits_read_col ( fits_file_ptr, TFLOAT, 13, 1, 1, *Num_History_Recs_ptr, &fnulval, *Velocity_Y_ptr_to_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 13 with Velocity Y.\n" );
      exit (36);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for Velocity Y!\n" );


   //  Read column 14 = Velocity Z

   fnulval = -7.77E+37;
   fits_read_col ( fits_file_ptr, TFLOAT, 14, 1, 1, *Num_History_Recs_ptr, &fnulval, *Velocity_Z_ptr_to_ptr, &anynul, &status );
   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to read Column 14 with Velocity Z.\n" );
      exit (37);
   }
   if ( anynul != 0 ) printf ( "\nNull value encountered for Velocity Z!\n" );



   #ifdef DEBUG_READ_SCPOS
      MET_ptr = *MET_ptr_to_ptr;

      QSJ_1_ptr = *QSJ_1_ptr_to_ptr;
      QSJ_2_ptr = *QSJ_2_ptr_to_ptr;
      QSJ_3_ptr = *QSJ_3_ptr_to_ptr;
      QSJ_4_ptr = *QSJ_4_ptr_to_ptr;

      WSJ_1_ptr = *WSJ_1_ptr_to_ptr;
      WSJ_2_ptr = *WSJ_2_ptr_to_ptr;
      WSJ_3_ptr = *WSJ_3_ptr_to_ptr;

      Position_X_ptr = *Position_X_ptr_to_ptr;
      Position_Y_ptr = *Position_Y_ptr_to_ptr;
      Position_Z_ptr = *Position_Z_ptr_to_ptr;

      Velocity_X_ptr = *Velocity_X_ptr_to_ptr;
      Velocity_Y_ptr = *Velocity_Y_ptr_to_ptr;
      Velocity_Z_ptr = *Velocity_Z_ptr_to_ptr;

      printf ( "\n1st Record: time: %lf\n", MET_ptr [0] );
      printf ( "Quaternion: %lf  %lf  %lf  %lf\n", QSJ_1_ptr [0], QSJ_2_ptr [0], QSJ_3_ptr [0], QSJ_4_ptr [0] );
      printf ( "Angular Velocity: %lf  %lf  %lf\n", WSJ_1_ptr [0], WSJ_2_ptr [0], WSJ_3_ptr [0] );
      printf ( "Position: %f  %f  %f\n", Position_X_ptr [0], Position_Y_ptr [0], Position_Z_ptr [0] );
      printf ( "Velocity: %f  %f  %f\n", Velocity_X_ptr [0], Velocity_Y_ptr [0], Velocity_Z_ptr [0] );

      printf ( "\n2nd Record: time: %lf\n", MET_ptr [1]  );
      printf ( "Quaternion: %lf  %lf  %lf  %lf\n", QSJ_1_ptr [1], QSJ_2_ptr [1], QSJ_3_ptr [1], QSJ_4_ptr [1] );
      printf ( "Angular Velocity: %lf  %lf  %lf\n", WSJ_1_ptr [1], WSJ_2_ptr [1], WSJ_3_ptr [1] );
      printf ( "Position: %f  %f  %f\n", Position_X_ptr [1], Position_Y_ptr [1], Position_Z_ptr [1] );
      printf ( "Velocity: %f  %f  %f\n", Velocity_X_ptr [1], Velocity_Y_ptr [1], Velocity_Z_ptr [1] );

      printf ( "\nLast record: time: %lf\n", MET_ptr [*Num_History_Recs_ptr-1] );
      printf ( "Quaternion: %lf  %lf  %lf  %lf\n", QSJ_1_ptr [*Num_History_Recs_ptr-1], QSJ_2_ptr [*Num_History_Recs_ptr-1],
           QSJ_3_ptr [*Num_History_Recs_ptr-1], QSJ_4_ptr [*Num_History_Recs_ptr-1] );
      printf ( "Angular Velocity: %lf  %lf  %lf\n", WSJ_1_ptr [*Num_History_Recs_ptr-1], WSJ_2_ptr [*Num_History_Recs_ptr-1],
           WSJ_3_ptr [*Num_History_Recs_ptr-1] );
      printf ( "Position: %f  %f  %f\n",
          Position_X_ptr [*Num_History_Recs_ptr-1], Position_Y_ptr [*Num_History_Recs_ptr-1], Position_Z_ptr [*Num_History_Recs_ptr-1] );
      printf ( "Velocity: %f  %f  %f\n",
          Velocity_X_ptr [*Num_History_Recs_ptr-1], Velocity_Y_ptr [*Num_History_Recs_ptr-1], Velocity_Z_ptr [*Num_History_Recs_ptr-1] );
   #endif


   //  Close the FITS file

   fits_close_file ( fits_file_ptr, &status );

   if ( status != 0 ) {
      Output_StatusValue_as_Msg (status);
      printf ( "==> Failed to close file '%s'\n", InputFileName );
      exit (41);
   }


   return;

}  // Read_Lvl1_SC_PosHist ()



// ****************************************************************************************

//  Call this routine when the data vectors are no longer needed --
//  it uses the C-routine "free" to release the memory.

void  Cleanup_Read_Lvl1_SC_PosHist (

   // pointers to pointers vectors...

   double ** MET_ptr_to_ptr,

   double ** QSJ_1_ptr_to_ptr,
   double ** QSJ_2_ptr_to_ptr,
   double ** QSJ_3_ptr_to_ptr,
   double ** QSJ_4_ptr_to_ptr,

   double ** WSJ_1_ptr_to_ptr,    // rad / s
   double ** WSJ_2_ptr_to_ptr,
   double ** WSJ_3_ptr_to_ptr,

   float ** Position_X_ptr_to_ptr,
   float ** Position_Y_ptr_to_ptr,
   float ** Position_Z_ptr_to_ptr,

   float ** Velocity_X_ptr_to_ptr,
   float ** Velocity_Y_ptr_to_ptr,
   float ** Velocity_Z_ptr_to_ptr

) {

   free ( *MET_ptr_to_ptr );

   free ( *QSJ_1_ptr_to_ptr );
   free ( *QSJ_2_ptr_to_ptr );
   free ( *QSJ_3_ptr_to_ptr );
   free ( *QSJ_4_ptr_to_ptr );

   free ( *WSJ_1_ptr_to_ptr );
   free ( *WSJ_2_ptr_to_ptr );
   free ( *WSJ_3_ptr_to_ptr );

   free ( *Position_X_ptr_to_ptr );
   free ( *Position_Y_ptr_to_ptr );
   free ( *Position_Z_ptr_to_ptr );

   free ( *Velocity_X_ptr_to_ptr );
   free ( *Velocity_Y_ptr_to_ptr );
   free ( *Velocity_Z_ptr_to_ptr );

   return;

}  // Cleanup_Read_Lvl1_SC_PosHist ()



// ****************************************************************************************

//  Call this if an error has occurred on a call to a cfitsio routine -- this routine
//  will convert the status numeric value to a string and output the string.


void  Output_StatusValue_as_Msg (

   const int  status

) {

    char ErrorText [30];


   // * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   fits_get_errstatus  ( status, ErrorText );

   printf ( "\nA cfitsio routine returned an error.  The numeric value is %d\n", status );
   printf ( "This translates to: '%s'\n", ErrorText );

   return;

}  // Output_StatusValue_as_Msg ()

