!  Copyright Michael S. Briggs, University of Alabama in Huntsville.
!  For NASA's Fermi Gamma-ray Burst Monitor (GBM).
!  copyright  2010 Feb.

!  minor rev. 2010 Feb 26.
!  Moved interpolation code to subroutine, 2010 May 9.

!   Revised 2013 Jan 4 to have 3 input modes and the
!   capability to automatically read the correct SC Pos Hist file.

!  Calculates position of Fermi by reading Level 1 Spacecraft Position
!  History File.    "Interpolates" between
!  records in the SC Position History file by finding the closest record
!  and "extending" the position with the velocity from the file.
!  The step is so small that it not risky like most extrapolations.
!  It compares the positions from the obtained by stepping forward and
!  backwards from the two records which bracket the requested MET.
!  The SC Position is output as altitude, Latitude & Longitude.


module SGC_mod

   use MSB_Constants

   implicit none

contains


! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

subroutine  Answer_from_Requested_File_and_Time ( EastLongitude_WGS84, Latitude_WGS84, Altitude_WGS84 )

   use Read_Lvl1_SC_PosHist_mod

   use Time_and_Coord_Routines, only: Cartesian_to_GeoCoord_WGS84

   use, intrinsic :: iso_c_binding, only: c_char, c_float, c_double, C_NULL_CHAR

   real (real64), intent (out) :: EastLongitude_WGS84    ! degrees
   real (real64), intent (out) :: Latitude_WGS84    !   degrees
   real (real64), intent (out) :: Altitude_WGS84    !   km

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   character (kind=c_char, len=FILE_NAME_MAX_LEN) :: InputFileName

   real (c_double) :: Request_MET      ! Mission Elapsed Time, s
   real (c_double) :: Before_MET
   real (c_double) :: After_MET

   real (c_double), dimension (4) :: Before_Quaternion
   real (c_double), dimension (4) :: After_Quaternion

   real (c_double), dimension (3) :: Before_AngVelocity    !  rad / s
   real (c_double), dimension (3) :: After_AngVelocity

   real (c_float), dimension (3) :: Before_Position    !  m
   real (c_float), dimension (3) :: After_Position

   real (c_float), dimension (3) :: Before_Velocity    !  m / s
   real (c_float), dimension (3) :: After_Velocity

   real (c_float), dimension (3) :: Position   ! m

   real (real64), dimension (4) :: Quaternion

   character (len=50) string
   logical :: Success
   character (len=26) :: FailReason


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   call GET_COMMAND_ARGUMENT ( 1, InputFileName )
   InputFileName = trim (InputFileName) // C_NULL_CHAR

   call GET_COMMAND_ARGUMENT ( 2, string )
   read (string, *)  Request_MET


   !  Get two times which bracket the requested time, along with the corresponding
   !  positions and velocities:

   call Read_Lvl1_SC_PosHist_SV  ( InputFileName, Request_MET, Before_MET, After_MET,      &
              Before_Quaternion, After_Quaternion, Before_AngVelocity, After_AngVelocity,  &
              Before_Position, After_Position, Before_Velocity, After_Velocity )

   call Interpolate_Convert_SC_PosHist  ( Request_MET, Before_MET, After_MET,                     &
                   Before_Quaternion, After_Quaternion, Before_AngVelocity, After_AngVelocity,    &
                   Before_Position, After_Position, Before_Velocity, After_Velocity,              &
                   Position, EastLongitude_WGS84, Latitude_WGS84, Altitude_WGS84, Quaternion, Success, FailReason )

   if ( .NOT. Success ) then
      write (*, '(// "Program SC_GeoCoord: Interpolate_Convert_SC_PosHist fails:" )' )
      write (*, '(A)' )  FailReason
      stop
   end if

#ifdef SPHEREEARTH
   call Cartesian_to_GeoCoord_WGS84 ( Position, Latitude_WGS84, Altitude_WGS84 )
#endif


end subroutine  Answer_from_Requested_File_and_Time


! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


!   Returns TRUE if the string is consistent with a floating point number and
!   FALSE otherwise.
!   The string is assumed to be left-adjusted so that there are no leading blanks.

function IsNumeric ( string )

  logical :: IsNumeric

  character (len=*) :: string

  ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

  integer (int32) :: i, j
  logical :: Char_OK

  character (len=1), dimension (1:15) :: test =   &
      [ '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '.', 'E', 'e', '-', '+' ]

  ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

  IsNumeric = .TRUE.

  do i=1, len_trim (string)

     Char_OK = .FALSE.

     do j=1, size (test)
        if ( string (i:i) == test (j) )  Char_OK = .TRUE.
     end do

     if ( .NOT. Char_OK )  then
        IsNumeric = .FALSE.
        return      !  return as soon as we find one bad character !!!
     end if

  end do


end function IsNumeric


! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

end module SGC_mod


! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


program SC_GeoCoord

   use MSB_Constants

   use, intrinsic :: iso_c_binding, only: C_FLOAT

   use SGC_mod

   use Cont_TTE_Trigger_SC_PosHist_subs

   implicit none

   real (real128) :: MET

   real (real64) :: EastLongitude_WGS84    ! degrees
   real (real64) :: Latitude_WGS84    !   degrees
   real (real64) :: Altitude_WGS84    !   m

   character (len=26) :: Output_Status

   integer :: NumberCmdArguments
   character (len=50) string
   logical :: FileFound
   integer :: MET_LUN, OUT_LUN, ReadCode
   character (len=13) :: StatusString
   integer (int32) :: ReadCount = 0, GoodCount = 0

   real (c_float), dimension (3) :: Position    ! m
   real (real64), dimension (4) :: Quaternion


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   NumberCmdArguments = COMMAND_ARGUMENT_COUNT ()

   if ( NumberCmdArguments /= 1  .AND. NumberCmdArguments /= 2 ) then
      write (*, *) "Usage: any of:"
      write (*, *) "SC_GeoCoord  PositionHistoryFile.fit  GBM_MET_value"
      write (*, *) "SC_GeoCoord  GBM_MET_value"
      write (*, *) "SC_GeoCoord  FileName [of file listing MET values]"
      stop
   end if


   if ( NumberCmdArguments == 2 ) then   ! number command arguments

      !  two command arguments: user has provided name of SC Position History file
      !  and the MET value to process.   This case mostly for debugging / testing.

      call Answer_from_Requested_File_and_Time ( EastLongitude_WGS84, Latitude_WGS84, Altitude_WGS84 )

#ifdef SPHEREEARTH
      write (*, 1010)  EastLongitude_WGS84, Latitude_WGS84, Altitude_WGS84 / 1000.0_real64
#else
      write (*, 1020)  EastLongitude_WGS84, Latitude_WGS84, Altitude_WGS84 / 1000.0_real64
#endif

   else   ! number command arguments


      !  one command argument: either a single MET value to process, or the
      !  name of a file with a list of MET values to process.

      call GET_COMMAND_ARGUMENT ( 1, string )

      if ( IsNumeric (string) ) then   ! arg is MET or file ?

         !  The argument is consistent with a floating point number, so
         !  process it as an MET value.

         read (string, *)  MET

         call Position_from_SpacePositionHistoryFile ( MET, EastLongitude_WGS84, Latitude_WGS84,   &
                  Altitude_WGS84, Output_Status )

         if ( Output_Status /= "SUCCESS" ) then
            write (*, *) "Failed to obtain location because of SC Pos Hist file problem!"
            write (*, *) "Reason: ", Output_Status
         else

#ifdef SPHEREEARTH
         write (*, 1010)  EastLongitude_WGS84, Latitude_WGS84, Altitude_WGS84 / 1000.0_real64
#else
         write (*, 1020)  EastLongitude_WGS84, Latitude_WGS84, Altitude_WGS84 / 1000.0_real64
#endif

         end if

      else   ! arg is MET or file ?

         !  The argument is not consistent with a floating point number, so
         !  process it as a filename that specifies a file containing a list of
         !  MET values.

         inquire ( file=string, exist=FileFound )

         if ( .NOT. FileFound ) then
            write (*, '( "Processing command argument as file, but file not found:" / A )' )  trim (string)
            stop    !  ABORT for ERROR !
         end if

         open (file=string, newunit=MET_LUN, status="old", access="sequential", form="formatted", action="read" )

         open (file="Fermi_positions.txt", newunit=OUT_LUN, status="replace",    &
               access="sequential", form="formatted", action="write" )


         ReadLoop: do

            read (MET_LUN, *, iostat=ReadCode)  MET

            if ( ReadCode /= 0 ) then
               if ( ReadCode == iostat_end ) then
                  exit ReadLoop
               else
                  write ( *, '( / "error reading line: ", I0 )' )  ReadCode
                  stop    !  ABORT for ERROR !
               end if
            end if


            ReadCount = ReadCount + 1
            write (*, '( / "For ", F18.6, " :" )' )  MET

            call Position_from_SpacePositionHistoryFile ( MET, EastLongitude_WGS84, Latitude_WGS84,   &
                     Altitude_WGS84, Output_Status, Position, Quaternion )

            StatusString = " "
            if ( Output_Status == "SUCCESS" ) then
               GoodCount = GoodCount + 1
            else

               write (*, *) "Failed to obtain location because of SC Pos Hist file problem!"
               write (*, *) "Reason code: ", Output_Status

               StatusString = "  NO LOCATION"


#ifdef SPHEREEARTH
               write (*, 1010)  EastLongitude_WGS84, Latitude_WGS84, Altitude_WGS84 / 1000.0_real64
#else
               write (*, 1020)  EastLongitude_WGS84, Latitude_WGS84, Altitude_WGS84 / 1000.0_real64
#endif

            end if


            write (OUT_LUN,     &
              '( F18.6, 2X, F8.4, 2X, SP, F8.4, SS, 2X, F6.2, 3X, 3(1X, F9.0), 3X, 4(1X, F20.16), A )' ) &
                     MET, EastLongitude_WGS84, Latitude_WGS84, Altitude_WGS84 / 1000.0_real64,    &
                     Position, Quaternion, trim (StatusString)


         end do ReadLoop


         write (*, '( // "read ", I0, " MET values." )' )  ReadCount
         write (*, '( "Obtained ", I0, " locations." )' )  GoodCount


      end if    ! arg is MET or file ?


   end if   ! number command arguments



#ifdef SPHEREEARTH
   1010 format ( " Spherical Earth: longitude:", F9.4, ", latitude:", SP, F9.4, SS, " and altitude: ", F7.2, " km." )
#else
   1020 format ( " WGS84: longitude:", F9.4, ", latitude:", SP, F9.4, SS, " and altitude: ", F7.2, " km." )
#endif


   stop

end program SC_GeoCoord
