!  Copyright Michael S. Briggs, University of Alabama in Huntsville.
!  For NASA's Fermi Gamma-ray Burst Monitor (GBM).
!  2013 December 7 -- 14.  Revised 2016 Jan 27.

!  Routines designed to read Fermi Space Position History from GBM
!  FITS Space Position History Files.  Automatically locates the correct
!  position history file for a time.  Will reuse the file for successive
!  times on the same day, until a time for a different day is requested
!  (i.e., it only buffers one day of SC Position History data).
!  The files are searched for on the directory indicated by environment
!  variable PATH_OF_SC_POSHIST_FILES.

!  Split off from module Cont_TTE_Trigger_subs of file Cont_TTE_Trigger_subs.F90
!  to make these routines more easily available to other programs.

!  Position_from_SpacePositionHistoryFile is not used by the TGF search
!  program -- it provides a convenient interface for obtaining position
!  information from SC Position History files, for use by other programs.
!  It uses the other three routines of this file.


module  Cont_TTE_Trigger_SC_PosHist_subs

   use MSB_Constants

   implicit none

   PRIVATE   ! change default to PRIVATE

   PUBLIC ::  Read_SC_PosHist, Calc_Current_SC_Position, Cleanup_SC_PosHist_arrays,   &
              Position_from_SpacePositionHistoryFile,                                 &
              Position_from_SpacePositionHistoryFile_C


   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

contains


! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

!  This version designed to be called from C.
!  It works by calling the standard version.
!  Unlike the standard version, all arguments are required.

!  Note: C_FLOAT128 is a GNU extension.  If this is not available,
!  replace with C_LONG_DOUBLE.   There will be a lost of accuracy, but the
!  motion of the spacecraft during that time is completely insignificant.

subroutine Position_from_SpacePositionHistoryFile_C ( TGF_MET_C, EastLongitude_C, Latitude_C,   &
               altitude_C, Output_Status_C, Position_C, Quaternion_C )                          &
         bind (  C, name = "Position_from_SpacePositionHistoryFile_C" )

   use, intrinsic :: iso_c_binding, only: C_CHAR, C_FLOAT, C_DOUBLE, C_LONG_DOUBLE, C_NULL_CHAR, C_FLOAT128

   real (C_FLOAT128), intent (in), VALUE :: TGF_MET_C
   real (C_DOUBLE), intent (out) :: EastLongitude_C    ! degrees
   real (C_DOUBLE), intent (out) :: Latitude_C         ! degrees
   real (C_DOUBLE), intent (out) :: altitude_C      ! m about Earth's mean radius


   !  This variable indicates whether it was possible to calculate the
   !  position of Fermi for the requested time.  The possibilities are:
   !  SUCCESS -- the position information (EastLongitude, Latitude, altitude) is valid.
   !  NO_FILE -- position information not valid, the appropriate Level 1 GBM
   !        Spacecraft Position History file was not found.
   !  GAP -- position information not valid, the Space Position History file was
   !        found, but it has a time gap that covers the requested time and the time
   !        gap is too large to safely interpolate.

   character (kind=C_CHAR, len=1), dimension (27), intent (out) :: Output_Status_C

   real (C_FLOAT), dimension (3), intent (out) :: Position_C    ! m
   real (C_DOUBLE), dimension (4), intent (out) :: Quaternion_C


   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   real (real64) :: EastLongitude    ! degrees
   real (real64) :: Latitude         ! degrees
   real (real64) :: altitude      ! m about Earth's mean radius
   character (len=26) :: Output_Status
   real (c_float), dimension (3) :: Position    ! m
   real (real64), dimension (4) :: Quaternion

   integer (int32) :: i

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


   call Position_from_SpacePositionHistoryFile ( real (TGF_MET_C, real128), EastLongitude, Latitude,   &
               altitude, Output_Status, Position, Quaternion )

   EastLongitude_C = EastLongitude
   Latitude_C = Latitude
   altitude_C = altitude
   Position_C = Position
   Quaternion_C = Quaternion

   do i=1, 26
      Output_Status_C (i) = Output_Status (i:i)
   end do

   Output_Status_C ( len_trim (Output_Status) + 1 ) = C_NULL_CHAR


end subroutine Position_from_SpacePositionHistoryFile_C


! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


!  Obtains position of Fermi for time "TGF_MET" from GBM Spacecraft Position
!  History file (Level 1 FITS).   It will locate and read the appropriate SC Position
!  History file, as necessary, searching for SC Position History files on the directory
!  indicted by environment variable PATH_OF_SC_POSHIST_FILES.   If consecutive "TGF_MET"
!  values are from the same day, the file is read only for the first MET of that day --
!  the contents of the SC Position History file are retained until the MET for a different
!  day is requested.

!  This routine calls Calc_Current_SC_Position which calls Interpolate_Convert_SC_PosHist.
!  That routine updates the position with the velocity and the quaternion with the
!  angular velocity, to obtain higher accuracy than the one second resolution of the
!  spacecraft position history files, or to bridge short data gaps.

!  This code is similar to routine Info_for_GroundTrigger of the TGF search program.

subroutine Position_from_SpacePositionHistoryFile ( TGF_MET, EastLongitude, Latitude,   &
               altitude, Output_Status, Position_ARG, Quaternion_ARG )

   use, intrinsic :: iso_c_binding, only: C_BOOL, C_DOUBLE, C_FLOAT, C_LONG

   use Time_and_Coord_Routines, only: MET_to_UTC, UT_to_LocalSolarTime


   real (real128), intent (in) :: TGF_MET
   real (real64), intent (out) :: EastLongitude    ! degrees
   real (real64), intent (out) :: Latitude         ! degrees
   real (real64), intent (out) :: altitude      ! m about Earth's mean radius


   !  This variable indicates whether it was possible to calculate the
   !  position of Fermi for the requested time.  The possibilities are:
   !  SUCCESS -- the position information (EastLongitude, Latitude, altitude) is valid.
   !  NO_FILE -- position information not valid, the appropriate Level 1 GBM
   !        Spacecraft Position History file was not found.
   !  GAP -- position information not valid, the Space Position History file was
   !        found, but it has a time gap that covers the requested time and the time
   !        gap is too large to safely interpolate.

   character (len=26), intent (out) :: Output_Status


   real (c_float), dimension (3), intent (out), OPTIONAL :: Position_ARG    ! m
   real (real64), dimension (4), intent (out), OPTIONAL :: Quaternion_ARG


   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   integer (int32) :: JulianDay_INT
   integer (int32) :: year
   integer (int32) :: month
   integer (int32) :: day
   integer (int32) :: hours
   integer (int32) :: minutes
   real (real64) :: seconds
   character (len=9) :: seconds_string

   real (real128) :: JulianDay

   real (c_float), dimension (3) :: Position    ! m
   real (real64), dimension (4) :: Quaternion

   logical :: Good_SC_Position
   character (len=26) :: FailReason


   logical, SAVE :: Have_SC_PosHist_Data = .FALSE.
   integer (int32), SAVE :: prev_JulianDay = -9999


   integer (C_LONG), SAVE :: Num_History_Recs

   real (C_DOUBLE), dimension (:), pointer, SAVE :: METs_of_Records      ! Mission Elapsed Time, s

   real (C_DOUBLE), dimension (:), pointer, SAVE :: Quaternion_1_vs_t
   real (C_DOUBLE), dimension (:), pointer, SAVE :: Quaternion_2_vs_t
   real (C_DOUBLE), dimension (:), pointer, SAVE :: Quaternion_3_vs_t
   real (C_DOUBLE), dimension (:), pointer, SAVE :: Quaternion_4_vs_t

   real (C_DOUBLE), dimension (:), pointer, SAVE :: AngVelocity_1_vs_t    !  rad / s
   real (C_DOUBLE), dimension (:), pointer, SAVE :: AngVelocity_2_vs_t
   real (C_DOUBLE), dimension (:), pointer, SAVE :: AngVelocity_3_vs_t

   real (C_FLOAT), dimension (:), pointer, SAVE :: X_position_vs_t    !  m
   real (C_FLOAT), dimension (:), pointer, SAVE :: Y_position_vs_t
   real (C_FLOAT), dimension (:), pointer, SAVE :: Z_position_vs_t

   real (C_FLOAT), dimension (:), pointer, SAVE :: X_velocity_vs_t    !  m / s
   real (C_FLOAT), dimension (:), pointer, SAVE :: Y_velocity_vs_t
   real (C_FLOAT), dimension (:), pointer, SAVE :: Z_velocity_vs_t


   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


   call MET_to_UTC ( TGF_MET, JulianDay, year, month, day, hours, minutes, seconds, seconds_string )

   JulianDay_INT = int ( JulianDay - 0.5_real128, int32 )

   if ( JulianDay_INT /= prev_JulianDay ) then

      prev_JulianDay = JulianDay_INT

      write (*, 1130)  TGF_MET, year, month, day
      1130 format ( ">>> MET ", F18.6, " is ", I4, "-", I2.2, "-", I2.2 )

      !  If we previously read the SC Position History, free that memory:
      !  (An aside: An inconsequential memory leak -- after the last call to this subroutine,
      !  we don't free this memory....)
      if ( Have_SC_PosHist_Data )  call Cleanup_SC_PosHist_arrays ( METs_of_Records,           &
               Quaternion_1_vs_t, Quaternion_2_vs_t, Quaternion_3_vs_t, Quaternion_4_vs_t,     &
               AngVelocity_1_vs_t, AngVelocity_2_vs_t, AngVelocity_3_vs_t,                     &
               X_position_vs_t, Y_position_vs_t, Z_position_vs_t,                              &
               X_velocity_vs_t, Y_velocity_vs_t, Z_velocity_vs_t )

      call Read_SC_PosHist ( TGF_MET, Have_SC_PosHist_Data,   &
            Num_History_Recs, METs_of_Records,                            &
            Quaternion_1_vs_t, Quaternion_2_vs_t, Quaternion_3_vs_t, Quaternion_4_vs_t,      &
            AngVelocity_1_vs_t, AngVelocity_2_vs_t, AngVelocity_3_vs_t,                      &
            X_position_vs_t, Y_position_vs_t, Z_position_vs_t,     &
            X_velocity_vs_t, Y_velocity_vs_t, Z_velocity_vs_t  )

   end if


   !  Determine Spacecraft position, using the previously read SC Position History.

   if ( Have_SC_PosHist_Data )  call Calc_Current_SC_Position ( real (TGF_MET, C_DOUBLE),        &
               Num_History_Recs,  METs_of_Records,                                               &
               Quaternion_1_vs_t, Quaternion_2_vs_t, Quaternion_3_vs_t, Quaternion_4_vs_t,       &
               AngVelocity_1_vs_t, AngVelocity_2_vs_t, AngVelocity_3_vs_t,                       &
               X_position_vs_t, Y_position_vs_t, Z_position_vs_t,                                &
               X_velocity_vs_t, Y_velocity_vs_t, Z_velocity_vs_t,                                &
               EastLongitude, Latitude, altitude, Position, Quaternion, Good_SC_Position,        &
               FailReason )


   if ( .NOT. Have_SC_PosHist_Data  .OR.  .NOT.  Good_SC_Position ) then  ! failure ?

      EastLongitude = 999.99
      Latitude = -99.99
      altitude = -99.9

      if ( present ( Position_ARG ) )  Position_ARG = -9999999.0_c_float
      if ( present ( Quaternion_ARG ) )  Quaternion_ARG = -9.999999_real64


      if ( .NOT. Have_SC_PosHist_Data ) then  ! failure type ?
         Output_Status = "NO_FILE"
      else  ! failure type ?
         Output_Status = FailReason
      end if  ! failure type ?

   else  ! failure ?

      Output_Status = "SUCCESS"

      if ( present ( Position_ARG ) )  Position_ARG = Position
      if ( present ( Quaternion_ARG ) )  Quaternion_ARG = Quaternion


   end if  ! failure ?


end subroutine Position_from_SpacePositionHistoryFile


! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


subroutine  Read_SC_PosHist ( GBM_MET, Success, Num_History_Recs, METs_of_Records,     &
      Quaternion_1_vs_t, Quaternion_2_vs_t, Quaternion_3_vs_t, Quaternion_4_vs_t,      &
      AngVelocity_1_vs_t, AngVelocity_2_vs_t, AngVelocity_3_vs_t,                      &
      X_position_vs_t, Y_position_vs_t, Z_position_vs_t,                               &
      X_velocity_vs_t, Y_velocity_vs_t, Z_velocity_vs_t  )

   use, intrinsic :: iso_c_binding, only: C_LONG, C_DOUBLE, C_FLOAT, C_NULL_CHAR

   use Time_and_Coord_Routines, only: MET_to_UTC

   use Read_Lvl1_SC_PosHist_mod, only: Read_Lvl1_SC_PosHist_Fort

   real (real128), intent (in) :: GBM_MET

   logical, intent (out) :: Success

   integer (C_LONG), intent (out) :: Num_History_Recs

   real (C_DOUBLE), dimension (:), pointer :: METs_of_Records      ! Mission Elapsed Time, s

   real (C_DOUBLE), dimension (:), pointer :: Quaternion_1_vs_t
   real (C_DOUBLE), dimension (:), pointer :: Quaternion_2_vs_t
   real (C_DOUBLE), dimension (:), pointer :: Quaternion_3_vs_t
   real (C_DOUBLE), dimension (:), pointer :: Quaternion_4_vs_t

   real (C_DOUBLE), dimension (:), pointer :: AngVelocity_1_vs_t    !  rad / s
   real (C_DOUBLE), dimension (:), pointer :: AngVelocity_2_vs_t
   real (C_DOUBLE), dimension (:), pointer :: AngVelocity_3_vs_t

   real (C_FLOAT), dimension (:), pointer :: X_position_vs_t    !  m
   real (C_FLOAT), dimension (:), pointer :: Y_position_vs_t
   real (C_FLOAT), dimension (:), pointer :: Z_position_vs_t

   real (C_FLOAT), dimension (:), pointer :: X_velocity_vs_t    !  m / s
   real (C_FLOAT), dimension (:), pointer :: Y_velocity_vs_t
   real (C_FLOAT), dimension (:), pointer :: Z_velocity_vs_t


   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   real (real128) :: JulianDay

   integer (int32) :: year
   integer (int32) :: month
   integer (int32) :: day
   integer (int32) :: hours
   integer (int32) :: minutes
   real (real64) :: seconds


   character (len=255) :: SC_PosHist_PATH
   character (len=30) :: SC_PosHist_FileOnly
   character (len=286) :: SC_PosHist_FileName
   integer :: VerNum
   integer :: len

   logical :: SC_PosHist_Found

   integer, parameter :: FILE_MAX_VER_NUM = 7

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   Success = .FALSE.


   !  Locate and open the appropriate Level 1 Spacecraft Position History File.


   call MET_to_UTC ( GBM_MET, JulianDay, year, month, day, hours, minutes, seconds )

   !  Obtain the path of the directory that contains the position history files:
   call get_environment_variable ( "PATH_OF_SC_POSHIST_FILES", SC_PosHist_PATH )


   !  make sure that the path ends in a slash:
   len = len_trim (SC_PosHist_PATH)
   if ( len > 0 ) then
      if ( SC_PosHist_PATH (len:len) /= "/" )  SC_PosHist_PATH = trim (SC_PosHist_PATH) // "/"
   end if


   !  Identify the version number of the SC Position History file by trying version numbers starting from
   !  FILE_MAX_VER_NUM down to zero.  This way the highest version number file is used in preference to an
   !  older file.

   WhichVerNum: do VerNum=FILE_MAX_VER_NUM, 0, -1

      write ( SC_PosHist_FileOnly, '( "glg_poshist_all_", 3I2.2, "_v", I2.2, ".fit" )' )  year - 2000, month, day, VerNum

      SC_PosHist_FileName = trim (SC_PosHist_PATH) // trim (SC_PosHist_FileOnly)

      inquire ( file=SC_PosHist_FileName, exist=SC_PosHist_Found )

      if ( SC_PosHist_Found )  exit WhichVerNum

   end do WhichVerNum



   if ( .NOT. SC_PosHist_Found ) then

      ! In this case of not finding the file we obviously return w/o reading the data.

      write (*, '( / "ERROR: Unable to locate a SC Position History File." )' )
      write (*, '( "Tried files with several version numbers, including: ", A )' )  trim (SC_PosHist_FileName )

   else   ! SC_PosHist_Found ?


      write (*, '( / "Found SC Position History file: ", A )' )  trim (SC_PosHist_FileName )

      if ( VerNum == FILE_MAX_VER_NUM )   &
         write (*, '( / "** WARNING: Using version number ", I0, " -- check whether a higher version number exists!" )' )  &
             FILE_MAX_VER_NUM


      !  Having located the file, read the data:

      SC_PosHist_FileName = trim (SC_PosHist_FileName) //  C_NULL_CHAR

      call Read_Lvl1_SC_PosHist_Fort  ( SC_PosHist_FileName, Num_History_Recs, METs_of_Records,        &
                 Quaternion_1_vs_t, Quaternion_2_vs_t, Quaternion_3_vs_t, Quaternion_4_vs_t,           &
                 AngVelocity_1_vs_t, AngVelocity_2_vs_t, AngVelocity_3_vs_t,                           &
                 X_position_vs_t, Y_position_vs_t, Z_position_vs_t,                                    &
                 X_velocity_vs_t, Y_velocity_vs_t, Z_velocity_vs_t  )

      Success = .TRUE.

   end if


   return

end subroutine  Read_SC_PosHist


! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


!  changed 2013 Dec 17: only call if SC Position History data is in the arrays.

subroutine  Calc_Current_SC_Position ( Request_MET, Num_History_Recs,  METs_of_Records,   &
                     Quaternion_1_vs_t, Quaternion_2_vs_t, Quaternion_3_vs_t, Quaternion_4_vs_t,       &
                     AngVelocity_1_vs_t, AngVelocity_2_vs_t, AngVelocity_3_vs_t,                       &
                     X_position_vs_t, Y_position_vs_t, Z_position_vs_t,                                &
                     X_velocity_vs_t, Y_velocity_vs_t, Z_velocity_vs_t,                                &
                     EastLongitude, Latitude, altitude, Position, Quaternion, Good_SC_Position, FailReason )

   use, intrinsic :: iso_c_binding, only: C_LONG, C_DOUBLE, C_FLOAT, C_NULL_CHAR

   use Read_Lvl1_SC_PosHist_mod, only: Interpolate_Convert_SC_PosHist


   real (C_DOUBLE), intent (in) :: Request_MET      ! Mission Elapsed Time, s

   integer (C_LONG), intent (in) :: Num_History_Recs

   real (C_DOUBLE), dimension (:), pointer :: METs_of_Records      ! Mission Elapsed Time, s

   real (C_DOUBLE), dimension (:), pointer :: Quaternion_1_vs_t
   real (C_DOUBLE), dimension (:), pointer :: Quaternion_2_vs_t
   real (C_DOUBLE), dimension (:), pointer :: Quaternion_3_vs_t
   real (C_DOUBLE), dimension (:), pointer :: Quaternion_4_vs_t

   real (C_DOUBLE), dimension (:), pointer :: AngVelocity_1_vs_t    !  rad / s
   real (C_DOUBLE), dimension (:), pointer :: AngVelocity_2_vs_t
   real (C_DOUBLE), dimension (:), pointer :: AngVelocity_3_vs_t

   real (C_FLOAT), dimension (:), pointer :: X_position_vs_t    !  m
   real (C_FLOAT), dimension (:), pointer :: Y_position_vs_t
   real (C_FLOAT), dimension (:), pointer :: Z_position_vs_t

   real (C_FLOAT), dimension (:), pointer :: X_velocity_vs_t    !  m / s
   real (C_FLOAT), dimension (:), pointer :: Y_velocity_vs_t
   real (C_FLOAT), dimension (:), pointer :: Z_velocity_vs_t

   real (real64), intent (out) :: EastLongitude    ! degrees
   real (real64), intent (out) :: Latitude         ! degrees
   real (real64), intent (out) :: altitude      ! m

   real (c_float), dimension (3), intent (out) :: Position    ! m
   real (real64), dimension (4), intent (out) :: Quaternion

   logical, intent (out) :: Good_SC_Position
   character (len=26), intent (out) :: FailReason

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   real (c_double), dimension (4) :: Before_Quaternion
   real (c_double), dimension (4) :: After_Quaternion

   real (c_double), dimension (3) :: Before_AngVelocity    !  rad / s
   real (c_double), dimension (3) :: After_AngVelocity

   real (c_float), dimension (3) :: Before_Position    !  m
   real (c_float), dimension (3) :: After_Position

   real (c_float), dimension (3) :: Before_Velocity    !  m / s
   real (c_float), dimension (3) :: After_Velocity

   integer (int32) :: i_beg, i_end, i_try

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


   !  Binary search to find the two consecutive elements of METs_of_Records that
   !  bracket Request_MET.
   !  At the end of the loop, we will have:
   !  i_end = i_beg + 1,
   !  METs_of_Records (i_end) > Request_MET,
   !  METs_of_Records (i_beg) <= Request_MET.

   !  There is one case that causes a violation of the above rule: if
   !  Request_MET > METs_of_Records (Num_History_Recs), i.e., if the TTE time
   !  is after the time of the last record in the SC Position History file
   !  (which have been observed to be several seconds before the end of the
   !  day, e.g., 23 h 59 m 56.94 s for 2012 Dec 18.
   !  In this case:
   !  i_end will be the last entry in the position history file and i_beg the
   !  preceding one.

   i_beg = 1
   i_end = int ( Num_History_Recs, int32 )

   LocateRecs: do

      if ( i_end - i_beg == 1 )  exit LocateRecs

      i_try = i_beg + (i_end - i_beg) / 2

      if ( METs_of_Records (i_try) > Request_MET ) then
         i_end = i_try
      else
         i_beg = i_try
      end if

   end do LocateRecs


   !  Having located the index values that bracket the requested time,
   !  we now obtain the Quaternion, Angular Velocity, position and velocity measurements
   !  that bracket the requested time:

   Before_Quaternion =   &
      [ Quaternion_1_vs_t (i_beg), Quaternion_2_vs_t (i_beg), Quaternion_3_vs_t (i_beg), Quaternion_4_vs_t (i_beg) ]
   After_Quaternion =    &
      [ Quaternion_1_vs_t (i_end), Quaternion_2_vs_t (i_end), Quaternion_3_vs_t (i_end), Quaternion_4_vs_t (i_end) ]

   Before_AngVelocity = [ AngVelocity_1_vs_t (i_beg), AngVelocity_2_vs_t (i_beg), AngVelocity_3_vs_t (i_beg) ]
   After_AngVelocity =  [ AngVelocity_1_vs_t (i_end), AngVelocity_2_vs_t (i_end), AngVelocity_3_vs_t (i_end) ]

   Before_Position =  [ X_position_vs_t (i_beg), Y_position_vs_t (i_beg), Z_position_vs_t (i_beg) ]
   After_Position =   [ X_position_vs_t (i_end), Y_position_vs_t (i_end), Z_position_vs_t (i_end) ]

   Before_Velocity = [ X_velocity_vs_t (i_beg), Y_velocity_vs_t (i_beg), Z_velocity_vs_t (i_beg) ]
   After_Velocity =  [ X_velocity_vs_t (i_end), Y_velocity_vs_t (i_end), Z_velocity_vs_t (i_end) ]


   !  And we interpolate on those bracketing values to obtain an accurate spacecraft position
   !  and quaternion:
   !  In the special case described above (end of day), we might do a slight extrapolation forward.

   call Interpolate_Convert_SC_PosHist  ( Request_MET, METs_of_Records (i_beg), METs_of_Records (i_end),      &
                   Before_Quaternion, After_Quaternion, Before_AngVelocity, After_AngVelocity,                &
                   Before_Position, After_Position, Before_Velocity, After_Velocity,                          &
                   Position, EastLongitude, Latitude, altitude, Quaternion, Good_SC_Position, FailReason )


   return

end subroutine  Calc_Current_SC_Position


! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


subroutine  Cleanup_SC_PosHist_arrays ( METs_of_Records,                                       &
               Quaternion_1_vs_t, Quaternion_2_vs_t, Quaternion_3_vs_t, Quaternion_4_vs_t,     &
               AngVelocity_1_vs_t, AngVelocity_2_vs_t, AngVelocity_3_vs_t,                     &
               X_position_vs_t, Y_position_vs_t, Z_position_vs_t,                              &
               X_velocity_vs_t, Y_velocity_vs_t, Z_velocity_vs_t )

   use, intrinsic :: iso_c_binding, only: C_LONG, C_DOUBLE, C_FLOAT, C_NULL_CHAR

   use Read_Lvl1_SC_PosHist_mod, only: Cleanup_Read_Lvl1_SC_PosHist_Fort

   real (C_DOUBLE), dimension (:), pointer :: METs_of_Records      ! Mission Elapsed Time, s

   real (C_DOUBLE), dimension (:), pointer :: Quaternion_1_vs_t
   real (C_DOUBLE), dimension (:), pointer :: Quaternion_2_vs_t
   real (C_DOUBLE), dimension (:), pointer :: Quaternion_3_vs_t
   real (C_DOUBLE), dimension (:), pointer :: Quaternion_4_vs_t

   real (C_DOUBLE), dimension (:), pointer :: AngVelocity_1_vs_t    !  rad / s
   real (C_DOUBLE), dimension (:), pointer :: AngVelocity_2_vs_t
   real (C_DOUBLE), dimension (:), pointer :: AngVelocity_3_vs_t

   real (C_FLOAT), dimension (:), pointer :: X_position_vs_t    !  m
   real (C_FLOAT), dimension (:), pointer :: Y_position_vs_t
   real (C_FLOAT), dimension (:), pointer :: Z_position_vs_t

   real (C_FLOAT), dimension (:), pointer :: X_velocity_vs_t    !  m / s
   real (C_FLOAT), dimension (:), pointer :: Y_velocity_vs_t
   real (C_FLOAT), dimension (:), pointer :: Z_velocity_vs_t


   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


   !  This deallocates the memory allocate in C.  Those addresses stored in
   !  variables local to the C code:

   call Cleanup_Read_Lvl1_SC_PosHist_Fort ()

   !  For thoroughness, nullify the Fortran pointers:

   nullify ( METs_of_Records )

   nullify ( Quaternion_1_vs_t )
   nullify ( Quaternion_2_vs_t )
   nullify ( Quaternion_3_vs_t )
   nullify ( Quaternion_4_vs_t )

   nullify ( AngVelocity_1_vs_t )
   nullify ( AngVelocity_2_vs_t )
   nullify ( AngVelocity_3_vs_t )

   nullify ( X_position_vs_t )
   nullify ( Y_position_vs_t )
   nullify ( Z_position_vs_t )

   nullify ( X_velocity_vs_t )
   nullify ( Y_velocity_vs_t )
   nullify ( Z_velocity_vs_t )

   return

end subroutine Cleanup_SC_PosHist_arrays


! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


end module  Cont_TTE_Trigger_SC_PosHist_subs
