!   Michael S. Briggs,  Michael.Briggs@uah.edu.
!   University of Alabama in Huntsville.
!   copyright 2013 & 2014.  Revised 2014 July 18.
!   Revised 2016.

module Time_and_Coord_Routines

   use MSB_Constants

   implicit none


   !  Parameters of the World Geodetic System 1984
   ! semi-major axis:
   real (real64), parameter :: WGS84_a = 6378137.0_real64   ! m
   ! reciprocal of flattening:
   real (real64), parameter :: WGS84_1overf = 298.257223563_real64



   PRIVATE    ! makes PRIVATE the default.

   PUBLIC :: JulianDay_to_Date_and_Time, Date_and_Time_to_JulianDay,      &
       Date_and_Time_to_JulianDay_midnight, UT_to_LocalSolarTime,         &
       Equation_of_Time, UTC_to_MET, MET_to_UTC, MET_to_UT1,              &
       Cartesian_to_SphericalGeoCoord, SphericalGeoCoord_to_Cartesian,    &
       GeoCoord_WGS84_to_Cartesian, Cartesian_to_GeoCoord_WGS84,          &
       WGS84_a, WGS84_1overf, seconds_2_string,             &
       Calc_Greenwich_Mean_SiderealTime,                    &
       Cartesian_to_Accurate_Longitude



   !  Note: MJD = Julian Day Number - MJD_OFFSET.   MJD_OFFSET is 2400000.5

   integer (int32), dimension (1:2), parameter :: MODVAR_MJD_range = [ 54600, 61000 ]

   real (real64), dimension ( MODVAR_MJD_range (1): MODVAR_MJD_range (2) ) :: MODVAR_UT1_minus_UTC_ms    ! ms

   !  status of data in MODVAR_UT1_minus_UTC_ms:
   !  0 --> not available
   !  1 --> prediction read from Bulletin A
   !  2 --> true value read from Bulletin B
   integer (int8), dimension ( MODVAR_MJD_range (1): MODVAR_MJD_range (2) ) :: MODVAR_State_of_UT1_minus_UTC



   !  PRIVATE module variable that store the MET values of leap seconds:

   type LeapSecond_type

      integer (int64) :: MET_of_LeapSecond
      integer (int32) :: year
      integer (int32) :: month
      integer (int32) :: day

   end type LeapSecond_type

   type (LeapSecond_type), dimension (:), allocatable, SAVE :: LeapSecondData





contains


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  This is the easy direction, the conversion from WGS84 Longitude, Latitude and
!  Altitude to Cartesian coordinates.  Simple analytic equations are used, as listed
!  in http://en.wikipedia.org/wiki/Geodetic_datum.
!  The output is Cartesian coordinates fixed on the Earth, and not the intertial Cartesian
!  coordinates used by Fermi.

subroutine GeoCoord_WGS84_to_Cartesian ( Longitude_WGS84, Latitude_WGS84, Altitude_WGS84, Position )

   real (real64), intent (in) :: Longitude_WGS84    !   degrees
   real (real64), intent (in) :: Latitude_WGS84    !   degrees
   real (real64), intent (in) :: Altitude_WGS84    !   m

   real (real64), dimension (3), intent (out) :: Position   ! m

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   real (real64) :: x, y, z
   real (real64) :: f, e_sq
   real (real64) :: phi
   real (real64) :: lambda
   real (real64) :: N

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


   f = 1.0_real64 / WGS84_1overf
   e_sq = 2.0_real64 * f  -  f**2

   phi = Latitude_WGS84 * DEGREES_TO_RADIANS_DBL
   lambda = Longitude_WGS84 * DEGREES_TO_RADIANS_DBL

   N = WGS84_a  /  sqrt ( 1.0_real64 - e_sq * ( sin (phi) )**2 )

   x = ( N + Altitude_WGS84 ) * cos (phi) * cos (lambda)

   y = ( N + Altitude_WGS84 ) * cos (phi) * sin (lambda)

   z = ( N * (1.0_real64 - e_sq) + Altitude_WGS84 ) * sin (phi)

   Position =  [ x, y, z ]


end subroutine GeoCoord_WGS84_to_Cartesian



! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  Based on Cartesian_to_SphericalGeoCoord.  But this routine only calculates
!  longitude.  The main difference is that this routine uses a more accurate
!  calculation of sidereal time based on UT1 instead of UTC.  In principle Longitude
!  is the same for spherical and ellipsoidal models of the Earth since both are
!  rotational symmetric.  However, WGS84 uses a difference longitude coordinate
!  system with a slight shift of the Prime Meridian (see below).

!  Converts Position in Inertial (i.e., non-rotating) Earth-centered Cartesian
!  coordinates (as used by Fermi) Position (a 3-vector X, Y, Z) to the
!  WGS84 geodetic geographic coordinate EastLongitude.  (Another routine handles
!  the other two WGS84 geodetic coordinates Latitude and Altitude, which do not
!  involve solving for the rotation of the Earth.)
!  Time is a necessary input, to rotate the Earth under Fermi.   The rotation is done
!  via Sidereal (star referenced) time, which is calculated from the input value of
!  Fermi MET.

!  Michael S. Briggs, UA Huntsville, 2010 February & March 10.
!  Revised 2014 Feb 7.


subroutine  Cartesian_to_Accurate_Longitude  ( GBM_MET, Position, EastLongitude, Success )

   use, intrinsic :: iso_c_binding, only: c_float


   real (real128), intent (in) :: GBM_MET          ! Mission Elapsed Time, s
   real (c_float), dimension (3), intent (in) :: Position   ! m, from SC Position History file

   real (real64), intent (out) :: EastLongitude    ! degrees

   logical, intent (out) :: Success


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   real (real64) :: InertialLongitude

   real (real128) :: Greenwich_Mean_SiderealTime, Greenwich_Apparent_SiderealTime

   real (real128) :: theta_deg

   real (real128) :: JulianDay
   integer (int32) :: year
   integer (int32) :: month
   integer (int32) :: day
   integer (int32) :: hours
   integer (int32) :: minutes
   real (real64) :: seconds

   real (real64) :: RA    ! RA, radians
   real (real64) :: Dec   ! Dec, radians

   real (real64) :: obliquity_ecliptic_deg   ! degrees

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   call Calc_Greenwich_Mean_SiderealTime ( GBM_MET, Greenwich_Mean_SiderealTime, Success )


   !  correct from Mean Sidereal Time to Apparent Sidereal Time via according to nutation.
   !  The correction is about 1 s of time.

   call MET_to_UTC ( GBM_MET, JulianDay, year, month, day, hours, minutes, seconds )

   call SunLoc_J2000 ( JulianDay, RA, Dec, obliquity_ecliptic_deg )

   Greenwich_Apparent_SiderealTime = Greenwich_Mean_SiderealTime +    &
         DeltaPsi_arcsec ( JulianDay ) * cos ( obliquity_ecliptic_deg * DEGREES_TO_RADIANS_DBL ) / 15.0_real128


   if ( Success ) then

      !  convert from seconds to degrees, remove multiples of 360 to leave
      !  fractional part of a rotation:
      !  Apparently this is Sidereal Time in units of solar seconds, so that the
      !  length of a day is 86400.0 seconds, not 86636.55536790872.
      !  The example 11.a on page 84 of "Astronomical Algorithms", by Jean Meeus
      !  shows that 86400 should be used.

      theta_deg = 360.0_real128 * Greenwich_Apparent_SiderealTime / 86400.0_real128   !86636.55536790872_real128

      theta_deg = mod (theta_deg, 360.0_real128)

      InertialLongitude = atan2 ( Position (2), Position (1) )

      EastLongitude = InertialLongitude * RADIANS_TO_DEGREES_DBL  -  real ( theta_deg, real64 )


      !  shift from Greenwich Meridian to IERS Reference Meridian, which is the WGS84 Meridian.
      !  5.3 seconds of arc East of astronomic Greenwich prime meridian.

      EastLongitude = EastLongitude - 1.4722222222222E-3_real64


      !  In degrees, InertialLongitude from ATAN2 can be as small as -180,
      !  and -theta_deg as small as -360, so that their sum, EastLongitude,
      !  can be as small as -540.
      !  Therefore converting to EastLongitude to non-negative value may require
      !  two additions of 360.   Also allow for subtraction of 5.3 arc seconds.

      if ( EastLongitude < 0.0 ) EastLongitude = EastLongitude + 360.0_real64
      if ( EastLongitude < 0.0 ) EastLongitude = EastLongitude + 360.0_real64
      if ( EastLongitude < 0.0 ) EastLongitude = EastLongitude + 360.0_real64

   else

      EastLongitude = huge ( 1.0_real64 )

   end if


   return

end subroutine Cartesian_to_Accurate_Longitude


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  Conversion from Cartesian Coordinates to WGS84 Latitude and Altitude.
!  The method is an iterative solution of Bowring's irrational geodetic-latitude equation,
!  as described in http://en.wikipedia.org/wiki/Geodetic_datum.
!  The original sources are given in Wikipedia as:
! 		 The irrational equation to be solved: Bowring, B. R. (1976),
! 		 "Transformation from Spatial to Geographical Coordinates". Surv. Rev. 23 (181): 323327,
! 		 doi:10.1179/003962676791280626.
! 		 The solution, obtained by NewtonRaphson iteration:
! 		  Fukushima, T. (1999), "Fast Transform from Geocentric to Geodetic Coordinates",
! 		  J. Geod. 73 (11): 603610. doi:10.1007/s001900050271, (Appendix B)
!  Unfortunately these are not available at the UAH library.

subroutine Cartesian_to_GeoCoord_WGS84 ( Position, Latitude_WGS84, Altitude_WGS84 )

   use, intrinsic :: iso_c_binding, only: c_float

   real (c_float), dimension (3), intent (in) :: Position    ! m

   real (real64), intent (out) :: Latitude_WGS84    !   degrees
   real (real64), intent (out) :: Altitude_WGS84    !   m

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   real (real64) :: x, y, z

   !  tests show sufficient convergence in 2 iterations, total convergence in 3
   integer, parameter :: i_max = 3
   integer :: i

   real (real64), dimension (0: i_max) :: kappa

   real (real64) :: f, rho, e_sq, c
   real (real64) :: phi
   real (real64) :: h

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   x = Position (1)
   y = Position (2)
   z = Position (3)

   rho = sqrt ( x**2 + y**2 )

   f = 1.0_real64 / WGS84_1overf
   e_sq = 2.0_real64 * f  -  f**2

   kappa (0) = 1.0_real64 / (1.0_real64 - e_sq)


   do i=1, i_max

      c = ( rho**2 + (1.0_real64 - e_sq) * z**2 * kappa (i-1)**2 ) ** 1.5_real64   /   &
            ( WGS84_a * e_sq )

      kappa (i) = ( c + (1.0_real64 - e_sq) * z**2 * kappa (i-1)**3 )  /  ( c - rho**2 )

   end do


   phi = atan ( kappa (i_max) * z / rho )

   h = ( 1.0_real64 / kappa (i_max)  -  1.0_real64 / kappa (0) )  *   &
        sqrt ( rho**2 + z**2 * kappa (i_max)**2 )    /    e_sq

   Latitude_WGS84 = phi * RADIANS_TO_DEGREES_DBL

   Altitude_WGS84 = h


end subroutine Cartesian_to_GeoCoord_WGS84


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!   Converts UTC to MET.   Not trivial because UTC has leap seconds and MET does not.
!   So use standard astronomical equations to convert from Julian Day Number to MET,
!   then correct for leap seconds according to a table.
!   M. S. Briggs, 2015 Dec 27.

function UTC_to_MET ( year_IN, month_IN, day_IN, hours_IN, minutes_IN, seconds_IN )

   real (real128) :: UTC_to_MET

   integer (int32), intent (in) :: year_IN
   integer (int32), intent (in) :: month_IN
   integer (int32), intent (in) :: day_IN
   integer (int32), intent (in) :: hours_IN
   integer (int32), intent (in) :: minutes_IN
   real (real64), intent (in) :: seconds_IN


   !  Julian Day number of MET = 0,
   !  which is 2001 Jan 01, 0 hr 0 min 0 s.
   real (real128) :: jd_of_MET0 = 2451910.50_real128

   real (real128) :: GBM_MET
   real (real128) :: jd_calc

   integer (int32) :: i

   ! * * * * *  * * * * *  * * * * *  * * * * *  * * * * *  * * * * *  * * * * *  * * * * *  * * * * *

   !  load in the module variable LeapSecondData with the MET
   !  values of leap seconds and the dates of the leap seconds.

   if ( .NOT. allocated (LeapSecondData) )  call Read_LeapSecondData ()

   !  Julian Day Number from date & time
   jd_calc = Date_and_Time_to_JulianDay  ( year_IN, month_IN, day_IN, hours_IN, minutes_IN, seconds_IN )

   !write (*, '( / "Julian Day is:" )' )
   !write (*, *) jd_calc

   !  Convert to MET using seconds in day and Julian Day value of MET epoch.
   GBM_MET = ( jd_calc - jd_of_MET0 ) * 86400.0_real128

   do i=1, size (LeapSecondData)
      if ( GBM_MET >= LeapSecondData (i) % MET_of_LeapSecond )  GBM_MET = GBM_MET + 1.0_real128
   end do

   !  The above advances too far for during a leap second, so correct back:
   if ( seconds_IN >= 60.0_real64 )  GBM_MET = GBM_MET - 1.0_real128

   UTC_to_MET = GBM_MET

   !  Confirm:
   !call MET_to_UTC ( UTC_to_MET, jd_2, year_2, month_2, day_2, hours_2, minutes_2, seconds_2 )
   !write (*, '(/ "y m d=", I4.4, 1X, I2.2, 1X, I2.2, 6X, "h, m, s=", I2.2, 1X, I2.2, 2X, F9.6 )' )       &
   !      year_2, month_2, day_2, hours_2, minutes_2, seconds_2


   return

end function UTC_to_MET


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  Subroutine to convert Julian Days into Calendar dates and Time.
!  Adapted from a C program of mine, dated 2004 Feb 17 -- jd_to_cal.c

!  From the algorithm on page 8 of Practical Astronomy with Your Calculator,
!  3rd ed., Peter Duffett-Smith.

!  test: 2446113.75 gives 1985 Feb 17.25

! Note: Julian days being at 12h 00m UT; Output day is UT.

! Note: instructions on p. 7 of Duffet-Smith make it clear that "INT" function
! should be such that INT(-3.914) = -3.  The Fortran "int" works this way, "floor"
! does not.


subroutine  JulianDay_to_Date_and_Time  ( jd, year, month, day, hours, minutes, seconds )

   real (real128), intent (in) :: jd

   integer (int32), intent (out) :: year
   integer (int32), intent (out) :: month
   integer (int32), intent (out) :: day
   integer (int32), intent (out) :: hours
   integer (int32), intent (out) :: minutes
   real (real64), intent (out) :: seconds


   real (real128) :: jd_2
   integer (int32) :: I
   real (real128) :: F
   integer (int32) :: A
   integer (int32) :: B
   integer (int32) :: C
   integer (int32) :: D
   integer (int32) :: E
   integer (int32) :: G
   real (real64) :: d_small

   real (real64) :: tmpDBL
   real (real128) :: tmpQRK


   ! * * * * *  * * * * *  * * * * *  * * * * *  * * * * *  * * * * *  * * * * *  * * * * *  * * * * *

   !  1)
   jd_2 = jd + 0.5_real128
   I = int ( jd_2, int32 )
   F = jd_2 - I

   !  2)
   if ( I > 2299160 ) then
      tmpDBL =  ( I - 1867216.25_real64 ) / 36524.25_real64
      A = int ( tmpDBL, int32 )
      B = I + 1 + A - A / 4
   else
      A = 0  ! not used
      B = I
   end if

   !  3)
   C = B + 1524

   !  4)
   tmpDBL = ( C - 122.1_real64 ) / 365.25_real64
   D = int ( tmpDBL, int32 )

   !  5)
   tmpDBL = 365.25_real64 * D
   E = int ( tmpDBL, int32 )

   !  6)
   tmpDBL = ( C - E ) / 30.6001_real64
   G = int ( tmpDBL, int32 )

   !  7)
   tmpQRK = C - E + F - int ( 30.6001_real64 * G, int32 )
   d_small = real ( tmpQRK, real64 )

   !  8)
   if ( G <= 13 ) then
      month = G - 1
   else
      month = G - 13
   end if

   !  9)
   if ( month >= 3 ) then
      year = D - 4716
   else
      year = D - 4715
   end if


   day = int ( d_small, int32 )

   seconds = 86400.0_real64 * (d_small - day)

   hours = int ( seconds / 3600.0_real64, int32 )

   seconds = seconds - hours * 3600.0_real64

   minutes = int ( seconds / 60.0_real64, int32 )

   seconds = seconds - minutes * 60.0_real64


   return


end subroutine JulianDay_to_Date_and_Time


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  Calculates Greenwich mean sidereal time according to equation 13 of "The New
!  Definition of Universal Time" by S. Aoki et al., Astronomy and Astrophysics,
!  vol. 105, 359--361, (1982).
!  Also see Chapter 11 of "Astronomical Algorithms" by Jean Meeus.

!  Units are seconds.

!  Time input as Julian Days of UT1.   NOT UTC.
!  The input must be a Julian Day number that represents 0 hours UT, i.e., that represents
!  midnight, i.e., that ends with fractional part 0.5.

function Greenwich_Mean_SiderealTime_at_0hr ( JulianDay_midnight )

   real (real128) :: Greenwich_Mean_SiderealTime_at_0hr      ! seconds
   real (real128), intent (in) :: JulianDay_midnight

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   !  2000 Jan 1, 12 h UT1
   real (real128), parameter :: JD_2000 = 2451545.0_real128

   real (real128) :: T

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   T = ( JulianDay_midnight - JD_2000 ) / 36525.0_real128

   Greenwich_Mean_SiderealTime_at_0hr = 24110.54841_real128  +  8640184.812866_real128 * T    &
       + 0.093104_real128 * T**2  -  6.2E-6_real128 * T**3


   return

end function Greenwich_Mean_SiderealTime_at_0hr


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  Calculates Greenwich mean sidereal time at any time.  Uses
!  Greenwich_Mean_SiderealTime_at_0hr, which is only valid for 0 hours UT, then
!  applies a correction for the remaining hours.
!  Correction method given in Chapter 11 of "Astronomical Algorithms" by Jean
!  Meeus and example 7.b on page 40 of "Astronomical Formulae for Calculators",
!  4th ed., by Jean Meeus.

!  The constant 1.002737909350795 is the ratio sidereal time / solar time.
!  From equation 19 of "The New Definition of Universal Time" by S. Aoki et al.,
!  Astronomy and Astrophysics, vol. 105, 359--361, (1982).    The value is for 2000;
!  it has a slight time dependence which we neglect.  This value also used
!  in Chapter 11 cited above.


!  Inside this routine, MET is converted to solar time in UT1.   NOT UTC.
!  Then used to calculate sidereal time.


subroutine Calc_Greenwich_Mean_SiderealTime ( MET, Greenwich_Mean_SiderealTime, Success )

   real (real128), intent (in) :: MET

   real (real128), intent (out) :: Greenwich_Mean_SiderealTime      ! seconds

   logical, intent (out) :: Success


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   integer (int32) :: year
   integer (int32) :: month
   integer (int32) :: day
   integer (int32) :: hours
   integer (int32) :: minutes
   real (real64) :: seconds

   real (real128) :: JulianDay, JulianDay_midnight
   real (real128) :: Day_remainder
   real (real128) :: Seconds_remainder

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   call MET_to_UT1 ( MET, Success, JulianDay, year, month, day, hours, minutes, seconds )


   if ( Success ) then

      call Date_and_Time_to_JulianDay_midnight ( year, month, day, hours, minutes, seconds,    &
                    JulianDay_midnight, Day_remainder )

      Greenwich_Mean_SiderealTime = Greenwich_Mean_SiderealTime_at_0hr ( JulianDay_midnight )

      Seconds_remainder =  Day_remainder * 86400.0_real128

      Greenwich_Mean_SiderealTime = Greenwich_Mean_SiderealTime  +  1.002737909350795_real128 * Seconds_remainder

   else

      Greenwich_Mean_SiderealTime = huge ( 1.0_real128 )

   end if


   return

end subroutine Calc_Greenwich_Mean_SiderealTime


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  Converts a time to a Julian Day that is midnight, plus a remainder number of days.
!  A Julian Day Number that is midnight has a fractional part of 0.5.
!  This is needed for the more accurate sidereal time calculation, which has an
!  equation that is only valid for Julian Day Numbers that are midnight.

subroutine Date_and_Time_to_JulianDay_midnight ( year, month, day, hours, minutes, seconds,    &
                 JulianDay, Day_remainder )

   integer (int32), intent (in) :: year
   integer (int32), intent (in) :: month
   integer (int32), intent (in) :: day
   integer (int32), intent (in) :: hours
   integer (int32), intent (in) :: minutes
   real (real64), intent (in) :: seconds

   real (real128), intent (out) :: JulianDay
   real (real128), intent (out) :: Day_remainder

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   real (real128) :: JulianDayCalc
   integer (int32) :: JulianDay_int

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   JulianDayCalc = Date_and_Time_to_JulianDay  ( year, month, day, hours, minutes, seconds )

   JulianDay_int = int (JulianDayCalc)

   Day_remainder = JulianDayCalc - JulianDay_int

   if ( Day_remainder >= 0.5_real128 ) then
      JulianDay = JulianDay_int + 0.5_real128
      Day_remainder = Day_remainder - 0.5_real128
   else
      JulianDay = JulianDay_int - 0.5_real128
      Day_remainder = Day_remainder + 0.5_real128
   end if


end subroutine Date_and_Time_to_JulianDay_midnight


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  Based on the algorithm on page 7 of "Practical Astronomy with your Calculator"
!  One variation: that algorithm had day as a real number, to use fractional parts
!  to represent times finer than one day.   I use day as an integer, and add
!  variables for hours, minutes and seconds.   If you need to go finer than seconds,
!  the seconds variable is a real variable -- use fractional seconds.
!  M. S. Briggs, 2009 November 21.

function  Date_and_Time_to_JulianDay  ( year, month, day, hours, minutes, seconds )

   real (real128) :: Date_and_Time_to_JulianDay   ! function return

   integer (int32), intent (in) :: year
   integer (int32), intent (in) :: month
   integer (int32), intent (in) :: day
   integer (int32), intent (in) :: hours
   integer (int32), intent (in) :: minutes
   real (real64), intent (in) :: seconds

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   real (real128) :: jd


   integer (int32) :: year_prime
   integer (int32) :: month_prime
   integer (int32) :: A
   integer (int32) :: B
   integer (int32) :: C
   integer (int32) :: D
   real (real128) :: frac_day


   ! * * * * *  * * * * *  * * * * *  * * * * *  * * * * *  * * * * *  * * * * *  * * * * *  * * * * *


   year_prime = year
   month_prime = month

   !  2)
   if ( month == 1  .OR.  month == 2 ) then
      year_prime = year - 1_int32
      month_prime = month + 12_int32
   end if

   !  3)
   !  Is the date >= 1582 Oct 15 (i.e., Gregorian calender)
   if ( (year_prime > 1583) .OR.                       &
        (year_prime == 1582 .AND. month > 10) .OR.     &
        (year_prime == 1582 .AND. month == 10 .AND. day >= 15) ) then

      A = year_prime / 100_int32
      B = 2_int32 - A + A / 4_int32

   else
      B = 0_int32
   end if

   !  4)
   if ( year_prime < 0 ) then

      C = int ( 365.25_real64 * year_prime - 0.75_real64 )

   else

      C = int (365.25_real64 * year_prime)

   end if

   !  5)
   D = int ( 30.6001_real64 * (month_prime + 1_int32) )


   !  6)
   jd = B + C + D + day + 1720994.5_real128

   frac_day = minutes + seconds / 60.0_real128
   frac_day = hours + frac_day / 60.0_real128
   frac_day = frac_day / 24.0_real128

   jd = jd + frac_day

   Date_and_Time_to_JulianDay = jd

   return

end function Date_and_Time_to_JulianDay


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!   Adapted from a routine in the GBM GLAST Burst Monitor Flight Software: SunLoc.c
!   Michael S. Briggs, 2009 Nov 25.  UAH / NSSTC.
!   Further revised, 2009 Dec 20 -- now calculates apparent position.

!  MSB, GBM FSW version: 2005 August 25.

!   This version, "SunLoc", almost duplicates the calculation of the GBM FSW,
!   but with the improvement:
!   this code includes the corrections for nutation and aberration, and therefore
!   calculates apparent rather than true longitude lambda (an intermediate quantity).

!   Output is now apparent position.  The corrections for nutation and
!   aberrations are now included.    See p. 152 & 153 of J. Meeus.
!   The output positions are thus referred to the true equinox of the date,
!   not to J2000 -- see Meeus page 152.


!   One test: example 24.a of J. Meeus on page 153 -- verifies correct implementation
!   of the algorithm:
!   Quantity      RA           Dec
!   Example    198.38082    -7.78507
!   Output     198.380819   -7.785067
!   The agreement is exact, to the accuracy given by Meeus.


!   A test to another source, which is probably more accurate: the program MICA 2.2
!   of the U.S. Naval Observatory.
!   In MICA, compute for:
!   Origin = Geocentric, Place = Apparent, Reference Epoch = Date,
!   Reference Plane = Equator, and Coordinates to Spherical,
!   to obtain "Apparent Geocentric Positions / True Equator and Equinox of Date".

!   Time or prog         RA           Dec
!  2010 Feb 5 noon  =  JD 2455233.000
!     MICA            319.02622     -15.869650
!  This routine       319.031926    -15.867691

!   Time or prog         RA           Dec
!  2020 May 5 noon  =  JD 2458975.000
!     MICA             42.98460      +16.465760
!  This routine        42.981631     +16.464645

!   Time or prog         RA           Dec
!  2008 July 11 noon  =  JD 2454659.000
!     MICA            111.15982      +22.015135
!  This routine       111.159858     +22.015362

!   Conclusion: accuracy is probaby better than about 0.02 degree over the time
!   span of Fermi.


!  Unless otherwise stated, all equations, page and chapter numbers
!  are from Astronomical Algorithms by Jean Meeus, copyright 1991
!  by Willmann-Bell, Inc.

!  Caution: the equations in that book are in DEGREES!
!  Most of the variables in this routine follow the equations of
!  Meeus and have units of degrees -- these variables have "_deg"
!  appended to their names.

!  Also consulted: Practical Astronomy with your Calculator, 3rd ed.,
!  by Peter Duffett-Smith.


subroutine  SunLoc_Apparent ( JulianDay, RA, Dec, obliquity_ecliptic_deg_OPTIONAL_ARG )


   real (real128), intent (in) :: JulianDay

   real (real64), intent (out) :: RA    ! RA, radians
   real (real64), intent (out) :: Dec   ! Dec, radians

   !  The mean obliquity of the ecliptic.   Calculated below from Eq. 21.2 of J. Meeus,
   !  which originates from the IAU.  Error is stated as 1 arc sec over 2000 years.
   !  Optional argument because only some callers want this information.

   real (real64), intent (out), OPTIONAL :: obliquity_ecliptic_deg_OPTIONAL_ARG


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   real (real64) :: T_cen

   real (real64) :: LongMean_deg

   real (real64) :: MeanAnomaly_deg

   real (real64) :: eccen

   real (real64) :: equation_center_deg

   real (real64) :: TrueLong_deg

   real (real64) :: obliquity_ecliptic_deg

   real (real64) :: Omega

   real (real64) :: lambda

   real (real64) :: sin_Dec

   integer (int32) :: factor

   real (real128) :: tmp_QRK

   !  ************************************************************************************


   !  We neglect the correction from S/C Time (GPS Time) to
   !  Dynamical Time because it doesn't matter for the accuracy
   !  that we need.  The Sun (really the earth) only moves 360 degrees
   !  in one year, and the difference between UT or GPS times
   !  and Dynamical or Ephemeris Times will be roughly 60 s
   !  during the GLAST era.  The difference arises because the
   !  rotation of the Earth is slowing down -- UT is adjusted as
   !  the earth slows down, while dynamical (aka ephemeris)
   !  advances continuously and is more suitable for astronomical
   !  calculations.


   !  Equation 24.1, Julian centuries of 36525 ephemeris days from
   !  J2000.0 (2000 January 1.5 TD = 2000 Jan 01 noon):

   tmp_QRK = ( JulianDay - 2451545.0_real128 ) / 36525.0_real128
   T_cen = real ( tmp_QRK, real64 )


   !  Equation 24.2: Geometric Mean [ecliptic] Longitude of the Sun,
   !  in degrees.  "mean" means pretending a circular orbit.

   LongMean_deg = 280.46645_real64 + 36000.76983_real64 * T_cen + 0.0003032_real64 * T_cen **2
   factor = int ( LongMean_deg / 360.0_real64 )
   LongMean_deg = LongMean_deg - factor * 360.0_real64
   if ( LongMean_deg < 0.0 )  LongMean_deg = LongMean_deg + 360.0_real64


   !  Equation 24.3: Mean anomaly of Sun in degress.
   !  "anomaly" is the angle the sun has moved.
   !  [mean ==> viewed from the center of the putatively circularly orbit.]
   !  ["anomaly" is referenced to perigee.]

   MeanAnomaly_deg = 357.52910_real64 + 35999.05030_real64 * T_cen    &
                   - 0.0001559_real64 * T_cen **2                     &
                   - 4.8E-7_real64 * T_cen **3
   factor = int ( MeanAnomaly_deg / 360.0_real64 )
   MeanAnomaly_deg = MeanAnomaly_deg - factor * 360.0_real64
   if ( MeanAnomaly_deg < 0.0 )  MeanAnomaly_deg = MeanAnomaly_deg + 360.0_real64


   !  Equation 24.4: eccentricity of the Earth's orbit:

   eccen = 0.016708617_real64 - 0.000042037_real64 * T_cen - 1.236E-7_real64 * T_cen **2


   !  The equation of the center -- approx solution of ellipical orbit:

   equation_center_deg =   &
( 1.914600_real64 - 0.004817_real64 * T_cen - 0.000014_real64 * T_cen **2 ) * sin ( MeanAnomaly_deg * DEGREES_TO_RADIANS_DBL ) &
         + ( 0.019993_real64 - 0.000101_real64 * T_cen ) * sin ( 2.0_real64 * MeanAnomaly_deg * DEGREES_TO_RADIANS_DBL )       &
         + 0.000290_real64 * sin ( 3.0_real64 * MeanAnomaly_deg * DEGREES_TO_RADIANS_DBL )


   ! The Sun's True Longitude:

   TrueLong_deg = LongMean_deg + equation_center_deg


   !  The ecliptic latitude of the Sun, referenced to the Ecliptic of the date,
   !  never exceeds 1.2 arc secs, which we can surely neglect.


   !  To convert from Ecliptic to Equatorial Coordinates, we need the
   !  obliquity of the ecliptic, eq. 21.2 (but converted to decimal degrees):

   obliquity_ecliptic_deg = 23.4392911_real64  -  (46.8150_real64  / 3600.0_real64) * T_cen    &
                     -  (0.00059_real64  / 3600.0_real64) * T_cen **2                       &
                     +  (0.001813_real64 / 3600.0_real64) * T_cen **3

   if ( present ( obliquity_ecliptic_deg_OPTIONAL_ARG ) )           &
      obliquity_ecliptic_deg_OPTIONAL_ARG = obliquity_ecliptic_deg


   !  Calculation of apparent longitude lambda, referred to the true equinox of the date,
   !  correcting the true longitude TrueLong_Deg (aka Theta) for nutation and aberration.
   !  Also correcting the obliquity of the ecliptic.
   !  From equation on page 152 of J. Meeus, and equation 24.8 on page 153.

   Omega = 125.04_real64 - 1934.136_real64 * T_cen

   lambda = TrueLong_deg  -  0.00569_real64  -  0.00478_real64 * sin ( Omega * DEGREES_TO_RADIANS_DBL )

   obliquity_ecliptic_deg = obliquity_ecliptic_deg  +  0.00256_real64 * cos ( Omega * DEGREES_TO_RADIANS_DBL )


   !  The one bad case for rounding errors on the input for atan2 should be impossible --
   !  the Sun cannot be at either pole of the Ecliptic coordinate system:
   !  Inverse tangent of ( cos (obliquity_ecliptic_deg) * sin (lambda) ) /  cos (lambda) :

   RA = atan2 ( cos ( obliquity_ecliptic_deg * DEGREES_TO_RADIANS_DBL ) * sin ( lambda * DEGREES_TO_RADIANS_DBL ),   &
                     cos ( lambda * DEGREES_TO_RADIANS_DBL ) )

   !  Change range of RA from -PI to +PI to 0 to 2PI:

   if ( RA <  0.0_real64 )  RA = RA + 2.0_real64 * PI_DBL



   sin_Dec = sin ( obliquity_ecliptic_deg * DEGREES_TO_RADIANS_DBL ) * sin ( lambda * DEGREES_TO_RADIANS_DBL )

   if ( sin_Dec >  1.0000_real64  .AND.  sin_Dec <  1.000001_real64 )  sin_Dec =  1.0_real64
   if ( sin_Dec < -1.0000_real64  .AND.  sin_Dec > -1.000001_real64 )  sin_Dec = -1.0_real64

   Dec = asin ( sin_Dec )


   return

end subroutine SunLoc_Apparent


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!   Adapted from a routine in the GBM GLAST Burst Monitor Flight Software: SunLoc.c
!   Michael S. Briggs, 2009 Nov 25.  UAH / NSSTC.
!   Further revised, 2010 March 13 & 16 -- now calculates J2000 position.

!  MSB, GBM FSW version: 2005 August 25.

!   This version, "SunLoc_J2000", almost duplicates the calculation of the GBM FSW,
!   but with two differences.   (Version "FSW_SunLoc" has only one difference.)
!   1) First difference: a bug is corrected in the calcuation of the
!   obliquity of the ecliptic.   The minus signs of two terms are plus signs
!   in the FSW code of FSW versions 2.5 and earlier -- expected to be
!   corrected in FSW version 2.6.
!   2) Second difference from the FSW, and a difference from "FSW_SunLoc" and
!   "SunLoc_Apparent" -- the position is corrected to be for the coordinate
!   system J2000.   This is done for the longitude of the Sun using the
!   simple equation on page 152 of J. Meeus, which is implied to have an accuracy
!   of 0.01 degree between 1900 and 2100.   The latitude of the sun is essentially
!   zero.

!   Note: unlike "SunLoc_Apparent", this version does NOT include the corrections
!   for nutation and aberration.   Instead of correcting the true geometric position
!   the mean equinox of the date to the apparent position, it corrects to J2000
!   coordinates.

!   Output is now J2000.  The position is J2000, rather than the true geometric
!   position referred to the mean equinox of the date, or the apparent position
!   referred to the true equinox of the data -- see J. Meeus, page 152.

!   One test: example 24.a of J. Meeus on page 153 -- verifies correct implementation
!   of the algorithm:
!   Quantity      RA           Dec
!   Example    198.38082    -7.78507
!   Output     198.380819   -7.785067
!   The agreement is exact, to the accuracy given by Meeus.


!   A test to another source, which is probably more accurate: the program MICA 2.2
!   of the U.S. Naval Observatory.
!   In MICA, compute for:
!   Origin = Geocentric, Place = Astrometric, Reference Epoch = J2000,
!   Reference Plane = Equator, and Coordinates to Spherical,
!   to obtain "Astrometric Positions / Mean Equator and Equinox of J2000.0".

!   Time or prog         RA           Dec
!  2010 Feb 5 noon  =  JD 2455233.000
!     MICA            318.88711     -15.911173
!  This routine       318.893018    -15.909087

!   Time or prog         RA           Dec
!  2020 May 5 noon  =  JD 2458975.000
!     MICA             42.71200      +16.385899
!  This routine        42.708661     +16.384874

!   Time or prog         RA           Dec
!  2008 July 11 noon  =  JD 2454659.000
!     MICA            111.03499      +22.030154
!  This routine       111.035221     +22.030206

!   Conclusion: accuracy is probaby better than about 0.02 degree over the time
!   span of Fermi.


!  Unless otherwise stated, all equations, page and chapter numbers
!  are from Astronomical Algorithms by Jean Meeus, copyright 1991
!  by Willmann-Bell, Inc.

!  Caution: the equations in that book are in DEGREES!
!  Most of the variables in this routine follow the equations of
!  Meeus and have units of degrees -- these variables have "_deg"
!  appended to their names.

!  Also consulted: Practical Astronomy with your Calculator, 3rd ed.,
!  by Peter Duffett-Smith.


subroutine  SunLoc_J2000 ( JulianDay, RA, Dec, obliquity_ecliptic_deg_OPTIONAL_ARG )


   real (real128), intent (in) :: JulianDay

   real (real64), intent (out) :: RA    ! RA, radians
   real (real64), intent (out) :: Dec   ! Dec, radians

   !  The mean obliquity of the ecliptic.   Calculated below from Eq. 21.2 of J. Meeus,
   !  which originates from the IAU.  Error is stated as 1 arc sec over 2000 years.
   !  Optional argument because only some callers want this information.

   real (real64), intent (out), OPTIONAL :: obliquity_ecliptic_deg_OPTIONAL_ARG


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   real (real64) :: T_cen

   real (real64) :: LongMean_deg

   real (real64) :: MeanAnomaly_deg

   real (real64) :: eccen

   real (real64) :: equation_center_deg

   real (real64) :: TrueLong_deg

   real (real64) :: Long_J2000_deg

   real (real64) :: obliquity_ecliptic_deg

   real (real64) :: sin_Dec

   integer (int32) :: factor

   real (real128) :: tmp_QRK

   !  ************************************************************************************


   !  We neglect the correction from S/C Time (GPS Time) to
   !  Dynamical Time because it doesn't matter for the accuracy
   !  that we need.  The Sun (really the earth) only moves 360 degrees
   !  in one year, and the difference between UT or GPS times
   !  and Dynamical or Ephemeris Times will be roughly 60 s
   !  during the GLAST era.  The difference arises because the
   !  rotation of the Earth is slowing down -- UT is adjusted as
   !  the earth slows down, while dynamical (aka ephemeris)
   !  advances continuously and is more suitable for astronomical
   !  calculations.


   !  Equation 24.1, Julian centuries of 36525 ephemeris days from
   !  J2000.0 (2000 January 1.5 TD = 2000 Jan 01 noon):

   tmp_QRK = ( JulianDay - 2451545.0_real128 ) / 36525.0_real128
   T_cen = real ( tmp_QRK, real64 )


   !  Equation 24.2: Geometric Mean [ecliptic] Longitude of the Sun,
   !  in degrees.  "mean" means pretending a circular orbit.

   LongMean_deg = 280.46645_real64 + 36000.76983_real64 * T_cen + 0.0003032_real64 * T_cen **2
   factor = int ( LongMean_deg / 360.0_real64 )
   LongMean_deg = LongMean_deg - factor * 360.0_real64
   if ( LongMean_deg < 0.0 )  LongMean_deg = LongMean_deg + 360.0_real64


   !  Equation 24.3: Mean anomaly of Sun in degress.
   !  "anomaly" is the angle the sun has moved.
   !  [mean ==> viewed from the center of the putatively circularly orbit.]
   !  ["anomaly" is referenced to perigee.]

   MeanAnomaly_deg = 357.52910_real64 + 35999.05030_real64 * T_cen    &
                   - 0.0001559_real64 * T_cen **2                     &
                   - 4.8E-7_real64 * T_cen **3
   factor = int ( MeanAnomaly_deg / 360.0_real64 )
   MeanAnomaly_deg = MeanAnomaly_deg - factor * 360.0_real64
   if ( MeanAnomaly_deg < 0.0 )  MeanAnomaly_deg = MeanAnomaly_deg + 360.0_real64


   !  Equation 24.4: eccentricity of the Earth's orbit:

   eccen = 0.016708617_real64 - 0.000042037_real64 * T_cen - 1.236E-7_real64 * T_cen **2


   !  The equation of the center -- approx solution of ellipical orbit:

   equation_center_deg =   &
( 1.914600_real64 - 0.004817_real64 * T_cen - 0.000014_real64 * T_cen **2 ) * sin ( MeanAnomaly_deg * DEGREES_TO_RADIANS_DBL ) &
      + ( 0.019993_real64 - 0.000101_real64 * T_cen ) * sin ( 2.0_real64 * MeanAnomaly_deg * DEGREES_TO_RADIANS_DBL )          &
      + 0.000290_real64 * sin ( 3.0_real64 * MeanAnomaly_deg * DEGREES_TO_RADIANS_DBL )


   ! The Sun's True Longitude:

   TrueLong_deg = LongMean_deg + equation_center_deg


   !  The ecliptic latitude of the Sun, referenced to the Ecliptic of the date,
   !  never exceeds 1.2 arc secs, which we can surely neglect.


   !  To convert from Ecliptic to Equatorial Coordinates, we need the
   !  obliquity of the ecliptic, eq. 21.2 (but converted to decimal degrees):

   obliquity_ecliptic_deg = 23.4392911_real64  -  (46.8150_real64  / 3600.0_real64) * T_cen    &
                     -  (0.00059_real64  / 3600.0_real64) * T_cen **2                          &
                     +  (0.001813_real64 / 3600.0_real64) * T_cen **3

   if ( present ( obliquity_ecliptic_deg_OPTIONAL_ARG ) )           &
      obliquity_ecliptic_deg_OPTIONAL_ARG = obliquity_ecliptic_deg


   !  Correction of the Longitude to J2000 using equation on page 152 of J. Meeus.
   !  Accuracy is implied to be 0.01 degrees from 1900 to 2100.

   !  The equation of Meeus uses "Year - 2000", instead we use
   !  Julian Day Number minus the Julian Day Number of 2000 Jan 01 noon, divided
   !  by length of a year in days.   This avoids calculating the year....
   !  Testing showed, for example inputs from 1900 to 2100, the change in the
   !  output RA & Dec to be less than 0.0001 degrees.

   tmp_QRK = TrueLong_deg  - 0.01397_real128 * (JulianDay - 2451545.0_real128 ) / 365.25_real128
   Long_J2000_deg = real ( tmp_QRK, real64 )


   !  The one bad case for rounding errors on the input for atan2 should be impossible --
   !  the Sun cannot be at either pole of the Ecliptic coordinate system:
   !  Inverse tangent of ( cos (obliquity_ecliptic_deg) * sin (longitude) ) /  cos (longitude) :

   RA = atan2 ( cos ( obliquity_ecliptic_deg * DEGREES_TO_RADIANS_DBL ) * sin ( Long_J2000_deg * DEGREES_TO_RADIANS_DBL ),   &
                     cos ( Long_J2000_deg * DEGREES_TO_RADIANS_DBL ) )

   !  Change range of RA from -PI to +PI to 0 to 2PI:

   if ( RA <  0.0_real64 )  RA = RA + 2.0_real64 * PI_DBL



   sin_Dec = sin ( obliquity_ecliptic_deg * DEGREES_TO_RADIANS_DBL ) * sin ( Long_J2000_deg * DEGREES_TO_RADIANS_DBL )

   if ( sin_Dec >  1.0000_real64  .AND.  sin_Dec <  1.000001_real64 )  sin_Dec =  1.0_real64
   if ( sin_Dec < -1.0000_real64  .AND.  sin_Dec > -1.000001_real64 )  sin_Dec = -1.0_real64

   Dec = asin ( sin_Dec )


   return

end subroutine SunLoc_J2000


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!   apparent solar time = mean solar time + equation of time.
!   or: equation of time = apparent solar time - mean solar time.
!   "mean solar time" is solar time for the "fictious mean sun" that moves at a uniform rate.

!   Input is Julian Day.
!  Note: properly the input should be Julian Ephemeris Day, probably we are
!  ignoring this refinement.

!  Output, via the function return, is the "equation of time" in minutes.

!  Michael S. Briggs, 2010 January 16, UAH / NSSTC.

!  Tested against example 27.a of Meeus: input 1992 Oct 13, 0 h,
!  expected value 13 min 42.6 s, obtain 13.698967 minutes = 13 min 41.9 s --
!  so agreement within one second of time.
!  The USNO program MICA gives 13 min 42.4 s.
!  Also compared with several various "random" times with MICA -- difference ranges from
!  0.1 to 1.2 second.

!  2014 Feb 9: retest against example 27.a of Meeus, which is Julian Day number
!  2448908.50952083333323.    Obtain 13.7012596 which is 13 m 42.08 s.   Closer
!  than before to the expected values.


function  Equation_of_Time  ( JulianDay )

   real (real32) :: Equation_of_Time    ! function return.  units: minutes

   real (real128), intent (in) :: JulianDay

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   real (real64) :: RA    ! RA in J2000 coordinate system, radians
   real (real64) :: Dec   ! Dec in J2000 coordinate system, radians

   real (real64) :: obliquity_ecliptic_deg   ! mean obliquity of the ecliptic, degrees

   real (real128) :: tau

   real (real128) :: L0

   real (real32) :: E

   real (real128) :: tmp_QRK

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   call  SunLoc_Apparent ( JulianDay, RA, Dec, obliquity_ecliptic_deg )


   ! time in Julian millenia from J2000.0, i.e.,
   ! 365250 ephemeris days from J2000.0.

   tau = ( JulianDay - 2451545.0_real128 ) / 365250.0_real128


   ! Eq. 27.2 -- mean longitude of the sun

   L0 = 280.4664567_real128  +  360007.6982779_real128 * tau  + 0.03032028_real128 * tau**2    &
         + tau**3 / 49931.0_real128  - tau**4 / 15299.0_real128  -  tau**4 / 1988000.0_real128

   !  Remove multiples of 360 degrees to make the range of L0 be 0 to 360 degrees:
   L0 = mod ( L0, 360.0_real128 )
   if ( L0 < 0.0_real128 )  L0 = L0 + 360.0_real128


   ! Eq. 27.1, Equation of time in degrees.    The factor 1/3600 is to convert
   !  DeltaPsi from arc seconds to degrees.

   tmp_QRK = L0 - 0.0057183_real128  -  RA * RADIANS_TO_DEGREES_DBL +    &
                DeltaPsi_arcsec ( JulianDay ) / 3600.0_real128 * cos ( obliquity_ecliptic_deg * DEGREES_TO_RADIANS_DBL )
   E = real ( tmp_QRK, real32 )

   !  convert to minutes of time:
   E = 4.0_real32 * E

   !  The value should be less than 20 minutes, if not, correct by
   !  shifting by one day of minutes:

   if ( E < -30.0_real32 )  E = E + 1440.0_real32
   if ( E > +30.0_real32 )  E = E - 1440.0_real32


   Equation_of_Time = E


   return

end function  Equation_of_Time


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  Be careful with units, some equations use Delta_Psi in seconds of arc,
!  others in degrees.

!  Calculates the nutation in longitude, Delta Psi, caused mostly by the Moon.
!  Uses the equations of Chapter 21 of J. Meeus, Astronomial Algorithms.

!  Note: properly the input should be Julian Ephemeris Day, probably we are
!  ignoring this refinement.

function  DeltaPsi_arcsec ( JulianDay )


   real (real32) :: DeltaPsi_arcsec   ! function return.  units: arc sec, ''.

   real (real128), intent (in) :: JulianDay


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   real (real128) :: T_cen

   real (real128) :: Omega_deg, LongSun_deg, LongMoon_deg

   real (real128) :: tmp_QRK

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   ! Eq. 21.1, time in Julian centuries from Epoch J2000.0:

   T_cen = ( JulianDay - 2451545.0_real128 ) / 36525.0_real128

   ! Longitude of the ascending node of the Moon's mean orbit on the ecliptic,
   !  measured from the mean equinox of the date:

   Omega_deg = 125.04452_real128  -  1934.136261_real128 * T_cen  +  0.0020708_real128 * T_cen**2  +  T_cen**3 / 450000.0_real128

   LongSun_deg = 280.4665_real128  +  36000.7698_real128 * T_cen

   LongMoon_deg = 218.3165_real128  +  481267.8813_real128 * T_cen


   ! DeltaPsi in seconds of arc:

   tmp_QRK =  -17.20_real128 * sin ( Omega_deg * DEGREES_TO_RADIANS_QUAD )    &
              - 1.32_real128 * sin ( 2.0_real128 * LongSun_deg * DEGREES_TO_RADIANS_QUAD )    &
              - 0.23_real128 * sin ( 2.0_real128 * LongMoon_deg * DEGREES_TO_RADIANS_QUAD )   &
              + 0.21_real128 * sin ( 2.0_real128 * Omega_deg * DEGREES_TO_RADIANS_QUAD )
   DeltaPsi_arcsec = real ( tmp_QRK, real32 )

   !  convert to degrees:
   !DeltaPsi = DeltaPsi / 3600.0_real32


   return

end function  DeltaPsi_arcsec


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


subroutine  UT_to_LocalSolarTime  ( year, month, day, hours, minutes, seconds, East_Longitude_deg_ARG,    &
        Local_SolarTime_frac_of_day, LST_hours, LST_minutes )


   integer (int32), intent (in) :: year
   integer (int32), intent (in) :: month
   integer (int32), intent (in) :: day
   integer (int32), intent (in) :: hours
   integer (int32), intent (in) :: minutes
   real (real64), intent (in) :: seconds
   real (real64), intent (in) :: East_Longitude_deg_ARG

   !  output in two forms: redundant:
   real (real32), intent (out) :: Local_SolarTime_frac_of_day    ! units: fraction of day
   integer (int32), intent (out) :: LST_hours
   integer (int32), intent (out) :: LST_minutes

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   real (real32) :: East_Longitude_deg
   real (real32) :: Equation_of_Time_frac_of_day
   real (real32) :: East_Longitude_frac_of_day
   real (real32) :: UT_frac_of_day   !  UT: Universal Time.  units: fraction of day

   real (real32) :: tmpSGL
   real (real64) :: tmpDBL

   real (real128) :: JulianDay


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   !  The desired range of the input variable East_Longitude_deg is 0 to 360 degrees.
   !  If instead the input ranges from -180 to +180 degrees, correct...

   East_Longitude_deg = real ( East_Longitude_deg_ARG, real32 )
   if ( East_Longitude_deg < 0.0_real32 )  East_Longitude_deg = East_Longitude_deg + 360.0_real32

   East_Longitude_frac_of_day = East_Longitude_deg / 360.0_real32


   JulianDay = Date_and_Time_to_JulianDay  ( year, month, day, hours, minutes, seconds )


   !  1440: convert from minutes to days.
   Equation_of_Time_frac_of_day = Equation_of_Time  ( JulianDay ) / 1440.0_real32


   tmpDBL = ( ( (seconds / 60.0_real64) + minutes ) / 60.0_real64 + hours ) / 24.0_real64
   UT_frac_of_day = real ( tmpDBL, real32 )

   Local_SolarTime_frac_of_day = UT_frac_of_day + East_Longitude_frac_of_day + Equation_of_Time_frac_of_day

   if ( Local_SolarTime_frac_of_day  >  1.0_real32)    &
      Local_SolarTime_frac_of_day = Local_SolarTime_frac_of_day - 1.0_real32

   if ( Local_SolarTime_frac_of_day  <  0.0_real32)    &
      Local_SolarTime_frac_of_day = Local_SolarTime_frac_of_day + 1.0_real32

   tmpSGL = 24.0_real32 * Local_SolarTime_frac_of_day
   LST_hours = int ( tmpSGL, int32 )
   tmpSGL = 60.0_real32 * (tmpSGL - LST_hours)
   LST_minutes = nint (tmpSGL)
   if (LST_minutes == 60 ) then
      LST_minutes = 0
      LST_hours = LST_hours + 1
   end if

   return

end subroutine  UT_to_LocalSolarTime


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^

!  WARNING: THIS CALCULATION MAY NOT HAVE SUFFICIENT ACCURACY, ESPECIALLY DEPENDING
!  ON THE PRECISE TYPE OF "TIME" USED FOR JULIAN DAY.   MANY ARTICLES AND BOOKS ARE
!  IMPRECISE ON THAT ASPECT OF THE CALCULATION.   USING UTC IS NOT ACCURATE.

!  Returns Sidereal time at zero Longitude in degrees, from 0 to 360, on the current day,
!  whatever that day might be.
!  Input is time in Julian Days.

!  From p. 84 of J. Meeus, describing the result of eq. 11.4 (used herein):
!  "The sidereal time ... is the mean sidereal time, that is, the Greenwich hour
!   angle of the mean vernal point (the intersection of the ecliptic of the date
!   with the mean equator of the date)".

!  Adapted from GBM FSW routine GeoCoord -- only a portion of that routine.
!  Michael S. Briggs, 2010 Jan 16.

!  The UT to Sidereal time problem is solved by both P. Duffet-Smith
!  (Practical Astronomy with Your Calculator) and J. Meeus
!  (Astronomical Algorithms).   We prefer eq. 11.4 of Meeus (pp. 83-84)
!  because it is more general, capable of handling any time in a single step.
!  The other equations require a second step to deal with times that aren't
!  0 hours UT.

!  Also based upon an IDL routine for CGRO of Valerie Connaughton.


function  JulianDay_to_SiderealTime_Ver1  ( JulianDay )

   real (real64) :: JulianDay_to_SiderealTime_Ver1   ! function return

   real (real128), intent (in) :: JulianDay


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   real (real128) :: T_cen
   real (real128) :: theta_deg

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   !  Equation 11.1, Julian centuries of 36525 ephemeris days from J2000.0

   T_cen = ( JulianDay - 2451545.0_real128 ) / 36525.0_real128


   !  The basic idea is that a sidereal (star referenced) day is approx. 4 min
   !  shorter than a solar day.
   !  From P. Duffett-Smith (p. 16), there are 365.25 solar days in a year,
   !  with a year being the time that it takes the Sun to return to the
   !  same position w.r.t the stars.  During this time, the Earth makes
   !  366.25 rotations about its axis, i.e., there are 366.25 sidereal days
   !  in a year.
   !  I note that 366.25 / 365.25 * 360 = 1.002737851 * 360 = 360.985626,
   !  which explains the primary constant of the equation.
   !  The values 366 1/4 or 365 1/4 aren't quite exact, since the first
   !  constant is actually 1.00273790935 (as it appears in Duffett-Smith and
   !  Meeus), and the second 360.98564736626 (as it appears in Meeus and below).
   !  (Alternatively, 0.00273790935 * 24 hours = 3.94 minutes.)

   !  There is of course a "zero" point value to get the sidereal time of zero UT correct.

   !  Meeus attributes this equation to the IAU in 1982.


   !  Meeus Eq. 11.4 -- including the cube term that is omitted in the GBM FSW:
   !  theta_deg is Greenwich Sidereal Time in degrees.

   theta_deg = 280.46061837_real128  +  360.98564736629_real128 * ( JulianDay - 2451545.0_real128 )    &
                 + 0.000387933_real128 * T_cen**2  -  T_cen**3 / 38710000.0_real128


   !  Remove multiples of 360. degrees to return the sidereal time on the current day.

   theta_deg = mod ( theta_deg, 360.0_real128 )


   !  Another feature not present in the GBM FSW -- converts the results to 0 to 360 degrees
   !  for input times before approximately time 2000.0 years.

   if ( theta_deg < 0.0_real128 )  theta_deg = theta_deg + 360.0_real128


   JulianDay_to_SiderealTime_Ver1  =  real ( theta_deg, real64 )


   return


end function  JulianDay_to_SiderealTime_Ver1


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^

!  Subroutine to convert GLAST/Fermi Mission Elapsed Time (MET) to calendar date and time
!  (hours, minutes, seconds) UTC.   The MET must be in units of seconds, rather than raw
!  GBM values of tenths of seconds, 2 microseconds.

!  This routine now handles leap seconds correctly, even returning the special value
!  seconds = 60 for the leap second itself.
!  The only possible issue is that the JulianDay value might have a small glitch near
!  the leap seconds...    Need to think about this...
!  The UT time values [ year, month, day, hours, minutes, seconds ] should be correct for
!  all times from the GLAST Epoch of midnight, 2001 January 1 through what ever leap
!  seconds are recorded in the file MET_of_LeapSeconds.txt -- see subroutine
!  Read_LeapSecondData.

!  Michael S. Briggs, UAH / NSSTC, copyright 2009 August 14,
!  Major redesign 2010 February 21.
!  Revised 2012 June 15 for the 3rd leap second on 2012 June 30.


subroutine  MET_to_UTC ( GBM_MET, JulianDay, year, month, day, hours, minutes, seconds, seconds_string )


   ! * * * * *  * * * * *  * * * * *  * * * * *  * * * * *  * * * * *

   !  MET time has 9 digits before the decimal point, 6 after (to microseconds),
   !  for a total of 15 decimal digits.  This just matches double-precision, but I
   !  choose not to use double precision in order to have margin.  Therefore use
   !  quad precision.

   real (real128), intent (in) :: GBM_MET

   real (real128), intent (out) :: JulianDay

   integer (int32), intent (out) :: year
   integer (int32), intent (out) :: month
   integer (int32), intent (out) :: day
   integer (int32), intent (out) :: hours
   integer (int32), intent (out) :: minutes
   real (real64), intent (out) :: seconds

   character (len=9), intent (out), OPTIONAL :: seconds_string

   ! * * * * *  * * * * *  * * * * *  * * * * *  * * * * *  * * * * *

   !   Constants used by this routine:

   !  GLAST (now Fermi) epoch:
   !  The start of GLAST Mission Elapsed Time -- midnight, January 1, 2001.
   !  I converted this time to Julian Days in my memo "RollOverCalc.txt" of 2004 Feb 17 -- 19.
   !  (The 0.5 is because Julian Days start/end at noon, not midnight.)

   real (real128), parameter :: GLAST_Epoch_JulianDay = 2451910.5_real128


   real (real128), parameter :: seconds_per_day = 86400.0_real128


   ! * * * * *  * * * * *  * * * * *  * * * * *  * * * * *  * * * * *

   !   Local variables:

   integer (int64) :: MET_IntegerPortion

   real (real128) :: MET_Time_Corrected

   logical :: DoneFlag

   real (real128) :: tmp_QRK

   character (len=18) :: MET_string

   integer :: i


   ! * * * * *  * * * * *  * * * * *  * * * * *  * * * * *  * * * * *


   !  load in the module variable LeapSecondData with the MET
   !  values of leap seconds and the dates of the leap seconds.

   if ( .NOT. allocated (LeapSecondData) )  call Read_LeapSecondData ()


   ! Extract integer portion so that we don't have to worry about comparing
   ! floating point numbers.   For positive values, floor will strip off digits
   ! to the right of the decimal point.    IIIII.xxxx becomes  IIIII

   MET_IntegerPortion = floor ( GBM_MET, int64 )


   !  Special cases -- the leap seconds themselves -- these cause the seconds to
   !  have the otherwise impossible value of 60.xxx.

   DoneFlag = .FALSE.

   do i=1, size (LeapSecondData)

      if ( MET_IntegerPortion == LeapSecondData (i) % MET_of_LeapSecond ) then  ! exact leap second ?

         DoneFlag = .TRUE.

         year = LeapSecondData (i) % year
         month = LeapSecondData (i) % month
         day = LeapSecondData (i) % day
         hours = 23
         minutes = 59

         !  leap seconds is 60 seconds.  Then add on fractional part.
         tmp_QRK = 60.0_real64 + GBM_MET - MET_IntegerPortion
         seconds = real ( tmp_QRK, real64 )

      end if  ! exact leap second ?

   end do


   !   As needed, apply leap seconds corrections:

   MET_Time_Corrected = GBM_MET

   do i=1, size (LeapSecondData)
      if ( MET_IntegerPortion > LeapSecondData (i) % MET_of_LeapSecond )   &
         MET_Time_Corrected = MET_Time_Corrected - 1.0_real128
   end do


   JulianDay = GLAST_Epoch_JulianDay + MET_Time_Corrected / seconds_per_day


   !  Actual calculation of UT time, if we don't already have the values via a leap second special case:

   if ( .not. DoneFlag )  call  JulianDay_to_Date_and_Time  ( JulianDay, year, month, day, hours, minutes, seconds )



   !  Create the seconds as a string, if that optional argument is present.
   !  There are two issues to fix:
   !  1) If the seconds are less than 10.0, we want a leading 0, e.g., 07.801844.
   !  There is no way to request Fortran IO to do this for floating point
   !  numbers so we do this ourselves, encapsulated into routine seconds_2_string.
   !  2) We observe rare cases in which the last digit (microseconds) is off by 1.
   !  I haven't figured out the cause -- using Quad precision doesn't fix this --
   !  so we just patch the problem by copying from the MET at the string level.

   !  Times that have seconds epsilon below 60.0 can round to 60.000000.   This
   !  looks peculiar, but isn't totally wrong.  For the moment we don't fix this
   !  output glitch.   The "phase space" is about 5.0E-7 of a second.

   if ( present ( seconds_string ) ) then    ! optional argument ?

      seconds_string = seconds_2_string (seconds)

      write (MET_String, '(F18.6)' )  GBM_MET

      if ( MET_String (18:18) /= seconds_string (9:9) )  seconds_string (9:9) = MET_String (18:18)


   end if  ! optional argument ?



   return

end subroutine  MET_to_UTC


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^

!  Subroutine to convert GLAST/Fermi Mission Elapsed Time (MET) to calendar date and time
!  (hours, minutes, seconds) UT1.   The MET must be in units of seconds, rather than raw
!  GBM valuestest of tenths of seconds, 2 microseconds.

!   UT1 reflects the irregular rotation of the Earth.
!   Calculated from UTC via correction values that are read from "Bulletins" published
!   by the Earth Rotation Service.


subroutine  MET_to_UT1 ( GBM_MET, Success, JulianDay, year, month, day, hours, minutes, seconds, seconds_string )


   ! * * * * *  * * * * *  * * * * *  * * * * *  * * * * *  * * * * *

   !  MET time has 9 digits before the decimal point, 6 after (to microseconds),
   !  for a total of 15 decimal digits.  This just matches double-precision, but I
   !  choose not to use double precision in order to have margin.  Therefore use
   !  quad precision.

   real (real128), intent (in) :: GBM_MET

   logical, intent (out) :: Success

   real (real128), intent (out) :: JulianDay

   integer (int32), intent (out) :: year
   integer (int32), intent (out) :: month
   integer (int32), intent (out) :: day
   integer (int32), intent (out) :: hours
   integer (int32), intent (out) :: minutes
   real (real64), intent (out) :: seconds

   character (len=9), intent (out), OPTIONAL :: seconds_string

   ! * * * * *  * * * * *  * * * * *  * * * * *  * * * * *  * * * * *

   logical, SAVE :: FirstCall = .TRUE.

   integer (int32) :: mjd

   character (len=18) :: MET_string

   ! * * * * *  * * * * *  * * * * *  * * * * *  * * * * *  * * * * *


   !  Read the data with the differences between UT1 and UTC into module
   !  variables MODVAR_UT1_minus_UTC_ms and MODVAR_State_of_UT1_minus_UTC.

   if ( FirstCall ) then
      FirstCall = .FALSE.

      call Get_UT1_minus_UTC_data ( )

   end if

   !  get UTC, a uniform rate of time, except that it includes seconds.

   call MET_to_UTC ( GBM_MET, JulianDay, year, month, day, hours, minutes, seconds )

   mjd = int ( JulianDay - MJD_OFFSET )

   if ( mjd < MODVAR_MJD_range (1)  .OR.  mjd > MODVAR_MJD_range (2) ) then
      Success = .FALSE.
      return  !   ERROR RETURN   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
   else if ( MODVAR_State_of_UT1_minus_UTC (mjd) == 0_int8 ) then
      Success = .FALSE.
      return  !   ERROR RETURN   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
   else
      Success = .TRUE.
   end if



   !  seconds UTC to seconds UT1 via:
   !      seconds [UT1] = seconds [UTC] + MODVAR_UT1_minus_UTC_ms / 1000.0_real64
   !  But we don't want to do it that way due to the problems of carrying from
   !  seconds to minutes to hours to ..., so instead we correct the Julian Day
   !  and then calculate all of the other quantities:

   JulianDay = JulianDay + MODVAR_UT1_minus_UTC_ms (mjd) / ( 1000.0_real128 * 86400.0_real128 )

   call JulianDay_to_Date_and_Time  ( JulianDay, year, month, day, hours, minutes, seconds )


   !  Create the seconds as a string, if that optional argument is present.
   !  There are two issues to fix:
   !  1) If the seconds are less than 10.0, we want a leading 0, e.g., 07.801844.
   !  There is no way to request Fortran IO to do this for floating point
   !  numbers so we do this ourselves, encapsulated into routine seconds_2_string.
   !  2) We observe rare cases in which the last digit (microseconds) is off by 1.
   !  I haven't figured out the cause -- using Quad precision doesn't fix this --
   !  so we just patch the problem by copying from the MET at the string level.

   if ( present ( seconds_string ) ) then    ! optional argument ?

      seconds_string = seconds_2_string (seconds)

      write (MET_String, '(F18.6)' )  GBM_MET

      if ( MET_String (18:18) /= seconds_string (9:9) )  seconds_string (9:9) = MET_String (18:18)


   end if  ! optional argument ?


   return

end subroutine  MET_to_UT1


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  From spherical coordinates to Cartesian.
!  These are Cartesian coordinates fixed to the Earth, not the inertial coordinates
!  used by Fermi.

subroutine SphericalGeoCoord_to_Cartesian ( EastLongitude, Latitude, altitude, Position )

   real (real64), intent (in) :: EastLongitude    ! degrees
   real (real64), intent (in) :: Latitude         ! degrees
   real (real64), intent (in) :: altitude       ! m about Earth's mean radius

   real (real64), dimension (3), intent (out) :: Position   ! m, from SC Position History file


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   real (real64) :: radius

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   Position (1) = cos (Latitude * DEGREES_TO_RADIANS_DBL) * cos (EastLongitude * DEGREES_TO_RADIANS_DBL)
   Position (2) = cos (Latitude * DEGREES_TO_RADIANS_DBL) * sin (EastLongitude * DEGREES_TO_RADIANS_DBL)
   Position (3) = sin (Latitude * DEGREES_TO_RADIANS_DBL)

   radius = EARTH_MEAN_RADIUS_M_DBL + altitude
   Position = Position * radius

end subroutine SphericalGeoCoord_to_Cartesian


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  Renamed from Position_to_GeoCoord to Cartesian_to_SphericalGeoCoord on
!  2013 Nov 22.   To emphasize that the Longitude, Latitude and Altitude
!  are for a spherical Earth !!

!  Converts Position in Inertial (i.e., non-rotation) Earth-centered Cartesian
!  coordinates Position (a 3-vector X, Y, Z) to the position in geographic
!  coordinates EastLongitude, Latitude, altitude.   "altitude" is an optional output.
!  The time is a necessary input, to rotate the Earth.   The rotation is done via
!  Sidereal (star referenced) time, which is calculated from the input value of
!  Fermi MET.

!  Michael S. Briggs, UA Huntsville, 2010 February & March 10.

!  This code is extremely similar to routine GeoCoord.c


subroutine  Cartesian_to_SphericalGeoCoord  ( GBM_MET, Position, EastLongitude, Latitude, altitude )

   use, intrinsic :: iso_c_binding, only: c_float


   real (real128), intent (in) :: GBM_MET          ! Mission Elapsed Time, s
   real (c_float), dimension (3), intent (in) :: Position   ! m, from SC Position History file

   real (real64), intent (out) :: EastLongitude    ! degrees
   real (real64), intent (out) :: Latitude         ! degrees
   real (real64), intent (out), OPTIONAL :: altitude     ! m about Earth's mean radius


   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

   real (real64) :: radius

   real (real64) :: InertialLongitude

   real (real64) :: theta_deg

   real (real128) :: JulianDay
   integer (int32) :: year
   integer (int32) :: month
   integer (int32) :: day
   integer (int32) :: hours
   integer (int32) :: minutes
   real (real64) :: seconds

   ! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


   radius = sqrt ( dot_product ( real (Position, real64), real (Position, real64) ) )
   Latitude = asin ( Position (3) / radius ) * RADIANS_TO_DEGREES_DBL


   if ( present ( altitude ) )   altitude = radius - EARTH_MEAN_RADIUS_M_DBL


   call MET_to_UTC ( GBM_MET, JulianDay, year, month, day, hours, minutes, seconds )

   theta_deg = JulianDay_to_SiderealTime_Ver1  ( JulianDay )

   InertialLongitude = atan2 ( Position (2), Position (1) )

   EastLongitude = InertialLongitude * RADIANS_TO_DEGREES_DBL  -  theta_deg


   !  In degrees, InertialLongitude from ATAN2 can be as small as -180,
   !  and -theta_deg as small as -360, so that their sum, EastLongitude,
   !  can be as small as -540.
   !  Therefore converting to EastLongitude to non-negative value may require
   !  two additions of 360.

   if ( EastLongitude < 0.0 ) EastLongitude = EastLongitude + 360.0_real64
   if ( EastLongitude < 0.0 ) EastLongitude = EastLongitude + 360.0_real64


   return

end subroutine Cartesian_to_SphericalGeoCoord


! * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *


!  We want to output seconds as DD.DDDDDD with a leading zero 0D.DDDDDD if
!  the value is less than 10.   The floating point format F9.6 instead uses
!  a leading blank in this case and there is no way to override this behavior.
!  (Fortran provides a way of requesting leading zeros for integer formats,
!  but not for floating point.)   So we write the value to a string and test
!  the leading character, changing it to 0 if it is blank.

! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

function  seconds_2_string  ( seconds )

   character (len=9) :: seconds_2_string

   real (real64), intent (in) :: seconds

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   if ( seconds >= 61.000000_real64 ) then
      write (*, *) "seconds value too large:", seconds
      seconds_2_string = "TOOLARGE "
   else

      write (seconds_2_string, '(F9.6)' )  seconds
      if ( seconds_2_string (1:1) == " " )  seconds_2_string (1:1) = "0"

   end if


   return

end function  seconds_2_string


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


subroutine Get_UT1_minus_UTC_data ( )

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


   MODVAR_UT1_minus_UTC_ms = huge ( 1.0_real64 )
   MODVAR_State_of_UT1_minus_UTC = 0_int8


   !  Read Bulletins A to get predicted values into the future.
   !  Read it first so that actual values from Bulletin B will overwrite
   !  any overlap.

   call Read_Bulletins_A ( )


   !  Read several old-format Bulletin Bs.  These are necessary to obtain
   !  values from the start of the mission.  Rather inconvenient.
   !  Read them first to give preference to the new-format Bulletins
   !  for any overlap.

   call Read_OLD_Bulletin_Bs ( )


   !  Read the new-format Bulletin Bs.

   call Read_Bulletin_Bs ( )


end subroutine Get_UT1_minus_UTC_data


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


subroutine Read_Bulletins_A ( )

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   character (len=256) :: TimeFiles_PATH
   character (len=40) :: StartFileName
   character (len=281) :: FullName
   logical :: FileFound
   integer :: bull_LUN
   integer :: pos

   integer :: VolumeNum
   character (len=7) :: VolumeNum_as_RomanNumerals

   !  Generated using the Fortran 90 Roman Numerals package of Jeanne T. Martin
   !  of LLNL.   Since the range of Roman Numerals needed here is short, rather
   !  than using that package in this code, I code in the specific numerals needed.
   !  Also, we don't 99% of the features provided by that package, such as arithmetic
   !  on roman numerals.
   character (len=7), dimension (26:50), parameter :: RomanNumeralChars =        &
     [  "xxvi   ",  "xxvii  ",  "xxviii ",  "xxix   ",  "xxx    ",  "xxxi   ",   &
        "xxxii  ",  "xxxiii ",  "xxxiv  ",  "xxxv   ",  "xxxvi  ",  "xxxvii ",   &
        "xxxviii",  "xxxix  ",  "xl     ",  "xli    ",  "xlii   ",  "xliii  ",   &
        "xliv   ",  "xlv    ",  "xlvi   ",  "xlvii  ",  "xlviii ",  "xlix   ",   &
        "l      "  ]


   integer :: WeeklyNum
   character (len=8) :: WeeklyString

   logical :: SomeFileFound

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   SomeFileFound = .FALSE.

   call get_environment_variable ( "PATH_OF_TIME_FILES", TimeFiles_PATH )
   !  make sure that the path ends in a slash:
   pos = len_trim (TimeFiles_PATH)
   if ( pos > 0 ) then
      if ( TimeFiles_PATH (pos:pos) /= "/" )  TimeFiles_PATH = trim (TimeFiles_PATH) // "/"
   end if


   !  The earliest Bulletin A that we need is bulletina-xxvi-052.txt because it provides
   !  coverage that starts where the last Bulletin B available on 2014 Feb 8 ends.
   !  So the smallest Volume Number necessary is 26.   For Volume 26, the smallest (weekly) Number
   !  needed is the highest value, 52.   However, on some later volumes we might use smaller
   !  (weekly) Numbers, so we have to loop over all values.   Also, we might not download
   !  all Bulletins, so we don't abort the search loop as soon as we detect a missing file.

   do VolumeNum=26, 50

      !  Convert Volume number to equivalent Roman Number String:
      VolumeNum_as_RomanNumerals = RomanNumeralChars (VolumeNum)


      do WeeklyNum=1, 52

         StartFileName = "bulletina-" // trim (VolumeNum_as_RomanNumerals)

         write (WeeklyString, '( "-", I3.3, ".txt" )' )  WeeklyNum

         FullName = trim (TimeFiles_PATH) // trim (StartFileName) // trim (WeeklyString)

         inquire ( file=FullName, exist=FileFound )


         if ( FileFound ) then

            SomeFileFound = .TRUE.

            open ( file=FullName, newunit=bull_LUN, status="old", access="sequential",   &
                form="formatted", action="read" )

            call Read_One_BulletinA ( bull_LUN )

            close ( unit=bull_LUN )

         end if   ! FileFound ?


      end do   !  WeeklyNum

   end do   ! VolumeNum


   if ( .NOT. SomeFileFound ) then
      write (*, '( / "No Bulletins A were found !  Aborting.  Searched directory:" / A )' )  trim (TimeFiles_PATH)
      stop
   end if


end subroutine Read_Bulletins_A


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


subroutine Read_One_BulletinA ( bull_LUN )

   integer, intent (in) :: bull_LUN

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   character (len=65) :: line

   integer (int32) :: mjd
   real (real64) :: delta

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


   SkipHeaderLines: do

      read (bull_LUN, '(A)' )  line
      if ( trim (line) == "                    MJD      x(arcsec)   y(arcsec)   UT1-UTC(sec)" ) exit SkipHeaderLines

   end do SkipHeaderLines


   ReadData: do

      read (bull_LUN, '(A)' )  line
      if ( trim (line) == "      These predictions are based on all announced leap seconds." )  exit ReadData

      read (line, '( 19X, I5, 28X, F11.0 )' )  mjd, delta
      !write (*, '( I5, 2X, F10.1 )' )  mjd, delta

      if ( mjd >= MODVAR_MJD_range (1)  .AND.  mjd <= MODVAR_MJD_range (2) ) then

         MODVAR_UT1_minus_UTC_ms (mjd) = delta * 1000.0_real64   ! s to ms -- unit diff between Bull A and Bull B
         MODVAR_State_of_UT1_minus_UTC (mjd) = 1

      else
         write (*, '( / "ERROR 1: MJD ", I0, " is not in the expected range ", I0, " to ", I0 )' )    &
            mjd, MODVAR_MJD_range
      end if

   end do ReadData


end subroutine Read_One_BulletinA


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^

!  Opens old-format Bulletin B files and passes them to Read_One_OLD_BulletinB
!  for reading the data.

!  The approach for the old-format Bulletins is different from the new format.
!  As time advances, we will add more new-format Bulletins, so that routine is
!  flexible.  We know precisely which old-format Bulletins we need, so that
!  knowledge is built-in to this code, i.e., we need old-format Bulletin B files
!  246 to 251 and no others.

subroutine Read_OLD_Bulletin_Bs ( )

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   character (len=256) :: TimeFiles_PATH
   integer (int32) :: BulletinNum
   character (len=25) :: Bulletin_FileName
   character (len=281) :: FullName
   logical :: FileFound
   integer :: bull_LUN
   integer :: len

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


   call get_environment_variable ( "PATH_OF_TIME_FILES", TimeFiles_PATH )
   !  make sure that the path ends in a slash:
   len = len_trim (TimeFiles_PATH)
   if ( len > 0 ) then
      if ( TimeFiles_PATH (len:len) /= "/" )  TimeFiles_PATH = trim (TimeFiles_PATH) // "/"
   end if


   do BulletinNum=246, 251

      write ( Bulletin_FileName, '( "bulletinb_IAU1980-", I3, ".txt" )' )  BulletinNum

      FullName = trim (TimeFiles_PATH) // trim (Bulletin_FileName)

      inquire ( file=FullName, exist=FileFound )
      if ( .NOT. FileFound ) then
         write (*, '( "Required old Bulletin B file not found.  Aborting.  File: " / A )' )  trim (FullName)
         stop
      end if

      open ( file=FullName, newunit=bull_LUN, status="old", access="sequential",   &
             form="formatted", action="read" )

      call Read_One_OLD_BulletinB ( bull_LUN )

      close ( unit=bull_LUN )

   end do   ! BulletinNum


end subroutine Read_OLD_Bulletin_Bs


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


subroutine Read_One_OLD_BulletinB ( bull_LUN )

   integer, intent (in) :: bull_LUN

   character (len=70) :: line
   integer :: i
   integer (int32) :: mjd
   real (real64) :: delta

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


   SkipHeaderLines: do

      read (bull_LUN, '(A)' )  line
      if ( line == " 2 - SMOOTHED VALUES OF x, y, UT1, D, DPSI, DEPSILON (IERS EVALUATION)" ) exit SkipHeaderLines

   end do SkipHeaderLines

   do i=1, 5
     read (bull_LUN, '(A)' )  line
   end do


   ReadData: do

      read (bull_LUN, '(A)' )  line

      if ( line == "                                                             IERS, B 2" )  exit ReadData

      read (line, '( 12X, I5, 18X, F11.0 )' )  mjd, delta
      !write (*, '( I5, 2X, F11.6 )' )  mjd, delta

      if ( mjd >= MODVAR_MJD_range (1)  .AND.  mjd <= MODVAR_MJD_range (2) ) then

         MODVAR_UT1_minus_UTC_ms (mjd) = delta * 1000.0_real64   ! unit conversion from s to ms
         MODVAR_State_of_UT1_minus_UTC (mjd) = 2

      else
         write (*, '( / "ERROR 3: MJD ", I0, " is not in the expected range ", I0, " to ", I0 )' )    &
            mjd, MODVAR_MJD_range
      end if

   end do ReadData


end subroutine Read_One_OLD_BulletinB


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  Determines which new-format Bulletin B files are available, opens
!  the files and calls Read_One_BulletinB to read the data from each
!  file.  The earliest new-format Bulletin B is number 252.  The concept
!  is that additional Bulletin B files will be downloaded going forward
!  so the routine is flexible about the highest number available.  The routine
!  expects there to be no gaps in the sequence of Bulletins -- as soon
!  as it detects a missing Bulletin number it aborts the search.

subroutine Read_Bulletin_Bs ( )

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   character (len=256) :: TimeFiles_PATH
   integer (int32) :: BulletinNum
   character (len=14) :: Bulletin_FileName
   character (len=270) :: FullName, PREV_FullName
   logical :: FileFound, SomeFileFound
   integer :: bull_LUN
   integer :: len

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


   call get_environment_variable ( "PATH_OF_TIME_FILES", TimeFiles_PATH )
   !  make sure that the path ends in a slash:
   len = len_trim (TimeFiles_PATH)
   if ( len > 0 ) then
      if ( TimeFiles_PATH (len:len) /= "/" )  TimeFiles_PATH = trim (TimeFiles_PATH) // "/"
   end if


   SomeFileFound = .FALSE.
   FullName = " "
   BulletinNum = 252

   ReadFilesLoop: do

      PREV_FullName = FullName
      if ( BulletinNum <= 999 ) then
         write ( Bulletin_FileName, '( "bulletinb.", I3 )' )  BulletinNum
      else
         write ( Bulletin_FileName, '( "bulletinb.", I4 )' )  BulletinNum
      end if
      FullName = trim (TimeFiles_PATH) // trim (Bulletin_FileName)

      inquire ( file=FullName, exist=FileFound )
      if ( .NOT. FileFound )  exit ReadFilesLoop     !!! Loop termination !!!

      SomeFileFound = .TRUE.
      open ( file=FullName, newunit=bull_LUN, status="old", access="sequential",   &
             form="formatted", action="read" )

      call Read_One_BulletinB ( bull_LUN )

      close ( unit=bull_LUN )

      BulletinNum = BulletinNum + 1

   end do ReadFilesLoop


   if ( .NOT. SomeFileFound ) then
      write (*, '( / "No New Bulletin B files were found.  ABORTING.  Searched on directory:" / A )' )  trim (TimeFiles_PATH)
      stop
   end if


end subroutine Read_Bulletin_Bs


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


subroutine Read_One_BulletinB ( bull_LUN )

   integer, intent (in) :: bull_LUN

   character (len=64) :: line

   integer (int32) :: mjd
   real (real64) :: delta

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


   SkipHeaderLines: do

      read (bull_LUN, '(A)' )  line
      if ( line (1:20) == " Mean formal error  " ) exit SkipHeaderLines

   end do SkipHeaderLines
   read (bull_LUN, '(A)' )  line


   ReadData: do

      read (bull_LUN, '(A)' )  line
      if (len_trim (line) == 0 )  exit ReadData

      read (line, '( 15X, I5, 18X, F11.0 )' )  mjd, delta
      !write (*, '( I5, 2X, F10.4 )' )  mjd, delta

      if ( mjd >= MODVAR_MJD_range (1)  .AND.  mjd <= MODVAR_MJD_range (2) ) then

         MODVAR_UT1_minus_UTC_ms (mjd) = delta
         MODVAR_State_of_UT1_minus_UTC (mjd) = 2

      else
         write (*, '( / "ERROR 2: MJD ", I0, " is not in the expected range ", I0, " to ", I0 )' )    &
            mjd, MODVAR_MJD_range
      end if

   end do ReadData


end subroutine Read_One_BulletinB


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


!  Read the GBM-specific file with leap second information.  As of 2014 Feb 17 the
!  contents of the file MET_of_LeapSeconds.txt are:

!    MET        year  month   day
! 157766400     2005    12    31
! 252460801     2008    12    31
! 362793602     2012     6    30

!  This file will need to be updated whenever another leap second occurs.
!  Someone will have to deduce the MET value of the leap seconds.
!  This will have to be done in a consistent manner when Bulletin B / Bulletin A
!  files are updated to show the leap seconds update.    All of these files
!  are in the directory indicated by the environment variable PATH_OF_TIME_FILES.

!  The times in the file are the MET values of the leap seconds.  These are the
!  seconds that are 23 hours, 59 minutes, 60 seconds.

   subroutine Read_LeapSecondData ()

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

   character (len=256) :: TimeFiles_PATH
   character (len=278) :: FullName
   logical :: FileFound

   integer :: len
   integer :: leap_LUN
   character (len=10) :: DataLine
   integer :: ReadCode

   integer :: Number_of_LeapSeconds
   integer :: i

   ! - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


   call get_environment_variable ( "PATH_OF_TIME_FILES", TimeFiles_PATH )
   !  make sure that the path ends in a slash:
   len = len_trim (TimeFiles_PATH)
   if ( len > 0 ) then
      if ( TimeFiles_PATH (len:len) /= "/" )  TimeFiles_PATH = trim (TimeFiles_PATH) // "/"
   end if

   FullName = trim (TimeFiles_PATH) // "MET_of_LeapSeconds.txt"

   inquire ( file=FullName, exist=FileFound )
   if ( .NOT. FileFound ) then
      write (*, '( / "Leap seconds data file not found.  ABORTING.  Searched for:" / A )' )  trim (FullName)
      stop
   end if

   open ( file=FullName, newunit=leap_LUN, status="old", access="sequential",   &
          form="formatted", action="read" )


   Number_of_LeapSeconds = -1  !  -1 instead of 0 to not count the header line

   FindFileLen: do

      read (leap_LUN, '(A)', iostat=ReadCode)  DataLine

      if ( ReadCode /= 0 ) then
         if ( ReadCode == iostat_end ) then
            exit FindFileLen    ! end of file -- exit read loop
         else
            write ( *, '( / "error reading the file of Leap Seconds MET values: ", I0 )' )  ReadCode
            stop
         end if
      end if

      Number_of_LeapSeconds = Number_of_LeapSeconds + 1

   end do FindFileLen


   rewind (leap_LUN)
   read (leap_LUN, '(A)')  DataLine    ! skip header line


   allocate ( LeapSecondData (1: Number_of_LeapSeconds) )

   do i=1, Number_of_LeapSeconds
      read (leap_LUN, *)  LeapSecondData (i) % MET_of_LeapSecond,     &
          LeapSecondData (i) % year, LeapSecondData (i) % month, LeapSecondData (i) % day
   end do


   close (leap_LUN)


end subroutine Read_LeapSecondData


! ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^ ^


end module Time_and_Coord_Routines
